      module forcedirect
!****************************************************************
! This module is an addon component of VINE, copyright by the   *
! authors, as listed below. Please see the COPYING and LICENSE  *
! files in this directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!  This file contains the VINE module to calculate gravity      *
!  on particles, using a cache blocked direct summation         *
!  of the contributions of each particle on the others.         *
!                                                               *
!  The externally visible routine is:                           *
!     host_dirsum  : Do a direct summation gravity calculation  *
!                    on the host machine.                       *
!       Arguments:                                              *
!         npart    : Number of particles in the simulation      *
!         isoft    : Variable kernel, fixed kernel or Plummer   *
!                    softening                                  *
!         eps2     : Squared softening radius for Plummer       *
!                    softening option                           *
!         posmas   : Particle position (not used here)          *
!         vel      : Particle velocity (not used here)          *
!                    positions and velocities only needed for   *
!                    grape calculations. Included here so that  *
!                    the subroutine interfaces are identical.   *
!         acc      : Gravitational acceleration                 *
!         poten    : Gravitational potential                    *
!                                                               *
!                                                               *
! Implementation notes                                          *
! --------------------                                          *
!                                                               *
! 1) This module currently only contains stubs for actual       *
! direct force calculation routines. I made actual routines     *
! that were fairly sophisticated in their use of cache etc,     *
! but it turns out that in the end I outsmarted myself and      *
! made routines that were 1) buggy and 2) difficult to debug.   *
! I haven't had a chance to go back and fix any of these        *
! problems, so what you get are just the stubs. The interested  *
! coder can take a look at the routines that live in the        *
! grapeX.F modules and in the fakegrape6.F module. Some time    *
! ago, I had reason to write versions of those routines, and    *
! I am certain that they work. You should be able to take       *
! these routines and adapt them into working dirsum routines    *
! for both Plummer and kernel softening options with very       *
! little difficulty.                                            *
!                                                               *
! This comment references the old, broken version:              *
! 2) It is an odd fact that when you use the kernel softening   *
! option with direct summation (subroutine kernel_dirsum        *
! and kernel_fastdirsum),this code requires somewhat longer     *
! to run than when you do a tree traversal with the opening     *
! criterion set to open all nodes. This is because in the       *
! tree traversal version, part of the process is to separate    *
! near and far atoms, so that far atoms can be sent to the      *
! much simpler and faster version of the gravity routine that   *
! doesn't require that the kernel quantities be calculated      *
! since there is a guarantee that all atoms are far enough      *
! away not to require any softening.                            *
!                                                               *
! Module written by Andy Nelson                                 *
!                                                               *
!****************************************************************
      use dimens
      use treevars
      use constants
      use perform

      implicit none

      private
      public host_dirsum

      include 'INCLUDE/hwinfo'

      double precision, parameter:: gconst=ONE
!
! Parameters needed for kernel softening calculation
!
      double precision, parameter::twntysevths=20d0/7d0
      double precision, parameter::sixth=1d0/6d0 
      double precision, parameter::fiftnth=1d0/15d0
      double precision, parameter::thirtyth=1d0/30d0

!
! How much i and j memory we have:
!
      integer, parameter :: istorage=64
      integer, parameter :: jstorage=131072


      contains
!----------------------------------------------------------------------
      subroutine host_dirsum(npart,isoft,eps2,posmas,vel,acc,poten)
!*********************************************************************
!                                                                    *
! Driver for selecting options for direct summation                  *
! of gravitational accelerations and potentials.                     *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: npart,isoft
      double precision, intent(in)   :: eps2
      double precision, intent(in)   :: posmas(ndim1,npart)
      double precision, intent(in)   :: vel(ndim,npart)
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      integer i,j,nact,nact_sub

      character(len=20), parameter::where='host_dirsum'
      if(idbg.ge.3)print *,where


!      print *,'Host based direct summation is broken'
!      print *,'See comments in the header of module forcedirect.F'
!      stop'Host based direct summation is broken. Sorry.'

      nact=0
!$omp parallel default(none)
!$omp& shared(active,linksphtotree,nact,npart)
!$omp&  private(i,j,nact_sub)
      nact_sub=0
!$omp do
      do i=1,npart
         j=linksphtotree(i)
         if(active(j).eq.1)nact_sub=nact_sub+1
      enddo
!$omp critical
      nact=nact+nact_sub
!$omp end critical
!$omp end parallel


!      if(nact.gt.npart/2)then
!         select case(isoft)
!         case(id_Plummer)
!            call Plummer_fastdirsum(npart,eps2,acc,poten)
!         case(id_fixkern,id_varkern)
!            call kernel_fastdirsum (npart,acc,poten)
!         case default
!            stop 'Unknown softening option in host_dirsum'
!         end select
!      else
         select case(isoft)
         case(id_Plummer)
            call Plummer_dirsum(npart,posmas,eps2,acc,poten)
         case(id_fixkern,id_varkern)
            call kernel_dirsum (npart,acc,poten)
         case default
            stop 'Unknown softening option in host_dirsum'
         end select
!      endif

      return
      end subroutine host_dirsum
!----------------------------------------------------------------------
      subroutine Plummer_dirsum(npart,posmas,eps2,acc,poten)
!*********************************************************************
!                                                                    *
! Calculate the gravity on active particles due to all               *
! other particles in the simulation using an N**2 summation and      *
! a Plummer law softening for the gravitational acceleration.        *
!                                                                    *
! Basically, this routine is a cache optimized version of this       *
! do loop nest:                                                      *
!                                                                    *
!    do j=1,npart                                                    *
!       do i=1,npart                                                 *
!          get forces on i due to j                                  *
!       enddo                                                        *
!    enddo                                                           *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: npart
      double precision, intent(in   ):: eps2
      double precision, intent(in   ):: posmas(ndim1,npart)
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      integer isofar
      integer i,j,jj,ii,k
      integer ni,nj,njlow,njhigh
      double precision eps2use
!
! Data for j particles:
!
      integer jindx(jstorage)
      double precision xmj(ndim1,jstorage)
!
! Data for i particles:
!
      integer iindx(istorage)
      double precision xi   (ndim,istorage) 
      double precision acc_i(ndim,istorage)
      double precision phi_i(     istorage)

      double precision tme1,tme2
      character(len=20), parameter:: where =('Plummer_dirsum')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)


      eps2use=eps2

!$omp parallel do default(none)
!$omp& firstprivate(eps2use)
!$omp& shared(npart,posmas,acc,poten,active,linksphtotree)
!$omp& private(i,j,ii,jj,k,isofar,njlow,njhigh,ni,nj)
!$omp& private(xmj,jindx,xi,iindx,acc_i,phi_i)
      do j=1,npart,jstorage
!
! Cycle through groups of nj particles that contribute forces
!
         njlow  = j
         njhigh = min(j+jstorage-1,npart)

         nj=njhigh-njlow+1

         do jj=njlow,njhigh
            k=jj-njlow+1

            jindx(k)=jj
            xmj(ims,k)=posmas(ims,jj)
            xmj(1  ,k)=posmas(1  ,jj)
            xmj(2  ,k)=posmas(2  ,jj)
#if NDIM==3
            xmj(3  ,k)=posmas(3  ,jj)
#endif
         enddo

         isofar=0
         do while(isofar.lt.npart)
!
! Cycle through groups of ni particles that need forces
!
            ni=0
            do while(ni.lt.istorage .and. isofar.lt.npart)
               isofar=isofar+1
               ii=linksphtotree(isofar)
               if(active(ii).eq.0)cycle

               ni=ni+1
               iindx(ni)=isofar
            enddo

            do i=1,ni
               k=iindx(i)
               xi(1,i)=posmas(1,k)
               xi(2,i)=posmas(2,k)
#if NDIM==3
               xi(3,i)=posmas(3,k)
#endif
            enddo

            call Plummer_n2(ni,nj,iindx,jindx,xi,xmj,eps2use,
     &                     acc_i,phi_i)

!$omp critical
            do i=1,ni
               k=iindx(i)
               poten(  k) = poten(  k) - phi_i(  i)
               acc  (1,k) = acc  (1,k) + acc_i(1,i)
               acc  (2,k) = acc  (2,k) + acc_i(2,i)
#if NDIM==3
               acc  (3,k) = acc  (3,k) + acc_i(3,i)
#endif
            enddo
!$omp end critical

         enddo
      enddo

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_hostcalc)=timing(itme_hostcalc)+(tme2-tme1)
         ncalls(itme_hostcalc)=ncalls(itme_hostcalc)+1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine Plummer_dirsum
!----------------------------------------------------------------------
      subroutine kernel_dirsum(npart,acc,poten)
!*********************************************************************
!                                                                    *
! Calculate the gravity on active particles due to all               *
! other particles in the simulation using an N**2 summation and      *
! a W4 kernel softening for the gravitational acceleration.          *
!                                                                    *
! See note in header about speed. This routine would go a lot        *
! faster if we presorted neighbor/nonneighbor clumps. Could use      *
! code from the first bit of getneigh for this purpose.              *
!                                                                    *
! Basically, this routine is a cache optimized version of this       *
! do loop nest:                                                      *
!                                                                    *
!    do j=1,npart                                                    *
!       do i=1,npart                                                 *
!          get forces on i due to j                                  *
!       enddo                                                        *
!    enddo                                                           *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: npart
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      return
      end subroutine kernel_dirsum
!----------------------------------------------------------------------
      subroutine Plummer_fastdirsum(npart,eps2,acc,poten)
!*********************************************************************
!                                                                    *
! Calculate the gravity on active particles due to all               *
! other particles in the simulation using an N(N-1)/2 summation and  *
! a Plummer law softening for the gravitational acceleration.        *
!                                                                    *
! Basically, this routine is a cache optimized version of this       *
! do loop nest:                                                      *
!                                                                    *
!    do j=1,npart                                                    *
!       do i=1,j-1                                                   *
!          get forces on i and j due to each other.                  *
!       enddo                                                        *
!    enddo                                                           *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: npart
      double precision, intent(in)   :: eps2
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      return
      end subroutine Plummer_fastdirsum
!----------------------------------------------------------------------
      subroutine kernel_fastdirsum(npart,acc,poten)
!*********************************************************************
!                                                                    *
! Calculate the gravity on active particles due to all               *
! other particles in the simulation using an N(N-1)/2 summation and  *
! a W4 kernel softening for the gravitational acceleration.          *
!                                                                    *
! Basically, this routine is a cache optimized version of this       *
! do loop nest:                                                      *
!                                                                    *
!    do j=1,npart                                                    *
!       do i=1,j-1                                                   *
!          get forces on i and j due to each other.                  *
!       enddo                                                        *
!    enddo                                                           *
!                                                                    *
! See note in header about speed. This routine would go a lot        *
! faster if we presorted neighbor/nonneighbor clumps. Could use      *
! code from the first bit of getneigh for this purpose.              *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: npart
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      return
      end subroutine kernel_fastdirsum
!----------------------------------------------------------------------
      subroutine Plummer_n2(ni,nj,iindx,jindx,xi,xmj,eps2,
     &                      acc_i,phi_i)
!****************************************************************
!                                                               *
! This routine does a Plummer force calculation of nj particles *
! on ni particles.                                              *
!                                                               *
!****************************************************************

      integer, intent(in):: ni,nj
      integer, intent(in):: iindx(ni),jindx(nj)
      double precision, intent(in ):: eps2
      double precision, intent(in ):: xi (ndim ,ni)
      double precision, intent(in ):: xmj(ndim1,nj)
      double precision, intent(out):: acc_i(3,ni)
      double precision, intent(out):: phi_i(  ni)

      integer i,j
      integer nlo1,nlo2,nhi1,nhi2,ii
      double precision dx,dy,dz,r2,rinv,r3inv,emr3inv

      acc_i=ZERO ; phi_i=ZERO

      do j=1,nj

         ii=0
         do i=1,ni
            if(iindx(i).eq.jindx(j))ii=i   !avoid self calculation
         enddo

         nlo1=1 ; nhi1=ni
         nlo2=1 ; nhi2=0

         if(ii.ne.0)then
            nlo1=1    ; nhi1=ii-1
            nlo2=ii+1 ; nhi2=ni
         endif

         do i=nlo1,nhi1

            dx=xmj(1,j)-xi(1,i)
            dy=xmj(2,j)-xi(2,i)
            r2=dx**2 + dy**2 + eps2
#if NDIM==3
            dz=xmj(3,j)-xi(3,i)
            r2=r2 + dz**2
#endif
            rinv=ONE/sqrt(r2)
            r3inv=rinv**3

            emr3inv=xmj(ndim1,j)*r3inv

            phi_i(  i) = phi_i(  i) - xmj(ndim1,j)*rinv
            acc_i(1,i) = acc_i(1,i) + dx*emr3inv
            acc_i(2,i) = acc_i(2,i) + dy*emr3inv
#if NDIM==3
            acc_i(3,i) = acc_i(3,i) + dz*emr3inv
#endif
         enddo

         do i=nlo2,nhi2

            dx=xmj(1,j)-xi(1,i)
            dy=xmj(2,j)-xi(2,i)
            r2=dx**2 + dy**2 + eps2
#if NDIM==3
            dz=xmj(3,j)-xi(3,i)
            r2=r2 + dz**2
#endif
            rinv=ONE/sqrt(r2)
            r3inv=rinv**3

            emr3inv=xmj(ndim1,j)*r3inv

            phi_i(  i) = phi_i(  i) - xmj(ndim1,j)*rinv
            acc_i(1,i) = acc_i(1,i) + dx*emr3inv
            acc_i(2,i) = acc_i(2,i) + dy*emr3inv
#if NDIM==3
            acc_i(3,i) = acc_i(3,i) + dz*emr3inv
#endif
         enddo
      enddo

12    format(2i8,10(1pd24.16))
13    format(3i8,10(1pd24.16))
14    format(4i8,10(1pd24.16))
15    format(5i8,10(1pd24.16))

      return
      end subroutine Plummer_n2
!----------------------------------------------------------------------
      end module forcedirect
