      module analyze
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
! This file contains the VINE module to do various on the fly   *
! analysis on a running simulation.                             *
!                                                               *
!  Written by Andy Nelson, Markus Wetzstein and Thorsten Naab   *
!****************************************************************
      use dimens
      use partvars
      use constants
      use stubs
      use perform

      implicit none

      private
      public angmom,centmass,dumcentmass,toten,insync
      public synch2n

      contains
!----------------------------------------------------------------
      subroutine angmom
!************************************************************
!                                                           *
!  compute the total angular momentum of the system         *
!                                                           *
!************************************************************
      integer i
      double precision am1,am2,am3,xi,yi,zi,vxi,vyi,vzi
      character(len=20), parameter:: where =('angmom')
      if(idbg.ge.2)print *,where

      am1=zero ; am2=zero ; am3=zero
!
! compute total angular momentum of particles
!
!$omp parallel do default(none) private(i,xi,yi,zi,vxi,vyi,vzi)
!$omp&   shared(cmtot,vcmtot,npart,posmas,vel)
!$omp&   reduction(+:am1,am2,am3)
      do i=1,npart
         xi =posmas(1,i) - cmtot (1)
         yi =posmas(2,i) - cmtot (2)
         vxi=vel   (1,i) - vcmtot(1)
         vyi=vel   (2,i) - vcmtot(2)
         am3=am3 + posmas(ims,i)*(xi*vyi - vxi*yi)
#if NDIM==3
         zi=posmas(3,i) - cmtot (3)
         vzi=vel  (3,i) - vcmtot(3)
         am1=am1 + posmas(ims,i)*(yi*vzi - vyi*zi)
         am2=am2 + posmas(ims,i)*(zi*vxi - vzi*xi)
#endif
      enddo
      amompart(1)=am1
      amompart(2)=am2
      amompart(3)=am3
      amompart(4)=sqrt(amompart(1)**2 + amompart(2)**2 + amompart(3)**2)
!
! compute total angular momentum of point masses
!
      amompoim(:)=zero

      do i=1,npoim
         xi =posmaspm(1,i) - cmtot (1)
         yi =posmaspm(2,i) - cmtot (2)
         vxi=velpm   (1,i) - vcmtot(1)
         vyi=velpm   (2,i) - vcmtot(2)
         amompoim(3)=amompoim(3) + posmaspm(ims,i)*(xi*vyi - vxi*yi)
#if NDIM==3
         zi=posmaspm(3,i) - cmtot (3)
         vzi=velpm  (3,i) - vcmtot(3)
         amompoim(1)=amompoim(1) + posmaspm(ims,i)*(yi*vzi - vyi*zi)
         amompoim(2)=amompoim(2) + posmaspm(ims,i)*(zi*vxi - vzi*xi)
#endif
      enddo
      amompoim(4)=sqrt(amompoim(1)**2 + amompoim(2)**2 + amompoim(3)**2)

!
! compute total angular momentum of system
!
      amomtot(1)=amompart(1)+amompoim(1)
      amomtot(2)=amompart(2)+amompoim(2)
      amomtot(3)=amompart(3)+amompoim(3)

      amomtot(4)=sqrt(amomtot(1)**2 + amomtot(2)**2 + amomtot(3)**2)

      return
      end subroutine angmom
!----------------------------------------------------------------
      subroutine centmass
!************************************************************
!                                                           *
!  compute the center of mass of the system                 *
!                                                           *
!************************************************************
      integer i
      double precision cm1  ,cm2  ,cm3  ,vcm1  ,vcm2  ,vcm3
      double precision cm1sq,cm2sq,cm3sq,vcm1sq,vcm2sq,vcm3sq

      character(len=20), parameter:: where =('centmass')
      if(idbg.ge.2)print *,where
!
! initialize to zero. It would be nice if openmp 
!     reductions of array locations, not just scalars. 
!
         cm1=zero   ;    cm2=zero ;    cm3=zero
        vcm1=zero   ;   vcm2=zero ;   vcm3=zero
       cm1sq=zero   ;  cm2sq=zero ;  cm3sq=zero
      vcm1sq=zero   ; vcm2sq=zero ; vcm3sq=zero
      totmaspart=zero
!
! compute total center of mass of particles, its velocity
!  and the standard deviation around them (useful for rk error)
!
!$omp parallel do default(none) private(i),shared(posmas,vel,npart),
!$omp& reduction(+:totmaspart, cm1, cm2, cm3,
!$omp&                        vcm1,vcm2,vcm3,
!$omp&                        cm1sq  ,cm2sq  ,cm3sq,
!$omp&                        vcm1sq ,vcm2sq ,vcm3sq)
      do i=1,npart
         totmaspart=totmaspart+posmas(ims,i)
         cm1    = cm1    + posmas(ims,i)*posmas(1,i)
         cm2    = cm2    + posmas(ims,i)*posmas(2,i)
         cm1sq  = cm1sq  + posmas(ims,i)*posmas(1,i)**2
         cm2sq  = cm2sq  + posmas(ims,i)*posmas(2,i)**2
         vcm1   = vcm1   + posmas(ims,i)*vel(1,i)
         vcm2   = vcm2   + posmas(ims,i)*vel(2,i)
         vcm1sq = vcm1sq + posmas(ims,i)*vel(1,i)**2
         vcm2sq = vcm2sq + posmas(ims,i)*vel(2,i)**2
#if NDIM==3
         cm3    = cm3    + posmas(ims,i)*posmas(3,i)
         cm3sq  = cm3sq  + posmas(ims,i)*posmas(3,i)**2
         vcm3   = vcm3   + posmas(ims,i)*vel(3,i)
         vcm3sq = vcm3sq + posmas(ims,i)*vel(3,i)**2
#endif
      enddo

      cmpart (1)=cm1/totmaspart
      cmrms  (1)=sqrt( cm1sq/totmaspart -  cmpart(1)**2)
      cmpart (2)=cm2/totmaspart
      cmrms  (2)=sqrt( cm2sq/totmaspart -  cmpart(2)**2)
      vcmpart(1)=vcm1/totmaspart
      vcmrms (1)=sqrt(vcm1sq/totmaspart - vcmpart(1)**2)
      vcmpart(2)=vcm2/totmaspart
      vcmrms (2)=sqrt(vcm2sq/totmaspart - vcmpart(2)**2)
#if NDIM==3
      cmpart (3)=cm3/totmaspart
      cmrms  (3)=sqrt( cm3sq/totmaspart -  cmpart(3)**2)
      vcmpart(3)=vcm3/totmaspart
      vcmrms (3)=sqrt(vcm3sq/totmaspart - vcmpart(3)**2)
#endif
!
! Point masses
!
      totmaspoim=zero
      cmpoim (1)=zero ; cmpoim (2)=zero
      vcmpoim(1)=zero ; vcmpoim(2)=zero
#if NDIM==3
      cmpoim (3)=zero ; vcmpoim(3)=zero
#endif
      do i=1,npoim
         totmaspoim=totmaspoim + posmaspm(ims,i)
         cmpoim (1)= cmpoim(1) + posmaspm(ims,i)*posmaspm(1,i)
         cmpoim (2)= cmpoim(2) + posmaspm(ims,i)*posmaspm(2,i)
         vcmpoim(1)=vcmpoim(1) + posmaspm(ims,i)*velpm(1,i)
         vcmpoim(2)=vcmpoim(2) + posmaspm(ims,i)*velpm(2,i)
#if NDIM==3
         cmpoim (3)= cmpoim(3) + posmaspm(ims,i)*posmaspm(3,i)
         vcmpoim(3)=vcmpoim(3) + posmaspm(ims,i)*velpm(3,i)
#endif
      enddo
      cmpoim (1)=cmpoim (1)/(totmaspoim+small)
      cmpoim (2)=cmpoim (2)/(totmaspoim+small)
      vcmpoim(1)=vcmpoim(1)/(totmaspoim+small)
      vcmpoim(2)=vcmpoim(2)/(totmaspoim+small)
#if NDIM==3
      cmpoim (3)=cmpoim (3)/(totmaspoim+small)
      vcmpoim(3)=vcmpoim(3)/(totmaspoim+small)
#endif
!
! other species of particles/whatever else go here. 
!
!         none at the moment
!
! Then add all parts together
!
      totmas=totmaspart+totmaspoim
      cmtot (1)=(totmaspart*cmpart (1) + totmaspoim*cmpoim (1))/totmas
      cmtot (2)=(totmaspart*cmpart (2) + totmaspoim*cmpoim (2))/totmas
      vcmtot(1)=(totmaspart*vcmpart(1) + totmaspoim*vcmpoim(1))/totmas
      vcmtot(2)=(totmaspart*vcmpart(2) + totmaspoim*vcmpoim(2))/totmas
#if NDIM==3
      cmtot (3)=(totmaspart*cmpart (3) + totmaspoim*cmpoim (3))/totmas
      vcmtot(3)=(totmaspart*vcmpart(3) + totmaspoim*vcmpoim(3))/totmas
#endif

      return
      end subroutine centmass
!----------------------------------------------------------------
      subroutine dumcentmass
!************************************************************
!                                                           *
!  compute the center of mass of the system, given that     *
!  particles may be at different points in time.            *
!                                                           *
!************************************************************
      use indivtsvars
      use miscvars, only: indts
      use indivts, only: getdt,save_activpm,restore_activpm
      use integ

      integer i
      double precision cm1,cm2,cm3,vcm1,vcm2,vcm3

      double precision tme1,tme2
      character(len=20), parameter:: where =('dumcentmass')
      if(idbg.ge.2)print *,where

      if(indts.eq.1)then
!
! synchronize particles to the same (lastts_end) time.
! Only need to do this if indts is on. If not, the dum*
! variables are still filled with their latest (end ts)
! values anyway and haven't been overwritten yet.
!
         call save_activpm(npoim)   !need active in pmstore

         call synch2n(lastts_end)

         call restore_activpm(npoim)

      endif

      if(itme.ge.1)call cpu_time(tme1) 
!
!--initialize to zero
!
       cm1=zero ;  cm2=zero ;  cm3=zero
      vcm1=zero ; vcm2=zero ; vcm3=zero
      dumtotmaspart=zero
!
!--compute total center of mass of particles, its velocity
!  and the standard deviation around them (useful for rk error)
!
!$omp parallel do default(none) private(i)
!$omp& shared(dumposmas,dumvel,npart)
!$omp& reduction(+:dumtotmaspart, cm1, cm2, cm3,
!$omp&                           vcm1,vcm2,vcm3)
      do i=1,npart
         dumtotmaspart=dumtotmaspart+dumposmas(ims,i)
         cm1  = cm1  + dumposmas(ims,i)*dumposmas(1,i)
         cm2  = cm2  + dumposmas(ims,i)*dumposmas(2,i)
         vcm1 = vcm1 + dumposmas(ims,i)*dumvel(1,i)
         vcm2 = vcm2 + dumposmas(ims,i)*dumvel(2,i)
#if NDIM==3
         cm3  = cm3  + dumposmas(ims,i)*dumposmas(3,i)
         vcm3 = vcm3 + dumposmas(ims,i)*dumvel(3,i)
#endif
      enddo

      dumcmpart (1)=cm1/dumtotmaspart
      dumcmpart (2)=cm2/dumtotmaspart
      dumvcmpart(1)=vcm1/dumtotmaspart
      dumvcmpart(2)=vcm2/dumtotmaspart
#if NDIM==3
      dumcmpart (3)=cm3/dumtotmaspart
      dumvcmpart(3)=vcm3/dumtotmaspart
#endif
!
! Point masses
!
      dumtotmaspoim=zero
      dumcmpoim (1)=zero ; dumcmpoim (2)=zero 
      dumvcmpoim(1)=zero ; dumvcmpoim(2)=zero
#if NDIM==3
      dumcmpoim (3)=zero ; dumvcmpoim(3)=zero
#endif
      do i=1,npoim
         dumtotmaspoim=dumtotmaspoim+dumposmaspm(ims,i)
         dumcmpoim (1)= dumcmpoim(1)+dumposmaspm(ims,i)*dumposmaspm(1,i)
         dumcmpoim (2)= dumcmpoim(2)+dumposmaspm(ims,i)*dumposmaspm(2,i)
         dumvcmpoim(1)=dumvcmpoim(1)+dumposmaspm(ims,i)*dumvelpm(1,i)
         dumvcmpoim(2)=dumvcmpoim(2)+dumposmaspm(ims,i)*dumvelpm(2,i)
#if NDIM==3
         dumcmpoim (3)= dumcmpoim(3)+dumposmaspm(ims,i)*dumposmaspm(3,i)
         dumvcmpoim(3)=dumvcmpoim(3)+dumposmaspm(ims,i)*dumvelpm(3,i)
#endif
      enddo
      dumcmpoim (1)=dumcmpoim (1)/(dumtotmaspoim+small)
      dumcmpoim (2)=dumcmpoim (2)/(dumtotmaspoim+small)
      dumvcmpoim(1)=dumvcmpoim(1)/(dumtotmaspoim+small)
      dumvcmpoim(2)=dumvcmpoim(2)/(dumtotmaspoim+small)
#if NDIM==3
      dumcmpoim (3)=dumcmpoim (3)/(dumtotmaspoim+small)
      dumvcmpoim(3)=dumvcmpoim(3)/(dumtotmaspoim+small)
#endif
!
! other species of particles/whatever else go here. 
!
!         none at the moment

!
! Then add all parts together
!
      dumtotmas=dumtotmaspart+dumtotmaspoim
      dumcmtot (1)=(dumtotmaspart*dumcmpart (1) +
     &              dumtotmaspoim*dumcmpoim (1))/totmas
      dumcmtot (2)=(dumtotmaspart*dumcmpart (2) + 
     &              dumtotmaspoim*dumcmpoim (2))/totmas
      dumvcmtot(1)=(dumtotmaspart*dumvcmpart(1) +
     &              dumtotmaspoim*dumvcmpoim(1))/totmas
      dumvcmtot(2)=(dumtotmaspart*dumvcmpart(2) +
     &              dumtotmaspoim*dumvcmpoim(2))/totmas
#if NDIM==3
      dumcmtot (3)=(dumtotmaspart*dumcmpart (3) +
     &              dumtotmaspoim*dumcmpoim (3))/dumtotmas
      dumvcmtot(3)=(dumtotmaspart*dumvcmpart(3) +
     &              dumtotmaspoim*dumvcmpoim(3))/dumtotmas
#endif

      if(indts.eq.0)then
         totmas=dumtotmas
         cmtot (1)=dumcmtot(1)
         cmtot (2)=dumcmtot(2)
         vcmtot(1)=dumvcmtot(1)
         vcmtot(2)=dumvcmtot(2)
#if NDIM==3
         cmtot (3)=dumcmtot (3)
         vcmtot(3)=dumvcmtot(3)
#endif
      endif

      if(itme.ge.1)then
         call cpu_time(tme2) 
         timing(itme_analysis)=timing(itme_analysis)+
     &               (tme2-tme1)
!         ncalls(itme_analysis)=ncalls(itme_analysis)+1
      endif

      return
      end subroutine dumcentmass
!----------------------------------------------------------------
      subroutine synch2n(ntime)
!************************************************************
! Synchronize variables to the time corresponding to ntime  *
! Place results in the dum* arrays.                         *
! Needed for calling mtree at times when individual         *
! timesteps are used and particles are not synchronized     *
! in time. Also for on the fly analysis routines.           *
!                                                           *
!************************************************************
      use partvars, only: npoim,npart,npart_sph,dumposmas
      use miscvars, only: ibctype
      use integ   , only: halfupdate
      use indivts , only: getdt    
      use indivtsvars
#if PERIODIC        
      use ewald   , only: checkboxlimits
#endif              
      integer, intent(in):: ntime
      integer i
      double precision tme1,tme2
      if(itme.ge.1)call cpu_time(tme1) 
 
!$omp parallel do default(none) private(i)
!$omp& firstprivate(npart_sph) shared(npart,iactive)
!$omp& shared(iactive_s,iactive_n)
      do i=1,npart
         iactive  (i)=i
         if(i.le.npart_sph)then
            iactive_s(i)=i
         else
            iactive_n(i-npart_sph)=i 
         endif 
      enddo
         
      do i=1,npoim           
         iactive(i)=i        
      enddo

      nactive  =0 ; nactive_s =0 ; nactive_n =0 ; nactivepm =0   !In other words:
      nactivef =0 ; nactivef_s=0 ; nactivef_n=0 ; nactivefpm=0   !extrap everybody

      call getdt(npart,npoim,ntime)

      if(itme.ge.1)call cpu_time(tme2) 

      call halfupdate(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                nactive_s ,nactive_n ,nactive ,nactivepm ,
     &                iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                itstepbin,itstepbinpm,dtincr,
     &                deltat,deltatpm)

      if(ibctype.eq.id_perdbnd)then
         call checkboxlimits(npart,npoim,dumposmas,dumposmaspm)
      endif


      if(itme.ge.1)then
         timing(itme_analysis)=timing(itme_analysis)+tme2-tme1
!         ncalls(itme_analysis)=ncalls(itme_analysis)+1
      endif

      return
      end subroutine synch2n
!-----------------------------------------------------------------
      subroutine insync(i12step)
!************************************************************
! Compute some quantities for output at a time when all     *
! particles are synchronized in time.                       *
!************************************************************
      use indivtsvars, only: dtincr,lastts_end
      use miscvars, only: indts,iusebin,t,deltold
      use miscvars, only: igrav,etot,egrav,etherm,ekin
      use miscvars, only: t_unit,em_unit,u_unit
      use derivcalc, only: renew_deriv

      integer, intent(in):: i12step

      logical isopen,do_it,required

      logical, save :: first =.true.
      double precision fracTS,currtime
      double precision tme1,tme2
      character(len=20), parameter:: where =('insync')

      if(idbg.ge.2)print *,where

#if defined(LF)
      if(.not.first .and. indts.eq.1 .and. igrav.eq.1 .and. 
     &                    i12step.eq.1)then
!
! Give more exact gravitational potential energy for outputting
! to the world. Not required for the code to run correctly
! and will create possibly harmful side effects.
! Set do_it=.true. to do renewal.  .false. causes noop. 

         do_it=.false.
         required=.false.
         call renew_deriv(do_it,required)

      endif
      first=.false.
#endif

      if(itme.ge.1)call cpu_time(tme1) 

      call toten

      if(indts.eq.1)then
         fracTS=dtincr*dble(lastts_end)
      else
         if(iusebin.eq.0)then
            fracTS=deltold
         else
            fracTS=dtincr*dble(lastts_end)
         endif
      endif

      currtime=t+fracTS

      inquire(unit=32,opened=isopen)
      if(.not.isopen)then
         open(32,file='energy.log',position='append')
      endif

      write(32,10)t_unit*currtime,em_unit*u_unit*ekin  ,
     &                            em_unit*u_unit*egrav ,
     &                            em_unit*u_unit*etherm,
     &                            em_unit*u_unit*etot
      call bufflush(32)

      if(itme.ge.1)then
         call cpu_time(tme2) 
         timing(itme_analysis)=timing(itme_analysis)+tme2-tme1
         ncalls(itme_analysis)=ncalls(itme_analysis)+1
      endif

10    format(10(1pd20.12))
11    format(i7,10(1pd20.12))
      return
      end subroutine insync
!----------------------------------------------------------------
      subroutine toten
!************************************************************
!                                                           *
!  Compute all specificy energies                           *
!  kinetic, potential and internal.                         *
!                                                           *
!************************************************************
      use miscvars, only: etot,egrav,etherm,ekin

      integer i
      double precision v2pm
      character(len=20), parameter:: where =('toten')

      double precision velsq
      velsq(i)=vel(1,i)**2 + vel(2,i)**2
#if NDIM==3
     &                     + vel(3,i)**2
#endif

      if(idbg.ge.2)print *,where

      ekin=zero
      egrav=zero
      etherm=zero

!$omp parallel default(none) private(i) shared(npart,npart_sph)
!$omp&  shared(vel,posmas,u,poten)
!$omp&  reduction(+:ekin,egrav,etherm) 
!$omp do
      do i=1,npart
         ekin =ekin  + posmas(ims,i)*velsq(i)
         egrav=egrav + posmas(ims,i)*poten(i)
      enddo
!$omp do
      do i=1,npart_sph
         etherm=etherm + posmas(ims,i)*u(i)
      enddo
!$omp end parallel

      do i=1,npoim
         v2pm=velpm(1,i)**2 + velpm(2,i)**2
#if NDIM==3
     &       + velpm(3,i)**2
#endif
         ekin  =ekin  + posmaspm(ims,i)*v2pm
         egrav =egrav + posmaspm(ims,i)*potpm(i)
      enddo
!
!--normalizations
!
      ekin = half*ekin
      egrav=-half*egrav
      etot = ekin + egrav + etherm

      return
      end subroutine toten
!----------------------------------------------------------------
      end module analyze
