       module cosmo
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
! This file contains the VINE module to do cosmological              *
! calculations with the code. Note that only routines for the        *
! leapfrog integrator are contained, the Runge-Kutta doesn't work    *
! with cosmological expansion currently.                             *
!                                                                    *
!  Written by Thorsten Naab and Markus Wetzstein                     *
!*********************************************************************
      use dimens
      use miscvars
      use constants

      implicit none

      private
      public cosmo_vel,cosmic_time,cosmic_stepa,cosmo_ts
      public cosmo_dumptime

#ifndef COSMO
      cosmo module: COSMO preprocessor flag not set
#endif
#ifndef LF
      cosmo module: LF preprocessor flag not set (required for cosmo)
#endif

      contains
!--------------------------------------------------------------------------
      subroutine cosmo_vel(nactive,nactivepm,iactive,iactivepm,
     &                     npart,npoim,deltat,deltatpm,
     &                     vdot,vel,dumvel,vdotpm,velpm,dumvelpm)
!*********************************************************************
!                                                                    *
! Here the velocitiy updates due to cosmological expansion are       *
! computed. This update scheme explicitly assumes a leapfrog         *
! integrator!                                                        *
!                                                                    *
! Written by Thorsten Naab Feb2004                                   *
!                                                                    *
!*********************************************************************

      integer, intent(in):: npart,nactive  ,iactive  (nactive  )
      integer, intent(in):: npoim,nactivepm,iactivepm(nactivepm)
      double precision, intent(in   ) :: deltat  (npart)
      double precision, intent(in   ) :: vel     (ndim,npart)
      double precision, intent(in   ) :: vdot    (ndim,npart)
      double precision, intent(inout) :: dumvel  (ndim,npart)
      double precision, intent(in   ) :: deltatpm(npoim)
      double precision, intent(in   ) :: velpm   (ndim,npoim)
      double precision, intent(in   ) :: vdotpm  (ndim,npoim)
      double precision, intent(inout) :: dumvelpm(ndim,npoim)
      integer i,j
      double precision atothree,dtfull,v1now,v2now

      if(omegam.eq.1.d0 .and. omegal.eq.0.d0)then
! EdS
         ascale=(cosm_h0*tnow/twothd)**twothd
         atothree=ascale**(-three)
         redshift=(one/ascale)-one
         hnow=cosm_h0*(one+redshift)*sqrt(one+omegam*redshift)
      else
! general cosmologies
         atothree=ascale**(-three)
         hnow=hubble(cosm_h0,ascale,omegam,omegal)
         redshift=(one/ascale)-one
      endif

!$omp parallel do if(nactive .gt. 1024) default(none)
!$omp& firstprivate(hnow,atothree)
!$omp&  shared(nactive,iactive,vel,dumvel,vdot,deltat)
!$omp&  private(i,j,dtfull,v1now,v2now) 
      do j=1,nactive
         i=iactive(j)

         dtfull=two*deltat(i)                      !assumes leapfrog
         v1now=(one-hnow*dtfull)/(one+hnow*dtfull)
         v2now=dtfull*atothree/(one+hnow*dtfull)

         dumvel(1,i)=v1now*vel(1,i) + v2now*vdot(1,i)
         dumvel(2,i)=v1now*vel(2,i) + v2now*vdot(2,i)
#if NDIM==3
         dumvel(3,i)=v1now*vel(3,i) + v2now*vdot(3,i)
#endif
      enddo

      do j=1,nactivepm
         i=iactivepm(j)

         dtfull=two*deltatpm(i)                    !assumes leapfrog
         v1now=(one-hnow*dtfull)/(one+hnow*dtfull)
         v2now=dtfull*atothree/(one+hnow*dtfull)

         dumvelpm(1,i)=v1now*velpm(1,i) + v2now*vdotpm(1,i)
         dumvelpm(2,i)=v1now*velpm(2,i) + v2now*vdotpm(2,i)
#if NDIM==3
         dumvelpm(3,i)=v1now*velpm(3,i) + v2now*vdotpm(3,i)
#endif

      enddo

      return
      end subroutine cosmo_vel
!--------------------------------------------------------------------------
      subroutine cosmic_stepa(ascale,dtnext,cosm_h0,omegam,omegal)
!*********************************************************************
!                                                                    *
! Use Runge-Kutta method to step expansion factor ascale.            *
!                                                                    *
!*********************************************************************

      double precision, intent(in   ):: dtnext,cosm_h0,omegam,omegal
      double precision, intent(inout):: ascale
      integer, parameter:: nintegrate=16
      integer kk
      double precision HH,H6,a1,adot0,adot1,adot2

      print*,ascale,dtnext,cosm_h0,omegam,omegal

      if(omegam.eq.one.or.omegal.eq.zero)then
         print*,'For this cosmology there is currently no vine routine!'
         stop 'VINE cannot do this cosmology'
      endif

      HH = dtnext/dble(2*nintegrate)
      H6 = dtnext/dble(6*nintegrate)
     
      do kk=1,nintegrate
         adot0 = ascale*hubble(cosm_h0,ascale,omegam,omegal)     

         a1 = ascale+HH*adot0
         adot1 = a1*hubble(cosm_h0,a1,omegam,omegal)
     
         a1 = ascale+HH*adot1
         adot2 = a1*hubble(cosm_h0,a1,omegam,omegal)
     
         a1 = ascale+2.0d0*HH*adot2
         adot2=adot1+adot2
         adot1 = a1*hubble(cosm_h0,a1,omegam,omegal)
     
         ascale = ascale+H6*(adot0+adot1+2.0d0*adot2)
     
      enddo

      print*,ascale,dtnext,cosm_h0,omegam,omegal
      return
      end subroutine cosmic_stepa
!--------------------------------------------------------------------------
      subroutine cosmo_ts(nactive,nactivepm,iactive,iactivepm,
     &                    itstepbin,itstepbinpm,dtincr,
     &                    dtrat,dtratpm)
!*********************************************************************
!                                                                    *
! This routine limimts the maximum timestep to a 1% increase of      *
! the scale parameter.                                               *
!                                                                    *
! Written by Thorsten Naab, Feb 2004                                 *
!                                                                    *
!*********************************************************************

      integer, intent(in):: nactive  , iactive  (nactive  )
      integer, intent(in):: nactivepm, iactivepm(nactivepm)
      integer, intent(in):: itstepbin(idim),itstepbinpm(ipmdim)
      double precision, intent(in   ):: dtincr
      double precision, intent(inout):: dtrat(idim),dtratpm(ipmdim)

      integer i,j
      double precision acheck,redcheck,t_check,dtrat_cos

      character(len=20), parameter:: where =('cosmo_ts')
      if(idbg.ge.2)print *,where

      acheck=ascale+0.01d0*ascale
      redcheck=(one/acheck)-one
      call cosmic_time(redcheck,t_check,omegam,omegal,cosm_h0)
      dtcosmax=t_check-tnow
      print*,'ACHECK,ASCALE',
     $     acheck,ascale,redcheck,t_check,tnow,dtcosmax
c      afac=ascale**1.5d0

!$omp parallel do if(nactive.gt.1024) default(none)
!$omp& firstprivate(dtcosmax)
!$omp&  shared(nactive,iactive,dtincr,itstepbin,dtrat)
!$omp&  private(i,j,dtrat_cos) 
      do j=1,nactive
         i=iactive(j)
         dtrat_cos=dtcosmax/(dtincr*dble(itstepbin(i)))
         dtrat(i)=min(dtrat(i),dtrat_cos)
      enddo

      do j=1,nactivepm
         i=iactivepm(j)
         dtrat_cos=dtcosmax/(dtincr*dble(itstepbinpm(i)))
         dtratpm(i)=min(dtratpm(i),dtrat_cos)
      enddo

      return
      end subroutine cosmo_ts
!--------------------------------------------------------------------------
      subroutine cosmo_dumptime(nextoutstep,tnext)
!*********************************************************************
!                                                                    *
! Return the next time for a restart dump, for cosmological sims.    *
!                                                                    *
!*********************************************************************
      integer, intent(out):: nextoutstep
      double precision, intent(out):: tnext

      double precision rednext


      ired=ired+1
      rednext=redump(ired)
      call cosmic_time(rednext,tnext,omegam,omegal,cosm_h0)
      if(ired.eq.ncdumps)tstop=tnext
      nextoutstep=-1


      print *,'NEXT COSMO DUMP AT',tnext,'REDSHIFT',rednext

      return
      end subroutine cosmo_dumptime
!--------------------------------------------------------------------------
      subroutine cosmic_time(cosm_z,cosm_t,omegam,omegal,cosm_h0)
!*********************************************************************
!                                                                    *
! Calculate the cosmic time at redshift cosm_z, for the cosmological *
! model given.                                                       *
!                                                                    *
!*********************************************************************

      double precision cosm_z,cosm_t,omegam,omegal,cosm_h0

      double precision w(0:15),x(0:15)
      double precision step,h2,a0,a1,ak,adot
      double precision den0,den1,sum,ss
      integer n,nsteps,k

      data x/0.095012509837637440185d0, 0.281603550779258913230d0,
     &       0.458016777657227386342d0, 0.617876244402643748447d0,
     &       0.755404408355003033895d0, 0.865631202387831743880d0,
     &       0.944575023073232576078d0, 0.989400934991649932596d0,
     &      -0.095012509837637440185d0, -0.281603550779258913230d0,
     &      -0.458016777657227386342d0, -0.617876244402643748447d0,
     &      -0.755404408355003033895d0, -0.865631202387831743880d0,
     &      -0.944575023073232576078d0, -0.989400934991649932596d0/
      data w/0.189450610455068496285d0, 0.182603415044923588867d0,
     &       0.169156519395002538189d0, 0.149595988816576732081d0,
     &       0.124628971255533872052d0, 0.095158511682492784810d0,
     &       0.062253523938647892863d0, 0.027152459411754094852d0,
     &       0.189450610455068496285d0, 0.182603415044923588867d0,
     &       0.169156519395002538189d0, 0.149595988816576732081d0,
     &       0.124628971255533872052d0, 0.095158511682492784810d0,
     &       0.062253523938647892863d0, 0.027152459411754094852d0/

      a0 = 0.0D0
      a1 = 1.0D0/(1.0D0+cosm_z)
      den0 = omegam
      den1 = omegal
      nsteps = 10000
      step = (a1-a0)/nsteps
      h2 = step/2.0D0
      sum = 0.0D0
      do n=1,nsteps
        ss = 0.0D0
        do k=0,15
          ak = a0+h2*(1.0D0+x(k))
          adot = SQRT( den0/ak + den1*ak*ak + 1.0D0-den0-den1 )
          ss = ss + w(k)/adot
        enddo
        ss = ss*h2
        a0 = a0 + step
        sum = sum + ss
      enddo
      cosm_t = sum/cosm_h0

      return
      end subroutine cosmic_time
!--------------------------------------------------------------------------
      subroutine cosmic_dumptime(nextdumpstep,tnextdump)

      integer, intent(out):: nextdumpstep
      double precision, intent(out):: tnextdump

      double precision rednext

      ired=ired+1
      rednext=redump(ired)
      call cosmic_time(rednext,tnextdump,omegam,omegal,cosm_h0)
      if(ired.eq.ncdumps)tstop=tnextdump

      nextdumpstep=-1

      print *,'NEXT COSMO DUMP AT',tnextdump,'REDSHIFT',rednext

      return
      end subroutine cosmic_dumptime
!--------------------------------------------------------------------------
      double precision function hubble(cosm_h0,ascale,omegam,omegal)
!*********************************************************************
!                                                                    *
!        Computes Hubble constant H = a^\dot/a =                     *
!          H_0*(Omega_m/a^3 + Lambda + (1-Omega_m-Lambda)/a^2)^(1/2) *
!                                                                    *
!*********************************************************************
      double precision cosm_h0,ascale,omegam,omegal
        hubble = (cosm_h0/ascale)*sqrt(omegam/ascale + omegal*ascale**2
     $           + 1.0E0-omegam-omegal)
      return
      end function hubble
!--------------------------------------------------------------------------
      end module cosmo
