      module eostate
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!  This file contains the VINE module to do calculations of     *
!  the equation of state on particles, as needed for SPH.       *
!                                                               *
!  This routine should be used as a 'master' routine from       *
!  which you call your own, special purpose equations of        *
!  state. In its current form, it includes on a few basic       *
!  variants.                                                    *
!                                                               *
!  In its current form, this version calculates the eos         *
!  for ALL sph particles, whether or not they are active or     *
!  are a neighbor of an active particle (semi-active).          *
!  This means that all the work calculating the eos for         *
!  particles that are not at least semi-active is wasted.       *
!                                                               *
!  If an eos is added that is much more expensive than the      *
!  ones here now (say a nuclear eos), or the particle timestep  *
!  distribution is spread over many bins, the calculation will  *
!  need to be reduced to only active and semiactive particles,  *
!  otherwise the total calculation time will be significantly   *
!  impacted. A possible method for doing this easily would      *
!  be to create and integer array in sphvars.F with entries     *
!  corresponding to each SPH particle. These entries can be     *
!  set to a value of 1 in density when a given particle is      *
!  touched. Then the eos can check for this value when it       *
!  does its calculation. The entries can be set back to zero    *
!  either in the eos after the calculation, or if some other    *
!  routines have similar requirements, a separate routine could *
!  be written to do such a turn off at the end of the deriv     *
!  calculation.                                                 *
!                                                               *
!  Note that particles that are not themselves active, but      *
!  which are only neighbors of an active particle may need      *
!  updates too, depending on whether you are doing a full       *
!  extrapolation of all particle data or not. This makes        *
!  using the iactive* arrays as indicators of whether an        *
!  update is required somewhat useless.                         *
!                                                               *
!  Written by Andy Nelson                                       *
!****************************************************************
#ifndef SPH
      eostate module: SPH proprocessor flag not set
#endif
      use dimens
      use miscvars
      use sphvars
      use constants
      use perform

      implicit none

      private
      public eos

      contains
!--------------------------------------------------------------------
      subroutine eos(npart,npart_sph,npoim,rho,material,
     &               posmas,posmaspm,u)
!
!  compute the pressure and sound speed
!  according to various gas equations of state
!
      integer, intent(in):: npart,npart_sph,npoim
      integer, intent(in):: material(npart_sph)
      double precision, intent(in   ):: posmas  (ndim1,npart)
      double precision, intent(in   ):: posmaspm(ndim1,npoim)
      double precision, intent(in   ):: rho(npart_sph)
      double precision, intent(inout):: u  (npart_sph) !inout for iso

      integer i
      double precision cosmo_rho
      double precision gama1
      double precision tme1,tme2
      character(len=20), parameter:: where =('eos')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      cosmo_rho=one
      if(iexpand.eq.1)cosmo_rho=one/ascale**ndim

      gama1=gamma-ONE

      select case(ieos)
      
      case(id_idealgas)
!     
! ideal gas equation of state
!
!$omp parallel do default(none) private(i)
!$omp& firstprivate(gama1,gamma,cosmo_rho)
!$omp& shared(npart_sph,rho,u,thermo)
         do i=1,npart_sph 
            thermo(i_pres ,i)=gama1*u(i)*rho(i)*cosmo_rho
            thermo(i_sound,i)=sqrt(gamma*thermo(i_pres,i)/
     &                                    (rho(i)*cosmo_rho))
         enddo

      case(id_lagrgiso,id_euleriso)
!
! Isothermal
!
         if(gamma.ne.one)stop'Isothermal gas EOS needs gamma=1'

         if(ieos.eq.id_euleriso)then
            call eulerianiso(npart,npart_sph,npoim,posmaspm,posmas,u)
         endif

!$omp parallel do default(none) private(i)
!$omp& firstprivate(cosmo_rho)
!$omp& shared(npart_sph,rho,u,thermo)
         do i=1,npart_sph
            thermo(i_pres ,i)=u(i)*rho(i)*cosmo_rho
            thermo(i_sound,i)=sqrt(thermo(i_pres,i)/
     &                              (rho(i)*cosmo_rho))
         enddo

      case(id_polyeos)
!
! Polytropic eos
!
!$omp parallel do default(none) private(i)
!$omp& firstprivate(eosK,gama1,gamma,cosmo_rho)
!$omp& shared(npart_sph,rho,u,thermo)
         do i=1,npart_sph 
            u(i) = eosK*(rho(i)*cosmo_rho)**gama1
            thermo(i_pres ,i)= u(i)*rho(i)*cosmo_rho
            thermo(i_sound,i)=sqrt(gamma*thermo(i_pres,i)/
     &                                    (rho(i)*cosmo_rho))
         enddo

      case(id_entreos)
!
! Entropy is state variable
!
!$omp parallel do default(none) private(i)
!$omp& firstprivate(gama1,gamma,cosmo_rho)
!$omp& shared(npart_sph,rho,u,thermo)
         do i=1,npart_sph 
            thermo(i_pres ,i)= u(i)*(rho(i)*cosmo_rho)**gamma
            thermo(i_sound,i)=sqrt(gamma*thermo(i_pres,i)/
     &                                    (rho(i)*cosmo_rho))
         enddo

      case(id_bateeos)

!$omp parallel do default(none) private(i)
!$omp& shared(npart_sph,rho,u,thermo)
         do i=1,npart_sph
            thermo(i_pres ,i)=twothd*u(i)*rho(i)
            thermo(i_sound,i)=sqrt(thermo(i_pres,i)/rho(i))
         enddo

      case default
         stop'No known eos selected'
      end select

      if(idbg.ge.3)print *,'done ',where
      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_eos)=timing(itme_eos) + tme2-tme1
         ncalls(itme_eos)=ncalls(itme_eos) + 1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine eos
!--------------------------------------------------------------------
      subroutine eulerianiso(npart,npart_sph,npoim,posmaspm,posmas,u)

      integer, intent(in):: npart,npart_sph,npoim
      double precision, intent(in ):: posmas  (ndim1,npart)
      double precision, intent(in ):: posmaspm(ndim1,npoim)
      double precision, intent(out):: u(npart_sph)

      integer i
      double precision rcore2i,uexpon2,dx,dy,dz,ri2,ratio2,t1
      character(len=20), parameter:: where =('eulerianiso')
      if(idbg.ge.2)print *,where

      if(gamma.ne.one)stop'Eulerian Isothermal EOS needs gamma=1'

      rcore2i=one/rcore**2
      uexpon2=-half*uexpon
!$omp parallel do default(none) private(i,dx,dy,dz,ri2,ratio2,t1),
!$omp& firstprivate(uconst,uexpon2,rcore2i)
!$omp&  shared(npart_sph,posmaspm,posmas,u)
      do i=1,npart_sph
         dx = posmaspm(1,1) - posmas(1,i)
         dy = posmaspm(2,1) - posmas(2,i)
         ri2= dx*dx + dy*dy
#if NDIM==3
         dz = posmaspm(3,1) - posmas(3,i)
         ri2= ri2 + dz*dz
#endif
         ratio2=ri2*rcore2i
         t1=one+ratio2
         u(i)=uconst*t1**uexpon2
      enddo

      return
      end subroutine eulerianiso
!--------------------------------------------------------------------
      end module eostate
