      module errorcrit
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!  This file contains the VINE module to calculate error        *
!  criteria for the integration forward in time, for use in     *
!  setting the next timestep for each particle. The only        *
!  routine that needs to be externally visible is errors,       *
!  which returns the array dtrat, which contains the ratio      *
!  of the old and new timesteps.                                *
!                                                               *
! All error terms that need to be computed are computed         *
! here. If you need to add others, make a subroutine and        *
! call it from the errors subroutine. This is the place. If     *
! you want to delete (comment out) some that are already here,  *
! that is easy too.                                             *
!                                                               *
! Presently available error routines are:                       *
!                                                               *
!  rk_err      : RKF errors (RK only)                           *
!  vf_1step_ts : Velocity and Acceleration limitation (1step)   *
!  vf_2step_ts : Velocity and Acceleration limitation (2step)   *
!  f_1step_ts  : Acceleration limitation only (1step)           *
!  f_2step_ts  : Acceleration limitation only (2step)           *
!  cosmo_ts    : Cosmology timestep limitation (COSMO only)     *
!  cour_ts     : Courant condition (SPH only)                   *
!  hchange     : Limit change in smoothing length (SPH only)    *
!  dchange     : Limit change in density  (SPH only--and not    *
!                    currently functional-use hchange instead)  *
!  stability   : Stabililty against numerically generated       *
!                gravitational collapse (SPH only)              *
!                                                               *
!  Written by Andy Nelson, Markus Wetzstein and Thorsten Naab   *
!****************************************************************
      use dimens
      use partvars
      use integvars
      use miscvars, only: i_hts,iexpand,igrav,isoft
      use miscvars, only: ftol,vtol,ftolpm,vtolpm
      use miscvars, only: eps,dhmax,hmax,hmin
      use constants 
      use stubs

      implicit none

      private
      public errors

!
! The VAMULT value is a softener that causes the error criterion
!  dt_new = vtol*vmag/amag
! to be limited to occasions when the other v and a error criteria
! are no more than VAMULT larger. This avoids (I hope) the possibility
! that the particle/point mass is simply sitting at rest and therefore
! has no velocity.
!
      double precision, parameter :: VAMULT=1d2


      contains
!---------------------------------------------------------------
      subroutine errors(i12step,
     &                  nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                  iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                  itstepbin ,itstepbinpm, dtincr,
     &                  dtrat,dtratpm)
!*************************************************************
!  Driver for computing integration errors for particles     *
!  that have been advanced. Computes dtrat, the ratio        *
!  between the current timestep and what should be used      *
!  in the next timestep.                                     *
!                                                            *
!  If i12step.eq.1, then the caller of this routine          *
!  is expected to automatically replace nactivef* with       *
!  nactive*, as required (remember that by definition        *
!  nativef* are 0 in one step integrators--this replacement  *
!  sidesteps that difficulty)                                *
!                                                            *
!*************************************************************
      use perform
#ifdef SPH
      use sph, only:cour_ts
#endif
#if RK 
      use integ, only:rk_err
#endif
#if COSMO 
      use cosmo, only:cosmo_ts
#endif

      integer, intent(in):: i12step
      integer, intent(in):: nactivef_s, iactive_s(nactivef_s)
      integer, intent(in):: nactivef_n, iactive_n(nactivef_n)
      integer, intent(in):: nactivef  , iactive  (nactivef  )
      integer, intent(in):: nactivefpm, iactivepm(nactivefpm)
      integer, intent(in):: itstepbin(idim),itstepbinpm(ipmdim)
      double precision, intent(in   ):: dtincr
      double precision, intent(inout):: dtrat(idim),dtratpm(ipmdim)

      integer i,j

      double precision tme1,tme2
      character(len=20), parameter:: where =('errors')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

!$omp parallel do if(nactivef.gt.1024) default(none) private(i,j)
!$omp&   shared(nactivef,iactive,dtrat)
      do j=1,nactivef
         i=iactive(j)
         dtrat(i)=huge(1d0)
      enddo
      do j=1,nactivefpm
         i=iactivepm(j)
         dtratpm(i)=huge(1d0)
      enddo
!
!  It should be trivial to add other error conditions 
!  for additional one or two step integrators to this
!  subroutine, possibly using a preprocessor option 
!  as with the RK option here:
!
#ifdef RK
!
! RKF1(2) error
!
      call rk_err(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &            iactive_s ,iactive_n ,iactive ,iactivepm ,
     &            itstepbin,itstepbinpm,dtincr,dtrat,dtratpm)

#endif
!
! Accleration and Velocity variation limitation.
! Each uses a different error parameter, so setting one to
! a huge value effectively turns off that error criterion.
!
      select case(i12step)
      case(1)
         call vf_1step_ts(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                    iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                    itstepbin,itstepbinpm,dtincr,dtrat,dtratpm)
      case(2)
         call vf_2step_ts(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                    iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                    itstepbin,itstepbinpm,dtincr,dtrat,dtratpm)
      case default
         stop 'errors:Don''t know integrator stepping'
      end select 
!
! CFL criterion
!
      if(nactivef_s.gt.0)then
         call cour_ts(nactivef_s,iactive_s,itstepbin,dtincr,
     &                h,dumalf,dtrat)
      endif
!
! Smoothing length variation limitation 
!
      if(i_hts.eq.1)then
         call hchange(nactivef_s,iactive_s,itstepbin,dtincr,dtrat)
      endif
!
! Cosmology time step limitation
!
      if(iexpand.eq.1)then
         call cosmo_ts(nactivef,nactivefpm,iactive,iactivepm,
     &                 itstepbin,itstepbinpm,dtincr,dtrat,dtratpm)
      endif
!
! Numerical stability against artificial Jeans or Toomre collapse
!
!      if(igrav.eq.1)then
!      call stable(nactivef_s,iactive_s,npart,npart_sph,dumposmas,rho)  
!      endif

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_error)=timing(itme_error) + (tme2-tme1)
         ncalls(itme_error)=ncalls(itme_error) + 1
         timing(itme_intgratr)=timing(itme_intgratr) + (tme2-tme1)
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

! Move these up to replace the combined versions above if
! you are really fanatical about speed and only want one of the 
! criteria. Note the velocity only ones don't actually exist.
!
! Acceleration variation limitation (alone).
!
!      select case(i12step)
!      case(1)
!          call f_1step_ts(nactivef_s,nactivef_n,nactivef,nactivefpm,
!     &                    iactive_s ,iactive_n ,iactive ,iactivepm ,
!     &                    itstepbin,itstepbinpm,dtincr,dtrat,dtratpm)
!      case(2)
!          call f_2step_ts(nactivef_s,nactivef_n,nactivef,nactivefpm,
!     &                    iactive_s ,iactive_n ,iactive ,iactivepm ,
!     &                    itstepbin,itstepbinpm,dtincr,dtrat,dtratpm)
!      case default
!         stop 'errors:Don''t know integrator stepping'
!      end select 
!
! Velocity variation limitation (alone).
!
!      select case(i12step)
!      case(1)
!          call v_1step_ts(nactivef_s,nactivef_n,nactivef,nactivefpm,
!     &                    iactive_s ,iactive_n ,iactive ,iactivepm ,
!     &                    itstepbin,itstepbinpm,dtincr,dtrat,dtratpm)
!      case(2)
!          call v_2step_ts(nactivef_s,nactivef_n,nactivef,nactivefpm,
!     &                    iactive_s ,iactive_n ,iactive ,iactivepm ,
!     &                    itstepbin,itstepbinpm,dtincr,dtrat,dtratpm)
!      case default
!         stop 'errors:Don''t know integrator stepping'
!      end select 

      return
      end subroutine errors
!---------------------------------------------------------------
      subroutine f_1step_ts(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                      iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                      itstepbin, itstepbinpm, 
     &                      dtincr,dtrat,dtratpm)
!********************************************************************
!                                                                   *
! Acceleration variation limitation on timestep-1 step integrator   *
!                                                                   *
!********************************************************************
      integer, intent(in):: nactivef_s, iactive_s(nactivef_s)
      integer, intent(in):: nactivef_n, iactive_n(nactivef_n)
      integer, intent(in):: nactivef  , iactive  (nactivef  )
      integer, intent(in):: nactivefpm, iactivepm(nactivefpm)
      integer, intent(in):: itstepbin(idim),itstepbinpm(ipmdim)
      double precision, intent(in   ):: dtincr
      double precision, intent(inout):: dtrat(idim),dtratpm(ipmdim)

      integer i,j
      double precision accmag,dt_last,dt_new
      character(len=20), parameter:: where =('f_1step_ts')
!
!  statement function for squared acceleration 
!
      double precision acc1sqd,pmacc1sqd
#if NDIM==2
      acc1sqd  (i)=acc1  (1,i)**2 + acc1  (2,i)**2
      pmacc1sqd(i)=acc1pm(1,i)**2 + acc1pm(2,i)**2
#else
      acc1sqd  (i)=acc1  (1,i)**2 + acc1  (2,i)**2 + acc1  (3,i)**2
      pmacc1sqd(i)=acc1pm(1,i)**2 + acc1pm(2,i)**2 + acc1pm(3,i)**2
#endif

      if(idbg.ge.2)print *,where

      select case(isoft)
      case(id_varkern)
!
! use smoothing length as size over which
!  to measure acceleration variation 
!
!$omp parallel do if(nactivef.gt.1024) default(none)
!$omp& firstprivate(ftol)
!$omp& shared(nactivef,iactive,acc1,itstepbin,dtincr)
!$omp& shared(dumh,dtrat)
!$omp& private(i,j,accmag,dt_last,dt_new)
         do j=1,nactivef
            i=iactive(j)
            accmag   = sqrt(acc1sqd(i))
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new   = ftol*sqrt(dumh(i)/accmag)
            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo

      case(id_fixkern)
!
! use individual softening lengths as size over which
!  to measure acceleration variation for nbody particles,
! use min of fixed softening length and sph smoothing length
!  as size over which to measure acceleration for sph particles
!
!$omp parallel do if(nactivef_n.gt.1024) default(none)
!$omp& firstprivate(ftol)
!$omp& shared(nactivef_n,iactive_n,acc1,dtincr,itstepbin,dtrat,dumh)
!$omp& private(i,j,accmag,dt_last,dt_new) 
         do j=1,nactivef_n
            i=iactive_n(j)
            accmag   = sqrt(acc1sqd(i))
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new   = ftol*sqrt(dumh(i)/accmag)
            dtrat(i) = min(dtrat(i),dt_new/dt_last)

         enddo

!$omp parallel do if(nactivef_s.gt.1024) default(none)
!$omp& firstprivate(ftol,eps)
!$omp& shared(nactivef_s,iactive_s,acc1,dtincr,itstepbin)
!$omp& shared(dumh,dtrat)
!$omp& private(i,j,accmag,dt_last,dt_new)
         do j=1,nactivef_s
            i=iactive_s(j)
            accmag   = sqrt(acc1sqd(i))
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new   = ftol*sqrt(min(eps,dumh(i))/accmag)
            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo

      case(id_plummer)
!
! use fixed softening length as size over which
!  to measure acceleration variation for nbody particles,
! use min of fixed softening length and sph smoothing length
!  as size over which to measure acceleration for sph particles
!
!$omp parallel do if(nactivef_n.gt.1024) default(none)
!$omp& firstprivate(ftol,eps)
!$omp& shared(nactivef_n,iactive_n,acc1,dtincr,itstepbin,dtrat)
!$omp& private(i,j,accmag,dt_last,dt_new) 
         do j=1,nactivef_n
            i=iactive_n(j)
            accmag   = sqrt(acc1sqd(i))
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new   = ftol*sqrt(eps/accmag)
            dtrat(i) = min(dtrat(i),dt_new/dt_last)

         enddo

!$omp parallel do if(nactivef_s.gt.1024) default(none)
!$omp& firstprivate(ftol,eps)
!$omp& shared(nactivef_s,iactive_s,acc1,dtincr,itstepbin)
!$omp& shared(dumh,dtrat)
!$omp& private(i,j,accmag,dt_last,dt_new)
         do j=1,nactivef_s
            i=iactive_s(j)
            accmag   = sqrt(acc1sqd(i))
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new   = ftol*sqrt(min(eps,dumh(i))/accmag)
            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo

      case default

         stop 'Unknown softening option in f_1step_ts'

      end select
!
! Point masses don't have a lot of the above issues.
! Just use size of pm
!
      do j=1,nactivefpm
         i=iactivepm(j)
         accmag    = sqrt(pmacc1sqd(i))
         dt_last   = dtincr*dble(itstepbinpm(i))
         dt_new    = ftolpm*sqrt(hpm(i)/accmag)
         dtratpm(i)= min(dtratpm(i),dt_new/dt_last)
      enddo

      return
      end subroutine f_1step_ts
!---------------------------------------------------------------
      subroutine f_2step_ts(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                      iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                      itstepbin ,itstepbinpm,
     &                      dtincr  ,dtrat, dtratpm)
!********************************************************************
!                                                                   *
! Acceleration variation limitation on timestep-2 step integrator   *
!                                                                   *
!********************************************************************
      integer, intent(in):: nactivef_s, iactive_s(nactivef_s)
      integer, intent(in):: nactivef_n, iactive_n(nactivef_n)
      integer, intent(in):: nactivef  , iactive  (nactivef  )
      integer, intent(in):: nactivefpm, iactivepm(nactivefpm)
      integer, intent(in):: itstepbin(idim),itstepbinpm(ipmdim)
      double precision, intent(in   ):: dtincr
      double precision, intent(inout):: dtrat(idim),dtratpm(ipmdim)

      integer i,j
      double precision accmag,dt_last,dt_new,sizemin
      character(len=20), parameter:: where =('f_2step_ts')
!
!  statement function for squared acceleration, at
!  at the beginning and end of the timestep 
!
      double precision acc1sqd,acc2sqd,pmacc1sqd,pmacc2sqd
#if NDIM==2
      acc1sqd  (i)=acc1  (1,i)**2 + acc1  (2,i)**2
      acc2sqd  (i)=acc2  (1,i)**2 + acc2  (2,i)**2
      pmacc1sqd(i)=acc1pm(1,i)**2 + acc1pm(2,i)**2
      pmacc2sqd(i)=acc2pm(1,i)**2 + acc2pm(2,i)**2
#else
      acc1sqd  (i)=acc1  (1,i)**2 + acc1  (2,i)**2 + acc1  (3,i)**2
      acc2sqd  (i)=acc2  (1,i)**2 + acc2  (2,i)**2 + acc2  (3,i)**2
      pmacc1sqd(i)=acc1pm(1,i)**2 + acc1pm(2,i)**2 + acc1pm(3,i)**2
      pmacc2sqd(i)=acc2pm(1,i)**2 + acc2pm(2,i)**2 + acc2pm(3,i)**2
#endif

      if(idbg.ge.2)print *,where


      select case(isoft)
      case(id_varkern)
!
! use smoothing length as size over which 
!  to measure acceleration variation 
!
!$omp parallel do if(nactivef.gt.1024) default(none)
!$omp& firstprivate(ftol)
!$omp& shared(nactivef,iactive,acc1,acc2,dtincr,itstepbin)
!$omp& shared(dumh,dtrat)
!$omp& private(i,j,accmag,dt_last,dt_new)
         do j=1,nactivef
            i=iactive(j)
            accmag   = sqrt(max(acc1sqd(i), acc2sqd(i)))+small
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new   = ftol*sqrt(dumh(i)/accmag)
            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo

      case(id_fixkern)
!
! use individual softening lengths as size over which
!  to measure acceleration variation for nbody particles,
! use min of fixed softening length and sph smoothing length
!  as size over which to measure acceleration for sph particles
!
!$omp parallel do if(nactivef_n.gt.1024) default(none)
!$omp& firstprivate(ftol,eps)
!$omp& shared(nactivef_n,iactive_n,acc1,acc2,dtincr,itstepbin)
!$omp& shared(dtrat,dumh)
!$omp&    private(i,j,accmag,dt_last,dt_new)
         do j=1,nactivef_n
            i=iactive_n(j)
            accmag   = sqrt(max(acc1sqd(i), acc2sqd(i)))+small
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new   = ftol*sqrt(dumh(i)/accmag)
            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo
!$omp parallel do if(nactivef_s.gt.1024) default(none)
!$omp& firstprivate(ftol,eps)
!$omp& shared(nactivef_s,iactive_s,acc1,acc2,dumh)
!$omp& shared(dtincr,itstepbin,dtrat)
!$omp& private(i,j,accmag,dt_last,dt_new,sizemin)
         do j=1,nactivef_s
            i=iactive_s(j)
            accmag   = sqrt(max(acc1sqd(i), acc2sqd(i)))+small
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new   = ftol*sqrt(min(eps,dumh(i))/accmag)
            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo

      case(id_plummer)
!
! use fixed softening length as size over which
!  to measure acceleration variation for nbody particles,
! use min of fixed softening length and sph smoothing length
!  as size over which to measure acceleration for sph particles
!
!$omp parallel do if(nactivef_n.gt.1024) default(none)
!$omp& firstprivate(ftol,eps)
!$omp& shared(nactivef_n,iactive_n,acc1,acc2,dtincr,itstepbin,dtrat)
!$omp&    private(i,j,accmag,dt_last,dt_new)
         do j=1,nactivef_n
            i=iactive_n(j)
            accmag   = sqrt(max(acc1sqd(i), acc2sqd(i)))+small
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new   = ftol*sqrt(eps/accmag)
            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo
!$omp parallel do if(nactivef_s.gt.1024) default(none)
!$omp& firstprivate(ftol,eps)
!$omp& shared(nactivef_s,iactive_s,acc1,acc2,dumh)
!$omp& shared(dtincr,itstepbin,dtrat)
!$omp& private(i,j,accmag,dt_last,dt_new)
         do j=1,nactivef_s
            i=iactive_s(j)
            accmag   = sqrt(max(acc1sqd(i), acc2sqd(i)))+small
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new   = ftol*sqrt(min(eps,dumh(i))/accmag)
            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo

      case default

         stop 'Unknown softening option in f_2step_ts'

      end select
!
! Point masses don't have a lot of the above issues.
! Just use size of pm
!
      do j=1,nactivefpm
         i=iactivepm(j)
         accmag    =sqrt(pmacc1sqd(i))+small
         dt_last   =dtincr*dble(itstepbinpm(i))
         dt_new    =ftolpm*sqrt(hpm(i)/accmag)
         dtratpm(i)=min(dtratpm(i),dt_new/dt_last)
      enddo

      return
      end subroutine f_2step_ts
!---------------------------------------------------------------
      subroutine vf_1step_ts(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                       iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                       itstepbin ,itstepbinpm,
     &                       dtincr   ,dtrat, dtratpm)
!********************************************************************
!                                                                   *
! Acceleration and velocity variation limitation on timestep        *
! 1 step integrator                                                 *
!                                                                   *
! Note that for the dt_new3 criterion below, we require that        *
! the dt value is not less than VAMULT smaller than any of the      *
! other dt_newX values. This avoids (I hope) the possibility that   *
! the particle/point mass is simply sitting at rest and therefore   *
! has no velocity. -AFN                                             *
!                                                                   *
!********************************************************************
      integer, intent(in):: nactivef_s, iactive_s(nactivef_s)
      integer, intent(in):: nactivef_n, iactive_n(nactivef_n)
      integer, intent(in):: nactivef  , iactive  (nactivef  )
      integer, intent(in):: nactivefpm, iactivepm(nactivefpm)
      integer, intent(in):: itstepbin(idim),itstepbinpm(ipmdim)
      double precision, intent(in   ):: dtincr
      double precision, intent(inout):: dtrat(idim),dtratpm(ipmdim)

      integer i,j
      double precision dt_new,dt_new1,dt_new2,dt_new3,dt_last
      double precision accmag,velmag,sizemin
      character(len=20), parameter:: where =('vf_1step_ts')
!
!  statement function for squared acceleration and velocity
!  at the end of the timestep
!
      double precision acc1sqd,velsqd,pmacc1sqd,pmvelsqd
#if NDIM==2
      acc1sqd  (i)=acc1  (1,i)**2 + acc1  (2,i)**2
      velsqd   (i)=dumvel(1,i)**2 + dumvel(2,i)**2
      pmacc1sqd(i)=acc1pm(1,i)**2 + acc1pm(2,i)**2
      pmvelsqd (i)=velpm (1,i)**2 + velpm (2,i)**2
#else
      acc1sqd  (i)=acc1  (1,i)**2 + acc1  (2,i)**2 + acc1  (3,i)**2
      velsqd   (i)=dumvel(1,i)**2 + dumvel(2,i)**2 + dumvel(3,i)**2
      pmacc1sqd(i)=acc1pm(1,i)**2 + acc1pm(2,i)**2 + acc1pm(3,i)**2
      pmvelsqd (i)=velpm (1,i)**2 + velpm (2,i)**2 + velpm (3,i)**2
#endif

      if(idbg.ge.2)print *,where

      select case(isoft)
      case(id_varkern)
!
! use updated smoothing length as size over which
!  to measure variation 
!
!$omp parallel do if(nactivef.gt.1024) default(none)
!$omp& firstprivate(ftol,vtol)
!$omp& shared(nactivef,iactive,acc1,dtincr,itstepbin)
!$omp& shared(dumh,dumvel,dtrat) private(i,j,accmag,velmag)
!$omp& private(dt_new,dt_new1,dt_new2,dt_new3,dt_last)
         do j=1,nactivef
            i=iactive(j)
            accmag  = sqrt(acc1sqd(i))+small
            velmag  = sqrt(velsqd (i))+small
            dt_last = dtincr*dble(itstepbin(i))
            dt_new1 = vtol*dumh(i)/velmag
            dt_new2 = ftol*sqrt(dumh(i)/accmag)
            dt_new  = min(dt_new1,dt_new2)

            dt_new3 = vtol*velmag/accmag
            if(dt_new.lt.VAMULT*dt_new3)dt_new=min(dt_new,dt_new3)

            dtrat(i)= min(dtrat(i),dt_new/dt_last)
         enddo

      case(id_fixkern)
!
! use individual softening lengths as size over which
!  to measure acceleration variation for nbody particles,
! use min of fixed softening length and sph smoothing length
!  as size over which to measure acceleration for sph particles
!
!$omp parallel do if(nactivef_n.gt.1024) default(none)
!$omp& firstprivate(ftol,vtol)
!$omp& shared(nactivef_n,iactive_n,acc1,dtincr,itstepbin)
!$omp& shared(dumvel,dtrat,dumh) private(i,j,accmag,velmag)
!$omp& private(dt_last,dt_new,dt_new1,dt_new2,dt_new3) 
         do j=1,nactivef_n
            i=iactive_n(j)
            accmag  = sqrt(acc1sqd(i))+small
            velmag  = sqrt(velsqd (i))+small
            dt_last = dtincr*dble(itstepbin(i))
            dt_new1 = vtol*dumh(i)/velmag
            dt_new2 = ftol*sqrt(dumh(i)/accmag)
            dt_new  = min(dt_new1,dt_new2)

            dt_new3 = vtol*velmag/accmag
            if(dt_new.lt.VAMULT*dt_new3)dt_new=min(dt_new,dt_new3)

            dtrat(i)= min(dtrat(i),dt_new/dt_last)
         enddo

!$omp parallel do if(nactivef_s.gt.1024) default(none)
!$omp& firstprivate(ftol,vtol,eps)
!$omp& shared(nactivef_s,iactive_s,acc1,dtincr,itstepbin)
!$omp& shared(dumh,dumvel,dtrat) private(i,j,accmag,velmag)
!$omp& private(dt_last,dt_new,dt_new1,dt_new2,dt_new3,sizemin)
         do j=1,nactivef_s
            i=iactive_s(j)
            accmag  = sqrt(acc1sqd(i))+small
            velmag  = sqrt(velsqd (i))+small
            dt_last = dtincr*dble(itstepbin(i))
            sizemin = min(dumh(i),eps)
            dt_new1 = vtol*sizemin/velmag
            dt_new2 = ftol*sqrt(sizemin/accmag)
            dt_new  = min(dt_new1,dt_new2)

            dt_new3 = vtol*velmag/accmag
            if(dt_new.lt.VAMULT*dt_new3)dt_new=min(dt_new,dt_new3)

            dtrat(i)= min(dtrat(i),dt_new/dt_last)
         enddo

      case(id_plummer)
!
! use fixed softening length as size over which
!  to measure acceleration variation for nbody particles,
! use min of fixed softening length and sph smoothing length
!  as size over which to measure acceleration for sph particles
!
!$omp parallel do if(nactivef_n.gt.1024) default(none)
!$omp& firstprivate(ftol,vtol,eps)
!$omp& shared(nactivef_n,iactive_n,acc1,dtincr,itstepbin)
!$omp& shared(dumvel,dtrat) private(i,j,accmag,velmag)
!$omp& private(dt_last,dt_new,dt_new1,dt_new2,dt_new3)
         do j=1,nactivef_n
            i=iactive_n(j)
            accmag  = sqrt(acc1sqd(i))+small
            velmag  = sqrt(velsqd (i))+small
            dt_last = dtincr*dble(itstepbin(i))
            dt_new1 = vtol*eps/velmag
            dt_new2 = ftol*sqrt(eps/accmag)
            dt_new  = min(dt_new1,dt_new2)

            dt_new3 = vtol*velmag/accmag
            if(dt_new.lt.VAMULT*dt_new3)dt_new=min(dt_new,dt_new3)

            dtrat(i)= min(dtrat(i),dt_new/dt_last)
         enddo

!$omp parallel do if(nactivef_s.gt.1024) default(none)
!$omp& firstprivate(ftol,vtol,eps)
!$omp& shared(nactivef_s,iactive_s,acc1,dtincr,itstepbin)
!$omp& shared(dumh,dumvel,dtrat) private(i,j,accmag,velmag)
!$omp& private(dt_last,dt_new,dt_new1,dt_new2,dt_new3,sizemin)
         do j=1,nactivef_s
            i=iactive_s(j)
            accmag  = sqrt(acc1sqd(i))+small
            velmag  = sqrt(velsqd (i))+small
            dt_last = dtincr*dble(itstepbin(i))
            sizemin = min(eps,dumh(i))
            dt_new1 = vtol*sizemin/velmag
            dt_new2 = ftol*sqrt(sizemin/accmag)
            dt_new  = min(dt_new1,dt_new2)

            dt_new3 = vtol*velmag/accmag
            if(dt_new.lt.VAMULT*dt_new3)dt_new=min(dt_new,dt_new3)

            dtrat(i)= min(dtrat(i),dt_new/dt_last)
         enddo

      case default
         stop 'Unknown softening option in vf_1step_ts' 

      end select
!
! Point masses don't have a lot of the above issues.
! Just use size of pm
!
      do j=1,nactivefpm
         i=iactivepm(j)
         accmag    =sqrt(pmacc1sqd(i))+small
         velmag    =sqrt(pmvelsqd (i))+small
         dt_last   =dtincr*dble(itstepbinpm(i))
         dt_new1   =vtolpm*hpm(i)/velmag
         dt_new2   =ftolpm*sqrt(hpm(i)/accmag)
         dt_new = min(dt_new1,dt_new2)

         dt_new3   =vtolpm*velmag/accmag
         if(dt_new.lt.VAMULT*dt_new3)dt_new=min(dt_new,dt_new3)

         dtratpm(i)=min(dtratpm(i),dt_new/dt_last)
      enddo

      return
      end subroutine vf_1step_ts
!---------------------------------------------------------------
      subroutine vf_2step_ts(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                       iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                       itstepbin ,itstepbinpm,
     &                       dtincr  ,dtrat, dtratpm)
!********************************************************************
!                                                                   *
! Acceleration and velocity variation limitation on timestep        *
! 2 step integrator                                                 *
!                                                                   *
!********************************************************************
      integer, intent(in):: nactivef_s, iactive_s(nactivef_s)
      integer, intent(in):: nactivef_n, iactive_n(nactivef_n)
      integer, intent(in):: nactivef  , iactive  (nactivef  )
      integer, intent(in):: nactivefpm, iactivepm(nactivefpm)
      integer, intent(in):: itstepbin(idim),itstepbinpm(ipmdim)
      double precision, intent(in   ):: dtincr
      double precision, intent(inout):: dtrat(idim),dtratpm(ipmdim)

      integer i,j
      double precision accmag,velmag,dt_last,sizemin
      double precision dt_new,dt_new1,dt_new2,dt_new3
      character(len=20), parameter:: where =('vf_2step_ts')
!
!  statement function for squared acceleration and velocity, at
!  at the beginning and end of the timestep 
!
      double precision   acc1sqd,  acc2sqd,  vel1sqd,  vel2sqd
      double precision pmacc1sqd,pmacc2sqd,pmvel1sqd,pmvel2sqd
#if NDIM==2
      acc1sqd  (i)=acc1    (1,i)**2 + acc1    (2,i)**2
      acc2sqd  (i)=acc2    (1,i)**2 + acc2    (2,i)**2
      vel1sqd  (i)=vel     (1,i)**2 + vel     (2,i)**2
      vel2sqd  (i)=dumvel  (1,i)**2 + dumvel  (2,i)**2
      pmacc1sqd(i)=acc1pm  (1,i)**2 + acc1pm  (2,i)**2
      pmacc2sqd(i)=acc2pm  (1,i)**2 + acc2pm  (2,i)**2
      pmvel1sqd(i)=velpm   (1,i)**2 + velpm   (2,i)**2
      pmvel2sqd(i)=dumvelpm(1,i)**2 + dumvelpm(2,i)**2

#else
      acc1sqd  (i)=acc1    (1,i)**2 +acc1    (2,i)**2 + acc1    (3,i)**2
      acc2sqd  (i)=acc2    (1,i)**2 +acc2    (2,i)**2 + acc2    (3,i)**2
      vel1sqd  (i)=vel     (1,i)**2 +vel     (2,i)**2 + vel     (3,i)**2
      vel2sqd  (i)=dumvel  (1,i)**2 +dumvel  (2,i)**2 + dumvel  (3,i)**2
      pmacc1sqd(i)=acc1pm  (1,i)**2 +acc1pm  (2,i)**2 + acc1pm  (3,i)**2
      pmacc2sqd(i)=acc2pm  (1,i)**2 +acc2pm  (2,i)**2 + acc2pm  (3,i)**2
      pmvel1sqd(i)=velpm   (1,i)**2 +velpm   (2,i)**2 + velpm   (3,i)**2
      pmvel2sqd(i)=dumvelpm(1,i)**2 +dumvelpm(2,i)**2 + dumvelpm(3,i)**2
#endif

      if(idbg.ge.2)print *,where


      select case(isoft)
      case(id_varkern)
!
! use individual smoothing/softening lengths as size over which 
!  to measure acceleration and velocity variation 
!
!$omp parallel do if(nactivef.gt.1024) default(none)
!$omp& firstprivate(ftol,vtol)
!$omp& shared(nactivef,iactive,acc1,acc2,dtincr,itstepbin)
!$omp& shared(dumh,vel,dumvel,dtrat) private(i,j,accmag,velmag)
!$omp& private(dt_new,dt_new1,dt_new2,dt_new3,dt_last)
         do j=1,nactivef
            i=iactive(j)
            accmag   = sqrt(max(acc1sqd(i), acc2sqd(i)))+small
            velmag   = sqrt(max(vel1sqd(i), vel2sqd(i)))+small
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new1  = vtol*dumh(i)/velmag
            dt_new2  = ftol*sqrt(dumh(i)/accmag)
            dt_new   = min(dt_new1,dt_new2)

            dt_new3  = vtol*velmag/accmag
            if(dt_new.lt.VAMULT*dt_new3)dt_new=min(dt_new,dt_new3)

            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo

      case(id_fixkern)
!
! use individual softening lengths as size over which
!  to measure acceleration variation for nbody particles,
! use min of fixed softening length and sph smoothing length
!  as size over which to measure acceleration for sph particles
!
!$omp parallel do if(nactivef_n.gt.1024) default(none)
!$omp& firstprivate(ftol,vtol,eps)
!$omp& shared(nactivef_n,iactive_n,acc1,acc2,dtincr,itstepbin)
!$omp& shared(vel,dumvel,dtrat,dumh) private(i,j,accmag,velmag)
!$omp& private(dt_new,dt_new1,dt_new2,dt_new3,dt_last)
         do j=1,nactivef_n
            i=iactive_n(j)
            accmag   = sqrt(max(acc1sqd(i), acc2sqd(i)))+small
            velmag   = sqrt(max(vel1sqd(i), vel2sqd(i)))+small
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new1  = vtol*dumh(i)/velmag
            dt_new2  = ftol*sqrt(dumh(i)/accmag)

            dt_new3  = vtol*velmag/accmag
            if(dt_new.lt.VAMULT*dt_new3)dt_new=min(dt_new,dt_new3)

            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo
!$omp parallel do if(nactivef_s.gt.1024) default(none)
!$omp& firstprivate(ftol,vtol,eps)
!$omp& shared(nactivef_s,iactive_s,acc1,acc2,dtincr,itstepbin)
!$omp& shared(dumh,vel,dumvel,dtrat) private(i,j,accmag,velmag)
!$omp& private(dt_new,dt_new1,dt_new2,dt_new3,dt_last,sizemin)
         do j=1,nactivef_s
            i=iactive_s(j)
            accmag   = sqrt(max(acc1sqd(i), acc2sqd(i)))+small
            velmag   = sqrt(max(vel1sqd(i), vel2sqd(i)))+small
            dt_last  = dtincr*dble(itstepbin(i))
            sizemin  = min(dumh(i),eps)
            dt_new1  = vtol*sizemin/velmag
            dt_new2  = ftol*sqrt(sizemin/accmag)
            dt_new   = min(dt_new1,dt_new2)

            dt_new3  = vtol*velmag/accmag
            if(dt_new.lt.VAMULT*dt_new3)dt_new=min(dt_new,dt_new3)

            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo

      case(id_plummer)
!
! use fixed softening length as size over which
!  to measure acceleration variation for nbody particles,
! use min of fixed softening length and sph smoothing length
!  as size over which to measure acceleration for sph particles
!
!$omp parallel do if(nactivef_n.gt.1024) default(none)
!$omp& firstprivate(ftol,vtol,eps)
!$omp& shared(nactivef_n,iactive_n,acc1,acc2,dtincr,itstepbin)
!$omp& shared(vel,dumvel,dtrat) private(i,j,accmag,velmag)
!$omp& private(dt_last,dt_new,dt_new1,dt_new2,dt_new3)
         do j=1,nactivef_n
            i=iactive_n(j)
            accmag   = sqrt(max(acc1sqd(i), acc2sqd(i)))+small
            velmag   = sqrt(max(vel1sqd(i), vel2sqd(i)))+small
            dt_last  = dtincr*dble(itstepbin(i))
            dt_new1  = vtol*eps/velmag
            dt_new2  = ftol*sqrt(eps/accmag)
            dt_new   = min(dt_new1,dt_new2)

            dt_new3  = vtol*velmag/accmag
            if(dt_new.lt.VAMULT*dt_new3)dt_new=min(dt_new,dt_new3)

            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo
!$omp parallel do if(nactivef_s.gt.1024) default(none)
!$omp& firstprivate(ftol,vtol,eps)
!$omp& shared(nactivef_s,iactive_s,acc1,acc2,dtincr,itstepbin)
!$omp& shared(dumh,vel,dumvel,dtrat) private(i,j,accmag,velmag)
!$omp& private(dt_new,dt_new1,dt_new2,dt_new3,dt_last,sizemin)
         do j=1,nactivef_s
            i=iactive_s(j)
            accmag   = sqrt(max(acc1sqd(i), acc2sqd(i)))+small
            velmag   = sqrt(max(vel1sqd(i), vel2sqd(i)))+small
            dt_last  = dtincr*dble(itstepbin(i))
            sizemin  = min(dumh(i),eps)
            dt_new1  = vtol*sizemin/velmag
            dt_new2  = ftol*sqrt(sizemin/accmag)
            dt_new   = min(dt_new1,dt_new2)

            dt_new3  = vtol*velmag/accmag
            if(dt_new.lt.VAMULT*dt_new3)dt_new=min(dt_new,dt_new3)

            dtrat(i) = min(dtrat(i),dt_new/dt_last)
         enddo

      case default

         stop 'Unknown softening option in vf_2step_ts'
      end select
!
! Point masses don't have a lot of the above issues.
! Just use size of pm
!
      do j=1,nactivefpm
         i=iactivepm(j)
         accmag    = sqrt(max(pmacc1sqd(i),pmacc2sqd(i)))+small
         velmag    = sqrt(max(pmvel1sqd(i),pmvel2sqd(i)))+small
         dt_last   = dtincr*dble(itstepbinpm(i))
         dt_new1   = vtolpm*hpm(i)/velmag
         dt_new2   = ftolpm*sqrt(hpm(i)/accmag)
         dt_new   = min(dt_new1,dt_new2)

         dt_new3   = vtolpm*velmag/accmag
         if(dt_new.lt.VAMULT*dt_new3)dt_new=min(dt_new,dt_new3)

         dtratpm(i)= min(dtratpm(i),dt_new/dt_last)
      enddo

      return
      end subroutine vf_2step_ts
!---------------------------------------------------------------
      subroutine hchange(nactive_s,iactive_s,itstepbin,dtincr,dtrat)
!********************************************************************
!                                                                   *
! Timestep limitation from limiting the smoothing length change     *
! per timestep.                                                     *
!                                                                   *
! We allow a maximum fractional change of dhmax for each particle   *
! in one timestep: In other words, the maximum change in the value  *
! of h(i) is dhmax*h(i). We must limit this condition to particles  *
! that have not hit the limits hmin or hmax, because the hdot       *
! routine artificially resets the dhdt value to zero in this case.  *
!                                                                   *
! Note that the spacing routine below is a fortran intrinsic        *
! that returns the numerical spacing of adjacent reals or doubles   *
! near its argument. We use it here to avoid an equality comparison *
! of real valued quantities to each other. Always a bad thing to    *
! do in a computer.                                                 *
!                                                                   *
! Note that nactive_s could actually be nactivef_s in the           *
!  calling program. This is not a mistake. It is a swindle          *
!  to allow using one or two step integrators transparently.        *
!                                                                   *
!********************************************************************
      integer, intent(in):: nactive_s, iactive_s(nactive_s)
      integer, intent(in):: itstepbin(idim)
      double precision, intent(in   ):: dtincr
      double precision, intent(inout):: dtrat(idim)
      integer i,j
      double precision deltmax,dt_last,dtrat_h_extrap

!$omp parallel do if(nactive_s.gt.1024) default(none)
!$omp&  firstprivate(dhmax,hmax,hmin)
!$omp&    shared(nactive_s,iactive_s,dtincr,itstepbin,dtrat)
!$omp&    shared(dumh,dhdt1,dhdt2)
!$omp&    private(i,j,dtrat_h_extrap,dt_last,deltmax) 
      do j=1,nactive_s
         i=iactive_s(j)


#if defined(RK)
         if( ((dumh(i)-hmin).gt.10d0*(spacing(dumh(i)-hmin)) .and.
     &            (dhdt1(i).gt.zero .or. dhdt2(i).gt.zero))  .or.
     &       ((hmax-dumh(i)).gt.10d0*(spacing(hmax-dumh(i))) .and.
     &            (dhdt1(i).lt.zero .or. dhdt2(i).lt.zero)) )then

            deltmax=dhmax*dumh(i)/(abs(dhdt1(i)+small))
            deltmax=min(deltmax,dhmax*dumh(i)/(abs(dhdt2(i)+small)))
            dt_last=dtincr*dble(itstepbin(i))
            dtrat_h_extrap=deltmax/dt_last
            dtrat(i)=min(dtrat(i),dtrat_h_extrap)
         endif
#else
         if( ((dumh(i)-hmin).gt.10d0*(spacing(dumh(i)-hmin)) .and.
     &            dhdt1(i).gt.zero )  .or.
     &       ((hmax-dumh(i)).gt.10d0*(spacing(hmax-dumh(i))) .and.
     &            dhdt1(i).lt.zero ) )then

            deltmax=dhmax*dumh(i)/(abs(dhdt1(i)+small))
            dt_last=dtincr*dble(itstepbin(i))
            dtrat_h_extrap=deltmax/dt_last
            dtrat(i)=min(dtrat(i),dtrat_h_extrap)
         endif
#endif
      enddo

      return
      end subroutine hchange
!---------------------------------------------------------------
      end module errorcrit
