      module ewald
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!  This file contains the VINE module to do gravity for         *
!  periodic boundaries using ewald summation.                   *
!                                                               *
!  Point masses are not included in the gravity calculation     *
!                                                               *
!  This module uses the FFTW version 3 libraries, which         *
!  are available from www.fftw.org                              *
!                                                               *
!  Written by Markus Wetzstein and Thorsten Naab                *
!****************************************************************
      use dimens
      use miscvars, only: idisk1,boxlims,boxlen,halfboxlen,eps,
     $                    isoft,hmax
      use constants
      use perform
#ifdef RECIPES
      use recipes
#endif

      implicit none
      save

      private
      public init_rand,init_ewald
      public apply_ewald,checkboxlimits
      public boxshift,shift_box_to_original,shift_box_to_current
      
#ifdef FFTW
      include 'INCLUDE/fftw3.f'   !this includes definitions for FFTW
#else
        ewald-module:-FFTW-preprocessor-flag-is-not-set
#endif

      integer, parameter:: ngrid=64            ! grid resolution for
                                               ! box with unit length
      integer, parameter:: n2grid =2*ngrid     ! grid resolution in 
                                               ! simulation box    
      integer, parameter:: ngridhalf=ngrid/2   ! half the grid size
      integer, parameter:: n2gridhalf=n2grid/2 ! half the grid size
#if NDIM==3
      integer, parameter:: ngrid_box =ngrid**3    ! 3d grid size for unit box
      integer, parameter:: n2grid_box=n2grid**3   ! 3d grid size for
                                                  ! simulation box 
#else
      integer, parameter:: ngrid_box =ngrid**2    ! 2d grid size for unit box
      integer, parameter:: n2grid_box=n2grid**2   ! 2d grid size for
                                                  ! simulation box 
#endif

      integer idummyrandom

      double precision randshifttot(3),randshift(3)
!
! use n2grid+2 for the def of the arrays below because they will be fourier
! transformed and the additional +2 are needed because of the output format
! of FFTW (cf. FFTW documentation)
!
      integer*8 fftw_fxcorr_unitm_r2c,fftw_fycorr_unitm_r2c,
     $    fftw_potcorr_unitm_r2c,fftw_fxk_corr_c2r,fftw_fyk_corr_c2r,
     $    fftw_potk_corr_c2r,fftw_massgrid_r2c
#if NDIM==3
      integer*8 fftw_fzcorr_unitm_r2c,fftw_fzk_corr_c2r
      
      double precision fxcorr_unitm(n2grid+2,n2grid,n2grid),
     $                 fycorr_unitm(n2grid+2,n2grid,n2grid),
     $                 fzcorr_unitm(n2grid+2,n2grid,n2grid),
     $                 potcorr_unitm(n2grid+2,n2grid,n2grid),
     $                 fxcorr_now(n2grid+2,n2grid,n2grid),
     $                 fycorr_now(n2grid+2,n2grid,n2grid),
     $                 fzcorr_now(n2grid+2,n2grid,n2grid),
     $                 potcorr_now(n2grid+2,n2grid,n2grid),
     $                 massgrid(n2grid+2,n2grid,n2grid),
     $                 gridbuffer(ngrid,ngrid,ngrid)

      double complex fxk_corr (ngrid+1,n2grid,n2grid),
     $               fyk_corr (ngrid+1,n2grid,n2grid),
     $               fzk_corr (ngrid+1,n2grid,n2grid),
     $               potk_corr(ngrid+1,n2grid,n2grid),
     $               fxk_corr_unitm (ngrid+1,n2grid,n2grid),
     $               fyk_corr_unitm (ngrid+1,n2grid,n2grid),
     $               fzk_corr_unitm (ngrid+1,n2grid,n2grid),
     $               potk_corr_unitm(ngrid+1,n2grid,n2grid),
     $               masskgrid(ngrid+1,n2grid,n2grid)
#else
!
! don't define z components, other arrays just 2d
!
      double precision fxcorr_unitm(n2grid+2,n2grid),
     $                 fycorr_unitm(n2grid+2,n2grid),
     $                 potcorr_unitm(n2grid+2,n2grid),
     $                 fxcorr_now(n2grid+2,n2grid),
     $                 fycorr_now(n2grid+2,n2grid),
     $                 potcorr_now(n2grid+2,n2grid),
     $                 massgrid(n2grid+2,n2grid),
     $                 gridbuffer(ngrid,ngrid)

      double complex fxk_corr (ngrid+1,n2grid),
     $               fyk_corr (ngrid+1,n2grid),
     $               potk_corr(ngrid+1,n2grid),
     $               fxk_corr_unitm (ngrid+1,n2grid),
     $               fyk_corr_unitm (ngrid+1,n2grid),
     $               potk_corr_unitm(ngrid+1,n2grid),
     $               masskgrid(ngrid+1,n2grid)
#endif
!
! Attention: The complex arrays declared up there are not really newly 
! allocated, but are equivalent to the existing arrays, see declaration
! below. The arrays declared dble will hold fourier transformed data, so they
! will actually contain complex data (two subsequent dbles make up one dble
! complex). Through the equivalence below, a complex array is associated 
! with each of them, so the multiplication in k space can be carried out
! as a simple product of two complex numbers, rather then computing the real
! and imaginary part separatedly, using the split up components.
! Note that the size of the complex arrays is half the size of the real ones,
! ngrid+1 instead of n2grid+2 is used for the first dimension.
!
#if NDIM==3
      equivalence (fxk_corr (1,1,1),fxcorr_now(1,1,1)),
     $            (fyk_corr (1,1,1),fycorr_now(1,1,1)),
     $            (fzk_corr (1,1,1),fzcorr_now(1,1,1)),
     $            (potk_corr(1,1,1),potcorr_now(1,1,1)),        
     $            (fxk_corr_unitm(1,1,1), fxcorr_unitm(1,1,1)),
     $            (fyk_corr_unitm(1,1,1), fycorr_unitm(1,1,1)),
     $            (fzk_corr_unitm(1,1,1), fzcorr_unitm(1,1,1)),
     $            (potk_corr_unitm(1,1,1),potcorr_unitm(1,1,1)),
     $            (masskgrid(1,1,1),massgrid(1,1,1))
#else
      equivalence (fxk_corr (1,1),fxcorr_now(1,1)),
     $            (fyk_corr (1,1),fycorr_now(1,1)),
     $            (potk_corr(1,1),potcorr_now(1,1)),        
     $            (fxk_corr_unitm(1,1), fxcorr_unitm(1,1)),
     $            (fyk_corr_unitm(1,1), fycorr_unitm(1,1)),
     $            (potk_corr_unitm(1,1),potcorr_unitm(1,1)),
     $            (masskgrid(1,1),massgrid(1,1))
#endif


#ifndef PERIODIC
      ewald module: PERIODIC preprocessor flag is not set.
#endif
      contains
!---------------------------------------------------------------
      subroutine load_ewaldtable(ewaldfile)
!**************************************************************
!                                                             *
!  Load existing table with ewald correction forces from disk *
!                                                             *
!**************************************************************
      character(len=29), intent(in):: ewaldfile

      integer nsize,i,j,k

      character(len=20), parameter:: where =('load_ewaldtable')
#if NDIM==2
      character(len=3) , parameter:: dimen=('**2')
#else
      character(len=3) , parameter:: dimen=('**3')
#endif
      if(idbg.ge.2)print *,where

      open(idisk1,file=ewaldfile,form='unformatted')
      read(idisk1,err=100)nsize
      close(idisk1)
      print*,'Size of grid in file:',nsize,dimen
      if(nsize .ne. ngrid)then
         print*,'Wrong ewald table, should have size',ngrid,dimen
         stop
      endif

      open(idisk1,file=ewaldfile,form='unformatted')
      read(idisk1,err=100)nsize,
#if NDIM==2
     $     ((gridbuffer(i,j),i=1,ngrid),j=1,ngrid)
#else
     $     (((gridbuffer(i,j,k),i=1,ngrid),j=1,ngrid),k=1,ngrid)
#endif
      close(idisk1)

      return
!
!--an error as occured while reading
!
100   write(*,*)'reading error in load_ewaldtable!'
      stop
      end subroutine load_ewaldtable
!---------------------------------------------------------------
      subroutine save_ewaldtable(ewaldfile)
!**************************************************************
!                                                             *
!  Save generated table with ewald correction forces to disk  *
!                                                             *
!**************************************************************
      character(len=29), intent(in):: ewaldfile

      integer i,j,k
      character(len=20), parameter:: where =('save_ewaldtable')
      if(idbg.ge.2)print *,where

      open(idisk1, file=ewaldfile,form='unformatted')
      write(idisk1,err=100) ngrid,
#if NDIM==2
     $     ((gridbuffer(i,j),i=1,ngrid),j=1,ngrid)
#else
     $     (((gridbuffer(i,j,k),i=1,ngrid),j=1,ngrid),k=1,ngrid)
#endif
      close(idisk1)

      return
!
!--an error as occured while writing
!
100   write(*,*)'writing error in save_ewaldtable!'
      stop
      end subroutine save_ewaldtable
!---------------------------------------------------------------
      subroutine checkboxlimits(npart,npoim,dumposmas,dumposmaspm)
!***********************************************************************
!                                                                      *
! Check if particles moved across box boundary. If so, then move them  *
! inside the box again on its other side                               *
!                                                                      *
! Subroutine by M. Wetzstein, Heidelberg, 2002                         *
!                                                                      *
!***********************************************************************
      integer, intent(in):: npart,npoim
      double precision, intent(inout)::  dumposmas  (ndim1,npart)
      double precision, intent(inout)::  dumposmaspm(ndim1,npoim)

      integer j,k
      character(len=20), parameter:: where =('checkboxlimits')
      if(idbg.ge.2)print *,where

! check all particles, either they moved
! by integration or extrapolation
      do j=1,ndim
         print*,boxlims(1,j),boxlims(2,j),boxlen(j)
      enddo
!$omp parallel do default(none) private(j,k)
!$omp& shared(npart,dumposmas,boxlims,boxlen)
      do j=1,npart
         do k=1,ndim
            if(dumposmas(k,j).lt.boxlims(1,k))then
               dumposmas(k,j) = dumposmas(k,j) + boxlen(k)
            endif
            if(dumposmas(k,j).gt.boxlims(2,k))then
               dumposmas(k,j) = dumposmas(k,j) - boxlen(k)
            endif
         enddo
      enddo

      do j=1,npoim
         do k=1,ndim
            if(dumposmaspm(k,j).lt.boxlims(1,k))then
               dumposmaspm(k,j) = dumposmaspm(k,j) + boxlen(k)
            endif
            if(dumposmaspm(k,j).gt.boxlims(2,k))then
               dumposmaspm(k,j) = dumposmaspm(k,j) - boxlen(k)
            endif
         enddo
      enddo

      do j=1,npart
         do k=1,ndim
            if(abs(dumposmas(k,j)).gt.1d0)then
               print*,j,dumposmas(1,j),dumposmas(2,j),dumposmas(3,j)
               stop'Particle outside of box!'
            endif
         enddo
      enddo
      return
      end subroutine checkboxlimits
!---------------------------------------------------------------
      subroutine subtract_box
!************************************************************************
!                                                                       *
! Substract the contribution of the box itself (isolated solution) from *
! the tabulated values, so that only a real correction remains, taking  *
! only the periodic correction into account.                            *
!                                                                       *
!************************************************************************

      use partvars, only: npart,npart_sph,h

      double precision, parameter::sixth=1d0/6d0
      double precision, parameter::fiftnth=1d0/15d0
      double precision, parameter::thirtyth=1d0/30d0

      integer i,j,k
      double precision, parameter::twntysevths=20d0/7d0
      double precision unitmass,conversionfac,dx,dy,dz,r2,r,ri,rj,pot,
     $     hsum,emdi,emdi2,emdi3,di,d2,di2,em,v,v2,v3,v4,v5,emfrac,eps2,
     $     hmean,vinv,v6,sum
#if NDIM==2
      double precision acc(2)
#else
      double precision acc(3)
#endif
      character(len=20), parameter:: where =('subtract_box')
      if(idbg.ge.2)print *,where

      unitmass=one
      eps2=eps*eps
      conversionfac=boxlen(1)/dble(n2grid)
#if NDIM==2
      select case(isoft)
         case(id_Plummer)
            do j=1,n2grid
               do i=1,n2grid
                  
                  dx =dble(i-1)*conversionfac 
                  if(dx.gt.halfboxlen(1)) dx=dx-boxlen(1) 
                  
                  dy =dble(j-1)*conversionfac
                  if(dy.gt.halfboxlen(2)) dy=dy-boxlen(2)
                  r2 = dx*dx + dy*dy+ eps2
                  
                  r = sqrt(r2)
                  if(r.gt.zero)then
                     ri = one/(r*r2)
                     rj = one/r
                  else
                     ri = zero
                     rj = n2grid/boxlen(1) !assumes boxlen(1)=boxlen(2)
                  endif
                  
                  acc(1)=dx*unitmass*ri
                  acc(2)=dy*unitmass*ri
                  pot     =   unitmass*rj
                  
                  fxcorr_unitm(i,j) =acc(1)+ fxcorr_unitm(i,j)
                  fycorr_unitm(i,j) =acc(2)+ fycorr_unitm(i,j)
                  potcorr_unitm(i,j)=-pot    +potcorr_unitm(i,j)
               enddo
            enddo
         case(id_varkern,id_fixkern)
!
! Spline kernel soft. Difficult for computing forces on grid in 
! isolated box, as h can encompass variable numbers of grid points.
! We use the maximum h of all nbodies and hmax here for soft. as the latter is
! the upper limit for sph part. (-> max. number of affected grid points).
!
            hsum=0d0
! find max h of nbodies (if they exist)
            do i=npart_sph+1,npart 
               hsum=max(hsum,h(i))
            enddo
! check if sph part.'s hmax is greater than nbody max
            hsum=max(hmax,hsum)
            hsum=hsum+hsum

            do j=1,n2grid
               do i=1,n2grid
                  
                  dx =dble(i-1)*conversionfac 
                  if(dx.gt.halfboxlen(1)) dx=dx-boxlen(1) 
                  
                  dy =dble(j-1)*conversionfac
                  if(dy.gt.halfboxlen(2)) dy=dy-boxlen(2)
                  
                  d2 = dx*dx + dy*dy+ small
                  di  =one/sqrt(d2)
                  di2 =di*di
                  if(d2.gt.hsum*hsum)then
                     em=unitmass
                     pot= em*di
                  else
                     v=one/(di*hsum*half)
                     v2=v*v
                     v4=v2*v2
                     v5=v4*v
                     if(v2.gt.one)then !v.le.2 guaranteed by d2.lt.hsum2
                        v3=v2*v
                        emfrac=-0.05d0*v5+0.375d0*v4-v3 + v2-0.05d0
                     else
                        emfrac= 0.15d0*v5-0.375d0*v4 + half*v2
                     endif
                     em=twntysevths*emfrac*unitmass
                     pot= em*di !not physically `correct'
                     
                  endif                  
                  emdi3= em*di2*di
                  acc(1)= emdi3*dx 
                  acc(2)= emdi3*dy
                  
                  fxcorr_unitm(i,j) =acc(1)+ fxcorr_unitm(i,j)
                  fycorr_unitm(i,j) =acc(2)+ fycorr_unitm(i,j)
                  potcorr_unitm(i,j)=-pot    +potcorr_unitm(i,j)
               enddo
            enddo

      end select

#else
      select case(isoft)
         case(id_Plummer)

            do k=1,n2grid
               do j=1,n2grid
                  do i=1,n2grid
     
                     dx =dble(i-1)*conversionfac 
                     if(dx.gt.halfboxlen(1)) dx=dx-boxlen(1) 
                     
                     dy =dble(j-1)*conversionfac
                     if(dy.gt.halfboxlen(2)) dy=dy-boxlen(2)
                     dz = dble(k-1)*conversionfac
                     if(dz.gt.halfboxlen(3)) dz=dz-boxlen(3)
                     r2 = dx*dx + dy*dy+ dz*dz + eps2
                     r = sqrt(r2)
                     if(r.gt.zero)then
                        ri = one/(r*r2)
                        rj = one/r
                     else
                        ri = zero
                        rj = n2grid/boxlen(1) !assumes boxlen(1)=boxlen(2 and 3)
                     endif
                     
                     acc(1)=dx*unitmass*ri
                     acc(2)=dy*unitmass*ri
                     acc(3)=dz*unitmass*ri
                     pot     =   unitmass*rj
                     
                     fxcorr_unitm(i,j,k) =acc(1)+ fxcorr_unitm(i,j,k)
                     fycorr_unitm(i,j,k) =acc(2)+ fycorr_unitm(i,j,k)
                     fzcorr_unitm(i,j,k) =acc(3)+ fzcorr_unitm(i,j,k)
                     potcorr_unitm(i,j,k)=-pot    +potcorr_unitm(i,j,k)
                  enddo
               enddo
            enddo

         case(id_varkern,id_fixkern)
!
! Spline kernel soft. Difficult for computing forces on grid in 
! isolated box, as h can encompass variable numbers of grid points.
! We use the maximum h of all nbodies and hmax here for soft. as the latter is
! the upper limit for sph part. (-> max. number of affected grid points).
!
            hsum=0d0
! find max h of nbodies (if they exist)
            do i=npart_sph+1,npart 
               hsum=max(hsum,h(i))
            enddo
! check if sph part.'s hmax is greater than nbody max
            hsum=max(hmax,hsum)
            hsum=hsum+hsum

            do k=1,n2grid
               do j=1,n2grid
                  do i=1,n2grid
     
                     dx =dble(i-1)*conversionfac 
                     if(dx.gt.halfboxlen(1)) dx=dx-boxlen(1) 
                     
                     dy =dble(j-1)*conversionfac
                     if(dy.gt.halfboxlen(2)) dy=dy-boxlen(2)
                     dz = dble(k-1)*conversionfac
                     if(dz.gt.halfboxlen(3)) dz=dz-boxlen(3)

                     d2  =dx*dx+dy*dy+dz*dz + small
                     di  =one/sqrt(d2)
                     di2 =di*di

                     if(d2.gt.hsum*hsum)then
                        em =unitmass
                        pot=em*di
                     else
                        hmean=half*hsum
                        vinv=hmean*di
                        v=one/vinv
                        v2=v*v
                        v3=v2*v
                        v4=v2*v2
                        v5=v3*v2
                        v6=v3*v3
                        if(v2.gt.one)then !v.le.2 guaranteed by d2.lt.hsum2
                           emfrac= -fiftnth+atethd*v3-three*v4+1.2d0*v5-
     $                             sixth*v6
                           sum= fiftnth*vinv -1.6d0 + forthd*v2 - v3 + 
     &                          0.3d0*v4 -thirtyth*v5
                           pot=unitmass*sum/hmean
                        else
                           emfrac=forthd*v3 - 1.2d0*v5 + half*v6
                           sum=-1.4d0 + twothd*v2 - 0.3d0*v4 + 0.1d0*v5
                           pot=unitmass*sum/hmean
                        endif
                        em=emfrac*unitmass
                        
                     endif
                     emdi3=em*di2*di
                     acc(1)= emdi3*dx 
                     acc(2)= emdi3*dy
                     acc(3)= emdi3*dz
                     
                     fxcorr_unitm(i,j,k) =acc(1)+ fxcorr_unitm(i,j,k)
                     fycorr_unitm(i,j,k) =acc(2)+ fycorr_unitm(i,j,k)
                     fzcorr_unitm(i,j,k) =acc(3)+ fzcorr_unitm(i,j,k)
                     potcorr_unitm(i,j,k)=-pot    +potcorr_unitm(i,j,k)
                  enddo
               enddo
            enddo
      end select
#endif

      return
      end subroutine subtract_box
!---------------------------------------------------------------
      subroutine apply_ewald(npart,posmas,acc,poten)
!**************************************************************************
!                                                                         *
! This routine applies the periodic correction based on the ewald method. *
! The particles are binned onto a grid for which the acceleration and     *
! potential  corrections already exist (init_ewald). A cloud-in-cell      *
! scheme is used  for the binning.                                        *
! The mass density is then fourier transformed, multiplied with the       *
! corrections and transfomred back to real space.                         *
!                                                                         *
!**************************************************************************
      use indivtsvars

      integer, intent(in):: npart
      double precision, intent(in   ):: posmas(ndim1,npart)
      double precision, intent(inout):: acc(ndim,npart)
      double precision, intent(inout):: poten(npart)

      integer i,j,ix1,ix2,iy1,iy2,ipart,jpart
      double precision gconst,acc_normfac,pot_normfac,factor
      double precision r2inv_n2grid_box,dngrid_box
      double precision x,y,dx1,dx2,dy1,dy2,potcorrection
#if NDIM==2
      double precision d11,d21,d12,d22,fcorrection(2)
#else
      integer k,iz1,iz2
      double precision d111,d211,d121,d112,d212,d122,d221,d222,
     $     z,dz1,dz2,fcorrection(3)
#endif
      double precision tme1,tme2
      character(len=20), parameter:: where =('apply_ewald')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

!     normalisation: tough since in greensfunction uses a cube
!     of length 2 [-1,1]
!     notation: cube is greensfct related and box is simulation
!     box.

!
! First, do sanity check for implementation correctness of
! the periodic boundaries.
! Must have boxlen(1)=boxlen(2)=boxlen(3), for this implementation
!
      if(abs(boxlen(1)-boxlen(2))/boxlen(1).gt.1d-15)then
         print *,'boxlen(1)',boxlen(1)
         print *,'boxlen(2)',boxlen(2)
         stop'Ewald code must have box sides equal'
      endif
#if NDIM==3
      if(abs(boxlen(2)-boxlen(3))/boxlen(3).gt.1d-15)then
         print *,'boxlen(2)',boxlen(2)
         print *,'boxlen(3)',boxlen(3)
         stop'Ewald code must have box sides equal'
      endif
#endif
#if NDIM==2
      print*,'ATTENTION: NORMALIZATION FOR 2D ISN''t FIXED YET!!!'
#endif
      gconst       =one
      acc_normfac=gconst*boxlen(1)     ! normalization, /boxlen**2 *
                                         ! boxlen**3 <=> *boxlen
      pot_normfac  =gconst*boxlen(1)**2  ! normalization, /boxlen *
                                         ! boxlen**3 <=> *boxlen**2
      r2inv_n2grid_box=(one/dble(n2grid_box))**2
      factor = dble(ngrid)/boxlen(1)     ! Box ranges from -1 to +1 
      print*,'111'
! initialize array for mass
      do i=1,n2grid+2
         do j=1,n2grid
#if NDIM==3
            do k=1,n2grid
               massgrid(i,j,k)=zero
            enddo
#else
            massgrid(i,j)=zero
#endif
         enddo
      enddo
      print*,'222'
!
! bin all particles inside box onto the grid
!
      dngrid_box=dble(ngrid_box)
      
#if NDIM==2
!$omp parallel do default(none)
!$omp& firstprivate(factor,dngrid_box)
!$omp&   shared(npart,posmas,halfboxlen,massgrid)
!$omp&   private(i,x,y,ix1,ix2,iy1,iy2,dx1,dx2)
!$omp&   private(dy1,dy2)
      do i=1,npart
         x    = (posmas(1,i) + halfboxlen(1)) * factor
         y    = (posmas(2,i) + halfboxlen(2)) * factor
         ix1  = int(x) + 1
         iy1  = int(y) + 1
         ix2  = ix1 + 1
         iy2  = iy1 + 1
         dx1  = dble(ix1)-x
         dy1  = dble(iy1)-y
         dx2  = -dx1+one
         dy2  = -dy1+one
         massgrid(ix1,iy1)=massgrid(ix1,iy1)+dx1*dy1*dngrid_box*
     $                                         posmas(ims,i)
         massgrid(ix2,iy1)=massgrid(ix2,iy1)+dx2*dy1*dngrid_box*
     $                                         posmas(ims,i)
         massgrid(ix1,iy2)=massgrid(ix1,iy2)+dx1*dy2*dngrid_box*
     $                                         posmas(ims,i)
         massgrid(ix2,iy2)=massgrid(ix2,iy2)+dx2*dy2*dngrid_box*
     $                                         posmas(ims,i)
      enddo
#else
!$omp parallel do default(none)
!$omp& firstprivate(factor,dngrid_box)
!$omp&   shared(npart,posmas,halfboxlen,massgrid)
!$omp&   private(i,x,y,z,ix1,ix2,iy1,iy2,iz1,iz2,dx1,dx2)
!$omp&   private(dy1,dy2,dz1,dz2)
      do i=1,npart
         x    = (posmas(1,i) + halfboxlen(1)) * factor
         y    = (posmas(2,i) + halfboxlen(2)) * factor
         z    = (posmas(3,i) + halfboxlen(3)) * factor
         ix1  = int(x) + 1
         iy1  = int(y) + 1
         ix2  = ix1 + 1
         iy2  = iy1 + 1
         iz1  = int(z) + 1
         iz2  = iz1 + 1
         dx1  = dble(ix1)-x
         dy1  = dble(iy1)-y
         dx2  = -dx1+one
         dy2  = -dy1+one
         dz1  = dble(iz1)-z
         dz2  = -dz1+one
!         print*,'---------------------------------------'
!         print*,i,posmas(1,i),posmas(2,i),posmas(3,i)
!         print*,ix1,ix2,iy1,iy2,iz1,iz2
         massgrid(ix1,iy1,iz1)=massgrid(ix1,iy1,iz1)+dx1*dy1*dz1*
     $                                         dngrid_box*posmas(ims,i)
         massgrid(ix2,iy1,iz1)=massgrid(ix2,iy1,iz1)+dx2*dy1*dz1*
     $                                         dngrid_box*posmas(ims,i)
         massgrid(ix1,iy2,iz1)=massgrid(ix1,iy2,iz1)+dx1*dy2*dz1*
     $                                         dngrid_box*posmas(ims,i)
         massgrid(ix2,iy2,iz1)=massgrid(ix2,iy2,iz1)+dx2*dy2*dz1*
     $                                         dngrid_box*posmas(ims,i)
         massgrid(ix1,iy1,iz2)=massgrid(ix1,iy1,iz2)+dx1*dy1*dz2*
     $                                         dngrid_box*posmas(ims,i)
         massgrid(ix2,iy1,iz2)=massgrid(ix2,iy1,iz2)+dx2*dy1*dz2*
     $                                         dngrid_box*posmas(ims,i)
         massgrid(ix1,iy2,iz2)=massgrid(ix1,iy2,iz2)+dx1*dy2*dz2*
     $                                         dngrid_box*posmas(ims,i)
         massgrid(ix2,iy2,iz2)=massgrid(ix2,iy2,iz2)+dx2*dy2*dz2*
     $                                         dngrid_box*posmas(ims,i)
      enddo
#endif
      print*,'333'
!
! transform to k space
!      
      call dfftw_execute(fftw_massgrid_r2c)
!
! now convolve
!
! attention: here is the only place where the complex arrays are used to
! make the multiplication easier
!

#if NDIM==2

!$omp parallel do default(none) private(i,j)
!$omp& firstprivate(acc_normfac,pot_normfac,r2inv_n2grid_box)
!$omp&  shared(fxk_corr,fyk_corr,potk_corr,fxk_corr_unitm)
!$omp&  shared(masskgrid,fyk_corr_unitm,potk_corr_unitm)
      do j=1,n2grid
         do i=1,ngrid+1
            fxk_corr(i,j) = acc_normfac * r2inv_n2grid_box *
     $           masskgrid(i,j) * fxk_corr_unitm(i,j)
            fyk_corr(i,j) = acc_normfac * r2inv_n2grid_box *
     $           masskgrid(i,j) * fyk_corr_unitm(i,j)
            potk_corr(i,j)= pot_normfac   * r2inv_n2grid_box *
     $           masskgrid(i,j) * potk_corr_unitm(i,j)
         enddo
      enddo

#else

!$omp parallel do default(none) private(i,j,k)
!$omp& firstprivate(acc_normfac,pot_normfac,r2inv_n2grid_box)
!$omp&  shared(fxk_corr,fyk_corr,fzk_corr,potk_corr,fxk_corr_unitm)
!$omp&  shared(masskgrid,fyk_corr_unitm,fzk_corr_unitm,potk_corr_unitm)
      do k=1,n2grid
         do j=1,n2grid
            do i=1,ngrid+1
               fxk_corr(i,j,k) = acc_normfac * r2inv_n2grid_box *
     $              masskgrid(i,j,k) * fxk_corr_unitm(i,j,k)
               fyk_corr(i,j,k) = acc_normfac * r2inv_n2grid_box *
     $              masskgrid(i,j,k) * fyk_corr_unitm(i,j,k)
               fzk_corr(i,j,k) = acc_normfac * r2inv_n2grid_box *
     $              masskgrid(i,j,k) * fzk_corr_unitm(i,j,k)
               potk_corr(i,j,k)= pot_normfac   * r2inv_n2grid_box *
     $              masskgrid(i,j,k) * potk_corr_unitm(i,j,k)
             enddo
         enddo
      enddo

#endif
      print*,'444'
!
! now transfrom back to real space to get the corrections
!
      call dfftw_execute(fftw_fxk_corr_c2r)
      call dfftw_execute(fftw_fyk_corr_c2r)
      call dfftw_execute(fftw_potk_corr_c2r)
#if NDIM==3
      call dfftw_execute(fftw_fzk_corr_c2r)
#endif
!
! compute the corrections from the grid using CIC, then apply the corrections
!
      
#if NDIM==2
!$omp parallel do default(none)
!$omp& firstprivate(factor)
!$omp& shared(nactive,iactive,halfboxlen,posmas)
!$omp& shared(fxcorr_now,fycorr_now,potcorr_now,acc,poten)
!$omp& private(ipart,jpart,x,y,ix1,ix2,iy1,iy2)
!$omp& private(dx1,dx2,dy1,dy2,d11,d21,d12,d22)
!$omp& private(fcorrection,potcorrection)
      do ipart=1,nactive
         jpart=iactive(ipart)
         x    = (posmas(1,jpart) + halfboxlen(1)) * factor
         y    = (posmas(2,jpart) + halfboxlen(2)) * factor
         ix1  = int(x) + 1
         iy1  = int(y) + 1
         ix2  = ix1 + 1
         iy2  = iy1 + 1
         dx1  = dble(ix1)-x
         dy1  = dble(iy1)-y
         dx2  = -dx1+one
         dy2  = -dy1+one
!
! compute weights in 2d
!
         d11=dx1*dy1
         d21=dx2*dy1
         d12=dx1*dy2
         d22=dx2*dy2
!
!     Correction for acceleration
!
         fcorrection(1) = d11 * fxcorr_now(ix1,iy1)
     &                  + d21 * fxcorr_now(ix2,iy1)
     &                  + d12 * fxcorr_now(ix1,iy2)
     &                  + d22 * fxcorr_now(ix2,iy2)

         fcorrection(2) = d11 * fycorr_now(ix1,iy1)
     &                  + d21 * fycorr_now(ix2,iy1)
     &                  + d12 * fycorr_now(ix1,iy2)
     &                  + d22 * fycorr_now(ix2,iy2)

!
! Correction for potential
!
         potcorrection = d11 *  potcorr_now(ix1,iy1)
     &                 + d21 *  potcorr_now(ix2,iy1)
     &                 + d12 *  potcorr_now(ix1,iy2)
     &                 + d22 *  potcorr_now(ix2,iy2)

         acc(1,jpart)   = acc(1,jpart) + fcorrection(1)
         acc(2,jpart)   = acc(2,jpart) + fcorrection(2)
         poten(jpart)     = poten(jpart)   + potcorrection
      enddo
#else
!$omp parallel do default(none)
!$omp& firstprivate(factor)
!$omp& shared(nactive,iactive)
!$omp& shared(halfboxlen,posmas)
!$omp& shared(fxcorr_now,fycorr_now,fzcorr_now,potcorr_now,acc,poten)
!$omp& private(ipart,jpart,x,y,z,ix1,ix2,iy1,iy2,iz1,iz2)
!$omp& private(dx1,dx2,dy1,dy2,dz1,dz2)
!$omp& private(d111,d211,d121,d112,d212,d122,d221,d222)
!$omp& private(fcorrection,potcorrection)
      do ipart=1,nactive
         jpart=iactive(ipart)
         x    = (posmas(1,jpart) + halfboxlen(1)) * factor
         y    = (posmas(2,jpart) + halfboxlen(2)) * factor
         z    = (posmas(3,jpart) + halfboxlen(3)) * factor
         ix1  = int(x) + 1
         iy1  = int(y) + 1
         ix2  = ix1 + 1
         iy2  = iy1 + 1
         iz1  = int(z) + 1
         iz2  = iz1 + 1
         dx1  = dble(ix1)-x
         dy1  = dble(iy1)-y
         dx2  = -dx1+one
         dy2  = -dy1+one
         dz1  = dble(iz1)-z
         dz2  = -dz1+one
!
! compute weights in 3d
!
         d111=dx1*dy1*dz1
         d211=dx2*dy1*dz1
         d121=dx1*dy2*dz1
         d221=dx2*dy2*dz1
         d112=dx1*dy1*dz2
         d212=dx2*dy1*dz2
         d122=dx1*dy2*dz2        
         d222=dx2*dy2*dz2    
!
!     Correction for acceleration
!
         fcorrection(1) = d111 * fxcorr_now(ix1,iy1,iz1)
     &                  + d211 * fxcorr_now(ix2,iy1,iz1)
     &                  + d121 * fxcorr_now(ix1,iy2,iz1)
     &                  + d221 * fxcorr_now(ix2,iy2,iz1)
     &                  + d112 * fxcorr_now(ix1,iy1,iz2)
     &                  + d212 * fxcorr_now(ix2,iy1,iz2)
     &                  + d122 * fxcorr_now(ix1,iy2,iz2)
     &                  + d222 * fxcorr_now(ix2,iy2,iz2)

         fcorrection(2) = d111 * fycorr_now(ix1,iy1,iz1)
     &                  + d211 * fycorr_now(ix2,iy1,iz1)
     &                  + d121 * fycorr_now(ix1,iy2,iz1)
     &                  + d221 * fycorr_now(ix2,iy2,iz1)
     &                  + d112 * fycorr_now(ix1,iy1,iz2)
     &                  + d212 * fycorr_now(ix2,iy1,iz2)
     &                  + d122 * fycorr_now(ix1,iy2,iz2)
     &                  + d222 * fycorr_now(ix2,iy2,iz2)

         fcorrection(3) = d111 * fzcorr_now(ix1,iy1,iz1)
     &                  + d211 * fzcorr_now(ix2,iy1,iz1)
     &                  + d121 * fzcorr_now(ix1,iy2,iz1)
     &                  + d221 * fzcorr_now(ix2,iy2,iz1)
     &                  + d112 * fzcorr_now(ix1,iy1,iz2)
     &                  + d212 * fzcorr_now(ix2,iy1,iz2)
     &                  + d122 * fzcorr_now(ix1,iy2,iz2)
     &                  + d222 * fzcorr_now(ix2,iy2,iz2)
!
! Correction for potential
!
         potcorrection = d111 *  potcorr_now(ix1,iy1,iz1)
     &                 + d211 *  potcorr_now(ix2,iy1,iz1)
     &                 + d121 *  potcorr_now(ix1,iy2,iz1)
     &                 + d221 *  potcorr_now(ix2,iy2,iz1)
     &                 + d112 *  potcorr_now(ix1,iy1,iz2)
     &                 + d212 *  potcorr_now(ix2,iy1,iz2)
     &                 + d122 *  potcorr_now(ix1,iy2,iz2)
     &                 + d222 *  potcorr_now(ix2,iy2,iz2)

         acc(1,jpart)   = acc(1,jpart) + fcorrection(1)
         acc(2,jpart)   = acc(2,jpart) + fcorrection(2)
         acc(3,jpart)   = acc(3,jpart) + fcorrection(3)
         poten(jpart)     = poten(jpart)   + potcorrection
      enddo
#endif
      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_ewald)=timing(itme_ewald)+(tme2-tme1)
         ncalls(itme_ewald)=ncalls(itme_ewald)+1
         if(itme.ge.2)print *,'where ',tme2-tme1
      endif

      return
      end subroutine apply_ewald
!---------------------------------------------------------------
      subroutine buildfullgrid(iswitch,ngrid,ngridhalf,n2grid,
     $     gridbuffer,fullgridbuffer)

!************************************************************************
!                                                                       *
! The content of gridbuffer is the correction table for distances from  *
! -ngrid to ngrid cells of the grid. The full correction table has to   *
! be twice as large to avoid artifacts of undesired periodicity of the  *
! grid which would affect the result through wrap-around during the     *
! convolution (see e.g. Numerical Recipes, ch. 13.1). So gridbuffer has *
! to be extended and therefore only represents one octant of the full   *
! grid that is to used for the convolution.                             *
! So it has to be sorted correctly and copied into the octants.         *
! iswitch denotes for which direction (x,y,z) the correction table is,  *
! cause gridbuffer was only made for x.                                 *
!                                                                       *
!************************************************************************
      integer, intent(in):: iswitch,ngrid,ngridhalf,n2grid

#if NDIM==2
      double precision, intent(in ):: gridbuffer(ngrid,ngrid)
      double precision, intent(out)::
     &                  fullgridbuffer(n2grid+2,n2grid)
#else
      double precision, intent(in ):: gridbuffer(ngrid,ngrid,ngrid)
      double precision, intent(out)::
     &                  fullgridbuffer(n2grid+2,n2grid,n2grid)
#endif
      integer i,j,k,ii,jj,kk,imult,jmult,kmult

      character(len=20), parameter:: where =('buildfullgrid')
      
      if(idbg.ge.2)print *,where
!
! gridbuffer was made such that origin of the box is in 'center' of array
! meaning at gridbuffer(ngrid/2+1,ngrid/2+1,ngrid/2+1)
! now shift origin to 'center' of simulation box, corresponding to
! fullgridbuffer(n2grid/2+1,n2grid/2+1,n2grid/2+1).
!
#if NDIM==3
      do k=1,ngrid
#endif
         do j=1,ngrid
            do i=1, ngrid
               ii=i+ngridhalf
               jj=j+ngridhalf
               if (ii.gt.ngrid) ii=ii-ngrid
               if (jj.gt.ngrid) jj=jj-ngrid
#if NDIM==3
               kk=k+ngridhalf
               if (kk.gt.ngrid) kk=kk-ngrid
               if(iswitch.eq.1.or.iswitch.eq.4)then !x correction table or potential
                  fullgridbuffer(ii,jj,kk) = gridbuffer(i,j,k)
               elseif(iswitch.eq.2)then !y correction table
                  fullgridbuffer(ii,jj,kk) = gridbuffer(j,i,k)
               elseif(iswitch.eq.3)then !z correction table
                  fullgridbuffer(ii,jj,kk) = gridbuffer(k,j,i)
               endif
#else
               if(iswitch.eq.1.or.iswitch.eq.4)then !x correction table or potential
                  fullgridbuffer(ii,jj) = gridbuffer(i,j)
               elseif(iswitch.eq.2)then !y correction table
                  fullgridbuffer(ii,jj) = gridbuffer(j,i)
               endif
#endif
            enddo
         enddo
#if NDIM==3
      enddo ! outermost do loop, only for 3d  
#endif
!
! if potential, init different before copying
!
      if(iswitch.eq.4)then
#if NDIM==2
         fullgridbuffer(1,1)  =-dble(n2grid)/boxlen(1)
#else
         fullgridbuffer(1,1,1)=-dble(n2grid)/boxlen(1)
#endif
      endif
!
! now replicate the correction table to other octants
!
#if NDIM==3
      do kmult=0,1
#endif
         do jmult=0,1
            do imult=0,1
#if NDIM==3
               do k =1,ngrid
#endif
                  do j=1,ngrid
                     do i=1,ngrid
                        ii=i + ngrid * imult !first fill upper half, then lower
                        jj=j + ngrid * jmult
#if NDIM==3
                        kk=k + ngrid * kmult
                        fullgridbuffer(ii,jj,kk)=fullgridbuffer(i,j,k)
#else
                        fullgridbuffer(ii,jj)=fullgridbuffer(i,j)
#endif
                     enddo
                  enddo
#if NDIM==3
               enddo
#endif
            enddo
         enddo
#if NDIM==3
      enddo
#endif

      return
      end subroutine buildfullgrid
!---------------------------------------------------------------
      subroutine init_rand(idummy)

      integer, intent(in):: idummy
      double precision forgetme

      idummyrandom=idummy
      forgetme=ran2(idummyrandom)

      if(idbg.gt.1)print *,'First random number is:',forgetme

      return
      end subroutine init_rand
!---------------------------------------------------------------
      subroutine init_ewald
!************************************************************************
!                                                                       *
! This supplies the table of corrections to the gravitational forces    *
! and potential , based on Ewald's method as described in Klessen       *
! MNRAS (1997). A table of corrections for the desired grid resolution  *
! already exists on disk or is newly generated and then saved to disk   *
! for future use.                                                       *
!                                                                       *
!************************************************************************

!
! The correction tables are stored in the directory ewalddir. The code
! checks if this directory exists, otherwise it tells the user to create
! it. This check works fine using the inquire routine, but not for every
! compiler. E.g. with the Intel Fortran compiler it used to work up to v7
! but v8 always returns .false. if you inquire about a directory. Thus
! we use some compiler specific functions with this version of the intel
! compiler in order to do the job. It's not very elegant to do so, but
! it works.
! We use the stat routine, which is part of the ifport module, to check
! for the existence of ewalddir. fstat returns an integer array which is
! an IOR combination of several flags. s_ifdir is the one indicating 
! directory status, so we check for that (see below).
!
#if __INTEL_COMPILER>=800
      use ifport
      integer io_info(12)
      logical err
#endif
      logical tobe
      character(len=29) ewaldfile
      character(len=17), parameter::ewaldbasename=('TABLES/ewaldtable')
      character(len=6 ), parameter::ewalddir=('TABLES')
      integer numcpus,idum,jdum,kdum,ii,jj,kk,i,j,k
      double precision eL,alpha,sqrtpi,pi2
      double precision dumfcorr,dx,dy,dz,r,r2,eps2
      double precision tme1,tme2
      character(len=20), parameter:: where =('init_ewald')
!
! On SGI Altix systems, it seems that the return type of dfftw_init_threads
! is int*8 instead of the conventional int, although the FFTW doc says this
! would not be the case. At least for the systems we tested, using int*8
! allowed us to use FFTW lib with threads, which would not work with normal int
!
#ifdef ITAN
      integer*8 ierror
#else
      integer ierror
#endif

      if(idbg.ge.2)print *,where
      eps2=eps*eps
!
! Create the plans for the transforms in the FFTW library (cf. FFTW
! documentation). This has to be done BEFORE the data is written into
! the arrays, as using anyting else then FFTW_ESTIMATE overwrites the
! input arrays when the plans are created ! FFTW_PATIENT for production runs!
!
#ifdef _OPENMP     !compute transforms in parallel
      call numprocessors(numcpus)
      call dfftw_init_threads(ierror)
      if(ierror.eq.0)then
         print *,'Error initializing FFTW for threads!',numcpus,ierror
         stop'Couldn''t initialize parallel FFTW'
      endif
      call dfftw_plan_with_nthreads(numcpus)
#endif

      call cpu_time(tme1)

#if NDIM==2
      call dfftw_plan_dft_r2c_2d(fftw_fxcorr_unitm_r2c,n2grid,n2grid,
     $     fxcorr_unitm,fxcorr_unitm,FFTW_MEASURE)
      call dfftw_plan_dft_r2c_2d(fftw_fycorr_unitm_r2c,n2grid,n2grid,
     $     fycorr_unitm,fycorr_unitm,FFTW_MEASURE)
      call dfftw_plan_dft_r2c_2d(fftw_potcorr_unitm_r2c,n2grid,n2grid,
     $     potcorr_unitm,potcorr_unitm,FFTW_MEASURE)
!
! now create the plans for the FFTs in apply_ewald
!
      call dfftw_plan_dft_r2c_2d(fftw_massgrid_r2c,n2grid,n2grid,
     $     massgrid,massgrid,FFTW_MEASURE)
      call dfftw_plan_dft_c2r_2d(fftw_fxk_corr_c2r,n2grid,n2grid,
     $     fxcorr_now,fxcorr_now,FFTW_MEASURE)
      call dfftw_plan_dft_c2r_2d(fftw_fyk_corr_c2r,n2grid,n2grid,
     $     fycorr_now,fycorr_now,FFTW_MEASURE)
      call dfftw_plan_dft_c2r_2d(fftw_potk_corr_c2r,n2grid,n2grid,
     $     potcorr_now,potcorr_now,FFTW_MEASURE)

#else

      call dfftw_plan_dft_r2c_3d(fftw_fxcorr_unitm_r2c,n2grid,n2grid,
     $     n2grid,fxcorr_unitm,fxcorr_unitm,FFTW_MEASURE)
      call dfftw_plan_dft_r2c_3d(fftw_fycorr_unitm_r2c,n2grid,n2grid,
     $     n2grid,fycorr_unitm,fycorr_unitm,FFTW_MEASURE)
      call dfftw_plan_dft_r2c_3d(fftw_fzcorr_unitm_r2c,n2grid,n2grid,
     $     n2grid,fzcorr_unitm,fzcorr_unitm,FFTW_MEASURE)
      call dfftw_plan_dft_r2c_3d(fftw_potcorr_unitm_r2c,n2grid,n2grid,
     $     n2grid,potcorr_unitm,potcorr_unitm,FFTW_MEASURE)
!
! now create the plans for the FFTs in apply_ewald
!
      call dfftw_plan_dft_r2c_3d(fftw_massgrid_r2c,n2grid,n2grid,
     $     n2grid,massgrid,massgrid,FFTW_MEASURE)
      call dfftw_plan_dft_c2r_3d(fftw_fxk_corr_c2r,n2grid,n2grid,
     $     n2grid,fxcorr_now,fxcorr_now,FFTW_MEASURE)
      call dfftw_plan_dft_c2r_3d(fftw_fyk_corr_c2r,n2grid,n2grid,
     $     n2grid,fycorr_now,fycorr_now,FFTW_MEASURE)
      call dfftw_plan_dft_c2r_3d(fftw_fzk_corr_c2r,n2grid,n2grid,
     $     n2grid,fzcorr_now,fzcorr_now,FFTW_MEASURE)
      call dfftw_plan_dft_c2r_3d(fftw_potk_corr_c2r,n2grid,n2grid,
     $     n2grid,potcorr_now,potcorr_now,FFTW_MEASURE)
#endif
      if(itme.eq.1)then
         call cpu_time(tme2)
         timing(itme_fftwplan)=timing(itme_fftwplan)+(tme2-tme1)
         ncalls(itme_fftwplan)=ncalls(itme_fftwplan)+1
         print *,'FFTW planning time ',tme2-tme1
      endif
!
! check if table with ewald correction forces exists, 
! if not then generate it and save it for future use      
!

! check first if directory TABLES/ exists

!
! See the lenghty comment above about the reason for not using inquire
! with the intel compiler v8 or higher.
!
#if __INTEL_COMPILER>=800
      err=stat(ewalddir,io_info)
      if(.not.err)then
         if(iand(s_ifdir,io_info(3)).ne.s_ifdir)then
            print*,'Directory ',ewalddir,' for storing ewald correction'
            print*,'tables doesn''t exist, please create it!!!'
            stop
         endif
      else
         print*,'Error getting information on directory',ewalddir
         stop
      endif
#else
      inquire(file=ewalddir,exist=tobe)
      if(.not.tobe)then
         print*,'Directory ',ewalddir,'for storing ewald correction'
         print*,'tables doesn''t exist, please create it!!!'
         stop
      endif
#endif
      
      write(ewaldfile,200)ewaldbasename,ngrid
#if NDIM==2
 200  format(a17,i3.3,'_force_2d')
#else
 200  format(a17,i3.3,'_force_3d')
#endif
      print*,'Checking if ewald force correction table exists:',
     $     ewaldfile

      inquire(file=ewaldfile,exist=tobe)

      if(.not.tobe)then
         print*,'Table doesn''t exist, generating new one...'
         call generate_ewaldforce
         call save_ewaldtable(ewaldfile)
      else
         print*,'Table exists, loading...'
         call load_ewaldtable(ewaldfile)
      endif
!
! copy information into arrays for x,y,z corrections
!
      call buildfullgrid(1,ngrid,ngridhalf,n2grid, !boxlen,
     $     gridbuffer,fxcorr_unitm)

      call buildfullgrid(2,ngrid,ngridhalf,n2grid, !boxlen,
     $     gridbuffer,fycorr_unitm)
#if NDIM==3
      call buildfullgrid(3,ngrid,ngridhalf,n2grid, !boxlen,
     $     gridbuffer,fzcorr_unitm)
#endif
!
! check if potential correction file exists, if not generate new one
!
      write(ewaldfile,300)ewaldbasename,ngrid
#if NDIM==2
 300  format(a17,i3.3,'_poten_2d')
#else
 300  format(a17,i3.3,'_poten_3d')
#endif
      print*,'Checking if ewald potential correction table exists:',
     $     ewaldfile

      inquire(file=ewaldfile,exist=tobe)

      if(.not.tobe)then
         print *,'Table doesn''t exist, generating new one...'
         call generate_ewaldpot
         call save_ewaldtable(ewaldfile)
      else
         print *,'Table exists, loading...'
         call load_ewaldtable(ewaldfile)
      endif
      print *,'Loading/Creating Ewald-Correction-Tables completed'
!
! potential correction is stored like x force correction, 
! so sort it the same way, but initialize different
!
      call buildfullgrid(4,ngrid,ngridhalf,n2grid, !boxlen,
     $     gridbuffer,potcorr_unitm)

#if NDIM==2
      select case (isoft)
         case (id_plummer)
            potcorr_unitm(1,1) = -1d0/sqrt(eps2)
         case (id_varkern)
            potcorr_unitm(1,1) = -1d0/sqrt(hmax)
         case (id_fixkern)
            potcorr_unitm(1,1) = -1d0/sqrt(hmax) 
      end select
#else
      select case (isoft)
         case (id_plummer)
            potcorr_unitm(1,1,1) = -1d0/sqrt(eps2) 
         case (id_varkern)
            potcorr_unitm(1,1,1) = -1d0/sqrt(hmax) 
         case (id_fixkern)
            potcorr_unitm(1,1,1) = -1d0/sqrt(hmax) 
      end select
#endif
!
! make sure only periodic correction is stored, 
! no contribution of matter from inside box
!
      call subtract_box

! transform the correction to k space, so that they can be used for
! multiplying with the (fourier transformed) mass density right away
! note that an in-place transform is used, the results are stored in
! the original arrays

      call cpu_time(tme1)
      call dfftw_execute(fftw_fxcorr_unitm_r2c)
      call dfftw_execute(fftw_fycorr_unitm_r2c)
#if NDIM==3
      call dfftw_execute(fftw_fzcorr_unitm_r2c)
#endif
      if(itme.eq.1)then
         call cpu_time(tme2) 
         timing(itme_fft)=(tme2-tme1)/3d0
         ncalls(itme_fft)=ncalls(itme_fft)+1
#if NDIM==2
         print *,'Time for one FFT:',(tme2-tme1)/2d0
#else
         print *,'Time for one FFT:',(tme2-tme1)/3d0
#endif
      endif
      call dfftw_execute(fftw_potcorr_unitm_r2c)

      return
      end subroutine init_ewald
!---------------------------------------------------------------
      subroutine generate_ewaldforce
!**************************************************************
!                                                             *
!  Generated table with ewald correction forces               *
!                                                             *
!**************************************************************

      integer idum,jdum,kdum,ii,jj,kk,i,j,k
      double precision eL,alpha,sqrtpi,pi2,dumfcorr
      double precision dx,dy,dz,r,r2,dummy,h1,h2
      double precision d(3)

      character(len=20), parameter:: where =('generate_ewaldforce')

      if(idbg.ge.2)print *,where

! define constants
      eL=one
      alpha=two/eL
      sqrtpi=sqrt(PI)
      idum=1
      jdum=1
      kdum=1
      pi2=PI**2
!
! looping over every grid point to get distances
! this time compute force correction only, potential comes later,
! so that only one array as buffer is needed
!
      do ii=-ngrid/2,ngrid/2-1
         do jj=-ngrid/2,ngrid/2-1
#if NDIM==3
            do kk=-ngrid/2,ngrid/2-1
#endif         
               d(1)=dble(ii)/dble(ngrid)
               d(2)=dble(jj)/dble(ngrid)
               r2=d(1)*d(1)+d(2)*d(2)
#if NDIM==3
               d(3)=dble(kk)/dble(ngrid)
               r2=r2+d(3)*d(3)
#endif         
               dumfcorr=zero

               if(r2.ne.zero)then
! compute long range part of Ewald sum                  

                  do i=-4,4,1
                     do j=-4,4,1
#if NDIM==3
                       do k=-4,4,1
#endif         
                           dx=(d(1)-eL*dble(i))
                           dy=(d(2)-eL*dble(j))
                           r=dx*dx+dy*dy
#if NDIM==3
                           dz=(d(3)-eL*dble(k))
                           r=r+dz*dz
#endif         
                           r=sqrt(r)
                           dummy=erfunctionc(alpha*r)+two*alpha*
     $                          r/sqrtpi*exp(-alpha*alpha*r*r)
                           if(r.gt.zero)then
                              dumfcorr=dumfcorr-dz/(r*r*r)*dummy
                           endif
#if NDIM==3
                        enddo !most inner do loop, only for 3d  
#endif         
                     enddo   
                  enddo
! compute short range part of Ewald sum                  

                  do i=-4,4,1
                     do j=-4,4,1
#if NDIM==3
                        do k=-4,4,1
#endif         
                           h1=d(1)*dble(i)+d(2)*dble(j)
                           h2=i*i+j*j                           
#if NDIM==3
                           h1=h1+d(3)*dble(k)
                           h2=h2+k*k
#endif         
                           if(h2.gt.zero)then
                              dummy=two/(eL*eL)/h2*
     $                             dexp(-pi2*h2/(eL*eL)/(alpha*alpha))*
     $                             dsin(two*PI/eL*h1)
                              dumfcorr=dumfcorr-k*dummy
                           endif
#if NDIM==3
                        enddo !most inner do loop, only for 3d  
#endif         
                     enddo   
                  enddo
               endif
! --- copying into the right order    

#if NDIM==3
               gridbuffer(idum,jdum,kdum)=dumfcorr
#else
               gridbuffer(idum,jdum)=dumfcorr
#endif         
               idum=idum+1
               if(idum.gt.ngrid)then
                  jdum=jdum+1
                  idum=idum-ngrid
               endif
#if NDIM==3
               if(jdum.gt.ngrid)then
                  kdum=kdum+1
                  jdum=jdum-ngrid
               endif
            enddo !most inner do loop, only for 3d
#endif         
         enddo
      enddo

      return
      end subroutine generate_ewaldforce
!---------------------------------------------------------------
      subroutine generate_ewaldpot
!**************************************************************
!                                                             *
!  Generated table with ewald correction potentials           *
!                                                             *
!**************************************************************

      integer idum,jdum,kdum,ii,jj,kk,i,j,k
      double precision eL,alpha,sqrtpi,pisqrd
      double precision dx,dy,dz,r,r2,dummy,h1,h2
      double precision potfac,dumpotcorr
      double precision d(3)
      character(len=20), parameter:: where=('generate_ewaldpot')

      if(idbg.ge.2)print *,where

! define constants
      eL=one
      alpha=two/eL
      sqrtpi=sqrt(PI)
      pisqrd=PI**2
      idum=1
      jdum=1
      kdum=1
      potfac=PI/(alpha*alpha*eL*eL*eL)
!
! looping over every grid point to get distances
! this time compute potential correction only
!
      do ii=-ngrid/2,ngrid/2-1
         do jj=-ngrid/2,ngrid/2-1
#if NDIM==3
            do kk=-ngrid/2,ngrid/2-1
#endif         
               d(1)=dble(ii)/dble(ngrid)
               d(2)=dble(jj)/dble(ngrid)
               r2=d(1)*d(1)+d(2)*d(2)
#if NDIM==3
               d(3)=dble(kk)/dble(ngrid)
               r2=r2+d(3)*d(3)
#endif         
               dumpotcorr=0.d0

               if(r2.ne.zero)then
! compute long range part of Ewald sum                  
                  do i=-4,4,1
                     do j=-4,4,1
#if NDIM==3
                       do k=-4,4,1
#endif         
                          dx=(d(1)-eL*dble(i))
                          dy=(d(2)-eL*dble(j))
                          r=dble(dx*dx+dy*dy)
#if NDIM==3
                          dz=(d(3)-eL*dble(k))
                          r=r+dble(dz*dz)
#endif         
                          r=sqrt(r)
                          dummy=erfunctionc(alpha*r)
                          if(r.gt.zero)then
                             dumpotcorr=dumpotcorr-dummy/r
                          endif
#if NDIM==3
                       enddo !most inner do loop, only for 3d  
#endif         
                    enddo   
                 enddo
! compute short range part of Ewald sum         
                  do i=-4,4,1
                     do j=-4,4,1
#if NDIM==3
                        do k=-4,4,1
#endif         
                           h1=d(1)*dble(i)+d(2)*dble(j)
                           h2=i*i+j*j
#if NDIM==3
                           h1=h1+d(3)*dble(k)
                           h2=h2+k*k
#endif         
                           if(h2.gt.zero)then
                              dummy=one/eL/h2/PI*
     $                           dexp(-pisqrd*h2/(eL*eL)/(alpha*alpha))*
     $                           dcos(two*PI/eL*h1)
                              dumpotcorr=dumpotcorr-dummy
                           endif
#if NDIM==3
                        enddo !most inner do loop, only for 3d  
#endif         
                     enddo   
                  enddo
               endif
! --- copying into the right order    
#if NDIM==3
               gridbuffer(idum,jdum,kdum)=potfac+dumpotcorr
#else
               gridbuffer(idum,jdum)=potfac+dumpotcorr
#endif
               idum=idum+1
               if(idum.gt.ngrid)then
                  jdum=jdum+1
                  idum=idum-ngrid
               endif
#if NDIM==3
               if(jdum.gt.ngrid)then
                  kdum=kdum+1
                  jdum=jdum-ngrid
               endif
            enddo !most inner do loop, only for 3d
#endif         
         enddo
      enddo

      return
      end subroutine generate_ewaldpot
!---------------------------------------------------------------
      subroutine shift_box_to_original(npart,npoim,posmas,posmaspm)
!
! correct for random shift before dump is written
!
      integer, intent(in):: npart,npoim
      double precision, intent(inout):: posmas  (ndim1,npart)
      double precision, intent(inout):: posmaspm(ndim1,npoim)
      integer i,k

      do i=1,npart
         do  k=1,ndim
            posmas(k,i) = posmas(k,i) - randshifttot(k)
            if (posmas(k,i).lt.boxlims(1,k)) posmas(k,i) = 
     &           posmas(k,i) + boxlen(k)
            if (posmas(k,i).gt.boxlims(2,k)) posmas(k,i) = 
     &           posmas(k,i) - boxlen(k)
         enddo
      enddo

      do i=1,npoim
         do  k=1,ndim
            posmaspm(k,i) = posmaspm(k,i) - randshifttot(k)
            if (posmaspm(k,i).lt.boxlims(1,k)) posmaspm(k,i) = 
     &           posmaspm(k,i) + boxlen(k)
            if (posmaspm(k,i).gt.boxlims(2,k)) posmaspm(k,i) = 
     &           posmaspm(k,i) - boxlen(k)
         enddo
      enddo

      return
      end subroutine shift_box_to_original
!---------------------------------------------------------------
      subroutine shift_box_to_current(npart,npoim,posmas,posmaspm)
!
! apply current random shift again after dump was written
!
      integer, intent(in):: npart,npoim
      double precision, intent(inout):: posmas  (ndim1,npart)
      double precision, intent(inout):: posmaspm(ndim1,npoim)
      integer i,k

      do  i=1,npart
         do  k=1,ndim
            posmas(k,i) = posmas(k,i) + randshifttot(k)
            if (posmas(k,i).lt.boxlims(1,k)) posmas(k,i) = 
     &           posmas(k,i) + boxlen(k)
            if (posmas(k,i).gt.boxlims(2,k)) posmas(k,i) = 
     &           posmas(k,i) - boxlen(k)
         enddo
      enddo

      do  i=1,npoim
         do  k=1,ndim
            posmaspm(k,i) = posmaspm(k,i) + randshifttot(k)
            if (posmaspm(k,i).lt.boxlims(1,k)) posmaspm(k,i) = 
     &           posmaspm(k,i) + boxlen(k)
            if (posmaspm(k,i).gt.boxlims(2,k)) posmaspm(k,i) = 
     &           posmaspm(k,i) - boxlen(k)
         enddo
      enddo

      return
      end subroutine shift_box_to_current
!---------------------------------------------------------------
      subroutine boxshift(npart,npoim,posmas,posmaspm)

!***********************************************************************
! For periodic boundaries:
! Shifts the simulation box randomly to avoid errors from
! boundary effects.
!***********************************************************************
      integer, intent(in)::npart,npoim
      double precision, intent(inout):: posmas  (ndim1,npart)
      double precision, intent(inout):: posmaspm(ndim1,npoim)

      integer  p,k

      character(len=20), parameter:: where =('boxshift')
      if(idbg.ge.2)print *,where

      do k=1,ndim
         randshift(k) = ran2(idummyrandom)*1.8d0 - 0.9d0
         randshifttot(k) = randshifttot(k) + randshift(k)
         if(randshifttot(k) .ge. 1d0) 
     &        randshifttot(k) = randshifttot(k) - boxlen(k)
         if(randshifttot(k) .le.-1d0) 
     &        randshifttot(k) = randshifttot(k) + boxlen(k)
         do p=1,npart
            posmas(k,p) = posmas(k,p) + randshift(k)
            if(posmas(k,p).LT.boxlims(1,k))then
               posmas(k,p) = posmas(k,p) + boxlen(k)
            endif
            if(posmas(k,p).GT.boxlims(2,k))then
               posmas(k,p) = posmas(k,p) - boxlen(k)
            endif
         enddo

         do p=1,npoim
            posmaspm(k,p) = posmaspm(k,p) + randshift(k)
            if(posmaspm(k,p).LT.boxlims(1,k))then
               posmaspm(k,p) = posmaspm(k,p) + boxlen(k)
            endif
            if(posmaspm(k,p).GT.boxlims(2,k))then
               posmaspm(k,p) = posmaspm(k,p) - boxlen(k)
            endif
         enddo
      enddo

      
      write (*,4567) randshifttot(1),randshifttot(2),randshifttot(3)
     &     ,randshift(1),randshift(2),randshift(3)
     &     ,idummyrandom
 4567 format ('Random shift vector: ',6(f12.4),i12)
      
      return
      end subroutine boxshift
!---------------------------------------------------------------
#ifndef RECIPES
      double precision function ran2(idum)
      integer, intent(in):: idum

      stop 'Numerical Recipes ran2 routine not installed'

      end function ran2
!---------------------------------------------------------------
      double precision function erfunctionc(x)
      double precision, intent(in):: x

      stop 'Numerical Recipes erfunctionc routine not installed'
      end function erfunctionc
#endif
!---------------------------------------------------------------
      end module ewald
