      module grape
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
! This file contains the VINE module to do gravity using the         *
! GRAPE 5 special purpose hardware boards. The routines here         *
! are called from the forcelaw.F module.                             *
!                                                                    *
!  Written by Markus Wetzstein, Andy Nelson and Thorsten Naab        *
!*********************************************************************

#if GRAPE==5
#ifndef SELFFORCE
      grape5 module: Grape flag set, but self force flag is not'
#endif
#else
      grape5 module:No or wrong grape preprocessor flag exists
#endif
      use dimens
      use treevars
      use indivtsvars
      use constants 
      use perform
#if FAKE
      use fakegrape
#endif

      implicit none
      save

      private
      public lock_grape,unlock_grape,grape_tree,grape_dirsum


!
! Functions from grape library: Comment when fakegrape is used
!
#if !FAKE
      integer  g5_get_number_of_pipelines_per_board,
     &         g5_get_number_of_boards,
     &         g5_get_firstcluster
      external g5_get_number_of_pipelines_per_board,
     &         g5_get_number_of_boards,
     &         g5_get_firstcluster
#endif

!
! Number of boards, pipes per board, and total possible pipes
! 
      integer, parameter :: maxboards=1
      integer, parameter :: npipes_per_board=96
      integer, parameter :: maxpipes=maxboards*npipes_per_board 
      integer, parameter :: NJMAX=131072
!
! Data for j particles to send:
!
      double precision g5_xj(3,NJMAX),g5_mj(NJMAX)
! 
! Data for i particles to send, then recieve
!
      double precision g5_xi (3,maxpipes)
      double precision g5_acc(3,maxpipes)
      double precision g5_phi(  maxpipes)


      double precision, parameter:: gconst=ONE

      contains
      subroutine lock_grape()
!*************************************************************************
! This acquires the GRAPE-5 boards. Currently all boards from the 1st one*
! up to maxboards are acquired.                                          *
!                                                                        *
! Written by M. Wetzstein, Aug 2002, modified Dec. 2002                  *
!                                                                        *
!*************************************************************************

      integer i,ic,ifirstg5

      ifirstg5=g5_get_firstcluster()
      do i=1,maxboards
         ic=ifirstg5+i-1
         call g5_openMC(ic)
      enddo

      return
      end subroutine lock_grape
!-----------------------------------------------------
      subroutine unlock_grape()
!*************************************************************************
! This frees all the GRAPE-5 boards formerly acquired                    *
!                                                                        *
! Written by M. Wetzstein, Aug 2002 , modified Dec. 2002                 *
!                                                                        *
!*************************************************************************

      integer i,ic,ifirstg5

      ifirstg5=g5_get_firstcluster()
      do i=1,maxboards
         ic=ifirstg5+i-1
         call g5_closeMC(ic)
      enddo

      return
      end subroutine unlock_grape
!-----------------------------------------------------
      subroutine grape_tree(ibunch,isoft,eps2,npart,nact,listact,
     &                      nla0,nla,latoms,vel,apcorr,acc,poten)
!*************************************************************************
! This computes grav. accelerations on GRAPE-5. Boards have to be opened *
! outside this routine. Designed for use in multi-threaded environments, *
! but can be used in serial version, too.                                *
!                                                                        *
! Written by M. Wetzstein, T. Naab, Aug 2001                             *
! Modified by M. Wetzstein, Aug 2002, Dec. 2002                          *
! Modified by A. Nelson, Apr 2005 (code cleanup, here and in dirsum)     *
!                                                                        *
!*************************************************************************

! vel,ibunch aren't used here, they are just passed because
! the call sequence has to be the same for all grape_* routines
! in the different grape module files

      integer, intent(in):: ibunch,isoft,npart
      integer, intent(in):: nact,listact(nact)
      integer, intent(in):: nla0,nla,latoms(nla)
      double precision, intent(in):: eps2
      double precision, intent(in   ):: vel(ndim,npart)
      double precision, intent(inout):: apcorr(0:ndim,nact)
      double precision, intent(inout):: acc(ndim,npart)
      double precision, intent(inout):: poten(npart)

      integer npipes,nboards,ifirstg5
      integer i,ii,j,ic
      integer njlow,njhigh,nj
      integer nilow,nihigh,ni
      integer ioffs,ioffr
      double precision g5xmin,g5xmax
      double precision emmin,emmax

      double precision tme1,tme2,tmeio1,tmeio2
      double precision tmecalc1,tmecalc2
      character(len=20), parameter:: where =('grape5_tree')
      if(itme.ge.1)call cpu_time(tme1)


      if(isoft.ne.id_Plummer)then
         print *,'You have not selected Plummer softening'
         print *,'and you have selected grape. This is not'
         print *,'possible.'
         stop 'Not using Plummer softening with Grape'
      endif

      ifirstg5=g5_get_firstcluster()
      nboards=g5_get_number_of_boards()
      npipes =g5_get_number_of_pipelines_per_board()
      npipes =min(maxpipes,npipes*nboards)

      emmin = 1d100  ; emmax = 0d0
      g5xmin= 1d100  ; g5xmax=-1d100
      do j=1,nla
         i=latoms(j)
         emmin=min(emmin,treept(ims,i))
         emmax=min(emmax,treept(ims,i))
         g5xmax=max(g5xmax,treept(1,i),treept(2,i),treept(3,i))
         g5xmin=min(g5xmin,treept(1,i),treept(2,i),treept(3,i))
      enddo

      call setresolution(ifirstg5,eps2,1d2,emmin,emmax,g5xmin,g5xmax)

      njlow=1 ; njhigh=min(nla,NJMAX)
      do while(njlow.le.nla)
         if(itme.ge.1)call cpu_time(tmeio1)
!
! Cycle through groups of nj particles/nodes on the interaction list 
!
         nj=njhigh-njlow+1

         do i=njlow,njhigh
            ii=latoms(i)
            j=i-njlow+1
            g5_xj(1,j) = treept(1  ,ii)
            g5_xj(2,j) = treept(2  ,ii)
            g5_xj(3,j) = treept(3  ,ii)
            g5_mj(  j) = treept(ims,ii)
         enddo

         do i=1,maxboards
            ic=ifirstg5+i-1
            call g5_set_nMC (ic,  nj)
            call g5_set_xjMC(ic,0,nj,g5_xj)
            call g5_set_mjMC(ic,0,nj,g5_mj)
         enddo
         if(itme.ge.1)then
            call cpu_time(tmeio2)
            timing(itme_grpIO)=timing(itme_grpIO)+(tmeio2-tmeio1)
            call cpu_time(tmecalc1)
         endif
!
! Cycle through groups of ni particles that need forces 
!
         nilow=1 ; nihigh = min(nact,npipes)
         do while(nilow .le. nact)

            do i=nilow,nihigh
               ii=i - nilow + 1
               j=listact(i)
               g5_xi(1,ii)=treept(1,j)
               g5_xi(2,ii)=treept(2,j)
               g5_xi(3,ii)=treept(3,j)
            enddo      

            ni=nihigh - nilow + 1

            call g5_calculate_force_on_x(g5_xi,g5_acc,g5_phi,ni)

            do i=nilow,nihigh
               j= i - nilow + 1
               ii=linktreetosph(listact(i))
               acc  (1,ii) = acc  (1,ii) + gconst*g5_acc(1,j)
               acc  (2,ii) = acc  (2,ii) + gconst*g5_acc(2,j)
               acc  (3,ii) = acc  (3,ii) + gconst*g5_acc(3,j)
               poten(  ii) = poten(  ii) - gconst*g5_phi(  j)
            enddo

            nilow=nihigh + 1
            nihigh= min(nihigh + npipes,nact)
         enddo
         if(itme.ge.1)then
            call cpu_time(tmecalc2)
            timing(itme_grpcalc)=timing(itme_grpcalc)+
     &                                (tmecalc2-tmecalc1)
         endif

         njlow = njhigh + 1
         njhigh= min(njhigh + NJMAX, nla)
      enddo

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_grape)=timing(itme_grape)+(tme2-tme1)
         ncalls(itme_grape)=ncalls(itme_grape)+1
         ncalls(itme_grpIO)=ncalls(itme_grpIO)+1
         ncalls(itme_grpcalc)=ncalls(itme_grpcalc)+1
      endif

      return
      end subroutine grape_tree
!-----------------------------------------------------------------
      subroutine grape_dirsum(npart,isoft,eps2,posmas,vel,acc,poten)

      integer, intent(in):: npart,isoft
      double precision, intent(in   ):: eps2
      double precision, intent(in   ):: posmas(ndim1,npart)
      double precision, intent(in   ):: vel(ndim,npart)
      double precision, intent(inout):: acc(ndim,npart)
      double precision, intent(inout):: poten(npart)

      integer nboards,npipes,ifirstg5
      integer ic
      integer i,j,ii
      integer njlow,njhigh,nj
      integer nilow,nihigh,ni
      double precision g5xmin,g5xmax
      double precision emmin,emmax

      double precision tme1,tme2,tmeio1,tmeio2
      double precision tmecalc1,tmecalc2
      character(len=20), parameter:: where =('grape5_dirsum')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      if(isoft.ne.id_Plummer)then
         print *,'You have not selected Plummer softening'
         print *,'and you have selected grape. This is not'
         print *,'possible.'
         stop 'Not using Plummer softening with Grape'
      endif

      ifirstg5=g5_get_firstcluster()
      nboards=g5_get_number_of_boards()
      npipes =g5_get_number_of_pipelines_per_board()
      npipes =min(maxpipes,npipes*nboards)

      emmin = 1d100  ; emmax = 0d0
      g5xmin= 1d100  ; g5xmax=-1d100
      do i=1,npart
         emmin=min(emmin,posmas(ims,i))
         emmax=min(emmax,posmas(ims,i))
         g5xmax=max(g5xmax,posmas(1,i),posmas(2,i),posmas(3,i))
         g5xmin=min(g5xmin,posmas(1,i),posmas(2,i),posmas(3,i))
      enddo

      call setresolution(ifirstg5,eps2,1d2,emmin,emmax,g5xmin,g5xmax)

      njlow=1 ; njhigh=min(npart,NJMAX)
      do while(njlow.le.npart)
         if(itme.ge.1)call cpu_time(tmeio1)
!
! Cycle through sets of nj particles that contribute forces
!
         nj=njhigh-njlow+1
         do i=njlow,njhigh
            j= i - njlow + 1
            g5_xj(1,j)=posmas(1  ,i)
            g5_xj(2,j)=posmas(2  ,i)
            g5_xj(3,j)=posmas(3  ,i)
            g5_mj(  j)=posmas(ims,i)
         enddo

         do i=1,maxboards
            ic=ifirstg5+i-1
            call g5_set_nMC (ic,  nj)
            call g5_set_xjMC(ic,0,nj,g5_xj)
            call g5_set_mjMC(ic,0,nj,g5_mj)
         enddo
         if(itme.ge.1)then
            call cpu_time(tmeio2)
            timing(itme_grpIO)=timing(itme_grpIO)+(tmeio2-tmeio1)
            call cpu_time(tmecalc1)
         endif

         nilow=1 ; nihigh=min(nactive,npipes)
         do while(nilow.le.nactive)
!
! Cycle through groups of ni particles that need forces
!
            do i=nilow,nihigh 
               ii=i-nilow+1
               j=iactive(i)
               g5_xi(1,ii)=posmas(1,j)
               g5_xi(2,ii)=posmas(2,j)
               g5_xi(3,ii)=posmas(3,j)
            enddo

            ni=nihigh-nilow+1

            call g5_calculate_force_on_x(g5_xi,g5_acc,g5_phi,ni)

            do i=nilow,nihigh
               ii=i-nilow+1
               j=iactive(i)
               acc(1,j)=acc(1,j) + gconst*g5_acc(1,ii)
               acc(2,j)=acc(2,j) + gconst*g5_acc(2,ii)
               acc(3,j)=acc(3,j) + gconst*g5_acc(3,ii)
               poten(j)=poten(j) - gconst*g5_phi(  ii)
            enddo

            nilow =nihigh + 1
            nihigh=min(nactive, nihigh + npipes)
         enddo
         if(itme.ge.1)then
            call cpu_time(tmecalc2)
            timing(itme_grpcalc)=timing(itme_grpcalc)+
     &                                (tmecalc2-tmecalc1)
         endif

         njlow  =     njhigh + 1
         njhigh = min(njhigh + NJMAX,npart)
      enddo
      
      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_grape)=timing(itme_grape)+(tme2-tme1)
         ncalls(itme_grape)=ncalls(itme_grape)+1
         ncalls(itme_grpIO)=ncalls(itme_grpIO)+1
         ncalls(itme_grpcalc)=ncalls(itme_grpcalc)+1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine grape_dirsum
!-----------------------------------------------------------------
      subroutine setresolution(ifirstg5,eps2,emfactor,
     &                           emmin,emmax,g5xmin,g5xmax)
!
! set resolution for the boards, use emfactor to make certain
! that the particles get into the correct mass range.
!
! G5 users manual p20 says m/g5mmin < (r/rmin)**2 to avoid overflow 
!    where: 
!            m    = largest mass sent to board
!            rmin = 1d-7*(g5xmax-g5xmin)
!            r    = largest distance between two particles
!                   in general: r=sqrt(3)*(g5xmax-g5xmin)
!
!    So need:   m/g5mmin < (sqrt(3)/1d-7)**2 for stability
!
      integer, intent(in):: ifirstg5
      double precision, intent(in):: emfactor,emmin,emmax
      double precision, intent(in):: eps2,g5xmin,g5xmax

      integer i,ic
      double precision g5mmin,g5eps

      g5mmin=emmin/emfactor

      if(emmax/g5mmin .gt. (sqrt(3d0)/1d-7)**2  )then
         print *,'Grape5 dynamic range exceeded'
         stop'Dynamic Range exceeded'
      endif
!
! send resolution to boards
!
      g5eps=sqrt(eps2)
      do i=1,maxboards
         ic=ifirstg5+i-1
         call g5_set_rangeMC(ic,g5xmin,g5xmax,g5mmin)
         call g5_set_eps_to_allMC(ic,g5eps)
      enddo

      return
      end subroutine setresolution
!-----------------------------------------------------------------
      end module grape
