      module io
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
! This file contains the VINE module to do input and output          *
! to and from files, including various auxilliary functions like     *
! determining the file names, and setting the units.                 *
!                                                                    *
! This module may require the FXDR libraries, which are              *
!  available from                                                    *
!  http://meteora.ucsd.edu/~pierce/fxdr_home_page.html               *
!  depending on whether or not you want to do io across              *
!  multiple systems portably.                                        *
!                                                                    *
!  Externally visible routines are:                                  *
!                                                                    *
!    readfile:  Reads in initial state of system from a text         *
!               input file (insph) and a dumpfile.                   *
!      Arguments:                                                    *
!        idump: what dump number we read                             *
!        konv2code: whether to convert to code units or not          *
!                                                                    *
!    writefiles:                                                     *
!      Arguments:                                                    *
!        idump: what dump number we want to write                    *
!        makecheckpoint: whether we should make a checkpoint dump    *
!                        file or a real one                          *
!        konv2phys: whether to convert to physical units or not      *
!                   in the dump file.                                *
!                                                                    *
!    setbase_time: Set the base time and time step number from       *
!                  which future restart/analysis times and time      *
!                  step numbers are calculated.                      *
!      Arguments:                                                    *
!        readfromchkpnt: Whether or not we read our restart dump     *
!                  from a checkpoint dump or from a regular dump.    *
!        reset:  Whether or not we want to override the previously   *
!                saved (in the tmp dump for example) dump time       *
!                and calculate it again                              *
!        dtfullstep: The input value for how much time we want       *
!                between output dumps                                *
!        dtdmp:  The input value for how much time we want between   *
!                output dumps. See comments below                    *
!        tlastdmp: The time of the last regular dump                 *
!        tnextdmp: The stored time of the next regular dump, used    *
!                  unless overridden.                                *
!        t       : The time now.                                     *
!        idump   : The dump number now                               *
!        t0      : The reference time for dumps                      *
!        idump0  : The reference dump number for dumps               *
!                                                                    *
!                                                                    *
!    setnextout_time: Set the next time to make an output/restart    *
!                     dump.                                          *
!                                                                    *
!      Arguments:                                                    *
!        maxsteps: The maximum number of time steps or dt_out        *
!           steps to allow before outputting.                        *
!        idump0  : Reference dump number (first since this           *
!                  executable started running)                       *
!        nstepnow: Number of time step, right now.                   *
!        idump   : Current dump number                               *
!        t0      : Reference time, from which we calculate           *
!                  future dump times.                                *
!        t       : Current time.                                     *
!        dt_out  : Time between output dumps                         *
!        tlast   : Time of last output dump                          *
!        tnext   : Time of next output dump                          *
!        nextoutstep: Next time step number for output.              *
!                                                                    *
!    setnextana_time: Set the next time to make an analysis output   *
!                                                                    *
!      Arguments:                                                    *
!        maxsteps: The maximum number of time steps allowed before   *
!                  making a dump                                     *
!        idump0  : Reference dump number (first since this           *
!                  executable started running)                       *
!        nstepnow: Number of time step, right now.                   *
!        idump   : Current dump number                               *
!        t0      : Reference time, from which we calculate           *
!                  future dump times.                                *
!        t       : Current time.                                     *
!        dt_ana  : Time between analysis dumps                       *
!        dtfull  : Allowed time increment between dumps              *
!                  (should be set to dt_ana in caller)               *
!        tlast   : Time of last analysis dump                        *
!        tnext   : Time of next analysis dump                        *
!        nextanastep: Next time step number for analysis.            *
!                                                                    *
!    pmstore: Store point mass and system information in a small     *
!             cache for a write to disk at some later time           *
!             (Thereby improve write efficiency with larger writes)  *
!      Arguments:                                                    *
!        i12step: Which integrator we are using. Needed to           *
!                 determine which pm particles are active this       *
!                 step                                               *
!    pmoutput: Do the output of the PM etc data to files for each    *
!              point mass.                                           *
!                                                                    *
!                                                                    *
!                                                                    *
!  Written by Andy Nelson, Markus Wetzstein and Thorsten Naab        *
!*********************************************************************
      use dimens
      use partvars
      use miscvars
      use indivtsvars
      use constants
      use stubs
      use perform

      implicit none
      private
      public options,lunit,getfname,rdump
      public readfile,writefiles
      public setbase_time,setnextout_time,setnextana_time
      public pmstore, pmoutput


      include 'INCLUDE/hwinfo'

!
! for storing header information in dump files
!
      integer, parameter:: iheadlength=1000
      double precision dheader(iheadlength)
      integer iheader(iheadlength)
!
! These are the indices of various values saved in
! the header part of the dump file. 
!
      integer,parameter::id_iheadlen= 1, id_npart  = 2, id_npart_sph = 3
      integer,parameter::id_nstep   = 4, id_idump  = 5, id_makechkpnt= 6
      integer,parameter::id_idbg    = 7, id_itme   = 8, id_ibctype   = 9
      integer,parameter::id_indts   =10, id_iusebin=11, id_ipres     =12
      integer,parameter::id_ipdv    =13, id_ieos   =14, id_icool     =15
      integer,parameter::id_iheat   =16, id_ivisc  =17, id_ivtime    =18
      integer,parameter::id_ibals   =19, id_ishok  =20, id_igrav     =21
      integer,parameter::id_isoft   =22, id_imac   =23, id_ihts      =24
      integer,parameter::id_iexpand =25, id_ncdumps=26,
     &                                               id_maxclumpsize=27
      integer,parameter::id_intgrtr =28, id_ishift =29, id_ndim     =30
      integer,parameter::id_io_fmt  =31, id_iunits =32,
     &                                               id_maxbunchsize=33
      integer,parameter::id_npoim   =34, id_ndt_ana=35, id_maxbuild =36
      integer,parameter::id_fullextrap=37,id_revise=38, id_n_dtmax  =39
      integer,parameter::id_n_ana    =40
      integer,parameter::id_lastint1=41  !1 after last id for integers 

      integer, parameter :: id_lf_dkd=1,id_rk=2,id_lf_kdk=3

#if defined(LF)
      integer, parameter :: id_myintg=id_lf_dkd
#elif defined(RK)
      integer, parameter :: id_myintg=id_rk
#else
      parameter(don't-know-my-integrator-in-io.F') 
#endif

      integer, parameter::id_t       = 1,id_dtmax   = 2,id_deltnew  = 3
      integer, parameter::id_gamma   = 4,id_ekin    = 5,id_egrav    = 6
      integer, parameter::id_etherm  = 7,id_ascale  = 8,id_cosbox   = 9
      integer, parameter::id_vlength =10,id_alfstar =11,id_betastar =12
      integer, parameter::id_tol     =13,id_cfl     =14,id_dhmax    =15
      integer, parameter::id_treeacc =16,id_sepmax  =17,id_hmin     =18
      integer, parameter::id_eps     =19,id_dtinit  =20,id_tstop    =21
      integer, parameter::id_dt_out  =22,id_uconst  =23,id_xmas     =24
      integer, parameter::id_ylen    =25,id_umin    =26,id_rcore    =27
      integer, parameter::id_h_0     =28,id_omegam  =29,id_gmasslim =30
      integer, parameter::id_hmax    =31,id_uexpon  =32,id_ftol     =33
      integer, parameter::id_clhfrac =34,id_xmin    =35,id_xmax     =36
      integer, parameter::id_ymin    =37,id_ymax    =38,id_zmin     =39
      integer, parameter::id_zmax    =40
      integer, parameter::id_eosK    =41,id_rhozero =42,id_ftolpm   =43
      integer, parameter::id_tolpm   =44,id_tnextout=45,id_alfmax   =46
      integer, parameter::id_vtol    =47,id_vtolpm  =48,id_dtmin    =49
      integer, parameter::id_tlastout=50,id_dt_ana  =51,id_bsepmax  =52
      integer, parameter::id_taucool =53
      integer, parameter::id_lastreal1=54 !1 after last id for real numbers 

!
! IO formats. Standard fortran, XDR or HDF
!
      integer, parameter::id_ftn=0, id_xdr=1, id_hdf=2
!
! Functions for the XDR libraries 
!
#if XDR
      integer initxdr,ixdrclose,ixdrimat,ixdrdmat,ixdrrewind
#endif

      contains
!---------------------------------------------------------------------
      subroutine readfile(idump,konv2code)
!**********************************************************************
!                                                                     *
! This routine reads a dump file for restarting the simulation        *
!                                                                     *
! Serves as a driver for the different getfname and rdump routines,   *
! and can be expanded to other files if necessary.                    *
!                                                                     *
! Written by M. Wetzstein, 2006                                       *
! Modified by A. Nelson    2006                                       *
!                                                                     *
!**********************************************************************
      integer, intent(in ):: konv2code
      integer, intent(out):: idump

      integer numcpus

      double precision tme1,tme2
      character(len=20), parameter:: where =('readfile')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)
!
! Get the name of the regular dump file, read it
! and check whether it is sane or not.
!
      call rdump(idump)
      call dumpfile_sanitycheck
!
! Get the last dump number idump from a file, because it will be 
! different than what the insph file has if we are reading from
! a checkpoint dump (idump<0).
!
      if(idump.lt.0)then
         readfromchkpnt=.true.
         call getlastdumpnr(idump)
      endif
!
! Print out some information to the screen
!
      print *,' '
      print *,' '
      print *,' '
      print *,'The time is:',t
      print *,'The maximum timestep is:',dtmax
      print *,'The minimum timestep is:',dtmin
      print *,'Number of Nbody particles is:',npart-npart_sph
      print *,'Number of SPH   particles is:',npart_sph
      print *,'Number of PM    particles is:',npoim
      call numprocessors(numcpus)
      print *,'Number of processors used is:',numcpus
      print *,' '
      print *,' '
      print *,' '
!
! Print out some cosmology information
!
      if(iexpand .eq. 1)then
         print *,' '
         print *,' '
         print *,'Start at redshift',redshift
         print *,'Simulation of a',cosbox,' Mpc cube'
         print *,'Omega_m =',omegam,', Lambda =',omegal
         print *,' '
         print *,' '
      endif
!
! Scale to code units if desired.
! 
      if(konv2code.eq.1)call phys2code

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_rdump)  =timing(itme_rdump)+tme2-tme1
         ncalls(itme_rdump)  =ncalls(itme_rdump)+1
         timing(itme_IOcalls)=timing(itme_IOcalls)+tme2-tme1
         ncalls(itme_IOcalls)=ncalls(itme_IOcalls)+1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine readfile
!--------------------------------------------------------------------
      subroutine writefiles(idump,makecheckpoint,konv2phys)
!**********************************************************************
!                                                                     *
! This routine writes the dump files and serves as a driver for       *  
! any other dump routines that you might want to include.             *
!                                                                     *
! Written by M. Wetzstein, 2006                                       *
! Modified by A. Nelson    2006                                       *
!                                                                     *
!**********************************************************************
      logical, intent(in):: makecheckpoint
      integer, intent(in):: idump,konv2phys

      character(250) absnrfile
      double precision tme1,tme2
      character(len=20), parameter:: where =('writefiles')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)
!        
! make a dump file name
!      
      if(makecheckpoint)then
         dumpfile=chkpntdumpname
         print *,'Writing checkpoint dump file'
      else
         call getfname(idump,1,1)

         if(makechkpnt.eq.1)then
!
! This is a regular dump and we are allowing checkpoint dumps. Therefore
! we must write the number of this regular dump file into nrfile,
! for later restart from a checkpoint dump.
!
            absnrfile = trim(absolutepath) // nrfile
            open(iterm,file=absnrfile,status='replace')
            write(iterm,*)idump
            close(iterm)
         endif
      endif

!
! Convert to physical units, fill dump file header,
! dump out, then convert back to internal units.
!
      if(konv2phys.eq.1)call code2phys
      call fillheader(idump)
      call wdump(idump,makecheckpoint)
      if(konv2phys.eq.1)call phys2code
!
! Drain the cache of saved up `on the fly' analysis data
! for point masses. If you have other saved analysis stuff
! that needs to by synchronized when restart happens, then
! dump it here too.
!
      if(npoim.gt.0)call pmoutput

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_wdump)=timing  (itme_wdump)+tme2-tme1
         ncalls(itme_wdump)=ncalls  (itme_wdump)+1
         timing(itme_IOcalls)=timing(itme_IOcalls)+tme2-tme1
         ncalls(itme_IOcalls)=ncalls(itme_IOcalls)+1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine writefiles
!--------------------------------------------------------------------
      subroutine setbase_time(readfromchkpnt,reset,dtfullstep,
     &                        dtdmp,tlastdmp,tnextdmp,t ,idump,
     &                        t0,idump0)
!**************************************************************
! Compute the reference times, from which output dump times   *
! are calculated. Account for the possibility that we are     *
! starting from a checkpoint dump and the dump time increment *
! has changed.                                                *
!                                                             *
! t0: Reference time from which to calculate future dump      *
!     times.                                                  *
! idump0: The initial dump number. Can be either a restart    *
!         dump index or some other sort of index if we are    *
!         working on times for on the fly analysis.           *
!                                                             *
!                                                             *
! The reason for having dtdmp and dtfullstep is that          *
!    first we want to be able to have restart dumps in        *
!       increments of the dtmax variable that appears as      *
!       the highest bin in the individual time step scheme.   *
!    second we want to be able to use different increments    *
!       for any analysis code, for which perhaps we want to   *
!       be able to output at smaller intervals than a full    *
!       maximum time step.                                    *
!                                                             *
! Setting dtdmp.ne.dtfullstep in the caller means forcing     *
! dump intervals at integer multiples of dtfullstep.          *
! For example, you would do this if you had a dtmax value     *
! that was small compared to the dump interval dt_out that    *
! you wanted. Then you would set dtfullstep=dtmax and         *
! dtdmp=dt_out (see insph file).                              *
!                                                             *
! Setting dtdmp.eq.dtfullstep in the caller means setting     *
! arbitrary dump time intervals.                              *
! For example, you would do this if you wanted analysis dump  *
! information on smaller intervals than full dtmax steps.     *
! Then you would set dtfullstep=dt_ana and dtdmp=dt_ana in    *
! the caller.                                                 *
!                                                             *
! Later calls to setnextdumptime using the references         *
! calculated and returned from this routine will give you     *
! later dump times as you specify.                            *
!                                                             *
! This routine and its cousins (setnext*time) are tested      *
! but not well tested.                                        *
!                                                             *
! Subroutine by A. Nelson                                     *
!**************************************************************
      logical, intent(in  ):: readfromchkpnt,reset
      integer, intent(in  ):: idump
      double precision, intent(in):: dtfullstep,dtdmp
      double precision, intent(in):: t,tlastdmp,tnextdmp

      integer, intent(out ):: idump0
      double precision, intent(out):: t0

      integer id
      double precision deltdump



      idump0=idump

      if(.not.readfromchkpnt)then
!
! Easy. Just set base time to now
!
         t0 =t

      else
!
! Need to do special things when reading from a checkpoint dump
!
         if( (t-tnextdmp).gt.ZERO)then     !next dump time in the past
                                           !something is wrong

            print *,'VINE just read data from a checkpoint restart'
            print *,'dump, but this dump file thinks its next output'
            print *,'time is in the past. Resetting the reference time'
            print *,'to the current time'

            t0 =t
         else

            if(iusebin.eq.0)then
               deltdump=dtdmp
            else
               deltdump= dtmax*nint(dtdmp/dtfullstep)
            endif

            if(.not.reset)then  
            
               t0=tnextdmp
            
            else
!            
! Don't want to use previously saved next dump time for
! whatever reason. Most commonly because we want to change
! the dump time interval. If so, it might be smaller and therefore
! t0+dt_out might yield a time in the past. It might also be 
! larger and still some time in the future. This fix
! accomodates both possibilities.
!

               t0 = tlastdmp
               id = 1
               do while( (t0+dble(id)*deltdump).lt.t )
                  id=id+1
               enddo
               t0 = t0 + dble(max(0,id-1))*deltdump

            endif

            t0 = t0 - deltdump  !subtract because setnext always
                                !adds, but we don't want this the
                                !first time. 
         endif
      endif

      return
      end subroutine setbase_time
!--------------------------------------------------------------------
      subroutine setnextout_time(maxsteps,idump0,nstepnow,idump,t0,
     &                           t,dt_out,tlast,tnext,nextoutstep)
!**************************************************************
! Compute the next analysis dump time.                        *
!                                                             *
! Notes:                                                      *
!   1) the unbinned ('continuous') timestep option can have   *
!      dumps at any time, but the binned option is restricted *
!      to dumps at the end of dtmax increments                *
!      -For binned timesteps: Use nextanastep to define the   *
!       maximum number of dtdmp timesteps to integrate over   *
!       before dumping again.                                 *
!      -For unbinned timesteps, nextanastep to define the     *
!       maximum number of simple timesteps to integrate over  *
!       before dumping again.                                 *
!                                                             *
!   2) Cosmology simulations can have their own dump times    *
!      set to some particular redshift.                       *
!      -Do cosmo dump stepping separately below               *
!                                                             *
!  Arguments                                                  *
!    maxsteps: The maximum number of time steps or dt_out     *
!       steps to allow before outputting.                     *
!    idump0  : Reference dump number (first since this        *
!              executable started running)                    *
!    nstepnow: Number of time step, right now.                *
!    idump   : Current dump number                            *
!    t0      : Reference time, from which we calculate        *
!              future dump times.                             *
!    t       : Current time.                                  *
!    dt_out  : Time between output dumps                      *
!    tlast   : Time of last output dump                       *
!    tnext   : Time of next output dump                       *
!    nextoutstep: Next time step number for output.           *
!                                                             *
! Note that if you set maxsteps too small, then nextoutstep   *
! will be set by maxsteps, but tnext will continue to         *
! increment by dt_ana each dump. It will therefore run out    *
! farther and farther ahead of the current time.              *
!                                                             *
! Note2: originally setnextana_time and setnextout_time       *
!        were suppused to be one routine, callable with       *
!        different data for the arguments so that you         *
!        could set as many different extra analysis routine   *
!        dump times as you wanted. That was the purpose of    *
!        having dtfull (like dtmax in setnextout_time)        *
!        for example.                                         *
!        This goal wasn't realized, and in the current        *
!        code, is also broken by the saved `first' variable   *
!        below.                                               *
!                                                             *
! Subroutine by A. Nelson                                     *
!                                                             *
!**************************************************************
#if COSMO
      use cosmo, only: cosmo_dumptime
#endif

      integer, intent(in   ):: idump0,nstepnow,idump,maxsteps
      integer, intent(inout):: nextoutstep
      double precision, intent(in   ):: t,t0,dt_out
      double precision, intent(inout):: tnext
      double precision, intent(  out):: tlast

      logical, save::first=.true.
      integer nd,ns

      if(iexpand.eq.0)then          !Non-cosmo
!
!    If we did not hit the number of steps limitation,
!        (i.e.nextoutstep>nstepnow) then we are here because it
!        was the right time for a dump. So increment both tnext 
!        and nextoutstep. 
!    Otherwise we didn't get all the way to our expected next dump
!        time. So leave tnext alone and increment only nextoutstep
!
         tlast = tnext

         if(iusebin.eq.0)then
            if(nextoutstep.ge.nstepnow .or. first)then
               tnext       = t0 + dble(idump-idump0+1)*dt_out
               nextoutstep = nstepnow + maxsteps
            else
               nextoutstep = nstepnow + maxsteps
            endif

         else

            if(nextoutstep.ge.nstepnow .or. first)then
               nd    = (idump-idump0+1)*nint(dt_out/dtmax)
               tnext = t0 + dtmax*dble(min(nd,maxsteps))

               ns =nint((tnext-t)/dtmax)
               nextoutstep = nstepnow + min(ns,maxsteps)
            else
               ns =nint((tnext-t)/dtmax)
               nextoutstep = nstepnow + min(ns,maxsteps)
            endif

         endif

      else
!===========================================
! Cosmo sets its own dump times
!
         tlast = tnext
         call cosmo_dumptime(nextoutstep,tnext)

      endif

      first=.false.

      return
      end subroutine setnextout_time
!--------------------------------------------------------------------
      subroutine setnextana_time(maxsteps,idump0,nstepnow,idump,t0,
     &                           t,dt_ana,dtfull,
     &                           tlast,tnext,nextanastep)
!**************************************************************
! Compute the next analysis dump time.                        *
!                                                             *
! Notes:                                                      *
!   1) the unbinned ('continuous') timestep option can have   *
!      dumps at any time, but the binned option is restricted *
!      to dumps at the end of dtmax increments                *
!      -For binned timesteps: Use nextanastep to define the   *
!       maximum number of dtdmp timesteps to integrate over   *
!       before dumping again.                                 *
!      -For unbinned timesteps, nextanastep to define the     *
!       maximum number of simple timesteps to integrate over  *
!       before dumping again.                                 *
!                                                             *
!   2) Cosmology simulations can have their own dump times    *
!      set to some particular redshift.                       *
!      -Do cosmo dump stepping separately below               *
!                                                             *
!  Arguments                                                  *
!    maxsteps: The maximum number of time steps or dt_ana     *
!       steps to allow before outputting.                     *
!    idump0  : Reference dump number (first since this        *
!              executable started running)                    *
!    nstepnow: Number of time step, right now.                *
!    idump   : Current dump number                            *
!    t0      : Reference time, from which we calculate        *
!              future dump times.                             *
!    t       : Current time.                                  *
!    dt_ana  : Time between analysis dumps                    *
!    dtfull  : Allowed time increment between dumps           *
!              (set to dt_ana in caller)                      *
!    tlast   : Time of last analysis dump                     *
!    tnext   : Time of next analysis dump                     *
!    nextanastep: Next time step number for analysis.         *
!                                                             *
! Note that if you set maxsteps too small, then nextanastep   *
! will be set by maxsteps, but tnext will continue to         *
! increment by dt_ana each dump. It will therefore run out    *
! farther and farther ahead of the current time.              *
!                                                             *
! Note2: originally setnextana_time and setnextout_time       *
!        were suppused to be one routine, callable with       *
!        different data for the arguments so that you         *
!        could set as many different extra analysis routine   *
!        dump times as you wanted. That was the purpose of    *
!        having dtfull (like dtmax in setnextout_time)        *
!        for example.                                         *
!        This goal wasn't realized, for reasons I forget now. *
!                                                             *
! Subroutine by A. Nelson                                     *
!                                                             *
!**************************************************************
#if COSMO
      use cosmo, only: cosmo_dumptime
#endif

      integer, intent(in   ):: idump0,nstepnow,idump,maxsteps
      integer, intent(inout):: nextanastep
      double precision, intent(in   ):: t,t0,dt_ana,dtfull
      double precision, intent(inout):: tnext
      double precision, intent(  out):: tlast

      logical, save::first=.true.
      integer nd,ns

      if(iexpand.eq.0)then          !Non-cosmo
!
!    If we did not hit the number of steps limitation,
!        (i.e.nextanastep>nstepnow) then we are here because it
!        was the right time for a dump. So increment both tnext 
!        and nextanastep. 
!    Otherwise we didn't get all the way to our expected next dump
!        time. So leave tnext alone and increment only nextanastep
!

         tlast = tnext

         if(iusebin.eq.0)then
!
! Unbinned ts: 
!
            if(nextanastep.ge.nstepnow .or. first)then
               tnext       = t0 + dble(idump-idump0+1)*dt_ana
               nextanastep = nstepnow + maxsteps
            else
               nextanastep = nstepnow + maxsteps
            endif
         else
!
! Binned (with or without indivts) ts: 
!
            if(t.ge.tnext .or. first)then                 !don't remember why is this difft
                                                          !from the out time version. arg.

               nd    = (idump-idump0+1)*nint(dt_ana/dtfull)
               tnext = t0 + dt_ana*dble(nd)

               ns =nint((tnext-t)/dtfull)
               nextanastep = nstepnow + min(ns,maxsteps)
            else

               ns =nint((tnext-t)/dtfull)
               nextanastep = nstepnow + min(ns,maxsteps)
            endif

         endif
         first=.false.

      else
!===========================================
! Cosmo sets its own dump times
!
         tlast = tnext
         call cosmo_dumptime(nextanastep,tnext)

      endif

      return
      end subroutine setnextana_time
!-----------------------------------------------------------------
      subroutine getfname(idump,icheck,iloud)
!************************************************************
!                                                           *
!  this routine determines the filename for dump files.     *
!                                                           *
!  For the first time through, if idump<0, assume we are    *
!  reading a checkpoint dump, and that the abs(idump) gives *
!  the last normal dump.                                    *
!                                                           *
!  If the abs(idump) dump does not exist, look (later) for  *
!  a file called 'lastdumpnr' and read idump from that      *
!  file. This is useful for batch queues etc. For example,  *
!  you might want to have checkpoint dumps and not know     *
!  which real dump you are going to get to before the queue *
!  time ends. So set idump=-10000 or something in your      *
!  insph file                                               *
!************************************************************
      integer, intent(in):: idump,icheck,iloud

      character(len=len_dumpdigits) dumpnum

      logical tobe
      logical, save:: first=.true.
      character(len=20), parameter:: where =('getfname')
      if(idbg.ge.2 .and. iloud.eq.1)print *,where

      if(first)then
         first=.false.
         if(idump.lt.0)then
            dumpfile = trim(absolutepath) // trim(chkpntdumpname)
         else
            write(dumpnum,300)idump
            dumpfile = trim(absolutepath) // trim(basenm) // dumpnum
            inquire(file=trim(dumpfile),exist=tobe)
            if(idump.eq.0 .and. .not.tobe)then
               print *,' '
               print *,'Dump file does not exist. '
               print *,'Tried to open file: ',trim(dumpfile)
               print *,' '
               print *,'Will try to open checkpoint dump file...'
               dumpfile = trim(absolutepath) // trim(chkpntdumpname)
               inquire(file=trim(dumpfile),exist=tobe)
               if(.not.tobe)then
                  print *,' '
                  print *,'but a checkpoint dump does not exist either!'
                  print *,'Tried to open file:',trim(dumpfile)
                  stop'ERROR: getfname-no dump file found'
               endif
            endif
         endif
      else
         write(dumpnum,300)idump
         dumpfile=trim(absolutepath) // trim(basenm) // dumpnum
         inquire(file=dumpfile,exist=tobe)
         if(tobe .and. icheck.eq.1)then
            print *,' '
            print *,'dump file exists:',trim(dumpfile)
            print *,' '
            stop'getfname: Dump file exists. Cannot overwrite'
         endif
      endif
      if(iloud.eq.1)then
         print *,' '
         print *,'dump file is:',trim(dumpfile)
         print *,' '
      endif

      if(idbg.ge.3 .and. iloud.eq.1)then
         print *,'done ',trim(where)
      endif

300   format(i4.4)   !would be nice if this depended on 
                     !len_dumpdigits explicitly somehow

      return
      end subroutine getfname
!---------------------------------------------------------------------
      subroutine pmoutput
!*********************************************************************
!                                                                    *
! output point mass information to a file                            *
!                                                                    *
! These files can be read by offline analysis routines.              *
!                                                                    *
!*********************************************************************

      character(len=2) cnum
      integer i,id,irecl,istore

      double precision tme1,tme2
      character(len=20), parameter:: where=('pmoutput')
      if(npoim.eq.0)return
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)


      do i=1,npoim

         write(cnum,'(i2.2)')i
         pmfile(i)= trim(basenm) // '.pm' // cnum

         irecl=2*( 4*ndim+ 2 + 3 )
         open(idisk4,file=pmfile(i),form='unformatted',
     &            access='direct',recl=IRECSIZE*irecl)

         do istore=1,nstore(i)

            write(idisk4,rec=(nsteppm(i)-nstore(i)+istore))
     &          (posmstore (id,i,istore), id=1,ndim1),
     &          (cmstore   (id,i,istore), id=1,ndim),
     &          (cmtotstore(id,i,istore), id=1,ndim),
     &          (velstore  (id,i,istore), id=1,ndim),
     &          (spinstore (id,i,istore), id=1,3),
     &          tmestore(1,istore)

         enddo
         close(idisk4)
         nstore(i)=0
      enddo

      if(idbg.ge.2)print *,'done ',where
      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_pmio)=timing(itme_pmio)+(tme2-tme1)
         ncalls(itme_pmio)=ncalls(itme_pmio)+1
         timing(itme_iocalls)=timing(itme_iocalls)+(tme2-tme1)
      endif

      return
      end subroutine pmoutput
!------------------------------------------------------------------
      subroutine pmstore(i12step)
!*********************************************************************
! Store pm information in a small cache for later output.            *
! Idea is to improve efficiency of IO by doing it in larger chunks.  *
!*********************************************************************

      integer, intent(in):: i12step

      integer i,j,loopindx
      double precision fracTS
      character(len=20), parameter:: where =('pmstore')

      if(npoim.eq.0)return
      if(idbg.ge.2)print *,where

      if(indts.eq.1)then
         fracTS=dble(lastts_end)*dtincr
      else
         if(iusebin.eq.0)then
            fracTS=ZERO         !time update done already
         else
            fracTS=dtincr*dble(lastts_end)
         endif
      endif

      loopindx=nactivepm
      if(i12step.eq.2)loopindx=nactivefpm

      do j=1,loopindx
         i=iactivepm(j)
         nstore(i)=nstore(i)+1
         tmestore  (    i,nstore(i))=   t_unit*(t + fracTS)
         cmstore   (1  ,i,nstore(i))=   d_unit*dumcmpart(1)
         cmstore   (2  ,i,nstore(i))=   d_unit*dumcmpart(2)
         cmtotstore(1  ,i,nstore(i))=   d_unit*dumcmtot (1)
         cmtotstore(2  ,i,nstore(i))=   d_unit*dumcmtot (2)

         posmstore (ims,i,nstore(i))=  em_unit*dumposmaspm(ims,i)
         posmstore (1  ,i,nstore(i))=   d_unit*dumposmaspm(1  ,i)
         posmstore (2  ,i,nstore(i))=   d_unit*dumposmaspm(2  ,i)
         velstore  (1  ,i,nstore(i))= vel_unit*dumvelpm   (1  ,i)
         velstore  (2  ,i,nstore(i))= vel_unit*dumvelpm   (2  ,i)
         spinstore (3  ,i,nstore(i))=amom_unit*spinpm     (3  ,i)
#if NDIM==3
         cmstore   (3  ,i,nstore(i))=   d_unit*dumcmpart(3)
         cmtotstore(3  ,i,nstore(i))=   d_unit*dumcmtot (3)
         posmstore (3  ,i,nstore(i))=   d_unit*dumposmaspm(3  ,i)
         velstore  (3  ,i,nstore(i))= vel_unit*dumvelpm   (3  ,i)
         spinstore (1  ,i,nstore(i))=amom_unit*spinpm     (1  ,i)
         spinstore (2  ,i,nstore(i))=amom_unit*spinpm     (2  ,i)
#endif
      enddo

      if(idbg.ge.2)print *,'done ',where

      return
      end subroutine pmstore
!------------------------------------------------------------------
      subroutine lunit 
!************************************************************
!                                                           *
!  this routine attributes logical unit numbers.            *
!                                                           *
!************************************************************
      character(len=20), parameter:: where =('lunit')

      if(idbg.ge.2)print *,where

      iterm=50
      idisk1=11
      idisk2=12
      idisk3=13
      idisk4=14

      return
      end subroutine lunit 
!---------------------------------------------------------------------
      subroutine options(idump)
!************************************************************
!                                                           *
!  this reads in all of the options desired for the run     *
!                                                           *
!************************************************************
       integer, intent(out):: idump

      character(len=80) txtxt
      character(len=38) txt
      character(len=20) label, label1
      integer i
      logical tobe
!
!--open input file
!
      inquire(file='insph',exist=tobe)
      if(.not.tobe)stop 'no insph file exists'
      open(iterm,file='insph')

      call getenv('PWD',absolutepath) !avoid annoying xlf/mac bug
      absolutepath = trim(absolutepath) // '/'

      read(iterm,*)txtxt
      read(iterm,*)txtxt
      read(iterm,*)txtxt
!
!--read name of file containing physical input
!
      read(iterm,*)txt,label,basenm
      label1 = 'basenm'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, trim(basenm)
!
!--read options
!
      read(iterm,*)txt,label,io_fmt
      label1 = 'io_fmt'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, io_fmt

      read(iterm,*)txt,label,idump
      label1 = 'idump'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, idump

      read(iterm,*)txt,label,makechkpnt
      label1 = 'makechkpnt'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, makechkpnt

      read(iterm,*)txt,label,override
      label1 = 'override'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, override

      read(iterm,*)txt,label,dtinit
      label1 = 'dtinit'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, dtinit

      read(iterm,*)txt,label,dtmax
      label1 = 'dtmax'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, dtmax

      read(iterm,*)txt,label,dtmin
      label1 = 'dtmin'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, dtmin

      read(iterm,*)txt,label,tstop
      label1 = 'tstop'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, tstop

      read(iterm,*)txt,label,nstepmax
      label1 = 'nstepmax'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, nstepmax

      read(iterm,*)txt,label,dt_out
      label1 = 'dt_out'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, dt_out

      read(iterm,*)txt,label,dt_ana
      label1 = 'dt_ana'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, dt_ana

      read(iterm,*)txt,label,nout_max
      label1 = 'nout_max'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, nout_max

      read(iterm,*)txt,label,nana_max
      label1 = 'nana_max'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, nana_max

      read(iterm,*)txt,label,idbg
      label1 = 'idbg'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, idbg

      read(iterm,*)txt,label,itme
      label1 = 'itme'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, itme

      read(iterm,*)txtxt
      read(iterm,*)txtxt
      read(iterm,*)txtxt

      read(iterm,*)txt,label,iunits
      label1 = 'iunits'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, iunits

      read(iterm,*)txt,label,xmas
      label1 = 'xmas'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, xmas

      read(iterm,*)txt,label,ylen
      label1 = 'ylen'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ylen

      read(iterm,*)txtxt
      read(iterm,*)txtxt
      read(iterm,*)txtxt

      read(iterm,*)txt,label,indts
      label1 = 'indts'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, indts

      read(iterm,*)txt,label,iusebin
      label1 = 'iusebin'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, iusebin

      read(iterm,*)txt,label,ifullextrap
      label1 = 'ifullextrap'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ifullextrap

      read(iterm,*)txt,label,ibctype
      label1 = 'ibctype'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ibctype

      read(iterm,*)txt,label,ishift
      label1 = 'ishift'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ishift

      read(iterm,*)txt,label,xmin
      label1 = 'xmin'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, xmin

      read(iterm,*)txt,label,xmax
      label1 = 'xmax'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, xmax

      read(iterm,*)txt,label,ymin
      label1 = 'ymin'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ymin

      read(iterm,*)txt,label,ymax
      label1 = 'ymax'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ymax

      read(iterm,*)txt,label,zmin
      label1 = 'zmin'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, zmin

      read(iterm,*)txt,label,zmax
      label1 = 'zmax'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, zmax

      read(iterm,*)txt,label,tol
      label1 = 'tol'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, tol

      read(iterm,*)txt,label,tolpm
      label1 = 'tolpm'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, tolpm

      read(iterm,*)txt,label,ftol
      label1 = 'ftol'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ftol

      read(iterm,*)txt,label,vtol
      label1 = 'vtol'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, vtol

      read(iterm,*)txt,label,ftolpm
      label1 = 'ftolpm'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ftolpm

      read(iterm,*)txt,label,vtolpm
      label1 = 'vtolpm'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, vtolpm

      read(iterm,*)txt,label,cfl
      label1 = 'cfl'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, cfl

      read(iterm,*)txt,label,i_hts
      label1 = 'i_hts'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, i_hts

      read(iterm,*)txt,label,dhmax
      label1 = 'dhmax'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, dhmax

      read(iterm,*)txtxt
      read(iterm,*)txtxt
      read(iterm,*)txtxt

      read(iterm,*)txt,label,hmin
      label1 = 'hmin'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, hmin

      read(iterm,*)txt,label,hmax
      label1 = 'hmax'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, hmax

      read(iterm,*)txt,label,umin
      label1 = 'umin'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, umin

      read(iterm,*)txt,label,ieos
      label1 = 'ieos'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ieos

      read(iterm,*)txt,label,ipres
      label1 = 'ipres'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ipres
      
      read(iterm,*)txt,label,ipdv
      label1 = 'ipdv'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ipdv

      read(iterm,*)txt,label,ishok
      label1 = 'ishok'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ishok

      read(iterm,*)txt,label,ivisc
      label1 = 'ivisc'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ivisc

      read(iterm,*)txt,label,ibals
      label1 = 'ibals'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ibals

      read(iterm,*)txt,label,ivtime
      label1 = 'ivtime'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, ivtime

      read(iterm,*)txt,label,vlength
      label1 = 'vlength'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, vlength

      read(iterm,*)txt,label,alphamax
      label1 = 'alphamax'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, alphamax

      read(iterm,*)txt,label,alphastar
      label1 = 'alphastar'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, alphastar

      read(iterm,*)txt,label,betastar
      label1 = 'betastar'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, betastar

      read(iterm,*)txt,label,icool
      label1 = 'icool'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, icool

      read(iterm,*)txt,label,taucool
      label1 = 'taucool'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, taucool

      read(iterm,*)txt,label,iheat
      label1 = 'iheat'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, iheat

      read(iterm,*)txtxt
      read(iterm,*)txtxt
      read(iterm,*)txtxt

      read(iterm,*)txt,label,igrav
      label1 = 'igrav'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, igrav

      read(iterm,*)txt,label,igrape
      label1 = 'igrape'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, igrape

      read(iterm,*)txt,label,idirect
      label1 = 'idirect'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, idirect

      read(iterm,*)txt,label,maxbuild
      label1 = 'maxbuild'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, maxbuild

      read(iterm,*)txt,label,irevise
      label1 = 'irevise'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, irevise

      read(iterm,*)txt,label,clhfrac
      label1 = 'clhfrac'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, clhfrac

      read(iterm,*)txt,label,imac
      label1 = 'imac'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, imac

      read(iterm,*)txt,label,treeacc
      label1 = 'treeacc'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, treeacc

      read(iterm,*)txt,label,isoft
      label1 = 'isoft'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, isoft

      read(iterm,*)txt,label,eps
      label1 = 'eps'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, eps

      read(iterm,*)txt,label,sepmax
      label1 = 'sepmax'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, sepmax

      read(iterm,*)txt,label,bsepmax
      label1 = 'bsepmax'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, bsepmax

      read(iterm,*)txt,label,igrapelim
      label1 = 'igrapelim'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, igrapelim

      read(iterm,*)txt,label,gmasslim
      label1 = 'gmasslim'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, gmasslim

      read(iterm,*)txtxt
      read(iterm,*)txtxt
      read(iterm,*)txtxt

      read(iterm,*)txt,label,irem
      label1 = 'irem'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, irem

      read(iterm,*)txt,label,idustev
      label1 = 'idustev'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, idustev

      read(iterm,*)txt,label,colfrac
      label1 = 'colfrac'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, colfrac

      read(iterm,*)txt,label,istrheat
      label1 = 'istrheat'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, istrheat

      read(iterm,*)txt,label,rdisk
      label1 = 'rdisk'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, rdisk

      read(iterm,*)txtxt
      read(iterm,*)txtxt
      read(iterm,*)txtxt

      read(iterm,*)txt,label,iexpand
      label1 = 'iexpand'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, iexpand

      if(iexpand.eq.1) then 
         read(iterm,*)txt,label,ncdumps 
         label1 = 'ncdumps'
         if(label .ne. label1) goto 1000
         write(6,*)txt, label, ncdumps

         read(iterm,*)txt,(redump(i),i=1,ncdumps)
         write(6,*)txt, label, (redump(i),i=1,ncdumps)
      else
         read(iterm,*)txtxt
         read(iterm,*)txtxt
      endif

      read(iterm,*)txt,label,h_0
      label1 = 'h_0'
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, h_0

      read(iterm,*)txt,label,omegam
      label1 = 'omegam'
      omegal=1.d0-omegam
      if(label .ne. label1) goto 1000
      write(6,*)txt, label, omegam

      close(iterm)

      return

1000  continue
      print *,'incorrect input file'
      write(6,1001)label,label1
1001  format(' got label = ',a20,'  expected label = ',a20)
      stop'Incorrect insph file'

      end subroutine options
!---------------------------------------------------------------------
      subroutine rdump(idump)
!**********************************************************************
!                                                                     *
!  this routine reads a dump into memory                              *
!                                                                     *
!**********************************************************************
      use vineutils

      integer, intent(in):: idump

      logical tobe
      integer i,id,numcpus
      integer id_xdrfile,ierr

      double precision tme1,tme2
      character(len=20), parameter:: where =('rdump')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      inquire(file=trim(dumpfile),exist=tobe)
      if(.not.tobe)then
         print *,'Attempted to open file: ',trim(dumpfile)
         stop' file does not exist'
      endif

      select case(io_fmt)
      case(id_xdr)

         id_xdrfile = initxdr(trim(dumpfile), 'r', .FALSE. )
         if(id_xdrfile.lt.0)then
            print *,'Error opening file',trim(dumpfile)
            stop'Error opening XDR file'
         endif

         ierr = ixdrimat(id_xdrfile,iheadlength,iheader)
         ierr = ixdrdmat(id_xdrfile,iheadlength,dheader)
         ierr = ixdrrewind(id_xdrfile)

         call readheader
         call params_sanitycheck(idump)
         call grabmem(npart)

         print *,'Reading data in...'

         ierr = ixdrimat(id_xdrfile,iheadlength,iheader)
         ierr = ixdrdmat(id_xdrfile,iheadlength,dheader)
         ierr = ixdrdmat(id_xdrfile,ndim1*npart,posmas)
         ierr = ixdrdmat(id_xdrfile,ndim*npart,vel)
         ierr = ixdrdmat(id_xdrfile,npart,h)
         ierr = ixdrdmat(id_xdrfile,npart_sph,u)
         ierr = ixdrdmat(id_xdrfile,npart_sph,rho)
         ierr = ixdrdmat(id_xdrfile,npart_sph,alpha)
         ierr = ixdrdmat(id_xdrfile,npart,poten)
         ierr = ixdrimat(id_xdrfile,npart,ipindx)
         ierr = ixdrimat(id_xdrfile,npart_sph,material)
         ierr = ixdrimat(id_xdrfile,npart,itstepbin)
         ierr = ixdrdmat(id_xdrfile,ndim1*npoim,posmaspm)
         ierr = ixdrdmat(id_xdrfile,ndim*npoim,velpm)
         ierr = ixdrdmat(id_xdrfile,3*npoim,spinpm)
         ierr = ixdrdmat(id_xdrfile,npoim,hpm)
         ierr = ixdrimat(id_xdrfile,npoim,itstepbinpm)
         ierr = ixdrimat(id_xdrfile,npoim,nsteppm)

         ierr = ixdrclose(id_xdrfile)

      case(id_ftn)

         open(idisk1,
     &        file=trim(dumpfile),status='old',form='unformatted')

         read(idisk1,err=100)(iheader(i),i=1,iheadlength),
     &                       (dheader(i),i=1,iheadlength)
         rewind(idisk1)
         call readheader
         call params_sanitycheck(idump)
         call grabmem(npart)

         print *,'Reading data in...'

         read(idisk1,err=100) 
     &        (iheader(i),i=1,iheadlength),
     &        (dheader(i),i=1,iheadlength),
     &        ((posmas(id,i),id=1,ndim1), i=1,npart),
     &        ((vel   (id,i),id=1,ndim ), i=1,npart),
     &        (h    (i),i=1,npart), 
     &        (u    (i),i=1,npart_sph),
     &        (rho  (i),i=1,npart_sph),
     &        (alpha(i),i=1,npart_sph),
     &        (poten(i),i=1,npart),
     &        (ipindx(i),i=1,npart),
     &        (material(i),i=1,npart_sph),
     &        (itstepbin(i),i=1,npart),
     &        ((posmaspm(id,i),id=1,ndim1),i=1,npoim),
     &        ((velpm   (id,i),id=1,ndim ),i=1,npoim),
     &        ((spinpm  (id,i),id=1,3    ),i=1,npoim),
     &        (hpm     (   i),i=1,npoim),
     &        (itstepbinpm(i),i=1,npoim),
     &        (nsteppm    (i),i=1,npoim)
         close(idisk1)

      case default
         print *,'Your selection for the IO format'
         print *,'has not been implemented'
         stop 'IO format not supported'
      end select

      return

100   write(*,*)'reading error in rdump!'
      stop'rdump: error reading dump'

      end subroutine rdump
!---------------------------------------------------------------------
      subroutine wdump(idump,makecheckpoint)
!**********************************************************************
!                                                                     *
!  this routine writes a dump on disk                                 *
!                                                                     *
!**********************************************************************
      logical, intent(in):: makecheckpoint
      integer, intent(in):: idump

      integer id_xdrfile,ierr,iochek,i,id
      logical tobe

      character(len=20), parameter:: where =('wdump')
      if(idbg.ge.2)print *,where

      select case(io_fmt)

      case(id_xdr)
!
! XDR only knows read, write and append as modes, so we can use the same
! mode for overwriting checkpoint dumps and writing normal dumps. But as we
! don't want to overwrite already existing normal dumps, we need to check
! for that first (XDR would simply overwrite it otherwise).
!
         inquire(file=trim(dumpfile),exist=tobe)
         if(tobe)then
            print *,trim(dumpfile),' already exists!'
            stop 'Attempt to overwrite existing dumpfile'
         endif

         id_xdrfile = initxdr(trim(dumpfile), 'w', .FALSE. )
         if(id_xdrfile.lt.0)then
            print *,'Error opening file',trim(dumpfile)
            stop 'Error opening file'
         endif

         ierr = ixdrimat(id_xdrfile,iheadlength,iheader)
         ierr = ixdrdmat(id_xdrfile,iheadlength,dheader)
         ierr = ixdrdmat(id_xdrfile,ndim1*npart,posmas)
         ierr = ixdrdmat(id_xdrfile,ndim*npart,vel)
         ierr = ixdrdmat(id_xdrfile,npart,h)
         ierr = ixdrdmat(id_xdrfile,npart_sph,u)
         ierr = ixdrdmat(id_xdrfile,npart_sph,rho)
         ierr = ixdrdmat(id_xdrfile,npart_sph,alpha)
         ierr = ixdrdmat(id_xdrfile,npart,poten)
         ierr = ixdrimat(id_xdrfile,npart,ipindx)
         ierr = ixdrimat(id_xdrfile,npart_sph,material)
         ierr = ixdrimat(id_xdrfile,npart,itstepbin)
         ierr = ixdrdmat(id_xdrfile,ndim1*npoim,posmaspm)
         ierr = ixdrdmat(id_xdrfile,ndim*npoim,velpm)
         ierr = ixdrdmat(id_xdrfile,3*npoim,spinpm)
         ierr = ixdrdmat(id_xdrfile,npoim,hpm)
         ierr = ixdrimat(id_xdrfile,npoim,itstepbinpm)
         ierr = ixdrimat(id_xdrfile,npoim,nsteppm)

         ierr = ixdrclose(id_xdrfile)

      case(id_ftn)
!
! checkpoint dump needs to overwrite a file that possibly already
! exists, but don't let a normal dump overwrite an existing file
!
         if(makecheckpoint)then
            open(idisk1,file=trim(dumpfile),status='replace',
     &                                      form='unformatted')
         else
            open(idisk1,file=trim(dumpfile),status='new',
     &                                      form='unformatted')
         endif

         write(idisk1,iostat=iochek,err=100)
     &        (iheader(i),i=1,iheadlength),
     &        (dheader(i),i=1,iheadlength),
     &        ((posmas(id,i),id=1,ndim1), i=1,npart),
     &        ((vel   (id,i),id=1,ndim ), i=1,npart),
     &        (h    (i),i=1,npart), 
     &        (u    (i),i=1,npart_sph),
     &        (rho  (i),i=1,npart_sph),
     &        (alpha(i),i=1,npart_sph),
     &        (poten(i),i=1,npart),
     &        (ipindx(i),i=1,npart),
     &        (material(i),i=1,npart_sph),
     &        (itstepbin(i),i=1,npart),
     &        ((posmaspm(id,i),id=1,ndim1),i=1,npoim),
     &        ((velpm   (id,i),id=1,ndim ),i=1,npoim),
     &        ((spinpm  (id,i),id=1,3    ),i=1,npoim),
     &        (hpm     (   i),i=1,npoim),
     &        (itstepbinpm(i),i=1,npoim),
     &        (nsteppm    (i),i=1,npoim)

         close(idisk1)
         if(iochek.lt.0)then
            print *,'ran out of data to write'
            print *,'something might be wrong'
         endif

      case default
         print *,'Your selection for the IO format'
         print *,'has not been implemented'
         stop 'IO format not supported'
      
      end select

      if(idbg.ge.0)print *,'a dump has occured at t=',t

      return
!
! an error occured while writing
!
100   write(*,*)'error in wdump dumpfile:',trim(dumpfile)
      stop 'wdump: error writing dump'

      end subroutine wdump
!---------------------------------------------------------------------
      subroutine params_sanitycheck(idump)
!***********************************************************************
!                                                                      *
! This routine checks (as far as possible) if the various combinations *
! of parameters in insph / the dump file and the selected compiler     *
! switches in the makefile are compatible with one another.            *
!                                                                      *
!***********************************************************************
      integer, intent(in):: idump

      character(len=12) oldintg

      print *,'=================================================='
      print *,' '
      print *,' '
      print *,'Checking for sane input parameters'
      print *,' '
      print *,' '

      if(iheadlength.ne.iheader(id_iheadlen))then
         print *,' '
         print *,'Dump believes header length is:',iheader(id_iheadlen)
         print *,'Code believes header length is:',iheadlength
         print *,' '
         stop'Inconsistent header length: Unable to run'
      endif
      if(ndim    .ne.iheader(id_ndim))then
         print *,' '
         print *,'Dump believes dimensionality is:',iheader(id_ndim)
         print *,'Code believes dimensionality is:',ndim
         print *,' '
         stop'Inconsistent dimensionality: Unable to run'
      endif
      if(npart_sph.gt.isphdim)then
         print *,' '
         print *,'Dump claims:',npart_sph,' SPH particles'
         print *,'Code max is:',isphdim,' (recompile for more)'
         print *,' '
         stop'Too many sph particles: Unable to run'
      endif
      if(npart-npart_sph.gt.nboddim)then
         print *,' '
         print *,'Dump claims:',npart-npart_sph,' N-body particles'
         print *,'Code max is:',nboddim,' (recompile for more)'
         print *,' '
         stop'Too many N-body particles: Unable to run'
      endif
      if(npart.gt.idim)then
         print *,' '
         print *,'Dump claims:',npart,' particles' 
         print *,'Code max is:',idim,' (recompile for more)'
         print *,' '
         stop'Too many particles: Unable to run'
      endif
      if(npoim.gt.ipmdim)then
         print *,' '
         print *,'Dump claims:',npoim ,' point mass particles'
         print *,'Code max is:',ipmdim,' (recompile for more)'
         print *,' '
         stop'Too many Point Mass particles: Unable to run'
      endif
      if(isoft.ne.id_Plummer .and. igrape.eq.1)then
         print *,' '
         print *,'Your insph file requests non-Plummer softening'
         print *,'but you are using the GRAPE hardware, which'
         print *,'can only use Plummer softening.'
         print *,' '
         stop'Incompatible softening and grape options'
      endif
      if(isoft.eq.id_varkern .and. npart_sph.eq.0)then
         print *,' '
         print *,'Your insph file requests variable kernel'
         print *,'softening but you have no SPH particles.'
         print *,'Only SPH particles can be variably softened'
         print *,' '
         stop'Incompatible softening and particle options'
      endif
      if(ibctype.eq.id_perdbnd .and. igrav.eq.1 .and.
     &        isoft.ne.id_Plummer)then
         print *,'The routine for the ewald method requires'
         print *,'that you use Plummer softening, no spline kernel'
         stop 'Incompatible softening for periodic boundaries'
      endif
      if((dheader(id_tnextout)-t).lt.zero)then
         print *,'*************************************************'
         print *,'The dump thinks the next dump time is before now!'
         print *,'Dump time from the file will be overridden when  '
         print *,'initially set, using the current time and dt_out '
         print *,'values from your insph file.                     '
         print *,'Dump thinks the current time is:',t
         print *,'Dump thinks next dump   time is:',dheader(id_tnextout)
         print *,'*************************************************'
! Note that this override is done in the setdumptime routine. Not here.
      endif
      if( abs(t+dt_out-dheader(id_tnextout))/dt_out.gt. 1d-10 .and.
     &          override)then
         print *,'The next expected dump time saved in the dump file'
         print *,'is different than what you have selected in your'
         print *,'insph file, and will therefore be overridden'
         print *,'Hope thats ok'
         print *,'Dump thinks next dump be  at t=',dheader(id_tnextout)
         print *,'This run will output next at t=',(t+dt_out)
         print *,' '
      endif
      if(dtmin.ge.dtmax)then
         print *,'Your dtmin is greater than your dtmax'
         print *,'dtmin:',dtmin
         print *,'dtmax:',dtmax
         stop 'Inconsistent time step limiters'
      endif
      if(iusebin.eq.1 .and. dt_out.lt.dtmax)then
         print *,'Your dump time increment is smaller than the'
         print *,'maximum timestep in your bin hierarchy.'
         print *,'For binned timesteps, this is not possible'
         print *,'dt_out=',dt_out
         print *,'dtmax =',dtmax
         stop 'Dump time increment is too small'
      endif
      if(iusebin.eq.1 .and. 
     &     (dt_out/dtmax - dble(nint(dt_out/dtmax))) .gt. 1d-5)then
         print *,'Your dump time increment (dt_out) is not an'
         print *,'integer multiple of the maximum time step dtmax'
         print *,'This means you will not get dumps at the times'
         print *,'you expect. Hope thats ok with you'
         print *,' '
         print *,'dtmax =',dtmax
         print *,'dt_out=',dt_out
         print *,' '
      endif
      if(iunits.ne.iheader(id_iunits))then
         print *,'Inconsistent units between'
         print *,'dump file and insph setting'
         print *,'insph file thinks iunits is:',iunits
         print *,'Dump  file thinks iunits is:',iheader(id_iunits)
         stop 'Inconsistent Unit settings'
      endif
      if(iunits.ne.0)then   !Scale free units don't need xmax/ylen
         if(abs(xmas-dheader(id_xmas))/xmas .gt. 1e-14)then
            print *,'Inconsistent units between'
            print *,'dump file and insph setting'
            print *,'insph file thinks xmas is:',xmas
            print *,'Dump  file thinks xmas is:',dheader(id_xmas)
            stop'Units don''t make sense'
         endif
         if(abs(ylen-dheader(id_ylen))/ylen .gt. 1e-14)then
            print *,'Inconsistent units between'
            print *,'dump file and insph setting'
            print *,'insph file thinks ylen is:',ylen
            print *,'Dump  file thinks ylen is:',dheader(id_ylen)
            stop'Units don''t make sense'
         endif
      endif
#ifndef SPH
      if(npart_sph.ne.0)then
         print *,' '
         print *,'Dump claims:',npart_sph,' SPH particles'
         print *,'but this executable does not have the SPH'
         print *,'module compiled in!'
         print *,'Recompile with the SPH makefile option set'
         stop'SPH not compiled into this executable'
      endif
#else
      if(ieos.eq.id_lagrgiso .or. ieos.eq.id_euleriso .or.
     &   ieos.eq.id_bateeos)then
         if(iunits.eq.1 .and. iexpand.eq.0 .and. 
     &               THREE.gt.Avemwt*Mprot/kBoltz*umin)then
            print *,'umin value of umin =',umin
            print *,'corresponds to a temperature of'
            print *,'T=',Avemwt*Mprot/kBoltz*umin,' K'
            print *,'which is lower than the cosmic background'
            print *,'VINE will reset it to the cosmic background'
            print *,'temperature of 3 Kelvin'

            umin = THREE/(Avemwt*Mprot/kBoltz)

            print *,'New umin value is:',umin,' in external units'
         endif
      else
         if(abs(dheader(id_gamma)-ONE).lt.1d-10)then
            print *,' '
            print *,'You have not selected an isothermal equation of'
            print *,'state, but have set the ratio of specific'
            print *,'heats, gamma, to unity!!!'
            print *,'gamma : ',dheader(id_gamma)
            stop'Inconsistent EOS settings'
         endif
         if(iunits.eq.1 .and. iexpand.eq.0 .and. THREE.gt.
     &         Avemwt*(dheader(id_gamma)-ONE)*Mprot/kBoltz*umin)then
            print *,'umin value of umin =',umin
            print *,'corresponds to a temperature of'
            print *,'T=',
     &         Avemwt*(dheader(id_gamma)-ONE)*Mprot/kBoltz*umin,' K'
            print *,'which is lower than the cosmic background'
            print *,'VINE will reset it to the cosmic background'
            print *,'temperature of 3 Kelvin'

            umin = THREE/(Avemwt*(dheader(id_gamma)-ONE)*Mprot/kBoltz)

            print *,'New umin value is:',umin,' in external units'
         endif
      endif
#endif
#ifndef PM
      if(npoim.ne.0)then
         print *,' '
         print *,'Your dump contains',npoim,' point masses '
         print *,'but this executable does not have the PM'
         print *,'module compiled in!'
         print *,'Recompile with the PM makefile option set'
         stop'Point masses compiled into this executable'
      endif
#endif
#ifndef PERIODIC
      if(ibctype.eq.id_perdbnd)then
         print *,'Periodic boundaries not compiled in this'
         print *,'executable. Recompile with the PERIODIC'
         print *,'makefile option set'
         stop'Periodic boundaries not compiled into this executable'
      endif
#endif
#ifndef COSMO
      if(iexpand.eq.1)then
         print *,'Cosmological expansion requested,'
         print *,'but not compiled into this version'
         print *,'Recompile with the COSMO makefile option set'
         stop'Cosmology not compiled into this executable'
      endif
#else
! in this version of vine, the cosmological corrections to the timestep updates
! as well as some of the unit conversions have been corrupted. 
      if(iexpand.eq.1)then
         print*,'In this VINE version, the cosmo module is not'
         print*,'functional. Without fixes to unit conversion and'
         print*,'timestep updates, you can''t run cosmological'
         print*,'simulations!'
         stop'No cosmo run possible!'
      endif
! check for leapfrog integrator because currently only 
! this one works with cosmological expansion
#ifndef LF
      if(iexpand.eq.1)then
         print *,'You want to run a cosmological simulation with'
         print *,'the Runge-Kutta integrator. Such a module doesn''t'
         print *,'exist yet, please switch to the leapfrog integrator'
         stop'Incompatible integrator for cosmological simulations'
      endif
#endif
      if(npart_sph.gt.0.and.iexpand.eq.1)then
         print*,'You want to run a cosmological simulation with'
         print*,'gas (npart_sph/=0). This is currently not tested in'
         print*,'this version of VINE.'
         stop'Cosmology not tested for gas in this VINE version.'
      endif
#endif
#ifdef LF
      if(iusebin.ne.1)then
         print *,'Using unbinned timesteps with the leapfrog'
         print *,'integrator may not be advisable, due to'
         print *,'the repeated small offsets from having properly'
         print *,'centered timesteps and also the corrections in '
         print *,'potentials and densities. (see subr. lf_correct)'
         print *,'Caveat Emptor'
      endif
#endif
      if(iusebin.eq.1 .and. iheader(id_iusebin).eq.0)then
         print *,'Previously used unbinned timesteps, but now using'
         print *,'binned timesteps. This means your timestep will'
         print *,'start from the smallest timestep bins, unless'
         print *,'you are starting from t=0.'
         print *,'This should be fixed someday for other times. too'
      endif

      if(io_fmt.ne.id_ftn .and. io_fmt.ne.id_xdr .and.
     &   io_fmt.ne.id_hdf)then
         print 500,'io_fmt   ',io_fmt
         stop
      endif
      if(iunits.ne.0 .and. iunits.ne.1 .and.
     &   iunits.ne.2 .and. iunits.ne.3)then
         print 500,'iunits   ',iunits
         stop
      endif
      if(makechkpnt.ne.0 .and. makechkpnt.ne.1)then
        print 500,'makechkpnt',makechkpnt
        stop
      endif
      if(ibctype.ne.id_freebnd .and. 
     &       ibctype.ne.id_perdbnd .and. 
     &       ibctype.ne.id_reflbnd)then
         print 500,'ibctype  ',ibctype
         stop
      endif
      if(indts.ne.0 .and. indts.ne.1)then
         print 500,'indts    ',indts
         stop
      endif
      if(iusebin.ne.0 .and. iusebin.ne.1)then
         print 500,'iusebin  ',iusebin
         stop
      endif
      if(ifullextrap.ne.0 .and. ifullextrap.ne.1)then
         print 500,'ifullextrap',ifullextrap
         stop
      endif
      if(ipres.ne.0 .and.ipres.ne.1)then
         print 500,'ipres    ',ipres
         stop
      endif
      if(ipdv.ne.0 .and. ipdv.ne.1)then
         print 500,'ipdv     ',ipdv
         stop
      endif
      if(ieos.ne.id_idealgas .and. 
     &   ieos.ne.id_euleriso .and. 
     &   ieos.ne.id_lagrgiso .and.
     &   ieos.ne.id_polyeos  .and.
     &   ieos.ne.id_entreos  .and.
     &   ieos.ne.id_bateeos  )then
         print 500,'ieos     ',ieos
         stop
      endif
      if(icool.ne.0 .and. icool.ne.1)then
         print 500,'icool    ',icool
         stop
      endif 
      if(iheat.ne.0 .and. iheat.ne.1)then
         print 500,'iheat    ',iheat
         stop
      endif
      if(ivisc.ne.0 .and. ivisc.ne.1)then
         print 500,'ivisc    ',ivisc
         stop
      endif
      if(ivtime.ne.0 .and. ivtime.ne.1)then
         print 500,'ivtime   ',ivtime
         stop
      endif
      if(ibals.ne.0 .and. ibals.ne.1)then
         print 500,'ibals    ',ibals
      endif
      if(ishok.ne.0 .and. ishok.ne.1)then
         print 500,'ishok    ',ishok
         stop
      endif
      if(igrav.ne.0 .and. igrav.ne.1)then
         print 500,'igrav    ',igrav
         stop
      endif
      if(isoft.ne.id_varkern .and.
     &       isoft.ne.id_fixkern .and.
     &       isoft.ne.id_Plummer)then
         print 500,'isoft    ',isoft
         stop
      endif
      if(imac.ne.id_angle .and.
     &      imac.ne.id_SWabs .and.
     &      imac.ne.id_gadget)then
         print 500,'imac     ',imac
         stop
      endif
      if(maxbuild.lt.0)then
         print 500,'maxbuild ',maxbuild
         stop
      endif
      if(irevise.ne.0 .and. irevise.ne.1)then
         print 500,'irevise ',irevise
         stop
      endif
      if(i_hts.ne.0 .and. i_hts.ne.1)then
         print 500,'i_hts    ',i_hts
         stop
      endif
      if(iexpand.ne.0 .and. iexpand.ne.1)then
         print 500,'iexpand  ',iexpand
         stop
      endif
      if(maxclumpsize.lt.1)then
         print 500,'maxclumpsize',maxclumpsize
         stop
      endif
      if(maxbunchsize.lt.1)then
         print 500,'maxbunchsize',maxbunchsize
         stop
      endif
      if(ishift.ne.0 .and.ishift.ne.1)then
         print 600,'ishift   ',ishift
         stop
      endif
!
! Real variables
!
      if(vlength.lt.zero)then
         print 550,'vlength  ',vlength
         stop
      endif
      if(alphamax.lt.alphastar)then
         print 550,'alphamax',alphamax
         print *,'Because it is less than alphastar.'
         stop
      endif
      if(alphastar.lt.zero)then
         print 550,'alphastar',alphastar
         stop
      endif
      if(betastar.lt.zero)then
         print 550,'betastar ',betastar
         stop
      endif
      if(tol.lt.zero)then
         print 550,'tol      ',tol
         stop
      endif
      if(tolpm.lt.zero)then
         print 550,'tolpm    ',tolpm
         stop
      endif
      if(ftol.lt.zero)then
         print 550,'ftol     ',ftol
      endif
      if(vtol.lt.zero)then
         print 550,'vtol     ',vtol
      endif
      if(ftolpm.lt.zero)then
         print 550,'ftolpm   ',ftolpm
      endif
      if(vtolpm.lt.zero)then
         print 550,'vtolpm   ',vtolpm
      endif
      if(cfl.lt.zero .or. cfl.gt.one)then
         print 550,'cfl      ',cfl
         stop
      endif
      if(dhmax.lt.zero .or. dhmax.gt.one)then
         print 550,'dhmax    ',dhmax
         stop
      endif
      if(treeacc.lt.zero .or. 
     &       (imac.ne.id_SWabs .and. treeacc.gt.one))then
         print 550,'treeacc  ',treeacc
         stop
      endif
      if(sepmax.lt.zero)then
         print 550,'sepmax   ',sepmax
         stop
      endif
      if(bsepmax.lt.zero)then
         print 550,'bsepmax  ',bsepmax
         stop
      endif
      if(hmin.le.zero)then
         print 550,'hmin     ',hmin
         stop
      endif
      if(eps.lt.zero)then
         print 550,'eps      ',eps
         stop
      endif
      if(dtinit.le.zero)then
         print 550,'dtinit   ',dtinit
         stop
      endif
      if(dtmax.lt.zero)then
         print 550,'dtmax    ',dtmax
         stop
      endif
      if(dtmin.lt.zero)then
         print 550,'dtmin    ',dtmin
         stop
      endif
      if(dt_out.le.zero)then
         print 550,'dt_out   ',dt_out
         stop
      endif
      if(xmas.le.zero)then
         print 550,'xmas     ',xmas
         stop
      endif
      if(ylen.le.zero)then
         print 550,'ylen     ',ylen
         stop
      endif
      if(umin.lt.zero)then
        print 550,'umin     ',umin
        stop
      endif
      if(gmasslim.le.zero)then
         print 550,'gmasslim ',gmasslim
         stop
      endif
      if(clhfrac.lt.one)then
         print 550,'clhfrac  ',clhfrac
         stop
      endif
      if(xmin.ge.xmax)then
         print 550,'xmin     ',xmin
         stop
      endif
      if(xmax.le.xmin)then
         print 550,'xmax     ',xmax
         stop
      endif
      if(ymin.ge.ymax)then
         print 550,'ymin     ',ymin
         stop
      endif
      if(ymax.le.ymin)then
         print 550,'ymax     ',ymax
         stop
      endif
      if(zmin.ge.zmax)then
         print 550,'zmin     ',zmin
         stop
      endif
      if(zmax.le.zmin)then
         print 550,'zmax     ',zmax
         stop
      endif
     
      if(iexpand.eq.1)then
         if(h_0.lt.zero)then
            print 550,'h_0      ',h_0
            stop
         endif
         if(omegam.lt.zero)then
            print 550,'omegam   ',omegam
            stop
         endif
      endif

      print *,' '
      print *,' '
      print *,'Your parameters appear sane'
      print *,' '
      print *,' '
      print *,'=================================================='
!
! now check if the same code parameters are used currently 
! as are stored in the dump file
!
      print *,'=================================================='
      print *,' '
      print *,' '
      print *,'Comparing current code parameters to stored values'
      print *,' '
      print *,' '
      if(id_myintg .ne. iheader(id_intgrtr))then
         oldintg='unknown'
         if(iheader(id_intgrtr).eq.id_lf_dkd)oldintg='Leapfrog'
         if(iheader(id_intgrtr).eq.id_rk    )oldintg='Runge-Kutta'
      print *,'********************************************************'
      print *,'********************************************************'
      print *,'********************************************************'
      print *,'********************************************************'
         if(id_myintg.eq.id_lf_dkd)print 602,'Leapfrog',oldintg
         if(id_myintg.eq.id_rk    )print 602,'Runge-Kutta',oldintg
      print *,'********************************************************'
      print *,'********************************************************'
      print *,'********************************************************'
      print *,'********************************************************'
      else
#if defined(RK)
         print *,'We are using the Runge-Kutta Integrator'
#endif
#if defined(LF)
         print *,'We are using the Leapfrog Integrator'
#endif
         print *,' '
         print *,' '
      endif

      if(idbg    .ne.iheader(id_idbg  ))
     &        print 600,'idbg     ',idbg,iheader(id_idbg)
      if(itme    .ne.iheader(id_itme  ))
     &        print 600,'itme     ',itme,iheader(id_itme)
      if(idump   .ne.iheader(id_idump  ))
     &        print 600,'idump    ',idump,iheader(id_idump)
      if(io_fmt  .ne.iheader(id_io_fmt ))
     &        print 600,'io_fmt   ',io_fmt ,iheader(id_io_fmt )
      if(iunits  .ne.iheader(id_iunits ))
     &        print 600,'iunits   ',iunits,iheader(id_iunits)
      if(makechkpnt.ne.iheader(id_makechkpnt))
     &        print 600,'maketmp  ',makechkpnt,iheader(id_makechkpnt)
      if(ibctype .ne.iheader(id_ibctype))
     &        print 600,'ibctype  ',ibctype,iheader(id_ibctype)
      if(indts   .ne.iheader(id_indts  ))
     &        print 600,'indts    ',indts,iheader(id_indts)
      if(iusebin .ne.iheader(id_iusebin))
     &        print 600,'iusebin  ',iusebin,iheader(id_iusebin)
      if(ifullextrap .ne.iheader(id_fullextrap))
     &        print 600,'ifullextrap',ifullextrap,iheader(id_fullextrap)
      if(ipres   .ne.iheader(id_ipres  ))
     &        print 600,'ipres    ',ipres,iheader(id_ipres)
      if(ipdv    .ne.iheader(id_ipdv   ))
     &        print 600,'ipdv     ',ipdv ,iheader(id_ipdv )
      if(ieos    .ne.iheader(id_ieos   ))
     &        print 600,'ieos     ',ieos ,iheader(id_ieos )
      if(icool   .ne.iheader(id_icool  ))
     &        print 600,'icool    ',icool,iheader(id_icool)
      if(iheat   .ne.iheader(id_iheat  ))
     &        print 600,'iheat    ',iheat,iheader(id_iheat)
      if(ivisc   .ne.iheader(id_ivisc  ))
     &        print 600,'ivisc    ',ivisc,iheader(id_ivisc)
      if(ivtime  .ne.iheader(id_ivtime ))
     &        print 600,'ivtime   ',ivtime,iheader(id_ivtime)
      if(ibals   .ne.iheader(id_ibals  ))
     &        print 600,'ibals    ',ibals ,iheader(id_ibals )
      if(ishok   .ne.iheader(id_ishok  ))
     &        print 600,'ishok    ',ishok ,iheader(id_ishok )
      if(igrav   .ne.iheader(id_igrav  ))
     &        print 600,'igrav    ',igrav ,iheader(id_igrav )
      if(isoft   .ne.iheader(id_isoft  ))
     &        print 600,'isoft    ',isoft ,iheader(id_isoft )
      if(imac    .ne.iheader(id_imac   ))
     &        print 600,'imac     ',imac  ,iheader(id_imac  )
      if(maxbuild.ne.iheader(id_maxbuild))
     &        print 600,'maxbuild ',maxbuild,iheader(id_maxbuild)
      if(irevise .ne.iheader(id_revise))
     &        print 600,'irevise  ',irevise,iheader(id_revise)
      if(i_hts   .ne.iheader(id_ihts   ))
     &        print 600,'i_hts    ',i_hts,iheader(id_ihts)
      if(iexpand .ne.iheader(id_iexpand))
     &        print 600,'iexpand  ',iexpand,iheader(id_iexpand)
      if(ncdumps .ne.iheader(id_ncdumps))then
              print 600,'ncdumps  ',ncdumps,iheader(id_ncdumps)
              print *,'There may be confusion with dump numbers'
      endif
      if(maxclumpsize.ne.iheader(id_maxclumpsize))
     &   print 600,'maxclumpsize',maxclumpsize,iheader(id_maxclumpsize)
      if(maxbunchsize.ne.iheader(id_maxbunchsize))
     &   print 600,'maxbunchsize',maxbunchsize,iheader(id_maxbunchsize)
      if(ishift .ne.iheader(id_ishift))
     &        print 600,'ishift   ',ishift,iheader(id_ishift)
!
! Real variables
!
      if(vlength  .ne.dheader(id_vlength))
     &        print 700,'vlength  ',vlength,dheader(id_vlength)
      if(alphamax .ne.dheader(id_alfmax ))
     &        print 700,'alphamax ',alphamax,dheader(id_alfmax)
      if(alphastar.ne.dheader(id_alfstar))
     &        print 700,'alphastar',alphastar,dheader(id_alfstar)
      if(betastar .ne.dheader(id_betastar))
     &        print 700,'betastar ',betastar,dheader(id_betastar)
      if(tol      .ne.dheader(id_tol     ))
     &        print 700,'tol      ',tol    ,dheader(id_tol    )
      if(tolpm    .ne.dheader(id_tolpm   ))
     &        print 700,'tolpm    ',tolpm  ,dheader(id_tolpm  )
      if(ftol     .ne.dheader(id_ftol    ))
     &        print 700,'ftol     ',ftol   ,dheader(id_ftol   )
      if(vtol     .ne.dheader(id_vtol    ))
     &        print 700,'vtol     ',vtol   ,dheader(id_vtol   )
      if(ftolpm   .ne.dheader(id_ftolpm  ))
     &        print 700,'ftolpm   ',ftolpm ,dheader(id_ftolpm )
      if(vtolpm   .ne.dheader(id_vtolpm  ))
     &        print 700,'vtolpm   ',vtolpm ,dheader(id_vtolpm )
      if(cfl      .ne.dheader(id_cfl     ))
     &        print 700,'cfl      ',cfl    ,dheader(id_cfl    )
      if(dhmax    .ne.dheader(id_dhmax   ))
     &        print 700,'dhmax    ',dhmax  ,dheader(id_dhmax  )
      if(treeacc  .ne.dheader(id_treeacc ))
     &        print 700,'treeacc  ',treeacc,dheader(id_treeacc)
      if(sepmax   .ne.dheader(id_sepmax  ))
     &        print 700,'sepmax   ',sepmax ,dheader(id_sepmax )
      if(bsepmax  .ne.dheader(id_bsepmax ))
     &        print 700,'bsepmax  ',bsepmax,dheader(id_bsepmax)
      if(hmin     .ne.dheader(id_hmin    ))
     &        print 700,'hmin     ',hmin   ,dheader(id_hmin   )
      if(hmax     .ne.dheader(id_hmax    ))
     &        print 700,'hmax     ',hmax   ,dheader(id_hmax   )
      if(eps      .ne.dheader(id_eps     ))
     &        print 700,'eps      ',eps    ,dheader(id_eps    )
      if(dtinit   .ne.dheader(id_dtinit  ))
     &        print 700,'dtinit   ',dtinit ,dheader(id_dtinit )
      if(dtmax    .ne.dheader(id_dtmax   ))
     &        print 700,'dtmax    ',dtmax  ,dheader(id_dtmax  )
      if(dtmin    .ne.dheader(id_dtmin   ))
     &        print 700,'dtmin    ',dtmin  ,dheader(id_dtmin  )
      if(tstop    .ne.dheader(id_tstop   ))
     &        print 700,'tstop    ',tstop  ,dheader(id_tstop  )
      if(dt_out   .ne.dheader(id_dt_out  ))
     &        print 700,'dt_out   ',dt_out ,dheader(id_dt_out )
      if(dt_ana   .ne.dheader(id_dt_ana  ))
     &        print 700,'dt_ana   ',dt_ana ,dheader(id_dt_ana )
      if(xmas     .ne.dheader(id_xmas    ))
     &        print 700,'xmas     ',xmas   ,dheader(id_xmas   )
      if(ylen     .ne.dheader(id_ylen    ))
     &        print 700,'ylen     ',ylen   ,dheader(id_ylen   )
      if(umin     .ne.dheader(id_umin    ))
     &        print 700,'umin     ',umin   ,dheader(id_umin   )
      if(gmasslim .ne.dheader(id_gmasslim))
     &        print 700,'gmasslim ',gmasslim,dheader(id_gmasslim)
      if(clhfrac  .ne.dheader(id_clhfrac ))
     &        print 700,'clhfrac  ',clhfrac ,dheader(id_clhfrac)
      if(xmin     .ne.dheader(id_xmin    ))
     &        print 700,'xmin     ',xmin    ,dheader(id_xmin  )
      if(xmax     .ne.dheader(id_xmax    ))
     &        print 700,'xmax     ',xmax    ,dheader(id_xmax  )
      if(ymin     .ne.dheader(id_ymin    ))
     &        print 700,'ymin     ',ymin    ,dheader(id_ymin  )
      if(ymax     .ne.dheader(id_ymax    ))
     &        print 700,'ymax     ',ymax    ,dheader(id_ymax  )
      if(zmin     .ne.dheader(id_zmin    ))
     &        print 700,'zmin     ',zmin    ,dheader(id_zmin  )
      if(zmax     .ne.dheader(id_zmax    ))
     &        print 700,'zmax     ',zmax    ,dheader(id_zmax  )
      if(taucool  .ne.dheader(id_taucool ))
     &        print 700,'taucool  ',taucool ,dheader(id_taucool)

      if(iexpand.eq.1)then
         if(h_0      .ne.dheader(id_h_0     )) 
     &        print 700,'h_0      ',h_0    ,dheader(id_h_0    )
         if(omegam   .ne.dheader(id_omegam  )) 
     &        print 700,'omegam   ',omegam ,dheader(id_omegam )
      endif

      print *,' '
      print *,' '
      print *,'     Finished Old/New Comparisons'
      print *,' '
      print *,'=================================================='

500   format('Invalid Value ',a12,':',i6)
550   format('Invalid Value ',a12,':',1pd12.4)
600   format('Current ',a12,':',i6,' Overrides:',i6)
602   format('Current ',a12,' integrator',' overrides previous: ',a12)
700   format('Current ',a12,':',1pd12.4,' Overrides:',1pd12.4)
      return
      end subroutine params_sanitycheck
!---------------------------------------------------------------------
      subroutine dumpfile_sanitycheck

      use clmpvars, only: lstclumps
#if COSMO
      use cosmo, only: cosmic_time
#endif

      integer, parameter :: maxhvals=30       !some reasonably big number

      logical reset_material
      integer i,j,icount,numhvals,nperh(maxhvals)
      double precision hvals(maxhvals)
      double precision cosm_t

      print *,'====================================================='
      print *,' '
      print *,'Checking the sanity of your dump file'
      print *,' '
!
! Check for the same dtmax in the dump file as in the insph file. 
! If not, then override the dump file value with the insph file 
! value, and reset all of the particle time bins appropriately.
!
      if(abs(dtmax-dheader(id_dtmax)).gt.TEN*spacing(dtmax))then

         if(iusebin.eq.1)then

            print *,'WARNING WARNING WARNING WARNING WARNING'
            print *,'Overriding dtmax value in dump file'
            print *,'Old value: ',dheader(id_dtmax)
            print *,'New value: ',dtmax
            print *,'Spacing  : ',spacing(dtmax)
            print *,'Currently not working yet. Setting to old value'
            print *,'Currently not working yet. Setting to old value'

            dtmax=dheader(id_dtmax)

         endif
      endif
!
! check for zero mass particles
!
      do i=1,npart
         if(posmas(ims,i).le.tiny(ONE))then
            print *,'Particle number:',i
            print *,'Particles with zero mass detected! This would'
            print *,'lead to problems during the tree build and'
            print *,'many other places.'
            stop 'Zero mass particle!'
         endif
      enddo
!
! Check for positive internal energy
!
      do i=1,npart_sph
         if(u(i).le.zero)then
            print *,'Particle number:',i
            print *,'SPH particle with non positive internal energy'
            print *,'detected. Not physical. '
            stop 'Non-positive internal energy'
         endif
      enddo
!
! Check for ipindx being correctly initialized. (need for orderparticles)
! First, check if ipindx is initialized at all,
!  assume large integers or non positive integers for 
!  non initialized values. 
!
      icount=0
      do i=1,npart
         if(ipindx(i).le.0 .or. ipindx(i).gt.npart) icount=icount+1
      enddo
      if(icount.eq.npart)then
         print *,' '
         print *,' '
         print *,'ipindx not initialized. Resetting...'
         print *,' '
         print *,' '
         do i=1,npart
            ipindx(i)=i
         enddo
      elseif(icount.gt.0)then
         if(irem.ne.0)then
           print *,'Something is wrong with the values of ipindx!'
           print *,'Perhaps you have restarted from a dump from which'
           print *,'particles have previously been removed?'
           print *,'Caveat emptor ipindx!!!'
           print *,' '
         else
           print *,'Something is wrong with the values of ipindx!'
           stop 'Error in ipindx initialization'
         endif
      endif
!
! Second, check if every index occurs once. Use lstclumps as 
!   a temporary array to save memory. This array shouldn't be
!   used that early in the code.
!
! THIS DOESN'T WORK YET!!!!!
!      do i=1,npart
!         lstclumps(ipindx(i))=i ! I think this is the right logic
!      enddo
!      do i=1,npart
!         if(i.ne.lstclumps(i))then
!            print *,'Index ',i,' not found in ipindx!'
!            stop 'Error in ipindx initialization'
!         endif
!      enddo

!
! Check for valid material identifications:
!
       reset_material=.false.
       do i=1,npart_sph
          if(material(i).le.0)then
             material(i)=1
             reset_material=.true.
          endif
       enddo
       if(reset_material)then
          print *,' '
          print *,'You have particles with no material identifications'
          print *,'VINE has reset their ID to the value 1' 
          print *,' '
       endif
!
! check for zero mass point masses
!
      do i=1,npoim
         if(posmaspm(ims,i).le.tiny(ONE))then
            print *,'Point mass number:',i
            print *,'Point mass with a mass of zero detected!'
            stop 'Zero mass point mass!'
         endif
      enddo

      print *,' '
      print *,'Gravitational softening options used:'

      select case(isoft)
      case(id_varkern,id_fixkern)

         print *,' '
         if(npart_sph.gt.0)then
            print *,'    SPH particles  : SPH spline kernel'
         endif
         if(npart-npart_sph.gt.0)then
            print *,'    Nbody particles: SPH spline kernel'
         endif
         print *,' '
         print *,'Softening lengths are defined by'
         if(isoft.eq.id_varkern)then
            print *,' '
            if(npart_sph.gt.0)then
               print *,'    SPH  : individual time dependent h value'
            endif
            if(npart-npart_sph.gt.0)then
            print *,'    Nbody: individual, fixed h values in dump file'
            endif
            print *,' '
         else
            print *,' '
            if(npart_sph.gt.0)then
               print 123,'    SPH  : fixed at eps=',eps
123            format(a,1pd16.8)
               print *,'           (eps is in insph)'
            endif
            if(npart-npart_sph.gt.0)then
               print *,'    Nbody: indiv, fixed h values in dump file'
            endif
         endif
         print *,' '
      case(id_plummer)

         print *,' '
         if(npart_sph.gt.0)then
            print *,'   SPH particles  : Plummer softening'
         endif
         if(npart-npart_sph.gt.0)then
            print *,'   Nbody particles: Plummer softening'
         endif
         print *,' '
         print *,'All gravitational softening lengths'
         print 124,'are fixed at eps=',eps
124      format(a,1pd16.8)
         print *,'where eps is as defined in insph.'

      case default
         print *,'Unknown gravity softening option in'
         print *,'params_sanitycheck!!!'
         stop 'Unknown gravitational softening'
      end select
      print *,' '
      print *,' '
!
! if spline kernel softening, check how many different h values are used
! for nbody particles
!
      if(npart.ne.npart_sph .and. isoft.ne.id_Plummer)then
         numhvals=1
         nperh(1)=1
         hvals(1)=h(npart_sph+1)
         do i=npart_sph+2,npart
            j=numhvals
            do while(j.gt.0)
               if(abs(h(i) - hvals(j)).le.small)then
                  nperh(j)=nperh(j)+1
                  j=-1
               endif
               j=j-1
            enddo
            if(j.eq.0)then
               numhvals=numhvals+1
               nperh(numhvals)=1
               hvals(numhvals)=h(i)
               if(numhvals.gt.maxhvals)then
                  stop'Too many different Nbody smoothing lengths'
               endif
            endif
         enddo
         print *,'Nbody gravitational softening length distribution'
         print *,'Number of particles |Softening length(internal units)'
         do j=1,numhvals
            print 35,nperh(j),hvals(j)
         enddo
         print *,' '
35       format(i12,'        ',1pd16.8)
      endif
!
! if spline kernel softening, check for valid nbody smoothing lengths
!
      if(npart.ne.npart_sph.and.isoft.ne.id_Plummer)then
         do i=npart_sph+1,npart
            if(h(i).le.tiny(ONE))then
               print *,'You are trying to use spline kernel softening'
               print *,'for gravity but there are nbody particles'
               print *,'with zero smoothing length h!'
               stop 'Zero smoothing lengths for nbody particles!'
            endif
         enddo
      endif
!
! check for valid sph smoothing lengths (always needed for hydro)
!
      do i=1,npart_sph
         if(h(i).le.tiny(ONE))then
            print *,'Particle number:',i
            print *,'There are SPH particles with zero smoothing'
            print *,'length h! The SPH method for hydro breaks'
            print *,'down if no neighbour particles can be found.'
            stop 'Zero smoothing length for SPH particle!'
         endif
      enddo
!
! if cosmo, check if the time from the dump file is consistent with the one
! computed from the scale factor / redshift stored in the dump file
!
      if(iexpand.eq.1)then
! 
! compute time cosm_t from the redshift in the dump file, for
! consistency check 
!
         ired=0
         redshift=(one/ascale)-one
         call cosmic_time(redshift,cosm_t,omegam,omegal,cosm_h0)

         if ( abs(cosm_t - t).gt.1.d-6 ) then 
            print *,'The cosmological scale factor stored in the dump'
            print *,'file is inconsistent with the time stored there,'
            print *,'given the cosmological parameters specified.'
            print *,'This could also result from problems with the'
            print *,'time units.'
            print *,'dump time    :',t
            print *,'redshift time:',cosm_t
            stop 'Incompatible times'
         endif
!
! check if SPH particles are included, as currently the expansion is not included
! in the SPH modules
!
         if(npart_sph.gt.0)then
            print *,'You have selected cosmological expansion and'
            print *,'there are SPH particles. Currently, the expansion'
            print *,'is not included in the SPH modules and you can'
            print *,'only use dark matter, i.e. N-body particles.'
            stop 'Cosmological expansion not included for SPH!'
         endif
      endif

      print *,'====================================================='
      print *,' '
      print *,'Your dump file appears to be sane. Be happy.'
      print *,' '
      print *,'====================================================='
      return
      end subroutine dumpfile_sanitycheck
!---------------------------------------------------------------------
      subroutine readheader
!*************************************************************************
!                                                                        *
! This routine reads information stored in the header of the dump        *
!                                                                        *
!*************************************************************************
      integer i,id,nindx
!
! integer header. max handles case of no saved data 
!
      npart        = max(0,iheader(id_npart    ))
      npart_sph    = max(0,iheader(id_npart_sph))
      npoim        = max(0,iheader(id_npoim    ))
      nstep        = max(0,iheader(id_nstep    ))
      n_dtmax_steps= max(0,iheader(id_n_dtmax  ))
      n_ana_steps  = max(0,iheader(id_n_ana    ))
!
! same for double precision header
!
      t        = dheader(id_t       )
!      dtmax    = dheader(id_dtmax  )  !now overridable from insph 
      tnext_out= dheader(id_tnextout)  !used only if reading a tmp dump
      tlast_out= dheader(id_tlastout)  !used only if reading a tmp dump
      deltnew  = dheader(id_deltnew )
      gamma    = dheader(id_gamma   )
      ekin     = dheader(id_ekin    )
      egrav    = dheader(id_egrav   )
      etherm   = dheader(id_etherm  )
      ascale   = dheader(id_ascale  )
      cosbox   = dheader(id_cosbox  )
      uconst   = dheader(id_uconst  )
      uexpon   = dheader(id_uexpon  )
      rcore    = dheader(id_rcore   )

      rhozero  = dheader(id_rhozero )
      eosK     = dheader(id_eosK    )

      return     
      end subroutine readheader
!---------------------------------------------------------------------
      subroutine fillheader(idump)
!***************************************************************************
!                                                                          *
! This routine converts the various code parameters into the information   *
! stored in the header of dump files.                                      *
!                                                                          *
!***************************************************************************
      integer, intent(in):: idump
      integer i,id,nindx
!
! initially fill header with -100 and -huge(1d0)
!
      iheader = -100
      dheader = -huge(1d0)
!
! integer header
!
! first write compiled in values

      iheader(id_iheadlen) = iheadlength
      iheader(id_ndim    ) = ndim
      iheader(id_intgrtr ) = id_myintg
      iheader(id_maxclumpsize) = maxclumpsize
      iheader(id_maxbunchsize) = maxbunchsize
!
! then values that are not in the insph file
!      
      iheader(id_npart     ) = npart
      iheader(id_npart_sph ) = npart_sph
      iheader(id_npoim     ) = npoim
      iheader(id_nstep     ) = nstep
      iheader(id_n_dtmax   ) = n_dtmax_steps
      iheader(id_n_ana     ) = n_ana_steps
!
! now those from insph
!
      iheader(id_idump     ) = idump
      iheader(id_makechkpnt) = makechkpnt
      iheader(id_idbg      ) = idbg
      iheader(id_itme      ) = itme
      iheader(id_ibctype   ) = ibctype
      iheader(id_indts     ) = indts
      iheader(id_iusebin   ) = iusebin
      iheader(id_fullextrap) = ifullextrap
      iheader(id_revise    ) = irevise
      iheader(id_ipres     ) = ipres
      iheader(id_ipdv      ) = ipdv
      iheader(id_ieos      ) = ieos
      iheader(id_icool     ) = icool
      iheader(id_iheat     ) = iheat
      iheader(id_ivisc     ) = ivisc
      iheader(id_ivtime    ) = ivtime
      iheader(id_ibals     ) = ibals
      iheader(id_ishok     ) = ishok
      iheader(id_igrav     ) = igrav
      iheader(id_isoft     ) = isoft
      iheader(id_imac      ) = imac
      iheader(id_maxbuild  ) = maxbuild
      iheader(id_ihts      ) = i_hts
      iheader(id_iexpand   ) = iexpand
      iheader(id_ncdumps   ) = ncdumps
      iheader(id_ishift    ) = ishift 
      iheader(id_iunits    ) = iunits 
      iheader(id_io_fmt    ) = io_fmt
! 
! double precision header
!      
      dheader(id_t        ) = t
      dheader(id_dtmax    ) = dtmax
      dheader(id_dtmin    ) = dtmin
      dheader(id_dt_ana   ) = dt_ana
      dheader(id_deltnew  ) = deltnew
      dheader(id_gamma    ) = gamma
      dheader(id_rhozero  ) = rhozero
      dheader(id_eosK     ) = eosK
      dheader(id_ekin     ) = ekin
      dheader(id_egrav    ) = egrav
      dheader(id_etherm   ) = etherm
      if(iexpand.eq.1)then
         dheader(id_ascale   ) = ascale
         dheader(id_cosbox   ) = cosbox
      else
         dheader(id_ascale   ) = -huge(1d0)
         dheader(id_cosbox   ) = -huge(1d0)
      endif
! 
! now those from insph
!
      dheader(id_vlength  ) = vlength
      dheader(id_alfmax   ) = alphamax
      dheader(id_alfstar  ) = alphastar
      dheader(id_betastar ) = betastar
      dheader(id_tol      ) = tol
      dheader(id_tolpm    ) = tolpm
      dheader(id_ftol     ) = ftol
      dheader(id_vtol     ) = vtol
      dheader(id_ftolpm   ) = ftolpm
      dheader(id_vtolpm   ) = vtolpm
      dheader(id_cfl      ) = cfl
      dheader(id_dhmax    ) = dhmax
      dheader(id_treeacc  ) = treeacc
      dheader(id_sepmax   ) = sepmax
      dheader(id_bsepmax  ) = bsepmax
      dheader(id_hmin     ) = hmin
      dheader(id_hmax     ) = hmax
      dheader(id_eps      ) = eps
      dheader(id_dtinit   ) = dtinit
      dheader(id_tstop    ) = tstop
      dheader(id_dt_out   ) = dt_out
      dheader(id_dt_ana   ) = dt_ana
      dheader(id_tnextout ) = tnext_out
      dheader(id_tlastout ) = tlast_out
      dheader(id_uconst   ) = uconst
      dheader(id_uexpon   ) = uexpon
      dheader(id_rcore    ) = rcore
      dheader(id_xmas     ) = xmas
      dheader(id_ylen     ) = ylen
      dheader(id_umin     ) = umin
      dheader(id_xmin     ) = xmin
      dheader(id_xmax     ) = xmax
      dheader(id_ymin     ) = ymin
      dheader(id_ymax     ) = ymax
      dheader(id_zmin     ) = zmin
      dheader(id_zmax     ) = zmax
      dheader(id_taucool  ) = taucool
      if(iexpand.eq.1)then
         dheader(id_h_0   ) = h_0
         dheader(id_omegam) = omegam
      else
         dheader(id_h_0   ) = -huge(1d0)
         dheader(id_omegam) = -huge(1d0)
      endif
      dheader(id_gmasslim ) = gmasslim
      dheader(id_clhfrac  ) = clhfrac

      return
      end subroutine fillheader
!---------------------------------------------------------------------
      subroutine setunits
!********************************************************************
!  Set the system of units. The code works in a system of units     *
!  where G=1 (for example), but the I/O dumps may have units,       *
!  since presumably we want to connect the system to our own        *
!  physical reality.                                                *
!                                                                   *
!  The code expects the following to be available:                  *
!        em_unit           mass                                     *
!        d_unit            distance                                 *
!        t_unit            time                                     *
!        vel_unit          velocity                                 *
!        u_unit            specific energy                          *
!        sd_unit           surface density                          *
!        rho_unit          volume density                           *
!        p_unit            pressure                                 *
!        ent_unit          entropy: in other words, units of the    *
!                                      K in p = K rho**gamma        *
!        amom_unit         angular momentum                         *
!                                                                   *
!                                                                   *
!                                                                   *
!  for cosmological boundary conditions the code expects            *
!  input from the grafic2 software by Edmund Bertschinger:          *
!  masses in M_tot(=1)*Omega_m/npart, positions in comoving Mpc,    *
!  peculiar velocities in km/s.                                     *
!  This corresponds the following system of code units:             *
!  G=1,rho_crit=1,boxsize=2 for a periodic box from -1 to 1.        *
!                                                                   *
!  rho_crit=h^2 * 1.878820056 -29 g/cm^3                            *
!                                                                   *
!  d_unit= boxsize_in_Mpc * 0.5 * 3.085678022e24 cm                 *
!  vel_unit= h * boxsize_in_Mpc * 17.27470748 km/s                  *
!  t_unit=1/h * 8.93120195e17 s or 1/h * 2.830077793e10 years       *
!  em_unit= h^2/8 * (boxsize_in_Mpc)^3 * 2.774971701e11 M_sol       *
!                                                                   *
!  To use the output from grafic2 convert                           *
!                                                                   *
!  d_unit   = (r_grafic-boxhalf_in_Mpc)/boxhalf_in_Mpc,             *
!                               the box is periodic from -1 to 1!   *
!  vel_unit = v_grafic*ascale/v_unit                                *
!  em_unit  = 8 * m_grafic,                                         *
!            as the simulation volume which is boxlength^3 = 8      *
!                                                                   *
!********************************************************************
      character(len=20), parameter:: where =('setunits')
      if(idbg.gt.2)print *,where

      select case(iunits)
      
      case(0)                        !scale free units

         em_unit  =1d0
         d_unit   =1d0
         t_unit   =1d0
         vel_unit =1d0

      case(1)                        !an example set of physical units

         em_unit  =xmas*Msun
         d_unit   =ylen*OneAu
         t_unit   =sqrt(d_unit**3/em_unit/bigG)
         vel_unit =d_unit/t_unit

      case(2)                        !an example of cosmology units.
                                     !using output from Bertschingers Grafic2
         em_unit  = 1d0/8d0
         d_unit   = half*cosbox
         vel_unit = (17.27470748d0*cosbox*h_0)/(1d0+redshift)
!         t_unit   = 2.830077793d10/h_0
         t_unit   = 1d0
         if(abs(d_unit/vel_unit -t_unit).gt.1d-10)then
            print *,'  d_unit=',d_unit
            print *,'vel_unit=',vel_unit
            print *,'  t_unit=',t_unit
            print *,'one or more of d_unit,vel_unit or t_unit'
            print *,'are not self consistent. Hope thats ok'
         endif

      case(3)                        ! units from Bate
         em_unit = xmas*Msun
         d_unit  = ylen*parsec
         t_unit  = sqrt(d_unit**3/em_unit/bigG)
         vel_unit= d_unit/t_unit

      case default 
         stop'setunits: Unknown unit system'

      end select

      u_unit   = vel_unit**2
      sd_unit  = em_unit/d_unit**2
      rho_unit = em_unit/d_unit**3
      p_unit   = (em_unit*d_unit**2/t_unit**2)/d_unit**2 !force/area
      ent_unit = p_unit/rho_unit**gamma
      amom_unit= em_unit*d_unit*vel_unit

      return
      end subroutine setunits
!---------------------------------------------------------------------
      subroutine phys2code
!********************************************************************
!                                                                   *
! conversions between input and code units                          *
!                                                                   *
! Note about call to setunits: except for the first time,           *
!     calling this routine call is redundant, but harmless.         *
!     For cosmology, we do need to call regularly, because          *
!     the units depend on redshift.                                 *
!                                                                   *
!********************************************************************
      integer i,id
!
! first define the unit system
!

      if(iunits.eq.2)then

         cosm_h0 = SQRT(8d0*PI/3d0) ! h0 in code units assuming
                                    ! rho_crit=1, G=1
         redshift=(one/ascale)-one

         call setunits
!
! dtmax has to be set to (t_stop-t_init)/1000,
! at least 1000 steps are performed
!
         dtinit =dtmax/8d0

!$omp parallel do default(none)
!$omp& firstprivate(d_unit)
!$omp& private(i,id) shared(npart,posmas)
         do i=1,npart
            do id=1,ndim
               posmas(id,i)=(posmas(id,i)-d_unit)/d_unit
            enddo
         enddo
         do i=1,npoim
            do id=1,ndim
               posmaspm(id,i)=(posmaspm(id,i)-d_unit)/d_unit
            enddo
         enddo

      else
         call setunits

!$omp parallel do default(none)
!$omp& firstprivate(d_unit)
!$omp& private(i,id) shared(npart,posmas)
         do i=1,npart
            do id=1,ndim
               posmas(id,i)=posmas(id,i)/d_unit
            enddo
         enddo
         do i=1,npoim
            do id=1,ndim
               posmaspm(id,i)=posmaspm(id,i)/d_unit
            enddo
         enddo

      endif
!
! Miscellaneous input variables
!
      t        = t/t_unit
      deltnew  = deltnew/t_unit
      dtinit   = dtinit/t_unit
      dtmax    = dtmax/t_unit
      dtmin    = dtmin/t_unit
      tstop    = tstop/t_unit
      dt_out   = dt_out/t_unit
      dt_ana   = dt_ana/t_unit
      tnext_out= tnext_out/t_unit
      tlast_out= tlast_out/t_unit

      gmasslim=gmasslim/em_unit

      xmin=xmin/d_unit
      xmax=xmax/d_unit
      ymin=ymin/d_unit
      ymax=ymax/d_unit
      zmin=zmin/d_unit
      zmax=zmax/d_unit
      sepmax=sepmax/d_unit
      hmin =hmin/d_unit
      hmax =hmax/d_unit
      eps  =eps/d_unit
      rcore=rcore/d_unit
      rdisk=rdisk/d_unit

      ekin   = ekin/(u_unit*em_unit)
      egrav  = egrav/(u_unit*em_unit)
      etherm = etherm/(u_unit*em_unit)
      umin   = umin/u_unit
      uconst = uconst/u_unit

      rhozero = rhozero/rho_unit
      eosK    = eosK/ent_unit

!
! Arrays
!
!$omp parallel default(none)
!$omp& firstprivate(em_unit,vel_unit,d_unit,sd_unit,rho_unit,u_unit)
!$omp& shared(npart,npart_sph,posmas,vel,h,rho,poten,u)
!$omp& private(i,id)
!$omp do
      do i=1,npart
         h     (    i)=h(i)/d_unit
         poten (    i)=poten(i)/u_unit
         posmas(ims,i)=posmas(ims,i)/em_unit
         do id=1,ndim
            vel(id ,i)=vel(id,i)/vel_unit
         enddo
      enddo
!$omp do
      do i=1,npart_sph
         u  (i)=u  (i)/u_unit
#if NDIM==2
         rho(i)=rho(i)/sd_unit
#else
         rho(i)=rho(i)/rho_unit
#endif
      enddo
!$omp end parallel

      do i=1,npoim
         hpm     (    i)=hpm  (i)/d_unit
         potpm   (    i)=potpm(i)/u_unit
         posmaspm(ims,i)=posmaspm(ims,i)/em_unit
         do id=1,ndim
            velpm(id,i)=velpm(id,i)/vel_unit
            spinpm(id,i)=spinpm(id,i)/(em_unit*d_unit*vel_unit)
         enddo
      enddo

10    format(10(1pd18.10))
11    format(i7,10(1pd12.4))
      return
      end subroutine phys2code
!---------------------------------------------------------------------
      subroutine code2phys
!********************************************************************
!                                                                   *
! conversion to physical units.                                     *
!                                                                   *
! Note about call to setunits: except for the first time,           *
!     calling this routine call is redundant, but harmless.         *
!     For cosmology, we do need to call regularly, because          *
!     the units depend on redshift.                                 *
!                                                                   *
!********************************************************************
      integer i,id,iexit

      if(iexpand.eq.1)then

         call setunits

!$omp parallel do default(none)
!$omp& firstprivate(d_unit)
!$omp& private(i,id) shared(npart,posmas)
         do i=1,npart
         do id=1,ndim
            posmas(id,i)=d_unit+d_unit*posmas(id,i)
         enddo
         enddo
         do i=1,npoim
         do id=1,ndim
            posmaspm(id,i)=d_unit+d_unit*posmaspm(id,i)
         enddo
         enddo

      else

         call setunits

!$omp parallel do default(none)
!$omp& firstprivate(d_unit)
!$omp& private(i,id) shared(npart,posmas)
         do i=1,npart
         do id=1,ndim
            posmas(id,i)=d_unit*posmas(id,i)
         enddo
         enddo
         do i=1,npoim
         do id=1,ndim
            posmaspm(id,i)=d_unit*posmaspm(id,i)
         enddo
         enddo

      endif

      t      =t*t_unit
      deltnew=deltnew*t_unit
      dtinit =dtinit*t_unit
      dtmax  =dtmax*t_unit
      dtmin  =dtmin*t_unit
      tstop  =tstop*t_unit
      dt_out =dt_out*t_unit
      dt_ana =dt_ana*t_unit
      tnext_out=tnext_out*t_unit
      tlast_out=tlast_out*t_unit


      gmasslim=gmasslim*em_unit

      xmin=xmin*d_unit
      xmax=xmax*d_unit
      ymin=ymin*d_unit
      ymax=ymax*d_unit
      zmin=zmin*d_unit
      zmax=zmax*d_unit
      sepmax=sepmax*d_unit
      hmin =hmin*d_unit
      hmax =hmax*d_unit
      eps  =eps*d_unit
      rcore=rcore*d_unit
      rdisk=rdisk*d_unit

      ekin   = ekin*(u_unit*em_unit)
      egrav  = egrav*(u_unit*em_unit)
      etherm = etherm*(u_unit*em_unit)
      umin   = umin*u_unit
      uconst = uconst*u_unit

      rhozero = rhozero*rho_unit
      eosK    = eosK*ent_unit

      iexit=0

!$omp parallel default(none)
!$omp& firstprivate(em_unit,vel_unit,d_unit,sd_unit,rho_unit,u_unit)
!$omp& shared(iexit)
!$omp& shared(npart,npart_sph,posmas,h,vel,rho,poten,u)
!$omp& private(i,id)
!$omp do
      do i=1,npart
         if(posmas(ims,i).eq.0d0)iexit=1
         h     (    i)=d_unit*h(i)
         poten (    i)=u_unit*poten(i)
         posmas(ims,i)=em_unit*posmas(ims,i)
         do id=1,ndim
            vel(id,i)=vel_unit*vel(id,i)
         enddo
      enddo
!$omp do 
      do i=1,npart_sph
         u  (i)=u_unit*u(i)
#if NDIM==2
         rho(i)=sd_unit*rho(i)
#else
         rho(i)=rho_unit*rho(i)
#endif
      enddo
!$omp end parallel

      if(iexit.eq.1)stop'zero mass particle'

      do i=1,npoim
         if(posmaspm(ims,i).eq.0.d0) stop'zero mass point mass'
         hpm     (    i)=d_unit*hpm(i)
         potpm   (    i)=u_unit*potpm(i)
         posmaspm(ims,i)=em_unit*posmaspm(ims,i)
         do id=1,ndim
            velpm(id,i)=vel_unit*velpm(id,i)
            spinpm(id,i)=spinpm(id,i)*(em_unit*d_unit*vel_unit)
         enddo
      enddo


      return
      end subroutine code2phys
!--------------------------------------------------------------------
      subroutine getlastdumpnr(idump)
!**********************************************************************
!                                                                     *
!  this routine assumes that a checkpoint dump has been read and      *
!  finds the number of the last regular dump file that exists.        *
!                                                                     *
!  Note: If getfname gets a negative number for its idump argument,   *
!  it assumes the last dump number is abs(idump).                     *
!                                                                     *
!  This variant is good for script files and automated restarting     *
!  from batch systems, where you may not always have access to the    *
!  specific dump number ahead of time or access to editing the insph  *
!  file at the right time.                                            *
!**********************************************************************

      integer, intent(inout):: idump
      logical tobe

      character(len=250) absnrfile

      character(len=20), parameter:: where =('getlastdumpnr')

      if(idbg.ge.2)print *,where

      absnrfile = trim(absolutepath) // nrfile
      inquire(file=absnrfile,exist=tobe)
      if(.not.tobe)then
         print *,'Attempting to restart from checkpoint dump file'
         print *,'but the file: ',absnrfile
         print *,'doesn''t exist. So the number of the last '
         print *,'regular dump file is unknown!'
         stop'Last regular dump file not found'
      else
         open(iterm,file=absnrfile)
         read(iterm,*)idump
         close(iterm)
         print *,' '
         print *,' '
         print *,'Last regular dump file number was: ',idump
         print *,' '
         print *,' '
      endif
      return
      end subroutine getlastdumpnr
!--------------------------------------------------------------------
      end module io
