      module grape
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
! This file contains the VINE module to do gravity using the         *
! GRAPE 3 special purpose hardware boards. The routines here         *
! are called from the tree.F module.                                 *
!                                                                    *
!  Written by Markus Wetzstein, Andy Nelson and Thorsten Naab        *
!*********************************************************************
      use dimens
      use treevars
      use miscvars, only: ibctype
      use indivtsvars
      use constants

      implicit none
      save

      private
      public grape_tree,grape_dirsum,lock_grape,unlock_grape

!  GRAPE-3 parameters
!
!set ig3boards to number of g3 boards used
!
      integer, parameter :: ig3boards=1,ig3chips=ig3boards*8
      integer grapelock
!
! gravitational smoothing length for GRAPE
!
      real g3epsarr(3,3)

#if GRAPE==3

#else
      grape3 module: No or wrong grape preprocessor flag exists!
#endif

      contains
!-----------------------------------------------------
      subroutine lock_grape()

#ifdef _OPENMP
      print*,'You attempted to use both OpenMP and GRAPE!'
      print*,'Each GRAPE board must only be accessed by one thread.'
      print*,'Thus using OpenMP and GRAPE together would result in a'
      print*,'crash, as in this version of VINE the boards are'
      print*,'shared between the threads.'
      stop
#endif

      call g3init()
      
      return
      end subroutine lock_grape
!-----------------------------------------------------

      subroutine unlock_grape()

      call g3free()

      return
      end subroutine unlock_grape

!-----------------------------------------------------

      subroutine grape_tree(iclump,eps2,npart,nact,listact,nla,
     $            latoms,vel,acc,poten)

! vel,npart isn't used here, it is just passed because the G6 routine needs it
! and the call sequence has to be the same for all grape_* routines
! in the different files

      integer, intent(in):: iclump,npart
      integer, intent(in):: nact, listact(nact)
      integer, intent(in):: nla , latoms(nla)
      double precision, intent(in):: eps2
      double precision, intent(in):: vel(ndim,npart)
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      integer g3nchips,g3nboards,g3nchipsperboard,g3maxptcl
      double precision gconst,g3massmin,g3massmax,dxmax,xcl,ycl,zcl,dx
      real g3pos(3,ig3chips),g3acc(3,ig3chips),g3pot(ig3chips)

      integer ichip,nchips,nchipsboard,nchipsloop,nchipremain
      integer npartmax,nboards,i,ii,j,k
      integer nlimlow,nlimhigh,jg3mem
      real g3eps2,g3spcres,g3massres,g3massmem,g3res,g3masslim
      real scalemass
      character(len=20), parameter:: where=('grape_tree')

!     set parameters for GRAPE3     
      gconst=one
      nchips      = g3nchips()        
      g3eps2      = real(eps2)    ! set grav. smoothing length square
      nchipsboard = g3nchipsperboard() ! get number of chips per
      npartmax    = g3maxptcl() ! get maximum number of particles
      nboards     = g3nboards() ! get number of boards (usually 1)

      if (nchips .ne. ig3chips) then 
         print *,'nchips',nchips
         print *,'ig3chips',ig3chips
         print *,'Wrong number of chips in ',where
         print *,'Check module grape3.F'
         stop
      endif

!     set spatial and mass resolution       
      xcl=treept(1,iclump)
      ycl=treept(2,iclump)
      zcl=treept(3,iclump)

      g3massmin=treept(ims,latoms(1))
      g3massmax=g3massmin
      dxmax=1d-6
      DO i=1,nla
         ii=latoms(i)
         dx = abs(treept(1,ii)-xcl)
         dxmax=max(dx,dxmax)
         dx = abs(treept(2,ii)-ycl)
         dxmax=max(dx,dxmax)
         dx = abs(treept(3,ii)-zcl)
         dxmax=max(dx,dxmax)
         g3massmin=min(treept(ims,ii),g3massmin)
         g3massmax=max(treept(ims,ii),g3massmax)
      ENDDO

      dxmax=dxmax+two*treept(ihnode,iclump)
      g3spcres=real(dxmax)/262144e0
      g3massres=real(g3massmin)

      g3res = sqrt(g3eps2) / g3spcres
      g3masslim = g3massres * (64.0 * MIN(g3res,g3res**2.))
      if(g3masslim.lt.g3massmax)then
         print*,'Upper mass limit too low:',g3masslim,' Should be:',
     $        g3massmax
         print*,'Try decreasing gmasslim parameter in insph.'
         stop
!         g3massres=g3massmax/(64.0 * MIN(g3res,g3res**2))
!         g3spcres=max(g3massres*64.*sqrt(eps2)/g3massmax,
!     $        sqrt(eps2*g3massres*64./g3massmax))
      endif

      CALL g3setscales(g3spcres,g3massres) 
!      
! mass correction factor for tree code
!
      if(npart.GT.100000) then
        scalemass=1e5/real(npart)
      else
         scalemass=one
      endif 

      CALL g3setmasscor(scalemass)
!
! acc computation
!      
      nlimlow=1
      nlimhigh=npartmax            
 1000 continue 
      if (nla .LE. nlimhigh) nlimhigh = nla

      CALL g3setn(nlimhigh-nlimlow+1) ! set number of particles used
      jg3mem=0

      do i=nlimlow,nlimhigh
         ii=latoms(i)
         g3pos(1,1) = real(treept(1,ii))
         g3pos(2,1) = real(treept(2,ii))
         g3pos(3,1) = real(treept(3,ii))
         g3massmem  = real(treept(ims,ii))
         CALL g3setxj(jg3mem,g3pos(1,1))
         CALL g3setmj(jg3mem,g3massmem)
         jg3mem=jg3mem+1
      enddo

      do i=1,nchips             ! set ALL chips to zero
         CALL g3seth2(0e0,i-1)  ! including the broken one.
      enddo  
      CALL g3seteps2(g3eps2)     ! set gravitational smoothing length
      
      nchipsloop = nchips
      do ichip=1,nact,nchipsloop
         nchipremain=nact-ichip+1
         if(nchipremain .lt. nchips) then 
            nchipsloop = nchipremain
         endif
         do i=1,nchipsloop 
            j=ichip+i-1
            k=listact(j)
            g3pos(1,i)=real(treept(1,k))
            g3pos(2,i)=real(treept(2,k))
            g3pos(3,i)=real(treept(3,k))
         enddo

         CALL g3frc(g3pos,g3acc,g3pot,nchipsloop)         
         
         do i=1,nchipsloop      ! loop all chips
            j=ichip+i-1         ! correct for offset
            ii=linktreetosph(listact(j))
            IF (ii .EQ. 0)then
               print*,'Wrong index in listact in file g3_acc!!!'
               stop
            endif
            acc(1,ii) = gconst*dble(g3acc(1,i))+acc(1,ii)   ! they are stored 
            acc(2,ii) = gconst*dble(g3acc(2,i))+acc(2,ii)   ! at real particle
            acc(3,ii) = gconst*dble(g3acc(3,i))+acc(3,ii)   ! numbers.
            poten(ii)=poten(ii)+gconst * dble(g3pot(i))     ! ditto
         enddo
      enddo

      if (nla .gt. nlimhigh) then ! Loop back for another shift.
         nlimlow  = nlimlow  + npartmax
         nlimhigh = nlimhigh + npartmax
         goto 1000
      endif
      
      return
      end subroutine grape_tree

!-----------------------------------------------------

      subroutine grape_dirsum(npart,eps2,posmas,vel,acc,poten)

      integer, intent(in):: npart
      double precision, intent(in   ):: eps2
      double precision, intent(in   ):: posmas(ndim1,npart)
      double precision, intent(in   ):: vel   (ndim ,npart)
      double precision, intent(inout):: acc(ndim,npart),poten(npart)

      integer g3nchips,g3nboards,g3nchipsperboard,g3maxptcl
      real g3posmem(3),g3pos(3,ig3chips),g3acc(3,ig3chips)
     $     ,g3pot(ig3chips) 

      double precision gconst,g3massmin,g3massmax,dxmax
      integer ichip,nchips,nchipsboard,nchipsloop,nchipremain
      integer nboards,nlimlow,nlimhigh,jg3mem
      integer i,ii,j,k,npartmax
      real g3eps2,g3massres,g3massmem,g3masslim,scalemass
      real g3spcres,g3res

      character(len=20), parameter:: where=('grape_dirsum')
      if(idbg.ge.2)print *,where

      gconst=one

      nchips      = g3nchips()        
      g3eps2      = real(eps2)       ! set grav. smoothing length square
      nchipsboard = g3nchipsperboard() ! get number of chips per
      npartmax    = g3maxptcl() ! get maximum number of particles
      nboards     = g3nboards() ! get number of boards (usually 1)
      if (nchips .ne. ig3chips) then 
         print*,'nchips',nchips
         print*,'ig3chips',ig3chips
         print*,'Wrong number of chips in ',where
         print*,'Check module grape3.F'
         stop
      endif
      
      g3massmin=posmas(ndim1,1)
      g3massmax=g3massmin         
      do i=1,npart
         g3massmin=min(g3massmin,posmas(ndim1,i))
         g3massmax=max(g3massmax,posmas(ndim1,i))
      enddo

      if(ibctype.eq.0)then
         dxmax=128.
      else
         dxmax=sqrt(12.)
      endif
      
      g3spcres=dxmax/262144.
      g3massres=real(g3massmin)
      g3res = sqrt(g3eps2) / g3spcres
      g3masslim = g3massres * (64.0 * MIN(g3res,g3res**2))

!      print*,'g3masslim:',g3masslim,' g3massres:',g3massres,
!     $     ' g3spcres:',g3spcres,' g3res:',g3res

      if(g3masslim.lt.g3massmax)then
         print*,'Upper mass limit too low:',g3masslim,' Should be:',
     $        g3massmax
         stop
!     g3massres=g3massmax/(64.0 * MIN(g3res,g3res**2))
!     g3spcres=max(g3massres*64.*sqrt(eps2)/g3massmax,
!     $        sqrt(eps2*g3massres*64./g3massmax))
      endif

      CALL g3setscales(g3spcres,g3massres) 
!      
! mass correction factor for tree code
!
      scalemass=1.
      CALL g3setmasscor(scalemass)
!     
! acceleration computation
!      
      nlimlow=1
      nlimhigh=npartmax            

 1000 continue 
      if (npart .LE. nlimhigh) nlimhigh = npart
      
      CALL g3setn(nlimhigh-nlimlow+1) ! set number of particles used
      jg3mem=0
      do i=nlimlow,nlimhigh
         g3posmem(1) = real(posmas(1,i))
         g3posmem(2) = real(posmas(2,i))
         g3posmem(3) = real(posmas(3,i))
         g3massmem   = real(posmas(ndim1,i))
         CALL g3setxj(jg3mem,g3posmem(1))
         CALL g3setmj(jg3mem,g3massmem)
         jg3mem=jg3mem+1
      enddo
      do i=1,nchips             ! set ALL chips to zero
         CALL g3seth2(0.0,i-1) ! including the broken one.
      enddo  

      CALL g3seteps2(g3eps2)     ! set gravitational smoothing length
      
      nchipsloop = nchips

      do ichip=1,nactive,nchipsloop
         nchipremain=nactive-ichip+1
         if(nchipremain .lt. nchips) then 
            nchipsloop = nchipremain
         endif
         do i=1,nchipsloop 
            j=ichip+i-1
            k=iactive(j)
            g3pos(1,i)=real(posmas(1,k))
            g3pos(2,i)=real(posmas(2,k))
            g3pos(3,i)=real(posmas(3,k))
         enddo

         CALL g3frc(g3pos,g3acc,g3pot,nchipsloop)         
         
         do i=1,nchipsloop      ! loop all chips
            j=ichip+i-1         ! correct for offset
            ii=iactive(j)
            IF (ii .EQ. 0) stop
            acc(1,ii)=gconst*dble(g3acc(1,i))+acc(1,ii)   ! they are stored 
            acc(2,ii)=gconst*dble(g3acc(2,i))+acc(2,ii)   ! at real particle
            acc(3,ii)=gconst*dble(g3acc(3,i))+acc(3,ii)   ! numbers.
            poten(ii)=poten(ii)+gconst*dble(g3pot(i))     ! ditto
         enddo
      enddo

      return
      end subroutine grape_dirsum

      end module grape
