      module grape
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
! This file contains the VINE module to do gravity using the         *
! GRAPE 5 special purpose hardware boards. The routines here         *
! are called from the tree.F module.                                 *
!                                                                    *
!  Written by Markus Wetzstein, Andy Nelson and Thorsten Naab        *
!*********************************************************************
      use dimens
      use treevars
      use indivtsvars
      use constants 

      implicit none
      save

      private
      public lock_grape,unlock_grape,grape_tree,grape_dirsum


#include "gp5util.h"
#define NJMAX (JMEMSIZE)

! Nr. of boards to use
      integer, parameter :: maxg5boards=1


! Set g5xmin,g5xmax to maximum extend of simulated system. This is
! very important for the scaling to the internal GRAPE-5 number format.
! Wrong settings here can lead to wrong forces!!! See the G5 manual for
! more information.

	double precision, parameter :: g5xmin=-200.d0, g5xmax=200.0d0

#if GRAPE==5

#else
      grape5 module:No or wrong grape preprocessor flag exists
#endif

      contains
      subroutine lock_grape()

!*************************************************************************
! This acquires the GRAPE-5 boards. Currently all boards from the 1st one*
! up to maxg5boards are acquired.                                        *
!                                                                        *
! Written by M. Wetzstein, Aug 2002, modified Dec. 2002                  *
!                                                                        *
!*************************************************************************

      integer i,ic,ifirstg5
      integer g5_get_firstcluster

#ifdef _OPENMP
      print *,'You attempted to use both OpenMP and GRAPE!'
      print *,'Each GRAPE board must only be accessed by one thread.'
      print *,'Thus using OpenMP and GRAPE together would result in a'
      print *,'crash, as in this version of VINE the boards are'
      print *,'shared between the threads.'
      stop
#endif

      ifirstg5=g5_get_firstcluster()
      do i=1,maxg5boards
         ic=ifirstg5+i-1
         call g5_openMC(ic)
      enddo

      return
      end subroutine lock_grape
!-----------------------------------------------------

      subroutine unlock_grape()

!*************************************************************************
! This frees all the GRAPE-5 boards formerly acquired                    *
!                                                                        *
! Written by M. Wetzstein, Aug 2002 , modified Dec. 2002                 *
!                                                                        *
!*************************************************************************

      integer i,ic,ifirstg5
      integer g5_get_firstcluster

      ifirstg5=g5_get_firstcluster()
      do i=1,maxg5boards
         ic=ifirstg5+i-1
         call g5_closeMC(ic)
      enddo

      return
      end subroutine unlock_grape
!-----------------------------------------------------

      subroutine grape_tree(iclump,eps2,npart,nact,listact,
     &                      nla,latoms,vel,acc,poten)


!*************************************************************************
! This computes grav. accelerations on GRAPE-5. Boards have to be opened *
! outside this routine. Designed for use in multi-threaded environments, *
! but can be used in serial version, too.                                *
!                                                                        *
! Written by M. Wetzstein, T. Naab, Aug 2001                             *
! Modified by M. Wetzstein, Aug 2002, Dec. 2002                          *
!                                                                        *
!*************************************************************************

! vel,iclump,nclump isn't used here, it is just passed because the G6 routine
! needs it and the call sequence has to be the same for all grape_*
! routines in the different files

      integer, intent(in):: iclump,npart
      integer, intent(in):: nact,listact(nact)
      integer, intent(in):: nla,latoms(nact)

      double precision, intent(in):: eps2
      double precision, intent(in   ):: vel(ndim,npart)
      double precision, intent(inout):: acc(ndim,npart)
      double precision, intent(inout):: poten(npart)

      integer npipes,nii,ifirstg5,npartmax
      integer i,ii,j,ic,nlimlow,nlimhigh,ng5part,ioffs,ioffr

      double precision emmin
      integer g5_get_number_of_pipelines_per_board,
     $        g5_get_firstcluster

      double precision g5posmem(3,NJMAX),g5_pos(3,maxbunchsize)
      double precision g5massmem(NJMAX),g5_acc(3,maxbunchsize)
      double precision g5_phi(maxbunchsize), gconst
      double precision g5eps,g5mmin

      gconst=one
      npipes=g5_get_number_of_pipelines_per_board()
      ifirstg5=g5_get_firstcluster()
      npartmax=NJMAX

! Determine mass resolution      
      
      emmin=1d9

      do i=1,nla
         j=latoms(i)
         emmin=min(emmin,treept(ims,j))
      enddo

! Important: this division of emmin needed for correct mass range!
      g5mmin=emmin/1.d2    
      g5eps=sqrt(eps2)
      do i=1,maxg5boards
         ic=ifirstg5+i-1
         call g5_set_rangeMC(ic,g5xmin,g5xmax,g5mmin)
      enddo             ! set spatial and mass resolution

      
! Write positions of active particles into array, so that they can be
! accessed consecutively
      do j=1,nact
         i=listact(j)
         g5_pos(1,j)=treept(1,i)
         g5_pos(2,j)=treept(2,i)
         g5_pos(3,j)=treept(3,i)
      enddo      

! outmost loop over pipelines, needed if nla > npartmax
      nlimlow=1
      nlimhigh=npartmax
 1000 continue
      if(nla .lt. nlimhigh) nlimhigh=nla
      ng5part=nlimhigh-nlimlow+1

! Write positions and masses of entries in interaction list into array, so
! that they can be accessed consecutively
      j=0
      do i=nlimlow,nlimhigh
         ii=latoms(i)
         j=j+1
         g5posmem(1,j) = treept(1,ii)
         g5posmem(2,j) = treept(2,ii)
         g5posmem(3,j) = treept(3,ii)
         g5massmem(j)  = treept(ims,ii)
      enddo

      do i=1,maxg5boards
         ic=ifirstg5+i-1
         call g5_set_nMC(ic,ng5part)              !compute ng5part particles
         call g5_set_mjMC(ic,0,ng5part,g5massmem) !write int list masses
         call g5_set_xjMC(ic,0,ng5part,g5posmem)  !write int list positions
         call g5_set_eps_to_allMC(ic,g5eps)       !write softening for pipes 
      enddo

! Write first set of particles which need accelerations and start their computation
      ioffs=0  !nr. of pipes used to compute so far
      i=1
      do while(i.le.maxg5boards .and. ioffs .lt.nact) 
            ! loop again if board available for uncomputed particles 
         ic=ifirstg5+i-1
         nii=npipes
         if(ioffs+nii .gt. nact) nii=nact-ioffs
         call g5_set_xiMC(ic,nii,g5_pos(1,ioffs+1))
         call g5_runMC(ic)
         ioffs=ioffs+nii
         i=i+1
      enddo

! Read accelerations from previous computation and start new computations if
! nact > nr. of pipelines available at once
      ioffr=0 ! nr. of pipes used to read accelerations off so far
      do while(ioffr .lt.nact)
         do i=1,maxg5boards
            ic=ifirstg5+i-1
            if(ioffr.lt.nact)then
               nii=npipes
               if(ioffr+nii .gt. nact) nii=nact-ioffr
               call g5_get_forceMC(ic,nii,g5_acc(1,ioffr+1),
     $              g5_phi(ioffr+1))
               do j=ioffr+1,ioffr+nii !sum partial accelerations
                  ii=linktreetosph(listact(j))
                  acc(1,ii) = gconst * g5_acc(1,j)+acc(1,ii)
                  acc(2,ii) = gconst * g5_acc(2,j)+acc(2,ii)
                  acc(3,ii) = gconst * g5_acc(3,j)+acc(3,ii) 
                  poten(ii)   = gconst * g5_phi(j)  +poten(ii)
               enddo
               ioffr=ioffr+nii
            endif
            if(ioffs .lt. nact)then ! uncomputed particles left?
               nii=npipes
               if(ioffs+nii .gt. nact) nii=nact-ioffs
               call g5_set_xiMC(ic,nii,g5_pos(1,ioffs+1))
               call g5_runMC(ic)
               ioffs=ioffs+nii
            endif
         enddo !board loop
      enddo !particles lop

      if (nla .gt. nlimhigh) then ! Loop back for another shift.
         nlimlow  = nlimlow  + npartmax
         nlimhigh = nlimhigh + npartmax
         goto 1000
      endif
      return
      end subroutine grape_tree

!-----------------------------------------------------------------

      subroutine grape_dirsum(npart,eps2,posmas,vel,acc,poten)

! ibctype can be used to adjust resolution scales for periodic boxes with
! known and fixed size

      integer, intent(in):: npart
      double precision, intent(in   ):: eps2
      double precision, intent(in   ):: posmas(ndim1,npart)
      double precision, intent(in   ):: vel(ndim,npart)
      double precision, intent(inout):: acc(ndim,npart)
      double precision, intent(inout):: poten(npart)

      double precision g5_pos(3,idim),g5_acc(3,idim),g5_phi(idim)
      double precision g5massmem(idim),emmin,gconst
      double precision g5mmin,g5eps,g5posmem(3,idim)

      integer npartmax,ifirstg5
      integer i,j,ic,nlimlow,nlimhigh,ng5part

      integer g5_get_firstcluster

      character(len=20), parameter:: where =('grape5_dirsum')
      if(idbg.ge.2)print *,where

      gconst=one
      npartmax=NJMAX
      ifirstg5=g5_get_firstcluster()

      do i=1,nactive
         j=iactive(i)
         g5_pos(1,i)=posmas(1,j)
         g5_pos(2,i)=posmas(2,j)
         g5_pos(3,i)=posmas(3,j)
      enddo
      do i=1,npart
         g5posmem(1,i)=posmas(1,i)
         g5posmem(2,i)=posmas(2,i)
         g5posmem(3,i)=posmas(3,i)
         g5massmem(i) =posmas(ndim1,i)
      enddo
!     
! set resolution 
!
      emmin=1.d8
      do i=1,npart
         emmin=min(emmin,posmas(ndim1,i))
      enddo
      
      g5mmin=emmin/1.d3    
      g5eps=sqrt(eps2)
      do i=1,maxg5boards
         ic=ifirstg5+i-1
         call g5_set_rangeMC(ic,g5xmin,g5xmax,g5mmin)
         call g5_set_eps_to_allMC(ic,g5eps)
      enddo

      nlimlow=1
      nlimhigh=npartmax

 1000 continue 
      if (npart .LT. nlimhigh) nlimhigh = npart
      ng5part=nlimhigh-nlimlow+1
      do i=1,maxg5boards
         ic=ifirstg5+i-1
         call g5_set_nMC(ic,ng5part) ! set number of particles used
         call g5_set_xjMC(ic,0,ng5part,g5posmem(1,nlimlow))
         call g5_set_mjMC(ic,0,ng5part,g5massmem)
      enddo

      call g5_calculate_force_on_x(g5_pos,g5_acc,g5_phi,nactive)

      do i=1,nactive
         j=iactive(i)
         acc(1,j)=acc(1,j) + gconst*g5_acc(1,i)
         acc(2,j)=acc(2,j) + gconst*g5_acc(2,i)
         acc(3,j)=acc(3,j) + gconst*g5_acc(3,i)
         poten(j)  =poten(j)   + gconst*g5_phi(i)
      enddo

      if (npart .gt. nlimhigh) then ! Loop back for another shift.
         nlimlow  = nlimlow  + npartmax
         nlimhigh = nlimhigh + npartmax
         goto 1000
      endif
      
      return
      end subroutine grape_dirsum
      end module grape
