      module grape
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
! This file contains the VINE module to do gravity using the         *
! GRAPE 6 special purpose hardware boards. The routines here         *
! are called from the tree.F module.                                 *
!                                                                    *
!  Written by Markus Wetzstein, Andy Nelson and Thorsten Naab        *
!*********************************************************************
      use dimens
      use treevars
      use indivtsvars
      use constants

      implicit none
      save

      private
      public unlock_grape,lock_grape,grape_tree,grape_dirsum

      integer, parameter :: npartmax=8*262144,ng6pipes=48

      double precision g6_accold(3,idim),g6_jold(3,idim),
     &                 g6_phiold(  idim)

#if GRAPE==6

#else
      grape6 module:No or wrong grape preprocessor flag exists!
#endif

      contains
!-----------------------------------------------------
      subroutine lock_grape()

#ifdef _OPENMP
      print*,'You attempted to use both OpenMP and GRAPE!'
      print*,'Each GRAPE board must only be accessed by one thread.'
      print*,'Thus using OpenMP and GRAPE together would result in a'
      print*,'crash, as in this version of VINE the boards are'
      print*,'shared between the threads.'
      stop
#endif
       
!-there is only one cluster in Heidelberg -> clusterid = 0           
      call g6_open(0)

      return
      end subroutine lock_grape
!-----------------------------------------------------

      subroutine unlock_grape()

!-there is only one cluster in Heidelberg -> clusterid = 0           
      call g6_close(0)

      return
      end subroutine unlock_grape

!-----------------------------------------------------

      subroutine grape_tree(iclump,eps2,npart,nact,listact,
     &                      nla,latoms,vel,acc,poten)

!*************************************************************************
! This computes grav. accelerations on GRAPE-6. Board has to be opened   *
! outside this routine.                                                  *
!                                                                        *
! Written by M. Wetzstein, T. Naab, Apr 2002                             *
! Modifications by M. Wetzstein, Apr 2002, Dec. 2002                     *
!                                                                        *
!*************************************************************************

! iclump,npart isn't used here, it is just passed because the G6 routine needs
! it the call sequence has to be the same for all grape_* routines
! in the different files

      integer, intent(in):: iclump,npart
      integer, intent(in):: nact,listact(nact)
      integer, intent(in):: nla ,latoms(nla)
      double precision, intent(in   ):: eps2
      double precision, intent(in   ):: vel(3,npart)
      double precision, intent(inout):: acc(ndim,npart)
      double precision, intent(inout):: poten(npart)

      double precision g6_pos(3,ng6pipes),g6_acc(3,ng6pipes)
      double precision g6_vel(3,ng6pipes),g6massmem
      double precision g6_phi(ng6pipes)
      double precision g6eps2,g6posmem(3),g6h(ng6pipes)
      double precision g6_j(3,ng6pipes),g6_jerk(3,idim)
      double precision g6_accold_temp(3,ng6pipes),
     &     g6_jold_temp(3,ng6pipes),g6_phiold_temp(ng6pipes)
      double precision gconst

      integer ig6ind(ng6pipes),g6calc_lasthalf
      integer i,ii,j,k,ierr,nharderr,nlimlow,nlimhigh,ng6part
      integer ioffs,npuse
      
      logical, save:: startup= .true.

      if(startup)then!init old accelerations, jerk, potential with educated guess
         do i=1,npart
            g6_accold(1,i)=one
            g6_accold(2,i)=one
            g6_accold(3,i)=one
            g6_jold(1,i)=1.d2
            g6_jold(2,i)=1.d2
            g6_jold(3,i)=1.d2
            g6_phiold(i)=one
         enddo
         startup=.false.
      endif

      do i=1,npart !init new jerk to 0
         g6_jerk(1,i)=zero
         g6_jerk(2,i)=zero
         g6_jerk(3,i)=zero
      enddo

      gconst=one
      nharderr=0

 100  continue  !entry point after hard board error, start from scratch

         call g6_set_xunit(51)  ! default 51 bits for representation of 
                                ! fractional part in position data
         call g6_set_tunit(51)  ! default 51 bits for representation of 
                                ! fractional part in time
         call g6_set_ti(0,zero)  ! time 0.d0, no prediction
         
! prepare softening and smoothing lengths
         g6eps2=eps2
         do i=1,ng6pipes
            g6h(i)=zero
         enddo

!         call g6_initialize_jp_buffer(0,100000)

         nlimlow=1
         nlimhigh=npartmax

 1000    continue 
         if (nla .LT. nlimhigh) nlimhigh = nla
         ng6part=nlimhigh-nlimlow+1
         
!
! write positions and masses of the interaction lists into grape usable
! arrays
!
         j=0
         do i=nlimlow,nlimhigh
            ii=latoms(i)

            g6posmem(1)=treept(1,ii)
            g6posmem(2)=treept(2,ii)
            g6posmem(3)=treept(3,ii)
            g6massmem  =treept(ims,ii)

            call g6_set_j_particle_multisend_mxfast(0,1,j,ii,
     $           g6massmem,g6posmem(1))
!            call g6_set_j_particle_mxonly(0,j,ii,
!     $           g6massmem,g6posmem(1))
            j=j+1
         enddo

!         call g6_flush_jp_buffer(0)

         ioffs=1
         do while(ioffs.le.nact)
            if(nact-ioffs+1 .ge.ng6pipes)then
               npuse=ng6pipes
            else
               npuse=nact-ioffs+1
            endif
            do i=1,npuse
               j=listact(ioffs+i-1)
               k=linktreetosph(j)
               ig6ind(i)=j

               g6_pos(1,i)=treept(1,j) !pos from tree arrays
               g6_pos(2,i)=treept(2,j)
               g6_pos(3,i)=treept(3,j)
               g6_vel(1,i)=vel(1,k)
               g6_vel(2,i)=vel(2,k)    ! rest in sph notation
               g6_vel(3,i)=vel(3,k)
               g6_jold_temp(1,i)=g6_jold(1,k)
               g6_jold_temp(2,i)=g6_jold(2,k)
               g6_jold_temp(3,i)=g6_jold(3,k)
               g6_accold_temp(1,i)=g6_accold(1,k)
               g6_accold_temp(2,i)=g6_accold(2,k)
               g6_accold_temp(3,i)=g6_accold(3,k)
               g6_phiold_temp  (i)=g6_phiold(k)
            enddo
            call g6calc_firsthalf(0,ng6part,npuse,ig6ind,g6_pos,g6_vel,
     $           g6_accold_temp,g6_jold_temp,g6_phiold_temp
     $           ,g6eps2,g6h)
            ierr=g6calc_lasthalf(0,ng6part,npuse,ig6ind,g6_pos,
     $           g6_vel,g6eps2,g6h,g6_acc,g6_j,g6_phi)
            if(ierr.eq.1)then
               nharderr=nharderr+1
               call g6_reinitialize(0)
               if(nharderr.le.20)then
                  goto 100      !start from scratch                  
               else
                  call g6_close(0)
                  stop'Too may hard errors on G6!!!'
               endif
            endif

            do i=1,npuse
               k=linktreetosph(listact(ioffs+i-1))
               acc(1,k)     = acc(1,k) + gconst*g6_acc(1,i)
               acc(2,k)     = acc(2,k) + gconst*g6_acc(2,i)
               acc(3,k)     = acc(3,k) + gconst*g6_acc(3,i)
               poten(k)     = poten(k) - gconst*g6_phi(i)
               g6_jerk(1,k) = g6_jerk(1,k) + g6_j(1,i)
               g6_jerk(2,k) = g6_jerk(2,k) + g6_j(2,i)
               g6_jerk(3,k) = g6_jerk(3,k) + g6_j(3,i)               
            enddo
            ioffs=ioffs+npuse
         enddo
         if (nla .gt. nlimhigh) then ! Loop back for another shift.
            nlimlow  = nlimlow  + npartmax
            nlimhigh = nlimhigh + npartmax
            goto 1000
         endif
         
!     store old quantities 

         do i=1,nact
            j=linktreetosph(listact(i))
            g6_accold(1,j)= acc(1,j)
            g6_accold(2,j)= acc(2,j)
            g6_accold(3,j)= acc(3,j)
            g6_phiold(j)  = poten(j)
            g6_jold(1,j)  = g6_jerk(1,j)
            g6_jold(2,j)  = g6_jerk(2,j)
            g6_jold(3,j)  = g6_jerk(3,j)
         enddo         
         if(nharderr.ne.0)then
            print*,'G6 hard errors:',nharderr
         endif
      return
      end subroutine grape_tree

!------------------------------------------------------

      subroutine grape_dirsum(npart,eps2,posmas,vel,acc,poten)

      integer, intent(in):: npart
      double precision, intent(in   ):: eps2
      double precision, intent(in   ):: posmas(ndim1,npart)
      double precision, intent(in   ):: vel(ndim,npart)
      double precision, intent(inout):: acc(ndim,npart)
      double precision, intent(inout):: poten(npart)

      double precision g6_pos(3,ng6pipes),g6_acc(3,ng6pipes),gconst
      double precision g6_phi(ng6pipes),g6_vel(3,ng6pipes),g6massmem
      double precision g6eps2,g6posmem(3),g6h(ng6pipes)
      double precision g6_j(3,ng6pipes),g6_jerk(3,idim)
      double precision g6_accold_temp(3,ng6pipes),
     &     g6_jold_temp(3,ng6pipes),g6_phiold_temp(ng6pipes)

      integer ig6ind(ng6pipes),g6calc_lasthalf

      integer i,ii,j,nlimlow,nlimhigh,ig6j_ind
      integer ierr,nharderr,ng6part,ioffs,npuse

      logical, save:: startup= .true.

      character(len=20), parameter:: where =('grape_dirsum')
      if(idbg.ge.2)print *,where

      gconst=one

      if(startup)then!init old accelerations, jerk, potential with educated guess
         do i=1,npart
            g6_accold(1,i)=one
            g6_accold(2,i)=one
            g6_accold(3,i)=one
            g6_jold(1,i)=1.d5
            g6_jold(2,i)=1.d5
            g6_jold(3,i)=1.d5
            g6_phiold(i)=one
         enddo
         startup=.false.
      endif
      
      nharderr=0
 100  continue
      
      do i=1,npart
         g6_jerk(1,i)=zero
         g6_jerk(2,i)=zero
         g6_jerk(3,i)=zero
      enddo
      
      call g6_set_xunit(51)     ! default 51 bits for representation of 
                                ! fractional part in position data
      call g6_set_tunit(51)     ! default 51 bits for representation of 
                                ! fractional part in time
      call g6_set_ti(0,zero)    ! time 0.d0, no prediction
         
! prepare softening and smoothing lengths
      g6eps2=eps2
      do i=1,ng6pipes
         g6h(i)=zero
      enddo

      nlimlow=1
      nlimhigh=npartmax
      
 1000 continue 
      if (npart .LT. nlimhigh) nlimhigh = npart
      ng6part=nlimhigh-nlimlow+1
         
      j=0
      do i=nlimlow,nlimhigh
         ii=linksphtotree(i)
         j=j+1
         
         g6posmem(1)=treept(1,ii)
         g6posmem(2)=treept(2,ii)
         g6posmem(3)=treept(3,ii)
         g6massmem  =treept(ims,ii)
         ig6j_ind=linktreetosph(ii)
         
         call g6_set_j_particle_mxonly(0,j-1,ig6j_ind,
     $        g6massmem,g6posmem(1))
      enddo

      ioffs=1
      do while(ioffs.le.nactive)
         if(nactive-ioffs+1 .ge.ng6pipes)then
            npuse=ng6pipes
         else
            npuse=nactive-ioffs+1
         endif
         do i=1,npuse
            j=iactive(ioffs+i-1)
            ig6ind(i)=j
            g6_pos(1,i)=posmas(1,j)
            g6_pos(2,i)=posmas(2,j)
            g6_pos(3,i)=posmas(3,j)
            g6_vel(1,i)=vel(1,j)
            g6_vel(2,i)=vel(2,j)
            g6_vel(3,i)=vel(3,j)
            g6_jold_temp(1,i)=g6_jold(1,j)
            g6_jold_temp(2,i)=g6_jold(2,j)
            g6_jold_temp(3,i)=g6_jold(3,j)
            g6_accold_temp(1,i)=g6_accold(1,j)
            g6_accold_temp(2,i)=g6_accold(2,j)
            g6_accold_temp(3,i)=g6_accold(3,j)
            g6_phiold_temp  (i)=g6_phiold(j)
         enddo
!         istart=iactive(ioffs)
         
         call g6calc_firsthalf(0,ng6part,npuse,ig6ind,g6_pos
     $        ,g6_vel,g6_accold_temp,g6_jold_temp,g6_phiold_temp
     $        ,g6eps2,g6h)
         ierr=g6calc_lasthalf(0,ng6part,npuse,ig6ind,g6_pos,
     $        g6_vel,g6eps2,g6h,g6_acc,g6_j,g6_phi)
         if(ierr.eq.1)then
            nharderr=nharderr+1
            call g6_reinitialize(0)
            if(nharderr.le.20)then
               
               goto 100         !start from scratch
               
            else
               call g6_close(0)
               stop'Too may hard errors on G6!!!'
            endif
         endif
         
         do i=1,npuse
            j=iactive(ioffs+i-1)
            acc(1,j)     = acc(1,j) + gconst*g6_acc(1,i)
            acc(2,j)     = acc(2,j) + gconst*g6_acc(2,i)
            acc(3,j)     = acc(3,j) + gconst*g6_acc(3,i)
            poten(j)     = poten(j) + gconst*g6_phi(i)
            g6_jerk(1,j) = g6_jerk(1,j) + g6_j(1,i)
            g6_jerk(2,j) = g6_jerk(2,j) + g6_j(2,i)
            g6_jerk(3,j) = g6_jerk(3,j) + g6_j(3,i)
            
         enddo
         ioffs=ioffs+npuse
      enddo

      if (npart .gt. nlimhigh) then ! Loop back for another shift.
         nlimlow  = nlimlow  + npartmax
         nlimhigh = nlimhigh + npartmax
         goto 1000
      endif
      
!     store old quantities 
         
      do i=1,nactive
         j=iactive(i)
         g6_accold(1,j)= acc(1,j)
         g6_accold(2,j)= acc(2,j)
         g6_accold(3,j)= acc(3,j)
         g6_phiold(j)  = poten(j)
         g6_jold(1,j)  = g6_jerk(1,j)
         g6_jold(2,j)  = g6_jerk(2,j)
         g6_jold(3,j)  = g6_jerk(3,j)
      enddo         
      
      return
      end subroutine grape_dirsum
      end module grape
