      module pm
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
!  This file contains the VINE module to do IO for point mass        *
!  information and to do sph/nbody particle accretion onto point     *
!  masses.                                                           *
!                                                                    *
!  Externally visible routines are:                                  *
!                                                                    *
!    remove: Remove SPH or Nbody particles for being tol close to    *
!            a point mass. Add them and their momenta to the         *
!            appropriate point mass.                                 *
!      Arguments:                                                    *
!        nremove: The number of particles removed, during this call  *
!                                                                    *
!  Written by Andy Nelson                                            *
!*********************************************************************
      use dimens
      use partvars
      use miscvars, only: irem
      use indivtsvars, only: nactivepm,nactivefpm,iactivepm
      use indivtsvars, only: it0pm,itstepbinpm,lastts_end
      use indivtsvars, only: it0,it1,itstepbin
      use perform
      use constants

      implicit none

      private
      public remove

      contains
!------------------------------------------------------------------
      subroutine remove(i12step,nremove)
!*******************************************************
!                                                      *
!  This routine removes particles when they cross a    *
!  fixed boundary, close to a point mass               *
!                                                      *
!  Someday this should be replaced by a tree search    *
!  based version. -AFN                                 *
!                                                      *
!*******************************************************

      integer, intent(in ):: i12step
      integer, intent(out):: nremove

      integer my_cpu,ncount

      integer i,j,k,jpoim,ipoim,loopindx
      integer ndel,numcpus,icpu,ip_low,ip_low1,ip_hi
      integer ilen(idim), nrem(maxcpus)
      double precision dcrit2

      double precision tme1,tme2
      character(len=20), parameter:: where =('remove')

!squared distance to point mass j statement function
      double precision d2
      d2(i,j)= (posmas(1,i)-posmaspm(1,j))**2 
     &        +(posmas(2,i)-posmaspm(2,j))**2
#if NDIM==3
     &        +(posmas(3,i)-posmaspm(3,j))**2
#endif

      if(npoim.eq.0)return
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      nremove=0
      select case(irem)
      case(1)
!
! determine list of particles to remove. The `if' statements restrict 
! removals to particles that were just active, or are on smaller
! time bins than the pm.  Not sure all the logic is right with these
! tests here. Also, should this inner loop be limited to looping over
! active particles only??? probably yes for better efficiency, but will
! we ever pick up particles in other ts bins this way? -AFN
!
         loopindx=nactivepm
         if(i12step.eq.2)loopindx=nactivefpm

         do jpoim=1,loopindx
            ipoim=iactivepm(jpoim)

            if(it0pm(ipoim).eq.lastts_end)then

               ncount=0

!$omp parallel default(none)
!$omp& shared(npart,nremove,ncount,nrem,ilen)
!$omp& shared(posmas,posmaspm,hpm,itstepbin,itstepbinpm,it0)
!$omp& firstprivate(idbg,ipoim,lastts_end)
!$omp& private(numcpus,ndel,icpu,i,ip_low,ip_low1,ip_hi,dcrit2,k)
!$omp& private(my_cpu)

               call numthreads(numcpus)            !Require exact 
               call get_mycpu(my_cpu)              !numthreads here
               ndel=npart/numcpus                  !for ilen array 
               dcrit2=hpm(ipoim)**2                !division below

               ip_low= 1 + (my_cpu-1)*ndel
               ip_low1=ip_low-1
               ip_hi = my_cpu*ndel
               if(my_cpu.eq.numcpus)ip_hi=npart
               nrem(my_cpu)=0
               do i=ip_low,ip_hi
                  if(d2(i,ipoim).lt.dcrit2)then
                     if(itstepbin(i).le.itstepbinpm(ipoim) .or.
     &                        it0(i).eq.lastts_end)then
                        nrem(my_cpu)=nrem(my_cpu) + 1
                        ilen(ip_low1+nrem(my_cpu))=i
                     endif
                  endif
               enddo

!$omp critical
               ncount=ncount+1              !kludge to get around apparent
               if(ncount.eq.numcpus)then    !ibm omp bug (par. region nesting)
!
! Merge the remove list
!
                  nremove=nremove+nrem(1)
                  do icpu=2,numcpus
                     nremove=nremove+nrem(icpu)
                     ip_low= 1 + (icpu-1)*ndel
                     do k=1,nrem(icpu)
                        i=ilen((ip_low-1)+k)
                        nrem(1)=nrem(1)+1
                        ilen(nrem(1))=i
                     enddo
                  enddo
                  if(nrem(1).gt.0)then
                     call mergeparts(ipoim,ilen,nrem)
                     call compactify(nrem(1),ilen)
                     if((idbg.ge.1 .and. nrem(1).gt.0) .or.
     &                   idbg.ge.3)then
                         print *,'removed',nrem(1),
     &                            ' particles close to PM#',ipoim
                     endif
                  endif
               endif

!$omp end critical
!$omp end parallel

            endif   !ipoim at endts
         enddo      !ipoim

      case(2)

         stop'forgot how to do bleeding particle removal'

      case default

         stop'don''t know how to remove particles'
      end select 

      if(idbg.ge.2)print *,'done ',where
      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_remove)=timing(itme_remove)+(tme2-tme1)
         ncalls(itme_remove)=ncalls(itme_remove)+1
         timing(itme_PM)=timing(itme_PM)+(tme2-tme1)
         ncalls(itme_PM)=ncalls(itme_PM)+1
      endif

10    format(    10(1pd18.10))
13    format(3i7,10(1pd12.4))
15    format(5i7,10(1pd12.4))
16    format(6i7,10(1pd12.4))

      return
      end subroutine remove
!------------------------------------------------------------------
      subroutine mergeparts(ipoim,ilen,nrem)
!*******************************************************
!                                                      *
!  This subroutine merges particles with               *
!  point mass ipoim.                                   *
!                                                      *
!*******************************************************
      integer, intent(in):: ipoim,ilen(idim),nrem(maxcpus)

      integer i,k
!
! Not sure if quad precision is really necessary, but it doesn't hurt
! (some compilers won't have it though...then default back to double)
! Could be better to use Kahan's algorithm here too
!
      real(qp) emnew,xnew,ynew,znew,vxnew,vynew,vznew
      real(qp) spinxnew,spinynew,spinznew,sx,sy,sz
      real(qp) dxi,dyi,dzi,dvx,dvy,dvz
      real(qp) dxpm,dypm,dzpm,dvxpm,dvypm,dvzpm

      character(len=20), parameter:: where=('merge')
      if(idbg.ge.2)print *,where
!
! find center of mass and momentum of PM and merging particles 
!  (can't be parallelized) Add PM after summing removed particles
!  to minimize round off error. (only does good if nrem >1 though)
!
      emnew =zero
      xnew  =zero  ; ynew  =zero  ; znew  =zero
      vxnew =zero  ; vynew =zero  ; vznew =zero

      do k=1,nrem(1)
         i=ilen(k)
         emnew=emnew + posmas(ims,i)
         xnew =xnew  + posmas(ims,i)*posmas(1  ,i)
         ynew =ynew  + posmas(ims,i)*posmas(2  ,i)
         vxnew=vxnew + posmas(ims,i)*vel   (1  ,i)
         vynew=vynew + posmas(ims,i)*vel   (2  ,i)
#if NDIM==3
         znew =znew  + posmas(ims,i)*posmas(3  ,i)
         vznew=vznew + posmas(ims,i)*vel   (3  ,i)
#endif
      enddo

      emnew= emnew + posmaspm(ims,ipoim)
      xnew = (xnew + posmaspm(ims,ipoim)*posmaspm(1,ipoim))/emnew
      ynew = (ynew + posmaspm(ims,ipoim)*posmaspm(2,ipoim))/emnew
      vxnew= (vxnew+ posmaspm(ims,ipoim)*velpm   (1,ipoim))/emnew
      vynew= (vynew+ posmaspm(ims,ipoim)*velpm   (2,ipoim))/emnew
#if NDIM==3
      znew = (znew + posmaspm(ims,ipoim)*posmaspm(3,ipoim))/emnew
      vznew= (vznew+ posmaspm(ims,ipoim)*velpm   (3,ipoim))/emnew
#endif
!
! adjust spin of PM
!
      spinxnew=zero ; spinynew=zero ; spinznew=zero

      do k=1,nrem(1)
         i=ilen(k)

         dxi= xnew  - posmas(1,i)  ; dyi= ynew  - posmas(2,i)
         dvx= vxnew - vel   (1,i)  ; dvy= vynew - vel   (2,i) 
#if NDIM==3
         dzi= znew  - posmas(3,i)  ; dvz= vznew - vel   (3,i) 
         sx = posmas(ims,i)*( dyi*dvz - dvy*dzi)
         sy = posmas(ims,i)*(-dxi*dvz + dvx*dzi)
         spinxnew=spinxnew + sx
         spinynew=spinynew + sy
#endif
         sz = posmas(ims,i)*( dxi*dvy - dvx*dyi)
         spinznew=spinznew + sz

         posmas(ims,i)=-huge(ONE)  !flag particle for subr. compactify
      enddo

      dxpm =xnew -posmaspm(1,ipoim) ; dvxpm=vxnew-velpm   (1,ipoim)
      dypm =ynew -posmaspm(2,ipoim) ; dvypm=vynew-velpm   (2,ipoim)
#if NDIM==3
      dzpm =znew -posmaspm(3,ipoim) ; dvzpm=vznew-velpm   (3,ipoim)
      spinxnew=spinxnew + posmaspm(ims,ipoim)*( dypm*dvzpm - dvypm*dzpm)
      spinynew=spinynew + posmaspm(ims,ipoim)*(-dxpm*dvzpm + dvxpm*dzpm)
#endif
      spinznew=spinznew + posmaspm(ims,ipoim)*( dxpm*dvypm - dvxpm*dypm)
!
! Update point mass quantities
!
      posmaspm(ims,ipoim)=emnew
      posmaspm(1  ,ipoim)= xnew  ; velpm   (1  ,ipoim)=vxnew
      posmaspm(2  ,ipoim)= ynew  ; velpm   (2  ,ipoim)=vynew
#if NDIM==3
      posmaspm(3  ,ipoim)= znew  ; velpm   (3  ,ipoim)=vznew
      spinpm  (1  ,ipoim)=spinpm(1,ipoim) + spinxnew
      spinpm  (2  ,ipoim)=spinpm(2,ipoim) + spinynew
#endif
      spinpm  (3  ,ipoim)=spinpm(3,ipoim) + spinznew

      return
      end subroutine mergeparts
!------------------------------------------------------------------
      subroutine compactify(nremove,ilen)
!*******************************************************
!                                                      *
!  This subroutine replaces particles somewhere inside *
!  a list with particles at the top of the list        *
!                                                      *
!*******************************************************
      use integvars

      integer, intent(in):: nremove, ilen(nremove)

      integer i,j,ilast
      character(len=20), parameter:: where =('compactify')

      if(idbg.ge.2)print *,where

      if(npart.ne.npart_sph .or.
     &      (npart.ne.npart_sph .and. npart_sph.ne.0) )then
         print *,'compactify can''t handle complications'
         print *,'of removing particles when both nbody'
         print *,'and sph particles exist--difficult to'
         print *,'do the list offsets correctly'
         stop'Rewrite compactify'
      endif

      ilast=npart
      do i=1,nremove
         j=ilen(i)
         if(idbg.ge.1)print *,'removing particle',j

         do while(posmas(ims,ilast).lt.zero) !is end of list particle
            ilast=ilast - 1                  !on remove list too?
         enddo
         if(ilast.gt.j)then

         u         (j)=u         (ilast) ; u         (ilast)=-huge(ONE)
         h         (j)=h         (ilast) ; h         (ilast)=-huge(ONE)
         alpha     (j)=alpha     (ilast) ; alpha     (ilast)=-huge(ONE)
         dudt1     (j)=dudt1     (ilast) ; dudt1     (ilast)=-huge(ONE)
         dhdt1     (j)=dhdt1     (ilast) ; dhdt1     (ilast)=-huge(ONE)
         dalfdt1   (j)=dalfdt1   (ilast) ; dalfdt1   (ilast)=-huge(ONE)
         rho       (j)=rho       (ilast) ; rho       (ilast)=-huge(ONE)
         rhosave   (j)=rhosave   (ilast) ; rhosave   (ilast)=-huge(ONE)
         divv      (j)=divv      (ilast) ; divv      (ilast)=-huge(ONE)
         divvsave  (j)=divvsave  (ilast) ; divvsave  (ilast)=-huge(ONE)
         ipindx    (j)=ipindx    (ilast) ; ipindx    (ilast)=-99999999
         it0       (j)=it0       (ilast) ; it0       (ilast)=-99999999
         it1       (j)=it1       (ilast) ; it1       (ilast)=-99999999
         itstepbin (j)=itstepbin (ilast) ; itstepbin (ilast)=-99999999
         posmas(ims,j)=posmas(ims,ilast) ; posmas(ims,ilast)=-huge(ONE)
         posmas(1  ,j)=posmas(1  ,ilast) ; posmas(1  ,ilast)=-huge(ONE)
         posmas(2  ,j)=posmas(2  ,ilast) ; posmas(2  ,ilast)=-huge(ONE)
         vel   (1  ,j)=vel   (1  ,ilast) ; vel   (1  ,ilast)=-huge(ONE)
         vel   (2  ,j)=vel   (2  ,ilast) ; vel   (2  ,ilast)=-huge(ONE)
         acc1  (1  ,j)=acc1  (1  ,ilast) ; acc1  (1  ,ilast)=-huge(ONE)
         acc1  (2  ,j)=acc1  (2  ,ilast) ; acc1  (2  ,ilast)=-huge(ONE)
#if NDIM==3
         posmas(3  ,j)=posmas(3  ,ilast) ; posmas(3  ,ilast)=-huge(ONE)
         vel   (3  ,j)=vel   (3  ,ilast) ; vel   (3  ,ilast)=-huge(ONE)
         acc1  (3  ,j)=acc1  (3  ,ilast) ; acc1  (3  ,ilast)=-huge(ONE)
#endif
#if defined(RK)
         rhohalf   (j)=rhohalf   (ilast) ; rhohalf   (ilast)=-huge(ONE)
         divvhalf  (j)=divvhalf  (ilast) ; divvhalf  (ilast)=-huge(ONE)
         velhaf(1  ,j)=velhaf(1  ,ilast) ; velhaf(1  ,ilast)=-huge(ONE)
         velhaf(2  ,j)=velhaf(2  ,ilast) ; velhaf(2  ,ilast)=-huge(ONE)
#if NDIM==3
         velhaf(3  ,j)=velhaf(3  ,ilast) ; velhaf(3  ,ilast)=-huge(ONE)
#endif
#endif
         endif
         ilast=ilast-1
      enddo
      npart=npart - nremove
      npart_sph=max(0,npart_sph-nremove)

      return
      end subroutine compactify
!------------------------------------------------------------------
      end module pm
