      module integ
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!*********************************************************************
!  This file contains the VINE module to do forward time             *
!  integrations on particles using a Runge-Kutta                     *
!  integrator. The specific implementation is the first order        *
!  integrator with second order error control discussed in:          *
!     Fehlberg, Erwin: Low-order classical Runge-Kutta               *
!         formulas with stepsize control and their application       *
!         to some heat transfer problems                             *
!         NASA TR R-315                                              *
!                                                                    *
!  The coefficients used here are as in Table XIV of Fehlberg        *
!  Specifically:                                                     *
!                                                                    *
!      alpha     beta(k,l)        c        c_hat                     *
!    l           0        1    |                                     *
!  k                           |                                     *
!  0    0                      | 1/256      1/512                    *
!  1    1/2     1/2            | 255/256  255/256                    *
!  2    1       1/256  255/256 |            1/512                    *
!                                                                    *
!  Applied to the equations this is:                                 *
!                                                                    *
!  y(n+1) = y(n) + (c0*kap0 + c1*kap1)*dt                            *
!                                                                    *
!  where                                                             *
!  tn   = time at step n                                             *
!  y    = the function to be advanced to time n+1 from time n        *
!         (position, velocity, smoothing length, int energ etc)      *
!  kap0 = f0(tn,y(n))                                                *
!  kap1 = f1(tn+alpha1*dt,y(n)+dt*beta10*f0)                         *
!  kap2 = f2(tn+alpha2*dt,y(n)+dt*beta20*f0+dt*beta21*f1)            *
!                                                                    *
!  kap2 is not used as is, but f2 is used for the error.             *
!  Lowest order truncation error is given as a difference            *
!  of the derivatives at the beginning and the end of                *
!  the timestep.                                                     *
!                                                                    *
!  TE = 1/512(f0 -f2)dt, where f0 and f2 are the derivatives         *
!                         at the beginning and end respectively      *
!                                                                    *
!  This gives absolute error, which is not usually particularly      *
!  useful. Need several steps to get useful error estimate.          *
!                                                                    *
!  1) We are interested in error magnitude, so do abs(TE)            *
!                                                                    *
!  2) We have 2nd order (in time) error control so,                  *
!                                                                    *
!      dtnew = dtold * sqrt(relevant error quantity)                 *
!                                                                    *
!  3) What is `relevant error quantity' (besides a black art)        *
!                                                                    *
!     Would usually like relative error magnitude. In other          *
!     words, get the solution to some tolerance `tol' per            *
!     timestep. That translates as                                   *
!                                                                    *
!         abs(f0-f2)*dtold                                           *
!     R=  ----------------- < 1  for an acceptable timestep          *
!         512*tol*abs(y(n))                                          *
!                                                                    *
!     Thus we can set dtnew from R as                                *
!                                                                    *
!     dtnew = dtold*sqrt(1/R)   (R too big, decrease)                *
!                               (R too sml, increase)                *
!                                                                    *
!  4) PROBLEM sometimes need absolute error                          *
!     Of course, this leaves out cases like the internal             *
!     energy in nuclear burning simulations, where we need a         *
!     very precise absolute measure of the error. In the             *
!     spirit of VINE's modularity/extensibility, the user            *
!     could either add a separate check in another subroutine        *
!     for such a pathological case, or redesign the                  *
!     RK error criterion for u in another more useful way.           *
!     Especially, now that the problem has been pointed out          *
!     here... :-) One way would be to replace y(n) above             *
!     with an appropriate constant.                                  *
!                                                                    *
!  5) ANOTHER PROBLEM quantities near zero                           *
!     As is, this is insufficient for quantities that                *
!     travel through zero, like position and velocity,               *
!     because R goes to infinity there. It is also bad for           *
!     situations where the particles are centered around             *
!     some non zero position or are moving with some system          *
!     velocity in some direction. The basic idea is that             *
!     the above error metric is position dependent. The              *
!     farther from the origin you are, the more relaxed the          *
!     error criterion gets. So what happens to the error             *
!     if for fun you decide to shift every particle to the           *
!     left by some large number of units?  Your simulation           *
!     result changes. Very bad. Very sad. Not good. You can          *
!     tell one reference frame from another. Einstein would          *
!     not be pleased.                                                *
!     On the other hand, a position (radius from origin)             *
!     dependent error criterion makes a lot of sense for             *
!     a Keplerian (for example) disk...the error criterion           *
!     is very precise near zero for positions, but relaxed           *
!     for velocities there (large values), while the situation       *
!     is reversed at large distances.                                *
!                                                                    *
!  6) ONE PARTIAL SOLUTION subtract center of mass                   *
!     Can solve these particular cases by subtracting off            *
!     a system CM or V_CM from the specific position or              *
!     velocity of a particle, but still have the basic               *
!     position/velocity dependence of the error criterion.           *
!     Particles far from the origin (or CM) have less                *
!     restrictive error criteria than those near it. This is         *
!     worse in systems like binary stars, where the particles        *
!     have small relative motions, large common motions, and         *
!     orbit around each other, far from the origin. Sigh.            *
!                                                                    *
!  7) ANOTHER PARTIAL SOLUTION include change error metric           *
!     Consulting that ever so useful and always correct              *
!     guide to numerics, called Numerical Recipes, we find           *
!     that they suggest replacing the magnitude of the               *
!     function with the sum                                          *
!              abs(y(n)) + abs(dtold*dydt(n))                        *
!                                                                    *
!     This sets a limit on the relative magnitude of                 *
!        -the error compared to the function value                   *
!        -the error compared to the function change, when            *
!         the function value is near zero.                           *
!                                                                    *
!     We can use this for scalar quantities like internal            *
!     energy and smoothing length, so that it avoids the             *
!     y(n) near zero problem unless y(n) is near zero and            *
!     is also changing slowly. Hopefully that will be more           *
!     rare. Thus for scalar quantities, we use an error              *
!     criterion of:                                                  *
!                                                                    *
!                     abs(f0-f2)*dtold                               *
!     R=  ----------------------------------------  < 1              *
!         512*tol*(abs(y(n)) + abs(dtold*(dydt(n)))                  *
!                                                                    *
!                                                                    *
!     Given the conditions in our point 6, we replace this           *
!     metric by                                                      *
!          abs(y(n)-y_cm) + abs(dtold*(dydt(n)-dydt_cm))             *
!     for positions and velocities, yielding:                        *
!                                                                    *
!                     abs(f0-f2)*dtold                               *
!     R=  ----------------------------------------  < 1              *
!         / 512*tol*(abs(y(n)-y_cm) +             \                  *
!         |         abs(dtold*(dydt(n)-dydt_cm))) |                  *
!         \                                       /                  *
!                                                                    *
!     In each case, we use dydt(n) = f0 as defined above.            *
!                                                                    *
!  8) ANOTHER PARTIAL SOLUTION add error softening term (i.e.        *
!                              make it absolute error-like           *
!     One way around the relative error near zero problem            *
!     is to add an error softening term. This is a very              *
!     problem dependent quantity however. In the code                *
!     below, we have put in an error softener of hmin for            *
!     the positions, so that instead of abs(y(n)-y_cm)               *
!     you have max(hmin,abs(y(n)-y_cm)).                             *
!                                                                    *
!  9) ANOTHER PARTIAL SOLUTION do error relative to rms              *
!                              position or velocity values.          *
!     Instead of having the specific value of a quantity             *
!     for a given particle used for the error for that               *
!     particle, it may be appropriate to have some RMS               *
!     value be used for all the particles. This is another           *
!     variant of an absolute error criterion.                        *
!                                                                    *
!                                                                    *
!  WARNING                                                           *
!     None of these partial solutions get around the fact            *
!     that the error can still be position dependent, but            *
!     they do acceptably well in certain problems.  With             *
!     enough documentation like what is written here, VINE           *
!     users can make sure they aren't shooting themselves            *
!     in the foot without knowing any better.                        *
!                                                                    *
!     Hopefully, you will only shoot yourselves in the foot          *
!     with full knowledge of the consequences of having a            *
!     hole in your foot. In other words:                             *
!                                                                    *
!     Caveat Emptor.                                                 *
!                                                                    *
!                                                                    *
! ============================================                       *
! Note about Nbody variable smoothing lengths.                       *
! ============================================                       *
! This code currently only implements smoothing lengths              *
! for nbody particles that are constant in time. Reasons             *
! for this are essentially that small scale random motions           *
! (when they occur) such as when bulk motions become                 *
! `thermalized' tend to couple to the smoothing length,              *
! causing a nasty resonance that oscillates faster and               *
! faster until the whole system turns into a quivering bowl          *
! of jelly. This is essentially an effect of the breakdown of        *
! the `collisionless' assumption in most nbody simulations.          *
! SPH particles don't suffer from this problem because the           *
! energy cascade from large scale to small ends up in heat,          *
! as small scale random motions are dissipated.                      *
!                                                                    *
! If you still want to try and include such effects, and this        *
! warning didn't scare you off, here is how:                         *
!                                                                    *
! 1) Redimension the arrays rho ,  rhosave,  rhohalf                 *
!                           divv, divvsave, divvhalf                 *
!                           dhdt1, dhdt2, nneightot                  *
!    to idim rather than isphdim. Further all invocations            *
!    of these arrays in various subroutines need the same            *
!    treatment. (i.e. in the i/o routines, you need to               *
!    output all npart particles, not just npart_sph)                 *
!                                                                    *
! 2) In subroutine deriv, move the if(npart_sph.eq.0) line           *
!    down to below the hdot call.                                    *
!                                                                    *
! 3) In subroutine density (in sph.F) change the value of iboth      *
!    to iboth=1, and in the storage loop at the end, save rho        *
!    and divv for all particles, not just sph particles.             *
!                                                                    *
! 4) Change the hdot subroutine arguments from the current           *
!    iactive_s and nactive_s to iactive and nactive.                 *
!                                                                    *
! 5) Move the h, rho and divv updates from the sph only loops in     *
!    of halfupdate and fullupdate, into the loops done for all       *
!    particles. Note that this means deleting lines like:            *
!    dumh(i)=h(i) because now you actually will change h, not        *
!    just copy it to the dummy/advanced time array.                  *
!                                                                    *
! 6) Move the h, rho and divv copyback values into the main          *
!    loop rather than their current location in the sph only loop    *
!    of copyback                                                     *
!                                                                    *
! 7) In the hchange subroutine, include all particles, not just      *
!    sph particles.                                                  *
!                                                                    *
! 8) Tell us if it works.                                            *
!                                                                    *
!  A VINE module written by                                          *
!  Andy Nelson, Markus Wetzstein and Thorsten Naab                   *
!                                                                    *
!*********************************************************************

      use dimens
      use perform
      use partvars
      use miscvars, only:umin,hmin,hmax,dhmax,alphastar,tol
      use miscvars, only:d_unit,vel_unit,tolpm
      use miscvars, only:ifullextrap
      use integvars
      use constants

      implicit none

      private
      public copyback,halfupdate,fullupdate,extrap_half
      public rk_err

!
! Make sure that the preprocessor flag for Runge-Kutta integrator
! is set, and all other integrator flags are not set, before going
! any further. In other words, attempt to catch stupid makefile
! errors.
!
#ifndef RK
      Runge-Kutta module: preprocessor flag RK not set
#endif

#ifdef LF
      Runge-Kutta module: preprocessor flag LF set
#endif

!replace this with some other integrator that
!    code has been written for
#ifdef YOUR_INTEGRATOR
      Runge-Kutta module: preprocessor flag YOUR_INTEGRATOR is set
#endif

      contains
!-----------------------------------------------------------------
      subroutine copyback(nactivef_s,nactivef,nactivefpm,
     &                    iactive_s ,iactive ,iactivepm)
!************************************************************
!                                                           *
!  copy temporary values and derivatives for particles      *
!  which have finished their time steps into the            *
!  beginning of time step arrays.                           *
!                                                           *
!  Copying rho to rhosave is for indiv timesteps. On a      *
!  whole or 1/2 timestep, the density is calculated and     *
!  returned in the variable rho. Need to save it for later  *
!  extrapolations. Same for divv.                           *
!                                                           *
!************************************************************
      integer, intent(in):: nactivef_s,iactive_s(nactivef_s)
      integer, intent(in):: nactivef  ,iactive  (nactivef)
      integer, intent(in):: nactivefpm,iactivepm(nactivefpm)

      integer i,j
      double precision tme1,tme2
      character(len=20), parameter:: where =('copyback')

      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

!$omp parallel do if(nactivef_s .gt. 1024) default(none)
!$omp& shared(nactivef_s,iactive_s,rho,rhosave,divv,divvsave)
!$omp& shared(h   ,u   ,alpha ,dhdt1,dudt1,dalfdt1)
!$omp& shared(dumh,dumu,dumalf,dhdt2,dudt2,dalfdt2)
!$omp& private(i,j)
      do j=1,nactivef_s
         i=iactive_s(j)

         rhosave (i)=rho    (i)
         divvsave(i)=divv   (i)
         h       (i)=dumh   (i)
         u       (i)=dumu   (i)
         alpha   (i)=dumalf (i)
         dhdt1   (i)=dhdt2  (i)
         dudt1   (i)=dudt2  (i)
         dalfdt1 (i)=dalfdt2(i)
      enddo

!$omp parallel do if(nactivef .gt. 1024) default(none) 
!$omp& shared(nactivef,iactive)
!$omp& shared(   posmas,   vel,acc1)
!$omp& shared(dumposmas,dumvel,acc2)
!$omp& private(i,j)
      do j=1,nactivef
         i=iactive(j)
         posmas (1,i)=dumposmas(1,i)
         posmas (2,i)=dumposmas(2,i)
         vel    (1,i)=dumvel   (1,i)
         vel    (2,i)=dumvel   (2,i)
         acc1   (1,i)=acc2     (1,i)
         acc1   (2,i)=acc2     (2,i)
#if NDIM==3
         posmas (3,i)=dumposmas(3,i)
         vel    (3,i)=dumvel   (3,i)
         acc1   (3,i)=acc2     (3,i)
#endif
      enddo

      do j=1,nactivefpm
         i=iactivepm(j)
         nsteppm (    i)=nsteppm(i)+1
         posmaspm(ims,i)=dumposmaspm(ims,i)
         posmaspm(1  ,i)=dumposmaspm(1  ,i)
         posmaspm(2  ,i)=dumposmaspm(2  ,i)
         velpm   (1  ,i)=dumvelpm   (1,  i)
         velpm   (2  ,i)=dumvelpm   (2,  i)

         acc1pm  (1  ,i)=acc2pm     (1,  i)
         acc1pm  (2  ,i)=acc2pm     (2,  i)
#if NDIM==3
         posmaspm(3  ,i)=dumposmaspm(3  ,i)
         velpm   (3  ,i)=dumvelpm   (3,  i)
         acc1pm  (3  ,i)=acc2pm     (3,  i)
#endif
      enddo

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_copyback)=timing(itme_copyback)+ (tme2-tme1)
         ncalls(itme_copyback)=ncalls(itme_copyback)+ 1
         timing(itme_intgratr)=timing(itme_intgratr)+ (tme2-tme1)
!        ncalls(itme_intgratr) increment only for fullstep calls
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine copyback
!-----------------------------------------------------------------
      subroutine extrap_half(deltat,deltatpm)
!************************************************************
! Subroutine stub needed for compatibility with other       *
! integrators like leapfrog, that need some variables       *
! to be advanced before beginning the integration.          *
!************************************************************
      double precision, intent(in):: deltat  (idim)
      double precision, intent(in):: deltatpm(ipmdim)

      print *,'No extrapolation needed or done for RK'

! Quickly do nothing and return

      return
      end subroutine extrap_half
!-----------------------------------------------------------------
      subroutine fullupdate(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                      iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                      deltat,deltatpm)
!************************************************************
!                                                           *
!  Update particles which are on the full time step         *
!  update list, using a Runge-Kutta-Fehlberg integrator     *
!  of second order: RK1(2)                                  *
!                                                           *
!************************************************************
      integer, intent(in):: nactivef_s, iactive_s(nactivef_s)
      integer, intent(in):: nactivef_n, iactive_n(nactivef_n)
      integer, intent(in):: nactivef  , iactive  (nactivef  )
      integer, intent(in):: nactivefpm, iactivepm(nactivefpm)
      double precision, intent(in):: deltat  (idim)
      double precision, intent(in):: deltatpm(ipmdim)

      integer i,j
      double precision dtf21,dtf22,dh,htmp,utmp,atmp

      double precision tme1,tme2
      character(len=20), parameter:: where =('fullupdate')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

!$omp parallel do if(nactivef_s .gt. 1024) default(none) 
!$omp&  firstprivate(umin,hmin,hmax,dhmax,alphastar)
!$omp&  shared(nactivef_s,iactive_s,deltat)
!$omp&  shared(h,u,alpha,dumh,dumu,dumalf)
!$omp&  shared(dhdt1,dhdt2,dudt1,dudt2,dalfdt1,dalfdt2)
!$omp&  private(i,j,dtf21,dtf22,atmp,utmp,htmp,dh)
      do j=1,nactivef_s
         i=iactive_s(j)

         dtf21 = deltat(i)*f21
         dtf22 = deltat(i)*f22

         dh=dtf21*dhdt1(i) + dtf22*dhdt2(i)
         dh= sign(min(abs(dh),dhmax*h(i)),dh)
         htmp=h(i)+dh
         dumh(i)=min(max(htmp,hmin),hmax)

         utmp=    u(i) + dtf21*  dudt1(i) + dtf22*  dudt2(i)
         atmp=alpha(i) + dtf21*dalfdt1(i) + dtf22*dalfdt2(i)
         dumu  (i)=max(utmp,umin)
         dumalf(i)=max(atmp,alphastar)
      enddo

!$omp parallel do if(nactivef .gt. 1024)  default(none)
!$omp&  shared(nactivef,iactive,deltat)
!$omp&  shared(   posmas,   vel,velhaf)
!$omp&  shared(dumposmas,dumvel,acc1,acc2)
!$omp&  private(i,j,dtf21,dtf22) 
      do j=1,nactivef
         i=iactive(j)
         dtf21 = deltat(i)*f21
         dtf22 = deltat(i)*f22

         dumposmas(ims,i)=posmas(ims,i)
         dumposmas(1,i)=posmas(1,i)+dtf21*vel (1,i)+dtf22*velhaf(1,i) 
         dumposmas(2,i)=posmas(2,i)+dtf21*vel (2,i)+dtf22*velhaf(2,i)
         dumvel   (1,i)=vel   (1,i)+dtf21*acc1(1,i)+dtf22*acc2(1,i)
         dumvel   (2,i)=vel   (2,i)+dtf21*acc1(2,i)+dtf22*acc2(2,i)
#if NDIM==3
         dumposmas(3,i)=posmas(3,i)+dtf21*vel (3,i)+dtf22*velhaf(3,i)
         dumvel   (3,i)=vel   (3,i)+dtf21*acc1(3,i)+dtf22*acc2(3,i)
#endif
      enddo

      do j=1,nactivefpm
         i=iactivepm(j)
         dtf21 = deltatpm(i)*f21
         dtf22 = deltatpm(i)*f22
         dumhpm     (    i)=hpm(i)
         dumposmaspm(ims,i)=posmaspm(ims,i)
         dumposmaspm(1  ,i)=posmaspm(1  ,i) +
     &                      (dtf21*velpm (1,i) + dtf22*dumvelpm(1,i))
         dumposmaspm(2  ,i)=posmaspm(2  ,i) +
     &                      (dtf21*velpm (2,i) + dtf22*dumvelpm(2,i))
         dumvelpm   (1  ,i)=velpm(1,i) +
     &                      (dtf21*acc1pm(1,i) + dtf22*acc2pm(1,i))
         dumvelpm   (2  ,i)=velpm(2,i) +
     &                      (dtf21*acc1pm(2,i) + dtf22*acc2pm(2,i))
#if NDIM==3
         dumposmaspm(3  ,i)=posmaspm(3  ,i) +
     &                      (dtf21*velpm (3,i) + dtf22*dumvelpm(3,i))
         dumvelpm   (3  ,i)=velpm(3,i) +
     &                      (dtf21*acc1pm(3,i) + dtf22*acc2pm(3,i))
#endif
      enddo

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_fullupdt)=timing(itme_fullupdt)+ (tme2-tme1)
         ncalls(itme_fullupdt)=ncalls(itme_fullupdt)+ 1
         timing(itme_intgratr)=timing(itme_intgratr)+ (tme2-tme1)
         ncalls(itme_intgratr)=ncalls(itme_intgratr)+ 1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine fullupdate
!-----------------------------------------------------------------
      subroutine halfupdate(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                      nactive_s ,nactive_n ,nactive ,nactivepm ,
     &                      iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                      itstepbin ,itstepbinpm,dtincr,
     &                      deltat,deltatpm)

!************************************************************
!                                                           *
!  Update particles which are on the half time step         *
!  update list, using a Runge-Kutta-Fehlberg integrator     *
!  of second order: RK1(2)                                  *
!                                                           *
!  For particles on individual timesteps that are currently *
!  inactive, density is extrapolated using Monaghan 1990    *
!  equation 3.9                                             *
!                                                           *
!************************************************************
      integer, intent(in):: nactive_s,nactive_n,nactive,nactivepm
      integer, intent(in):: nactivef_s, iactive_s(isphdim)
      integer, intent(in):: nactivef_n, iactive_n(nboddim)
      integer, intent(in):: nactivef  , iactive  (   idim)
      integer, intent(in):: nactivefpm, iactivepm( ipmdim)
      integer, intent(in):: itstepbin(idim),itstepbinpm(ipmdim)
      double precision, intent(in):: dtincr, deltat(idim)
      double precision, intent(in):: deltatpm(ipmdim)

      integer i,j,ntop,ntop_sph,ntoppm
      double precision dh,htmp,utmp,atmp,thalf,t2half

      double precision tme1,tme2
      character(len=20), parameter:: where =('halfupdate')
      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      if(ifullextrap.eq.0)then
         ntoppm  =nactivepm
         ntop_sph=nactive_s
         ntop    =nactive
      else
         ntoppm  =npoim
         ntop_sph=npart_sph
         ntop    =npart
      endif
!
! Predict variables at t=t+deltat(i), includes both half step
!  updates and extrapolations of inactive particles in the same
!  loops (if ifullextrap=1, otherwise just half step updates).
!
!$omp parallel do if(ntop-(nactivef+1) .gt. 1024)
!$omp&  default(none)
!$omp&  shared(ntop,nactivef,iactive,itstepbin,dtincr,deltat)
!$omp&  shared(   posmas,   vel,acc1,acc2,h)
!$omp&  shared(dumposmas,dumvel,velhaf,dumh)
!$omp&  private(i,j,thalf,t2half)
      do j=nactivef+1,ntop
         i=iactive(j)

         thalf=half*dble(itstepbin(i))*dtincr
         if(deltat(i).le.thalf)then

            dumh     (    i)=h(i)              !redo for sph below
            dumposmas(ims,i)=posmas(ims,i)
            dumposmas(  1,i)=posmas(1,i) + deltat(i)*vel (1,i)
            dumposmas(  2,i)=posmas(2,i) + deltat(i)*vel (2,i)
            dumvel   (  1,i)=   vel(1,i) + deltat(i)*acc1(1,i)
            dumvel   (  2,i)=   vel(2,i) + deltat(i)*acc1(2,i)
#if NDIM==3
            dumposmas(  3,i)=posmas(3,i) + deltat(i)*vel (3,i)
            dumvel   (  3,i)=vel   (3,i) + deltat(i)*acc1(3,i)
#endif
         else
            t2half=deltat(i)-thalf

            dumh     (    i)=h(i)              !redo for sph below
            dumposmas(ims,i)=posmas(ims,i)
            dumposmas(1,i)=posmas(1,i)+thalf*vel(1,i)+t2half*velhaf(1,i)
            dumposmas(2,i)=posmas(2,i)+thalf*vel(2,i)+t2half*velhaf(2,i)
            dumvel   (1,i)=vel   (1,i)+thalf*acc1(1,i)+t2half*acc2(1,i)
            dumvel   (2,i)=vel   (2,i)+thalf*acc1(2,i)+t2half*acc2(2,i)
#if NDIM==3
            dumposmas(3,i)=posmas(3,i)+thalf*vel(3,i)+t2half*velhaf(3,i)
            dumvel   (3,i)=vel   (3,i)+thalf*acc1(3,i)+t2half*acc2(3,i)
#endif
         endif
      enddo
!
! Now do point masses 
!
      do j=nactivefpm+1,ntop
         i=iactivepm(j)

         thalf=half*dble(itstepbinpm(i))*dtincr
         if(deltatpm(i).le.thalf)then

            dumhpm     (    i)=hpm(i)
            dumposmaspm(ims,i)=posmaspm(ims,i)
            dumposmaspm(  1,i)=posmaspm(1,i) + deltatpm(i)*velpm (1,i)
            dumposmaspm(  2,i)=posmaspm(2,i) + deltatpm(i)*velpm (2,i)
            dumvelpm   (  1,i)=   velpm(1,i) + deltatpm(i)*acc1pm(1,i)
            dumvelpm   (  2,i)=   velpm(2,i) + deltatpm(i)*acc1pm(2,i)
#if NDIM==3
            dumposmaspm(  3,i)=posmaspm(3,i) + deltatpm(i)*velpm (3,i)
            dumvelpm   (  3,i)=velpm   (3,i) + deltatpm(i)*acc1pm(3,i)
#endif
         else
            t2half=deltatpm(i)-thalf

            dumhpm     (    i)=hpm(i)
            dumposmaspm(ims,i)=posmaspm(ims,i)
            dumposmaspm(1  ,i)=posmaspm(1  ,i) +
     &                            thalf*velpm(1,i)+t2half*velhafpm(1,i)
            dumposmaspm(2  ,i)=posmaspm(2  ,i) +
     &                            thalf*velpm(2,i)+t2half*velhafpm(2,i)
            dumvelpm   (1  ,i)=velpm   (1  ,i) +
     &                            thalf*acc1pm(1,i)+t2half*acc2pm(1,i)
            dumvelpm   (2  ,i)=velpm   (2  ,i)+
     &                            thalf*acc1pm(2,i)+t2half*acc2pm(2,i)
#if NDIM==3
            dumposmaspm(3  ,i)=posmaspm(3  ,i) +
     &                            thalf*velpm(3,i)+t2half*velhafpm(3,i)
            dumvelpm   (3  ,i)=velpm   (3  ,i) +
     &                            thalf*acc1pm(3,i)+t2half*acc2pm(3,i)
#endif
         endif
      enddo
!
! Now do hydro quantities
!
!$omp parallel do if(ntop_sph-(nactivef_s+1) .gt. 1024) 
!$omp&  default(none)
!$omp&  firstprivate(umin,hmin,hmax,dhmax,alphastar)
!$omp&  shared(nactivef_s,ntop_sph,iactive_s,itstepbin)
!$omp&  shared(dtincr,deltat,rho,rhosave,divvsave,rhohalf,divvhalf)
!$omp&  shared(h,u,alpha,dhdt1,dudt1,dalfdt1,dhdt2,dudt2,dalfdt2)
!$omp&  shared(dumh,dumu,dumalf)
!$omp&  private(i,j,thalf,t2half)
!$omp&  private(utmp,atmp     ,htmp,dh)
      do j=nactivef_s+1,ntop_sph
         i=iactive_s(j)
         thalf=half*dble(itstepbin(i))*dtincr

         if(deltat(i).le.thalf)then

            rho(i)=rhosave(i) - deltat(i)*rhosave(i)*divvsave(i)

            dh     = deltat(i)*dhdt1(i)
            dh     = sign(min(abs(dh),dhmax*h(i)),dh)
            dumh(i)=min(hmax,max(h(i)+dh,hmin))

            utmp=u    (i) + deltat(i)*  dudt1(i)
            atmp=alpha(i) + deltat(i)*dalfdt1(i)
            dumu  (i)=max(utmp,umin)
            dumalf(i)=max(atmp,alphastar)
         else
            t2half=deltat(i)-thalf

            rho(i)=rhohalf(i) - t2half*rhohalf(i)*divvhalf(i)

            dh  = thalf*dhdt1(i) +t2half*dhdt2(i)
            dh  = sign(min(abs(dh),dhmax*h(i)),dh)
            htmp=h(i)+dh
            dumh(i)=min(hmax,max(htmp,hmin))

            utmp=u    (i)+thalf*dudt1  (i)+t2half*  dudt2(i)
            atmp=alpha(i)+thalf*dalfdt1(i)+t2half*dalfdt2(i)
            dumu  (i)=max(utmp,umin)
            dumalf(i)=max(atmp,alphastar)
         endif
      enddo

!
!  Save half time step velocities required for full time step
!  calculation.
!
!$omp parallel do if(npart-(nactivef+1) .gt. 1024) default(none)
!$omp&   shared(npart,iactive,nactivef,nactive,velhaf,dumvel)
!$omp&   private(i,j)
      do j=nactivef+1,nactive
         i=iactive(j)
         velhaf(1,i)=dumvel(1,i)
         velhaf(2,i)=dumvel(2,i)
#if NDIM==3
         velhaf(3,i)=dumvel(3,i)
#endif
      enddo

      do j=nactivefpm+1,nactivepm
         i=iactivepm(j)
         velhafpm(1,i)=dumvelpm(1,i)
         velhafpm(2,i)=dumvelpm(2,i)
#if NDIM==3
         velhafpm(3,i)=dumvelpm(3,i)
#endif
      enddo

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_halfupdt)=timing(itme_halfupdt)+ (tme2-tme1)
         ncalls(itme_halfupdt)=ncalls(itme_halfupdt)+ 1
         timing(itme_intgratr)=timing(itme_intgratr)+ (tme2-tme1)
!        ncalls(itme_intgratr) increment only for fullstep calls
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine halfupdate
!-----------------------------------------------------------------
      subroutine rk_err(nactivef_s,nactivef_n,nactivef,nactivefpm,
     &                  iactive_s ,iactive_n ,iactive ,iactivepm ,
     &                  itstepbin ,itstepbinpm,dtincr,
     &                  dtrat,dtratpm)
!***********************************************************
! RKF errors                                               *
!                                                          *
! See discussion in header of this file for information    *
! on these errors and how they work.                       *
!                                                          *
! The point masses here have hard coded error softeners.   *
! This avoids problems with the denominator of the error   *
! criterion going to zero. In other words, since the error *
! criterion is position and velocity dependent (yuck!)     *
! the error at position or velocity zero is infinite.      *
! Soften it to avoid this problem.                         *
!                                                          *
! Essentially this says: Don't care about position errors  *
! or velocity errors less than such and such value per     *
! timestep. This is a very problem dependent criterion     *
! and easy to get wrong. As such you should:               *
!                                                          *
!          Be afraid. Be very afraid.                      *
!                                                          *
!                                                          *
!***********************************************************
      integer, intent(in):: nactivef_s, iactive_s(nactivef_s)
      integer, intent(in):: nactivef_n, iactive_n(nactivef_n)
      integer, intent(in):: nactivef  , iactive  (nactivef  )
      integer, intent(in):: nactivefpm, iactivepm(nactivefpm)
      integer, intent(in):: itstepbin(idim),itstepbinpm(ipmdim)
      double precision, intent(in   ):: dtincr
      double precision, intent(inout):: dtrat(idim),dtratpm(ipmdim)

      integer i,j
      double precision  xerrmax, xdenom,xerr
      double precision  yerrmax, ydenom,yerr
      double precision  zerrmax, zdenom,zerr
      double precision vxerrmax,vxdenom,vxerr
      double precision vyerrmax,vydenom,vyerr
      double precision vzerrmax,vzdenom,vzerr
      double precision uerrmax  ,udenom,uerr
      double precision herrmax  ,hdenom,herr 
      double precision alferrmax,adenom,aerr

      double precision possoften,velsoften

      double precision dt_last, tolinv

! errmax is a locally defined array! 

      double precision errmax(idim),errmaxpm(ipmdim)

      character(len=20), parameter:: where =('rk_err')
      if(idbg.ge.2)print *,where

      do j=1,nactivef
         i=iactive(j)
         errmax(i)=zero
      enddo
      do j=1,nactivefpm
         i=iactivepm(j)
         errmaxpm(i)=zero
      enddo

      xerrmax  =zero ; yerrmax  =zero ; zerrmax  =zero
      vxerrmax =zero ; vyerrmax =zero ; vzerrmax =zero
      uerrmax  =zero ; herrmax  =zero ; alferrmax=zero

!$omp parallel do if(nactivef_s.gt.1024) default(none)
!$omp&  shared(dtincr)
!$omp&  shared(itstepbin,nactivef_s,iactive_s,errmax)
!$omp&  shared(dhdt1  ,dhdt2  ,h    ,hmin)      private(hdenom,herr)
!$omp&  shared(dudt1  ,dudt2  ,u    ,umin)      private(udenom,uerr)
!$omp&  shared(dalfdt1,dalfdt2,alpha,alphastar) private(adenom,aerr)
!$omp&  private(i,j,dt_last) 
!$omp&  reduction(max:herrmax,uerrmax,alferrmax)
      do j=1,nactivef_s
         i=iactive_s(j)
         dt_last=dtincr*dble(itstepbin(i))

         herr=zero ; uerr=zero ; aerr=zero

         if(    h(i).gt.hmin     )then
            hdenom=(abs(   h(i)) + abs(dt_last*dhdt1   (i)))
            herr=abs(dhdt1(i)-dhdt2(i))/hdenom
         endif
         if(    u(i).gt.umin     )then
            udenom=(abs   (u(i)) + abs(dt_last*dudt1   (i)))
            uerr  =abs(dudt1  (i)- dudt2  (i))/udenom
         endif
         if(alpha(i).gt.alphastar)then
            adenom=(abs(alpha(i)) + abs(dt_last*dalfdt1(i)))
            aerr  =abs(dalfdt1(i)- dalfdt2(i))/adenom
         endif
         errmax(i)=max(errmax(i),herr,uerr,aerr)
         herrmax  =max(herrmax  ,herr)
         uerrmax  =max(uerrmax  ,uerr)
         alferrmax=max(alferrmax,aerr)
      enddo

      tolinv=ONE/tol

!$omp parallel do if(nactivef.gt.1024)  default(none)
!$omp& firstprivate(tolinv,hmin)
!$omp& private(i,j,dt_last) shared(dtincr,errmax,dtrat)
!$omp& shared(itstepbin,nactivef,iactive)
!$omp& shared(posmas,vel,dumvel,acc1,acc2,cmrms,vcmrms,cmtot,vcmtot)
!$omp&       private(xdenom ,ydenom ,zdenom ,vxdenom ,vydenom ,vzdenom )
!$omp&       private(xerr   ,yerr   ,zerr   ,vxerr   ,vyerr   ,vzerr   )
!$omp& reduction(max:xerrmax,yerrmax,zerrmax,vxerrmax,vyerrmax,vzerrmax)
      do j=1,nactivef
         i=iactive(j)
         dt_last=dtincr*dble(itstepbin(i))

         xdenom=max(cmrms(1),hmin,abs(posmas(1,i)-cmtot(1))) +
     &                abs(dt_last*(vel(1,i)-vcmtot(1))) 
         ydenom=max(cmrms(2),hmin,abs(posmas(2,i)-cmtot(2))) +
     &                abs(dt_last*(vel(2,i)-vcmtot(2))) 
         vxdenom=max(vcmrms(1),abs(vel(1,i) - vcmtot(1))) + 
     &                abs(dt_last*acc1(1,i))
         vydenom=max(vcmrms(2),abs(vel(2,i) - vcmtot(2))) +
     &                abs(dt_last*acc1(2,i))
         xerr =abs(vel (1,i) - dumvel(1,i))/xdenom
         yerr =abs(vel (2,i) - dumvel(2,i))/ydenom
         vxerr=abs(acc1(1,i) - acc2  (1,i))/vxdenom
         vyerr=abs(acc1(2,i) - acc2  (2,i))/vydenom
         errmax(i)=max(errmax(i),xerr,yerr,vxerr,vyerr)
#if NDIM==3
         zdenom= max(cmrms(3),hmin,abs(posmas(3,i) - cmtot(3))) +
     &                abs(dt_last*(vel(3,i)-vcmtot(3)))
         vzdenom=max(vcmrms(3),abs(vel(3,i) - vcmtot(3))) +
     &                abs(dt_last*acc1(3,i))
         zerr=abs (vel (3,i) - dumvel(3,i))/zdenom
         vzerr=abs(acc1(3,i) - acc2  (3,i))/vzdenom
         errmax(i)=max(errmax(i),zerr,vzerr)
#endif 
!
!--RK present to next dt ratio 
!
         dtrat(i)=one/sqrt(dt_last*errmax(i)*e1*tolinv+small)

         xerrmax =max(xerrmax ,xerr)
         yerrmax =max(yerrmax ,yerr)
         vxerrmax=max(vxerrmax,vxerr)
         vyerrmax=max(vyerrmax,vyerr)
#if NDIM==3
         zerrmax =max(zerrmax ,zerr)
         vzerrmax=max(vzerrmax,vzerr)
#endif
      enddo


      if(idbg.ge.1)then
#if NDIM==2
         print 1001,xerrmax,yerrmax,vxerrmax,vyerrmax,
     &      uerrmax,herrmax,alferrmax
#else
         print 1001,xerrmax,yerrmax,zerrmax,vxerrmax,vyerrmax,vzerrmax,
     &      uerrmax,herrmax,alferrmax
#endif
         print *,' '
      endif

!
! Now do the error calculations for point masses
!
!
! Recall that I did say to be afraid of the 
! following two lines:
!
      possoften=3d5/d_unit     ! ignore this many cm/timestep
      velsoften=1d0/vel_unit   ! ignore this many cm/s/timestep

      xerrmax =zero  ; yerrmax =zero  ; zerrmax =zero
      vxerrmax=zero  ; vyerrmax=zero  ; vzerrmax=zero

      do j=1,nactivefpm
         i=iactivepm(j)
         dt_last=dtincr*dble(itstepbinpm(i))

         xdenom=( max(possoften, abs(posmaspm(1,i)-cmtot(1))) +
     &                abs(dt_last*(velpm(1,i)-vcmtot(1))) )
         ydenom=( max(possoften, abs(posmaspm(2,i)-cmtot(2))) +
     &                abs(dt_last*(velpm(2,i)-vcmtot(2))) )

         vxdenom=( max(velsoften, abs(velpm(1,i)-vcmtot(1))) +
     &                abs(dt_last*acc1pm(1,i)))
         vydenom=( max(velsoften, abs(velpm(2,i)-vcmtot(2))) +
     &                abs(dt_last*acc1pm(2,i)))

         xerr =abs(velpm (1,i) - dumvelpm(1,i))/xdenom
         yerr =abs(velpm (2,i) - dumvelpm(2,i))/ydenom
         vxerr=abs(acc1pm(1,i) - acc2pm  (1,i))/vxdenom
         vyerr=abs(acc1pm(2,i) - acc2pm  (2,i))/vydenom
         errmaxpm(i)=max(errmaxpm(i),xerr,yerr,vxerr,vyerr)
#if NDIM==3
         zdenom=( max(possoften, abs(posmaspm(3,i)-cmtot(3))) +
     &                abs(dt_last*(velpm(3,i)-vcmtot(3))))
         vzdenom=(max(velsoften, abs(velpm(3,i)-vcmtot(3))) +
     &                abs(dt_last*acc1pm(3,i)))
         zerr=abs (velpm (3,i) - dumvelpm(3,i))/zdenom
         vzerr=abs(acc1pm(3,i) - acc2pm  (3,i))/vzdenom
         errmaxpm(i)=max(errmaxpm(i),zerr,vzerr)
#endif
!
! RK present to next dt ratio
!
         dtratpm(i)=one/sqrt(dt_last*errmaxpm(i)*e1/tolpm+small)

         xerrmax =max(xerrmax ,xerr)
         yerrmax =max(yerrmax ,yerr)
         vxerrmax=max(vxerrmax,vxerr)
         vyerrmax=max(vyerrmax,vyerr)
#if NDIM==3
         zerrmax =max(zerrmax ,zerr)
         vzerrmax=max(vzerrmax,vzerr)
#endif
      enddo

      if(idbg.ge.1)then
#if NDIM==2
         print 1001,xerrmax,yerrmax,vxerrmax,vyerrmax
#else
         print 1001,xerrmax,yerrmax,zerrmax,vxerrmax,vyerrmax,vzerrmax
#endif
      endif


1001  format(15(1pd12.4))
      return
      end subroutine rk_err

      end module integ
