      module tree
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!  This file contains the VINE module to perform two important  *
!  tasks: Create lists of neighbors for a set of particles      *
!  given to it and resident in a previously constructed tree,   *
!  or to calculate interparticle forces on a set of particles   *
!  given to it and resident in the tree, due to all the other   *
!  particles in that tree. The force calculation is actually    *
!  done in an independent module called `forcelaw', through     *
!  calls to routines present there.                             *
!                                                               *
!  The tree module is aware of several options for calculating  *
!  the forces, through calling arguments in each routine.       *
!  It is possible to calculate forces:                          *
!     1) Directly, i.e. each particle's contribution on         *
!        each other particle is calculated and summed.          *
!        a) Using a GRAPE coprocessor                           *
!        b) Using the host processor                            *
!     2) Approximately, using a tree traversal, where tree      *
!        nodes passing some accuracy criterion are used as      *
!        approximate surrogates for the contributions of all    *
!        particles contained in those nodes. Currently,         *
!        traversals implementing several different node         *
!        opening criteria are included in the code. See         *
!        the code documentation and papers for details of       *
!        what these are.                                        *
!                                                               *
!  As written, calls are made to one or more of a set of        *
!  of routines living in the forcelaw module ('use'd by this    *
!  module), which actually perform the calculations.            *
!  At present, the code assumes the force law is the 1/r**2     *
!  force due to gravity.                                        *
!                                                               *
!  Two caveats:                                                 *
!                                                               *
!  1) In principle, this module could use any tree              *
!   (binary or oct). In practice, your mileage may vary.        *
!   See also the maketree.F file, for details of the tree       *
!   construction.                                               *
!                                                               *
!  2) In principle, this module could use/call any force law,   *
!   (gravity or other). In practice, your mileage may vary.     *
!   See also the forcelaw.F file, for details of the force law  *
!   used. The force law used will have a significant impact     *
!   on this module through changes in the node opening          *
!   criteria. Using the wrong criterion with the wrong force    *
!   will naturally produce large errors and perhaps nonsensical *
!   results.                                                    * 
!                                                               *
!                                                               *
!  WARNING: The exact tree that this module uses depends        *
!  on the number of processors used to build it and the         *
!  number of other things running on the system at the time.    *
!  This is normal and is due to the fact that tree nodes are    *
!  placed in the memory in segments that correspond to          *
!  nodes that were defined by one processor or another. The     *
!  order of the placement can vary, depending on which          *
!  processor gets done with its work first.                     *
!                                                               *
!  This has implications for tests of the accuracy of the       *
!  force calculation because of the intermittent reordering     *
!  that we do on the particles (see subroutine orderparticles). *
!  You have to make sure that you are determining the error of  *
!  the force on the same particle that you use for your         *
!  reference. Use the ipindx array for this purpose.            *
!                                                               *
!  The externally visible routines are:                         *
!     activate_tree: tell the tree which particles or groups    *
!                    of particles need accelerations            *
!       Arguments:                                              *
!         igrape   : Use grape or not                           *
!         numclumps: Number of clumps in the tree               *
!         lstclumps: Their location in the tree                 *
!         nactive  : Number of active particles                 *
!         iactive  : Their locations (in driver numeration)     *
!         nactive_c: Number of active clumps                    *
!         iactive_c: Their locations in the tree                *
!                                                               *
!     getparticles: get a list of active/inactive particles     *
!                   of the tree node `inode'                    *
!       Arguments:                                              *
!         inode    : tree node to search for particles          *
!         n        : Length of listpt_* arrays                  *
!         ntot_a   : Number of active particles in inode        *
!         ntot_i   : Number of inactive particles in inode      *
!         listpt_a : list of active particles in inode          *
!                    (driver numbering)                         *
!         listpt_i : list of inactive particles in inode        *
!                    (driver numbering)                         *
!                                                               *
!     gravity      : Calculate the gravitational forces on      *
!                    a set of particles. If you choose, you     *
!                    could replace the forcelaw module and      *
!                    some of the node opening criteria if you   *
!                    want to replace the gravity forces with    *
!                    something else.                            *
!       Arguments:                                              *
!         igrav    : Do gravity calculation or not.             *
!         ibctype  : Free, periodic or reflecting boundaries.   *
!         idirect  : Use direct summation to get gravity        *
!         igrape   : Use grape or not                           *
!         igrapelim: Minimum number of particles to calculate   *
!                    gravity for using grape.                   *
!         imac     : Which MAC to use.                          *
!         isoft    : Variable kernel, fixed kernel or Plummer   *
!                    softening                                  *
!         eps      : if isoft.eq.id_fixkern, spline soft. rad.  *
!                    if isoft.eq.id_Plummer, Plummer soft. rad. *
!         gmasslim : Maximum mass of acceptable node for        *
!                    gravity calculation (needed for grape)     *
!         nactive_c: Number of active clumps                    *
!         iactive_c: List of active clump locations in the tree *
!         nactive  : Number of active particles                 *
!         iactive  : List of active particle indices (driver    *
!                  : ordering)                                  *
!         npart    : Number of particles                        *
!         posmas   : Particle position (only needed for grape6) *
!         vel      : Particle velocity (only needed for grape6) *
!         acc      : Gravitational acceleration                 *
!         poten    : Gravitational potential                    *
!                                                               *
!     initgetneigh : Initialize some internal tree stuff for    *
!                    use in getneigh                            *
!                                                               *
!     getneigh: find the neighbors and assorted neighbor        *
!               information for particles                       *
!       Arguments:                                              *
!         ibctype  : Free, periodic or reflecting boundaries.   *
!         iboth    : Get neighbors for both nbody and sph parts.*
!                    Otherwise only SPH particles               *
!         iclump   : Clump that we're working on now.           *
!         ntot_a   : Number of active particles in iclump       *
!         ntot_i   : Number of inactive particles in iclump     *
!         listpt_a : list of active particles in iclump         *
!                    (driver numbering)                         *
!         listpt_i : list of inactive particles in iclump       *
!                    (driver numbering)                         *
!         nstart   : First clump to search for neighbors        *
!         nneigh   : number of neighbors                        *
!         listnay  : list index of neighbor (driver numbering)  *
!         distmas  : Quantities for sph calculations            *
!            idx   : xi - xj                                    *
!            idy   : yi - yj                                    *
!            idz   : zi - zj                                    *
!            ims   : mass of neighbor particle                  *
!            ir2   : squared distance                           *
!            ih2   : squared smoothing length                   *
!         moreneig : Found all neighbors or not?                *
!                                                               *
!  All other routines and variables should be private to this   *
!  module.                                                      *
!                                                               *
!  Written by Andy Nelson, Markus Wetzstein and Thorsten Naab   *
!                                                               *
!****************************************************************
      use dimens
      use treevars
      use constants
      use perform
      use stubs
#if SELFFORCE
      use forcelaw
#endif

      implicit none

      private
      public gravity
      public getparticles,getallparticles
      public initgetneigh,getneigh,getallneigh
      public activate_tree,deactivate_tree

!
! Someday this array will be an array for a clump/bunch of
! particles to use to store incremental corrections to the
! accumulated acceleration and potential for each particle
! using Kahan's algorithm. 
!
      double precision apcorr(0:ndim,max(maxclumpsize,maxbunchsize))

      contains
!----------------------------------------------------------------------
      subroutine activate_tree(igrape,numclumps,lstclumps,
     &                         nactive,iactive,nactive_c,iactive_c)
!***********************************************************************
!                                                                      *
! Subroutine to flag particles in the tree that need to be             *
!   updated etc. Also flag clumps, and return a list of                *
!   clumps that are active                                             *
!                                                                      *
!  Subroutine by A. Nelson                                             *
!                                                                      *
!***********************************************************************

      integer, intent(in) :: igrape
      integer, intent(in) :: nactive  ,iactive(nactive)
      integer, intent(in) :: numclumps,lstclumps(numclumps)
      integer, intent(out):: nactive_c,iactive_c(numclumps)
      integer i,me,my_dau,jclump,iclump,ibunch,jbunch
      integer icl_sib,ibnch_sib

      integer my_cpu,numcpus,icpu
      integer nperproc,nclo,nchi,ncact,nc_all(maxcpus)

      double precision tme1,tme2

      character(len=20), parameter:: where =('activate_tree')
      if(idbg.ge.3)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      nactive_c=0

!$omp parallel default(none)
!$omp& shared(nactive,natom,iactive,active,linksphtotree)
!$omp& shared(numclumps,lstclumps,nactive_c,iactive_c,link,nc_all)
!$omp& private(i,me,my_dau,iclump,jclump,icl_sib)
!$omp& private(icpu,my_cpu,numcpus,nperproc,nclo,nchi,ncact)
!$omp do 
      do i=1,nactive
         me=linksphtotree(iactive(i))
         active(me)=1
      enddo

      call get_mycpu(my_cpu)
      call numthreads(numcpus)

      nperproc=numclumps/numcpus

      nclo= 1 + nperproc*(my_cpu-1)
      nchi=     nperproc*my_cpu
      if(my_cpu.eq.numcpus)nchi=numclumps

      ncact=0
      do jclump=nclo,nchi
         iclump=lstclumps(jclump)

         icl_sib=abs(link(isib,iclump))

         if(icl_sib.eq.iclump+1 .and.
     &                active(iclump).eq.1)ncact=ncact+1

         do me=iclump+1,icl_sib-1
            my_dau=abs(link(idau,me))
            if(my_dau.eq.0)then
               active(iclump)= active(me)
               if(active(iclump).ne.0)then
                  ncact=ncact + 1
                  exit
               endif
            endif
         enddo
      enddo

      nc_all(my_cpu)=ncact

!$omp critical
      nactive_c=nactive_c+ncact
!$omp end critical
!$omp barrier

      ncact=0
      do icpu=1,my_cpu-1
         ncact =ncact + nc_all(icpu)
      enddo

      do jclump=nclo,nchi
         iclump=lstclumps(jclump)
         if(active(iclump).eq.1)then
            ncact=ncact+1
            iactive_c(ncact)=iclump
         endif
      enddo
!$omp end parallel

      if(igrape.eq.1)then
         do jbunch=1,numbunches
            ibunch=lstbunches(jbunch)
            if(link(idau,ibunch).ne.0)active(ibunch)=0

            ibnch_sib=abs(link(isib,ibunch))

            do me=ibunch+1,ibnch_sib-1
               my_dau=abs(link(idau,me))
               if(my_dau.eq.0)then
                  active(ibunch)= active(me)
                  if(active(ibunch).ne.0)exit
               endif
            enddo
         enddo
      endif

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_activate)=timing(itme_activate) + tme2-tme1
         ncalls(itme_activate)=ncalls(itme_activate)+1
         timing(itme_gravity)=timing(itme_gravity) + tme2-tme1
!         ncalls(itme_gravity)=ncalls(itme_gravity)+1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine activate_tree
!----------------------------------------------------------------------
      subroutine deactivate_tree(igrape,nactive,iactive,
     &                           nactive_c,iactive_c)
!***********************************************************************
!  Subroutine to deactivate nodes in the tree. To be used after        *
!  a timestep is completed and we want to get ready for another one.   *
!                                                                      *
!  Subroutine by A. Nelson                                             *
!                                                                      *
!***********************************************************************
      integer, intent(in):: igrape
      integer, intent(in):: nactive  ,iactive(nactive)
      integer, intent(in):: nactive_c,iactive_c(nactive_c)

      integer i,me,ibunch,jbunch

      double precision tme1,tme2
      character(len=20), parameter::where='deactivate_tree'
      if(itme.ge.1)call cpu_time(tme1)

!$omp parallel default(none)
!$omp& shared(nactive,iactive,nactive_c,iactive_c)
!$omp& shared(active,linksphtotree)
!$omp& shared(numbunches,lstbunches)
!$omp& private(i,me,ibunch,jbunch)
!$omp do 
      do i=1,nactive
         me=linksphtotree(iactive(i))
         active(me)=0
      enddo
!$omp do 
      do i=1,nactive_c
         active(iactive_c(i))=0
      enddo

!$omp do
      do jbunch=1,numbunches
         ibunch=lstbunches(jbunch)
         active(ibunch)=0
      enddo
!$omp end parallel


      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_activate)=timing(itme_activate) + tme2-tme1
!         ncalls(itme_activate)=ncalls(itme_activate)+1
         timing(itme_gravity)=timing(itme_gravity) + tme2-tme1
!         ncalls(itme_gravity)=ncalls(itme_gravity)+1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine deactivate_tree
!----------------------------------------------------------------------
      subroutine gravity(igrav,ibctype,idirect,igrape,igrapelim,imac,
     &    isoft,eps,gmasslim,nactive_c,iactive_c,nactive,iactive,
     &    npart,posmas,vel,acc,poten)
!***********************************************************************
!                                                                      *
!  Subroutine to traverse the tree, then use the results of that       *
!     traversal to calculate forces on particles using routines that   *
!     live in the forcelaw module. Currently set to be a gravitational *
!     interaction, hence the name of this routine. Could be any        *
!     interaction however, though you might need to modify some of     *
!     the traversals to have different node opening criteria.          *
!     Returns:                                                         *
!     acc, poten                                                       *
!          the acceleration and potential due to gravity for all       *
!          particles on all active particles.                          *
!                                                                      *
!  idirect=0: Use tree to winnow the lists of interactions             *
!  idirect=1: direct N**2 accel calculation (host or grape)            *
!                                                                      *
!  igrape=0: Don't use the grape hardware, even if it exists.          *
!  igrape=1: Use the grape hardware.                                   *
!                                                                      *
!  Subroutine by A. Nelson                                             *
!                                                                      *
!***********************************************************************
      integer, intent(in)::igrav,ibctype,idirect,igrape,igrapelim
      integer, intent(in)::imac,isoft,npart
      double precision, intent(in):: eps,gmasslim

      integer, intent(in):: nactive_c,iactive_c(nactive_c)
      integer, intent(in):: nactive  ,iactive  (nactive)
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      integer, parameter::lmax=max(maxclumpsize,maxbunchsize)
      integer nact,ntot,listpt(lmax)

! needed only for grape and/or Nbody direct summation 
      double precision, intent(in):: posmas(ndim1,npart),vel(ndim,npart)

      integer nla,nln,ncl,latoms(idim),lnodes(idim/2),neigclumps(idim)
      common/lists/nla,nln,ncl,latoms,lnodes,neigclumps
!$omp threadprivate (/lists/)

      integer numgroups,imacabs,jclump,iclump
      double precision eps2
      integer nglo,nghi,ng_inc

      double precision tme1,tme2,tme3,tme4
      character(len=20), parameter:: where =('gravity')
      if(idbg.ge.2)print *,where

      if(igrav  .eq.0) return
      if(nactive.eq.0) return
#ifndef SELFFORCE
      if(igrav  .eq.1)stop'gravity:Gravity wanted but not installed'
#endif

      if(itme.ge.1)call cpu_time(tme1)

      if(igrape.eq.1)then
         call lock_grape
         numgroups=numbunches
      else
         numgroups=nactive_c
      endif
      eps2=eps*eps
      imacabs=abs(imac)


      if(idirect.eq.1)then

         call dirsum(igrape,npart,isoft,eps2,posmas,vel,acc,poten)
      else

!$omp parallel do default(none) private(jclump,iclump)
!$omp&   private(tme3,tme4) shared(tme1)
!$omp&   private(ntot,nact,listpt)
!$omp&   shared (imacabs,igrape,isoft,igrapelim,gmasslim,eps2)
!$omp&   shared (npart,active,posmas,vel,acc,poten)
!$omp&   shared (ng_inc,nglo,nghi,numgroups,iactive_c,lstbunches) 
!$omp&   schedule(dynamic,32)
         do jclump=1,numgroups

            if(igrape.eq.0)then 
               iclump=iactive_c(jclump)
            else
               iclump=lstbunches(jclump)
            endif

            if(active(iclump).eq.1)then   !need this if for bunches

               call getallparticles(iclump,lmax,ntot,nact,listpt)

               call farwalk  (imacabs,iclump,ntot,     listpt,
     &                        nla,latoms,nln,lnodes,ncl,neigclumps,
     &                        igrape,gmasslim)
               call closewalk(imacabs,iclump,ntot,nact,listpt,
     &                        nla,latoms,nln,lnodes,ncl,neigclumps,
     &                        igrape,gmasslim,igrapelim,isoft,eps2,
     &                        npart,vel,acc,poten)
            endif
         enddo

      endif

      if(igrape.eq.1)call unlock_grape
!
! Save accelerations for use in Gadget (or other) MAC on next timestep
!  (also trigger on negative imac, to account for the first time through)
!
      if(imacabs.ne.imac .or. imac.eq.id_gadget)then
         call saveacc(nactive,iactive,npart,acc)
      endif
!
! If there are periodic boundaries apply Ewald correction.
!
      if(ibctype.eq.id_perdbnd)call apply_ewald(npart,posmas,acc,poten)


      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_gravacc)=timing(itme_gravacc) + tme2-tme1
         ncalls(itme_gravacc)=ncalls(itme_gravacc)+1
         timing(itme_gravity)=timing(itme_gravity) + tme2-tme1
         ncalls(itme_gravity)=ncalls(itme_gravity)+1
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      if(idbg.ge.3)print *,'done ',where

      return
      end subroutine gravity
!----------------------------------------------------------------------
      subroutine saveacc(nactive,iactive,npart,acc)
!*********************************************************************
!                                                                    *
! Save the accelerations for use the next time we want to update     *
! the accleration values with the gadget MAC.                        *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************

      integer, intent(in):: npart,nactive,iactive(nactive)
      double precision, intent(in):: acc(ndim,npart)
      integer i,j

!$omp parallel do if(nactive.gt.1024) default(none) 
!$omp&  shared(nactive,iactive,amag,acc,linksphtotree)
!$omp&  private(j,i)
      do j=1,nactive
         i=iactive(j)
         amag(i)=acc(1,i)**2 + acc(2,i)**2
#if NDIM==3
     &                       + acc(3,i)**2
#endif
        amag(i)=sqrt(amag(i))
      enddo

      return
      end subroutine saveacc
!----------------------------------------------------------------------
      subroutine closewalk(imac,iclump,ntot,nact,listpt,
     &                     nla,latoms,nln,lnodes,ncl,neigclumps,
     &                     igrape,gmasslim,igrapelim,isoft,eps2,
     &                     npart,vel,acc,poten)
!*********************************************************************
!                                                                    *
! Traverse through clumps in physical contact with jclump to find    *
! nodes, atoms and neighbors for each particle.                      *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in   ):: imac,iclump,igrape,igrapelim,isoft,npart
      integer, intent(in   ):: nact,ntot,listpt(ntot)
      integer, intent(inout):: nla,nln,latoms(idim),lnodes(idim/2)
      integer, intent(inout):: ncl,neigclumps(ncl+1)
      double precision, intent(in   ):: eps2,gmasslim
      double precision, intent(in   ):: vel(ndim,npart)
      double precision, intent(inout):: acc(ndim,npart), poten(npart)

      integer nla0,nln0,nlntmp,nla1,iact
      integer j,itreept,nclump,me,n,nonay
      double precision amagold

!
! Initialize Kahan's algorithm correction factor
!
      do n=1,nact
         apcorr(0,n)=ZERO
         apcorr(1,n)=ZERO
         apcorr(2,n)=ZERO
#if NDIM==3
         apcorr(3,n)=ZERO
#endif
      enddo
!
! Do calculation for distant nodes for whole group together 
! since its faster
!
      nla0=nla
      nla1=nla+1
      nln0=nln
      if(nact.gt.1)then
         call calcgravgroup(igrape,igrapelim,isoft,eps2,npart,iclump,
     &                      nact,ntot,listpt,nla0,nln0,latoms,lnodes,
     &                      vel,apcorr,acc,poten)
         nla0=0 ; nla1=nla0+1
         nln0=0
      endif
      if(igrape.eq.1)return
!
! Own clump, do n**2 calculation since its faster
!
      if(maxclumpsize.le.maxn2)then
        call acceln2(npart,nact,ntot,listpt,isoft,eps2,apcorr,acc,poten)
      else
         ncl=ncl+1
         neigclumps(ncl)=iclump
      endif
!
! Now sort out nodes and atoms from neighbor clumps
!
      iact=0
      do j=1,ntot
         itreept=listpt(j)
         if(active(itreept).ne.1)cycle
         iact=iact+1

         nla=nla0 ; nln=nln0
         do nclump=1,ncl
            me=neigclumps(nclump)
            if(imac.eq.id_gadget)then
               amagold=amag(linktreetosph(itreept))
               call getgad_atomnodes(amagold,gmasslim,me,itreept,
     &                               nla,latoms,nln,lnodes)
            else
               call get_atomnodes(gmasslim,me,itreept,
     &                            nla,latoms,nln,lnodes)
            endif
         enddo

         call calcgrav(itreept,iact,nact,isoft,npart,eps2,
     &                 nla0,nla1,nla,nln,latoms,lnodes,
     &                 apcorr,acc,poten)

      enddo

      return
      end subroutine closewalk
!----------------------------------------------------------------------
      subroutine farwalk(imac,iclump,ntot,listpt,
     &                   nla,latoms,nln,lnodes,ncl,neigclumps,
     &                   igrape,gmasslim)
!*********************************************************************
!                                                                    *
! Traverse the tree to find neigbor clumps and acceptable            *
!   nodes and atoms for clumps                                       *
!                                                                    *
! Specific implementation:                                           *
!                                                                    *
! Get the list of nodes that are acceptable for all particles        *
!  in iclump. Also get the list of possible neighbor clumps, then    *
!  filter out the distant (non-touching) clumps, placing them or     *
!  their children on interaction lists. Put the near (touching)      *
!  clumps on the neighbor clump list. Works by replacing a distant   *
!  clump in the middle of the list, with a near clump from the top   *
!  of the list, then lowering the top of the list. If grape then     *
!  put everything onto interaction lists.                            *
!                                                                    *
!   Subroutine by A. Nelson                                          *
!                                                                    *
!*********************************************************************

      integer, intent(in) :: imac,iclump,igrape
      integer, intent(in) :: ntot,listpt(ntot)
      
      integer, intent(out):: nla,latoms(idim),nln,lnodes(idim/2)
      integer, intent(out):: ncl,neigclumps(idim)
      double precision, intent(in):: gmasslim
      integer ntop,j,n,me,isphpt,itreept
      double precision amagold,bigmass

#include "treefuncs"

      nln =0 ; nla =0 ; ncl =0

      if(imac.eq.id_gadget)then
         amagold=huge(1d0)
         do j=1,ntot
            itreept=listpt(j)
            isphpt=linktreetosph(itreept)
            if(active(itreept).eq.1)amagold=min(amagold,amag(isphpt))
         enddo
         call getgad_clmpnodes(amagold,gmasslim,nroot,iclump,
     &                         nln,lnodes,ncl,neigclumps)
      else
         call get_clmpnodes(gmasslim,nroot,iclump,
     &                      nln,lnodes,ncl,neigclumps)
      endif
!
! filter clumps not in physical contact
!
      bigmass=huge(1d0)
      if(igrape.eq.0)then
         ntop=ncl
         do n=ncl,1,-1
            me=neigclumps(n)
            if(hsum2(me,iclump).lt.rr2(me,iclump))then
               if(imac.eq.id_gadget)then
                  call getgad_atomnodes(amagold,bigmass,me,iclump,
     &                                  nla,latoms,nln,lnodes)
               else
                  call get_atomnodes(bigmass,me,iclump,
     &                               nla,latoms,nln,lnodes)
               endif
               neigclumps(n)=neigclumps(ntop)
               ntop=ntop-1
            endif
         enddo
         ncl=ntop

      else
!
! For grape. Don't bother with filter. Just do traversals
! for all neighbor clumps and get lists of nodes and atoms.
!
         do n=ncl,1,-1
            me=neigclumps(n)
            if(imac.eq.id_gadget)then
               call getgad_atomnodes(amagold,bigmass,me,iclump,
     &                               nla,latoms,nln,lnodes)
            else
               call get_atomnodes(gmasslim,me,iclump,
     &                            nla,latoms,nln,lnodes)
            endif
         enddo
      endif

      return
      end subroutine farwalk
!----------------------------------------------------------------------
      subroutine initgetneigh
!*********************************************************************
!                                                                    *
! Initialize some values and stacks for use in the getneigh          *
! routine.                                                           *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************

      integer lclmp(idim),nc(2,maxtreelev),
     &        lwork(idim),nw(2,maxtreelev),
     &        mybranch(maxtreelev), my_lev
      common /worklists/lclmp, nc,
     &                  lwork, nw,
     &                  mybranch, my_lev
!$omp threadprivate (/worklists/)

      nw(1,1)=1  ; nw(2,1)=1
      nc(1,1)=1  ; nc(2,1)=0
      lwork(nw(1,1))=nroot

      my_lev=1

      return
      end subroutine initgetneigh
!----------------------------------------------------------------------
      subroutine getneigh(ibctype,iboth,iclump,ntot_a,ntot_i,
     &  listpt_a,listpt_i,nstart,nneigh,listnay,distmas,moreneig)
!*********************************************************************
!                                                                    *
! Get the particles which can contribute to the SPH smoothed         *
!  parameters like density.                                          *
!                                                                    *
!   iboth=1                                                          *
!            get SPH   neighbors for SPH   particles                 *
!        AND get Nbody neighbors for Nbody particles                 *
!   iboth=0                                                          *
!            get SPH   neighbors for SPH   particles                 *
!       DONT get Nbody neighbors for Nbody particles                 *
!                                                                    *
! Note that this version currently assumes that neighbors are        *
! defined as separations less than 2h. This won't necessarily        *
! be true with all kernels.                                          *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in   ):: ibctype,iboth,iclump
      integer, intent(in   ):: ntot_a,listpt_a(maxclumpsize)
      integer, intent(in   ):: ntot_i,listpt_i(maxclumpsize)
      integer, intent(inout):: nstart
      integer, intent(out  ):: nneigh(maxclumpsize)
      integer, intent(out  ):: listnay(maxbite,maxclumpsize)
      integer, intent(out  ):: moreneig
      double precision, intent(out):: 
     &           distmas(ndim3,maxbite,maxclumpsize)

      integer nla,nln,ncl,latoms(idim),lnodes(idim/2),neigclumps(idim)
      common/lists/nla,nln,ncl,latoms,lnodes,neigclumps
!$omp threadprivate (/lists/)

      integer i,jpart,naymax,node,inode


      naymax=0
      do jpart=1,ntot_a
         nneigh(jpart)=0
      enddo

      if(nstart.eq.0)then
!
!...deal with the node the particle is a member of
!
         if(maxclumpsize.le.maxn2)then
            call nayn2(iboth,ntot_a,listpt_a,
     &               ntot_i,listpt_i,nneigh,listnay,distmas)
            do i=1,ntot_a
               naymax=max(naymax,nneigh(i))
            enddo
         else
            call getnaybs(iboth,ntot_a,listpt_a,iclump,
     &                    naymax,nneigh,listnay,distmas)
         endif
         if(ibctype.eq.id_reflbnd)call reflneigh(iboth,ntot_a,
     &              listpt_a,naymax,nneigh,listnay,distmas)
!
! search for neighbor clumps
!
         call getnayclmps(iclump,ncl,neigclumps)
!
! return if no neighbor clumps
!
         if(ncl.eq.0)then
            moreneig=0
            return
         endif
      endif
!
! deal with neighbor clumps
!
      node=max(nstart,1)
      do while(naymax.le.maxsafe .and. node.le.ncl)
         inode=neigclumps(node)
         call getnaybs(iboth,ntot_a,listpt_a,inode,
     &                 naymax,nneigh,listnay,distmas)
        if(ibctype.eq.id_reflbnd)call reflneigh(iboth,ntot_a,
     &              listpt_a,naymax,nneigh,listnay,distmas)
         node=node+1
      enddo
      moreneig=0
      if(node-1.ne.ncl)moreneig=1
      nstart=node

      return
      end subroutine getneigh
!----------------------------------------------------------------------
      subroutine getallneigh(ibctype,iclump,ntot_a,ntot_i,
     &  listpt_a,listpt_i,nstart,nneigh,listnay,distmas,moreneig)
!*********************************************************************
!                                                                    *
! Get the particles which can contribute to the SPH smoothed         *
! parameters like density. Do not differentiate between types of     *
! of particles (i.e. an sph particle can have nbody particles as     *
! neighbors.                                                         *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in   ):: ibctype,iclump
      integer, intent(in   ):: ntot_a,listpt_a(maxclumpsize)
      integer, intent(in   ):: ntot_i,listpt_i(maxclumpsize)
      integer, intent(inout):: nstart
      integer, intent(out  ):: nneigh(maxclumpsize)
      integer, intent(out  ):: listnay(maxbite,maxclumpsize)
      integer, intent(out  ):: moreneig
      double precision, intent(out):: 
     &           distmas(ndim3,maxbite,maxclumpsize)

      integer nla,nln,ncl,latoms(idim),lnodes(idim/2),neigclumps(idim)
      common/lists/nla,nln,ncl,latoms,lnodes,neigclumps
!$omp threadprivate (/lists/)

      integer i,jpart,naymax,node,inode


      naymax=0
      do jpart=1,ntot_a
         nneigh(jpart)=0
      enddo

      if(nstart.eq.0)then
!
!...deal with the node the particle is a member of
!
         if(maxclumpsize.le.maxn2)then
            call allnayn2(ntot_a,listpt_a,
     &               ntot_i,listpt_i,nneigh,listnay,distmas)
            do i=1,ntot_a
               naymax=max(naymax,nneigh(i))
            enddo
         else
            call getallnaybs(ntot_a,listpt_a,iclump,
     &                    naymax,nneigh,listnay,distmas)
         endif
         if(ibctype.eq.id_reflbnd)call allreflneigh(ntot_a,
     &              listpt_a,naymax,nneigh,listnay,distmas)
!
! search for neighbor clumps
!
         call getnayclmps(iclump,ncl,neigclumps)
!
! return if no neighbor clumps
!
         if(ncl.eq.0)then
            moreneig=0
            return
         endif
      endif
!
! deal with neighbor clumps
!
      node=max(nstart,1)
      do while(naymax.le.maxsafe .and. node.le.ncl)

         inode=neigclumps(node)
         call getallnaybs(ntot_a,listpt_a,inode,
     &                 naymax,nneigh,listnay,distmas)
        if(ibctype.eq.id_reflbnd)call allreflneigh(ntot_a,
     &              listpt_a,naymax,nneigh,listnay,distmas)
         node=node+1
      enddo
      moreneig=0
      if(node-1.ne.ncl)moreneig=1
      nstart=node

      return
      end subroutine getallneigh
!----------------------------------------------------------------------
      subroutine reflneigh(iboth,ntot_a,listpt_a,
     &               naymax,nneigh,listnay,distmas)
!*********************************************************************
!  Get ghost particles that are on the other side of a reflecting    *
!  boundary. Currently not implemented, but here is how:             *
!                                                                    *
!  For each real neighbor reflect it outside the grid in all 6       *
!    directions. Then check to see if that ghost particle is a       *
!    neighbor. If so, then add it to the list.                       *
!                                                                    *
!  Probably will also need a flag array to say which neighbors       *
!    are real and which are ghosts. (needed for the velocities)      *
!    This array could be an integer or byte array that gets:         *
!      irefl(  ) = 0  real particle                                  *
!      irefl(  ) = 1  reflect the x direction                        *
!      irefl(  ) = 2  reflect the y direction                        *
!      irefl(  ) = 4  reflect the z direction                        *
!                                                                    *
!  Caveat 1: Will need special handling for `self' ghost particles.  *
!            Not done below.                                         *
!                                                                    *
!  Caveat 2: Will need to increase maxsafe by a factor 4 over its    *
!            current value (could have as many as 4 copies of a      *
!            single particle...the real one, plus 3 ghosts).         *
!                                                                    *
!  Caveat 3: the sketched out code below doesn't keep track          *
!            of whether we looked at a particle already, and         *
!            if it is in the list etc. Probably need some            *
!            organizational code to set loop indices correctly       *
!            to do this.                                             *
!                                                                    *
!  I haven't implemented this beyond what you see, so I don't        *
!  know how well it might work, or what else may need to go          *
!  into it.--AFN                                                     *
!                                                                    *
!   Subroutine by A. Nelson                                          *
!                                                                    *
!*********************************************************************

      integer, intent(in   ):: iboth
      integer, intent(in   ):: ntot_a,listpt_a(maxclumpsize)
      integer, intent(inout):: nneigh(maxclumpsize),naymax
      integer, intent(inout):: listnay(maxbite,maxclumpsize)
      double precision, intent(inout)::
     &           distmas(ndim3,maxbite,maxclumpsize)
      integer j,k,id,itreept,ktreept
      double precision d2

#include "treefuncs"

      stop 'Reflecting boundaries not installed in reflneigh'

      do id=1,ndim
         do j=1,listpt_a(ntot_a)
            itreept=listpt_a(j)
            do k=1,nneigh(j)
               ktreept=listnay(k,j)
               d2=(treept(id,ktreept)-bndbox(1,id))**2
               if(d2.lt.distmas(ih2,j,k))then
!
! flip particle's id coordinate across to the other
! side of bndbox and see if its ghost is a neighbor
! or not. If so, put on list
!
               endif
               d2=(bndbox(2,id)-treept(id,ktreept))**2
               if(d2.lt.distmas(ih2,j,k))then
!
! flip particle's id coordinate across to the other
! side of bndbox and see if its ghost is a neighbor
! or not. If so, put on list
!
               endif
            enddo
         enddo
      enddo

      return
      end subroutine reflneigh
!----------------------------------------------------------------------
      subroutine allreflneigh(ntot_a,listpt_a,
     &               naymax,nneigh,listnay,distmas)
!*********************************************************************
!  Get ghost particles that are on the other side of a reflecting    *
!  boundary. Currently not implemented, but here is how:             *
!                                                                    *
!  For each real neighbor reflect it outside the grid in all 6       *
!    directions. Then check to see if that ghost particle is a       *
!    neighbor. If so, then add it to the list.                       *
!                                                                    *
!  Probably will also need a flag array to say which neighbors       *
!    are real and which are ghosts. (needed for the velocities)      *
!    This array could be an integer or byte array that gets:         *
!      irefl(  ) = 0  real particle                                  *
!      irefl(  ) = 1  reflect the x direction                        *
!      irefl(  ) = 2  reflect the y direction                        *
!      irefl(  ) = 4  reflect the z direction                        *
!                                                                    *
!  Caveat 1: Will need special handling for `self' ghost particles.  *
!            Not done below.                                         *
!                                                                    *
!  Caveat 2: Will need to increase maxsafe by a factor 4 over its    *
!            current value (could have as many as 4 copies of a      *
!            single particle...the real one, plus 3 ghosts).         *
!                                                                    *
!  Caveat 3: the sketched out code below doesn't keep track          *
!            of whether we looked at a particle already, and         *
!            if it is in the list etc. Probably need some            *
!            organizational code to set loop indices correctly       *
!            to do this.                                             *
!                                                                    *
!  I haven't implemented this beyond what you see, so I don't        *
!  know how well it might work, or what else may need to go          *
!  into it.--AFN                                                     *
!                                                                    *
!   Subroutine by A. Nelson                                          *
!                                                                    *
!*********************************************************************

      integer, intent(in   ):: ntot_a,listpt_a(maxclumpsize)
      integer, intent(inout):: nneigh(maxclumpsize),naymax
      integer, intent(inout):: listnay(maxbite,maxclumpsize)
      double precision, intent(inout)::
     &           distmas(ndim3,maxbite,maxclumpsize)
      integer j,k,id,itreept,ktreept
      double precision d2

#include "treefuncs"

      stop 'Reflecting boundaries not installed in reflneigh'

      do id=1,ndim
         do j=1,listpt_a(ntot_a)
            itreept=listpt_a(j)
            do k=1,nneigh(j)
               ktreept=listnay(k,j)
               d2=(treept(id,ktreept)-bndbox(1,id))**2
               if(d2.lt.distmas(ih2,j,k))then
!
! flip particle's id coordinate across to the other
! side of bndbox and see if its ghost is a neighbor
! or not. If so, put on list
!
               endif
               d2=(bndbox(2,id)-treept(id,ktreept))**2
               if(d2.lt.distmas(ih2,j,k))then
!
! flip particle's id coordinate across to the other
! side of bndbox and see if its ghost is a neighbor
! or not. If so, put on list
!
               endif
            enddo
         enddo
      enddo

      return
      end subroutine allreflneigh
!----------------------------------------------------------------------
      subroutine findbranch(itreept,my_lev,levstart,mybranch)
!*********************************************************************
!  Get all of the parent nodes of itreept and the level (levstart)   *
!   where the branch is different from the previous branch we        *
!   worked on.                                                       *
!                                                                    *
!  This is duplicated code from the tree build file treemake.F       *
!  Duplicated to preserve the independence of code in                *
!  the two files treemake.F and tree.F.                              *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************
      integer, intent(in):: itreept
      integer, intent(inout):: my_lev,levstart,mybranch(maxtreelev)

      integer m,my_dau,my_dausib

      if(my_lev.gt.1)then

         do while(abs(          mybranch(my_lev)) .gt.itreept .or.
     &            abs(link(isib,mybranch(my_lev))).le.itreept)
            my_lev=my_lev-1
         enddo
      else
         mybranch(my_lev)=nroot
      endif

      levstart=my_lev

      do while(itreept.gt.mybranch(my_lev))
         my_dau=abs(link(idau,mybranch(my_lev)))
         my_dausib=abs(link(isib,my_dau))

         my_lev=my_lev+1
         if(itreept.ge.my_dausib)then
            mybranch(my_lev)=my_dausib
         else
            mybranch(my_lev)=my_dau
         endif
      enddo

      do m=my_lev,ntreelev+1
         mybranch(m)=mybranch(my_lev)
      enddo

      return
      end subroutine findbranch
!----------------------------------------------------------------------
      subroutine getnayclmps(itreept,ncl,neigclumps)
!*********************************************************************
!  You see a maze of twisted little treewalks, all different         *
!                                                                    *
!  Get the clumps that are physically touching itreept               *
!  Recursive in the sense that info from previous tree traversals    *
!  is used for the current traversal                                 *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************
      integer, intent(in):: itreept
      integer, intent(out):: ncl,neigclumps(idim)

      integer lclmp(idim),nc(2,maxtreelev),
     &        lwork(idim),nw(2,maxtreelev),
     &        mybranch(maxtreelev), my_lev
      common /worklists/lclmp, nc,
     &                  lwork, nw,
     &                  mybranch, my_lev
!$omp threadprivate (/worklists/)

      integer ilast,levm1,lev,levstart,n,i
      integer me,meabs,my_dau,my_dauabs
      integer negitreept

#include "treefuncs"

      call findbranch(itreept,my_lev,levstart,mybranch)
      lev=levstart

      do while(nw(1,lev).le.nw(2,lev))

         levm1=lev
         lev=lev+1
         nw(1,lev)=nw(2,levm1)+1  ; nw(2,lev)=nw(1,lev)-1
         nc(1,lev)=nc(2,levm1)+1  ; nc(2,lev)=nc(1,lev)-1

         do i=nw(1,levm1),nw(2,levm1)
            meabs=abs(lwork(i))

            if(lwork(i).ne.meabs)then      !node is a clump
               nc(2,lev)=nc(2,lev)+1
               lclmp(nc(2,lev))=meabs 

            elseif(rr2neig(lwork(i),mybranch(lev)) .lt.
     &               hsum2(lwork(i),mybranch(lev)) )then

               my_dau    =link(idau,meabs)
               my_dauabs =abs(my_dau)

               nw(2,lev)=nw(2,lev)+1
               lwork(nw(2,lev))=my_dau
               nw(2,lev)=nw(2,lev)+1
               lwork(nw(2,lev))=link(isib,my_dauabs)
            endif
         enddo
!
! remove self from the list of current work
!
         if(my_lev.eq.lev)then
            negitreept=-itreept
            ilast=lwork(nw(2,lev))
            do i=nw(1,lev),nw(2,lev)
               if(negitreept.eq.lwork(i))then
                  lwork(i)=ilast
               endif
            enddo
            nw(2,lev)=nw(2,lev)-1
         endif
      enddo
!
! filter the list of clumps for the contact neighbor clumps
!
      ncl=0
      do n=1,nc(2,lev)
         me=lclmp(n)
         if(hsum2(me,itreept).gt.rr2neig(me,itreept))then
            ncl=ncl+1
            neigclumps(ncl)=lclmp(n)
         endif
      enddo

      return
      end subroutine getnayclmps
!----------------------------------------------------------------------
      subroutine getgad_clmpnodes(amagold,gmasslim,inode,
     &                 itreept,nln,lnodes,ncl,neigclumps)
!*********************************************************************
!  You see trees walking in a twisty little maze, each differently   *
!                                                                    *
!  Get the nodes that are acceptable for itreept and also all        *
!  clumps that come up to be tested, that may be touching itreept    *
!  Use the Gadget MAC including the geometric safety test.           *
!                                                                    *
!  Split the walk into two parts to avoid putting itreept itself     *
!  on the list.                                                      *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************
      integer, intent(in):: inode,itreept
      integer, intent(inout):: nln,lnodes(idim/2)
      integer, intent(inout):: ncl,neigclumps(idim)
      double precision, intent(in):: amagold,gmasslim

      integer me,meabs,inode_sib,my_sib,my_dau,itreeptabs

#include "treefuncs"

      me=inode
      meabs=abs(me)
      inode_sib=abs(link(isib,me))
      itreeptabs=abs(itreept)
      
      do while(meabs.ne.itreeptabs)
         my_sib=link(isib,meabs)
         my_dau=link(idau,meabs)
         
         if(me.ne.meabs)then
            ncl=ncl+1
            neigclumps(ncl)=meabs
            me=my_sib
         elseif(rr2(me,itreept).lt.hsum2(me,itreept))then
            me=my_dau
         elseif(amagold*r6conv(me,itreept) .lt.
     &                      treept(iconvrg,me))then
            me=my_dau
         elseif(treept(ims,me).le.gmasslim)then
            nln=nln+1
            lnodes(nln)=me
            me=my_sib
         else
            me=my_dau
         endif
         meabs=abs(me)
      enddo
      
      me=link(isib,abs(me))
      meabs=abs(me)
      do while(meabs.ne.inode_sib)
         my_sib=link(isib,meabs)
         my_dau=link(idau,meabs)
         
         if(me.ne.meabs)then
            ncl=ncl+1
            neigclumps(ncl)=meabs
            me=my_sib
         elseif(rr2(me,itreept).lt.hsum2(me,itreept))then
            me=my_dau
         elseif(amagold*r6conv(me,itreept) .lt.
     &                      treept(iconvrg,me))then
            me=my_dau
         elseif(treept(ims,me).le.gmasslim)then
            nln=nln+1
            lnodes(nln)=me
            me=my_sib
         else
            me=my_dau
         endif
         meabs=abs(me)
      enddo
      
      return
      end subroutine getgad_clmpnodes
!----------------------------------------------------------------------
      subroutine getgad_atomnodes(amagold,gmasslim,inode,itreept,
     &                            nla,latoms,nln,lnodes)
!*********************************************************************
!  You see trees walking in a little, twisty maze, each differently  *
!                                                                    *
!  Get the atoms and nodes that are acceptable for itreept,          *
!  using the Gadget MAC, including the geometric safety test         *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************
      integer, intent(in):: inode,itreept
      integer, intent(inout):: nla,latoms(idim)
      integer, intent(inout):: nln,lnodes(idim/2)
      double precision, intent(in):: amagold,gmasslim

      integer me,inode_sib,my_sib,my_dau
      
#include "treefuncs"

      me=inode
      inode_sib=link(isib,me)
      
      if(itreept.lt.inode .or. itreept.ge.abs(inode_sib))then
!
! Not in my own clump
!
         do while(me.ne.inode_sib)
            my_sib=link(isib,me)
            my_dau=link(idau,me)
            if(my_dau.eq.0)then
               nla=nla+1
               latoms(nla)=me
               me=my_sib
            elseif(rr2(me,itreept).lt.hsum2(me,itreept))then
               me=my_dau
            elseif(amagold*r6conv(me,itreept) .lt.
     &              treept(iconvrg,me))then
               me=my_dau
            elseif(treept(ims,me).le.gmasslim)then
               nln=nln+1
               lnodes(nln)=me
               me=my_sib
            else
               me=my_dau
            endif
         enddo
      else
!
! In my own clump
!
         do while(me.ne.inode_sib)
            my_sib=link(isib,me)
            my_dau=link(idau,me)
            if(my_dau.eq.0)then
               if(me.ne.itreept)then
                  nla=nla+1
                  latoms(nla)=me
               endif
               me=my_sib
            elseif(rr2(me,itreept).lt.hsum2(me,itreept))then
               me=my_dau
            elseif(amagold*r6conv(me,itreept) .lt.
     &              treept(iconvrg,me))then
               me=my_dau
            elseif(treept(ims,me).le.gmasslim)then
               nln=nln+1
               lnodes(nln)=me
               me=my_sib
            else
               me=my_dau
            endif
         enddo
      endif
      
      return
      end subroutine getgad_atomnodes
!----------------------------------------------------------------------
      subroutine get_clmpnodes(gmasslim,inode,itreept,nln,lnodes,
     &                         ncl,neigclumps)
!*********************************************************************
!  You see a twisted maze of little treewalks, all different         *
!                                                                    *
!  Get the nodes that are acceptable for itreept and also all        *
!    clumps that come up to be tested, that may be touching itreept  *
!                                                                    *
!  Split the walk into two parts to avoid putting itreept itself     *
!  on the list.                                                      *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************
      integer, intent(in):: inode,itreept
      integer, intent(inout):: nln,lnodes(idim/2)
      integer, intent(inout):: ncl,neigclumps(idim)
      double precision, intent(in):: gmasslim

      integer me,meabs,inode_sib,my_sib,my_dau,itreeptabs

#include "treefuncs"

      me=inode
      meabs=abs(me)
      inode_sib=abs(link(isib,me))
      itreeptabs=abs(itreept)

      do while(meabs.ne.itreeptabs)
         my_sib=link(isib,meabs)
         my_dau=link(idau,meabs)

         if(me.ne.meabs)then
            ncl=ncl+1
            neigclumps(ncl)=meabs
            me=my_sib
         elseif(rr2(me,itreept).lt.r2conv(me,itreept))then
            me=my_dau
         elseif(treept(ims,me).le.gmasslim)then
            nln=nln+1
            lnodes(nln)=me
            me=my_sib
         else
            me=my_dau
         endif
         meabs=abs(me)
      enddo

      me=link(isib,abs(me))
      meabs=abs(me)
      do while(meabs.ne.inode_sib)
         my_sib=link(isib,meabs)
         my_dau=link(idau,meabs)

         if(me.ne.meabs)then
            ncl=ncl+1
            neigclumps(ncl)=meabs
            me=my_sib
         elseif(rr2(me,itreept).lt.r2conv(me,itreept))then
            me=my_dau
         elseif(treept(ims,me).le.gmasslim)then
            nln=nln+1
            lnodes(nln)=me
            me=my_sib
         else
            me=my_dau
         endif
         meabs=abs(me)
      enddo

      return
      end subroutine get_clmpnodes
!----------------------------------------------------------------------
      subroutine get_atomnodes(gmasslim,inode,itreept,
     &                         nla,latoms,nln,lnodes)
!*********************************************************************
!  You see a little maze of twisted treewalks, all different         *
!                                                                    *
!  Get the atoms and nodes that are acceptable for itreept           *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************
      integer, intent(in):: inode,itreept
      integer, intent(inout):: nla,latoms(idim)
      integer, intent(inout):: nln,lnodes(idim/2)
      double precision, intent(in):: gmasslim

      integer me,inode_sib,my_sib,my_dau

#include "treefuncs"

      me=inode
      inode_sib=link(isib,me)

      if(itreept.lt.me .or. itreept.ge.abs(inode_sib))then
!
! Not in my own clump
!
         do while(me.ne.inode_sib)
            my_sib=link(isib,me)
            my_dau=link(idau,me)
            if(my_dau.eq.0)then
               nla=nla+1
               latoms(nla)=me
               me=my_sib
            elseif(rr2(me,itreept).lt.r2conv(me,itreept))then
               me=my_dau
            elseif(treept(ims,me).le.gmasslim)then
               nln=nln+1
               lnodes(nln)=me
               me=my_sib
            else
               me=my_dau
            endif
         enddo

      else
!
! In my own clump
!
         do while(me.ne.inode_sib)
            my_sib=link(isib,me)
            my_dau=link(idau,me)
            if(my_dau.eq.0)then
               if(me.ne.itreept)then
                  nla=nla+1
                  latoms(nla)=me
               endif
               me=my_sib
            elseif(rr2(me,itreept).lt.r2conv(me,itreept))then
               me=my_dau
            elseif(treept(ims,me).le.gmasslim)then
               nln=nln+1
               lnodes(nln)=me
               me=my_sib
            else
               me=my_dau
            endif
         enddo

      endif

      return
      end subroutine get_atomnodes
!----------------------------------------------------------------------
      subroutine getnaybs(iboth,ntot_a,listpt_a,inode,
     &                          naymax,nneigh,listnay,distmas)
!*********************************************************************
!  You see a little twisted maze of treewalks, all different         *
!                                                                    *
!  Do a search for particle neighbors in node inode.                 *
!                                                                    *
!  It would be much nicer visually to put the outer loop of          *
!  this routine (do i=1,ntot_a) outside the subroutine call.         *
!  Doing so is much slower though because of call overhead.          *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************

      integer, intent(in):: iboth,inode
      integer, intent(in):: ntot_a,listpt_a(maxclumpsize)
      integer, intent(inout):: nneigh(maxclumpsize), naymax
      integer, intent(inout):: listnay(maxbite,maxclumpsize)
      double precision, intent(inout)::
     &           distmas(ndim3,maxbite,maxclumpsize)
      integer i,isphpt,itreept,nb1,nb2,inode_sib
      integer mesph,me,my_sib,my_dau
      double precision dx,dy,dz,d2,h2

#include "treefuncs"

      do i=1,ntot_a
         isphpt=listpt_a(i)
         nb1=0
         if(isphpt.gt.natom_sph)nb1=1
         if(iboth.eq.1 .or. nb1.eq.0)then 

            itreept=linksphtotree(isphpt)
            me=inode
            inode_sib=link(isib,inode)

            if(itreept.lt.me .or. itreept.ge.inode_sib)then

! Don't worry about filtering out self (not in my own inode)
!
               do while(me.ne.inode_sib)
                  my_sib=link(isib,me)
                  my_dau=link(idau,me)
                  dx=dxi(itreept,me)
                  dy=dyi(itreept,me)
                  d2 =dx**2 + dy**2
#if NDIM==3
                  dz=dzi(itreept,me)
                  d2 = d2 + dz**2
#endif 
                  h2=hsum2(itreept,me)
                  if(h2.gt.d2)then
                     if(my_dau.ne.0)then
                        me=my_dau
                     else
                        mesph=linktreetosph(me)
                        nb2=0
                        if(mesph.gt.natom_sph)nb2=1
                        if(mod(nb1+nb2,2).eq.0)then
                           nneigh (              i)= nneigh(i)+1
                           listnay(    nneigh(i),i)= mesph
                           distmas(idx,nneigh(i),i)= dx
                           distmas(idy,nneigh(i),i)= dy
                           distmas(ims,nneigh(i),i)= treept(ims,me)
                           distmas(ir2,nneigh(i),i)= d2
                           distmas(ih2,nneigh(i),i)= fourth*h2
#if NDIM==3
                           distmas(idz,  nneigh(i),i)= dz
#endif
                        endif
                        me=my_sib
                     endif
                  else
                     me=my_sib
                  endif
               enddo

            else
!
! Must filter out self when found (in my own inode)
!
               do while(me.ne.inode_sib)
                  my_sib=link(isib,me)
                  my_dau=link(idau,me)
                  dx=dxi(itreept,me)
                  dy=dyi(itreept,me)
                  d2 =dx**2 + dy**2
#if NDIM==3
                  dz=dzi(itreept,me)
                  d2 = d2 + dz**2
#endif 
                  h2=hsum2(itreept,me)
                  if(h2.gt.d2)then
                     if(my_dau.ne.0)then
                        me=my_dau
                     else
                        mesph=linktreetosph(me)
                        nb2=0
                        if(mesph.gt.natom_sph)nb2=1
                        if(mod(nb1+nb2,2).eq.0)then
                           if(me.ne.itreept)then
                              nneigh (              i)= nneigh(i)+1
                              listnay(    nneigh(i),i)= mesph
                              distmas(idx,nneigh(i),i)= dx
                              distmas(idy,nneigh(i),i)= dy
                              distmas(ims,nneigh(i),i)= treept(ims,me)
                              distmas(ir2,nneigh(i),i)= d2
                              distmas(ih2,nneigh(i),i)= fourth*h2
#if NDIM==3
                              distmas(idz,nneigh(i),i)= dz
#endif
                           endif
                        endif
                        me=my_sib
                     endif
                  else
                     me=my_sib
                  endif
               enddo

            endif  !filter out self or not
         endif     !iboth or sphparticle
         naymax=max(naymax,nneigh(i))
      enddo        !loop over active particles

      return
      end subroutine getnaybs
!----------------------------------------------------------------------
      subroutine getallnaybs(ntot_a,listpt_a,inode,
     &                          naymax,nneigh,listnay,distmas)
!*********************************************************************
!  You see a little twisted maze of treewalks, all different         *
!                                                                    *
!  Do a search for particle neighbors in node inode.                 *
!                                                                    *
!  It would be much nicer visually to put the outer loop of          *
!  this routine (do i=1,ntot_a) outside the subroutine call.         *
!  Doing so is much slower though because of call overhead.          *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************

      integer, intent(in):: inode
      integer, intent(in):: ntot_a,listpt_a(maxclumpsize)
      integer, intent(inout):: nneigh(maxclumpsize), naymax
      integer, intent(inout):: listnay(maxbite,maxclumpsize)
      double precision, intent(inout)::
     &           distmas(ndim3,maxbite,maxclumpsize)
      integer i,isphpt,itreept,inode_sib
      integer mesph,me,my_sib,my_dau
      double precision dx,dy,dz,d2,h2

#include "treefuncs"

      do i=1,ntot_a
         isphpt=listpt_a(i)
         itreept=linksphtotree(isphpt)
         me=inode
         inode_sib=link(isib,inode)

         if(itreept.lt.me .or. itreept.ge.inode_sib)then

! Don't worry about filtering out self (not in my own inode)
!
            do while(me.ne.inode_sib)
               my_sib=link(isib,me)
               my_dau=link(idau,me)
               dx=dxi(itreept,me)
               dy=dyi(itreept,me)
               d2 =dx**2 + dy**2
#if NDIM==3
               dz=dzi(itreept,me)
               d2 = d2 + dz**2
#endif
               h2=hsum2(itreept,me)
               if(h2.gt.d2)then
                  if(my_dau.ne.0)then
                     me=my_dau
                  else
                     mesph=linktreetosph(me)
                     nneigh (              i)= nneigh(i)+1
                     listnay(    nneigh(i),i)= mesph
                     distmas(idx,nneigh(i),i)= dx
                     distmas(idy,nneigh(i),i)= dy
                     distmas(ims,nneigh(i),i)= treept(ims,me)
                     distmas(ir2,nneigh(i),i)= d2
                     distmas(ih2,nneigh(i),i)= fourth*h2
#if NDIM==3
                     distmas(idz,  nneigh(i),i)= dz
#endif
                     me=my_sib
                  endif
               else
                  me=my_sib
               endif
            enddo

         else
!
! Must filter out self when found (in my own inode)
!
            do while(me.ne.inode_sib)
               my_sib=link(isib,me)
               my_dau=link(idau,me)
               dx=dxi(itreept,me)
               dy=dyi(itreept,me)
               d2 =dx**2 + dy**2
#if NDIM==3
               dz=dzi(itreept,me)
               d2 = d2 + dz**2
#endif
               h2=hsum2(itreept,me)
               if(h2.gt.d2)then
                  if(my_dau.ne.0)then
                     me=my_dau
                  else
                     mesph=linktreetosph(me)
                     if(me.ne.itreept)then
                        nneigh (              i)= nneigh(i)+1
                        listnay(    nneigh(i),i)= mesph
                        distmas(idx,nneigh(i),i)= dx
                        distmas(idy,nneigh(i),i)= dy
                        distmas(ims,nneigh(i),i)= treept(ims,me)
                        distmas(ir2,nneigh(i),i)= d2
                        distmas(ih2,nneigh(i),i)= fourth*h2
#if NDIM==3
                        distmas(idz,nneigh(i),i)= dz
#endif
                     endif
                     me=my_sib
                  endif
               else
                  me=my_sib
               endif
            enddo

         endif  !filter out self or not
         naymax=max(naymax,nneigh(i))
      enddo        !loop over active particles

      return
      end subroutine getallnaybs
!----------------------------------------------------------------------
      subroutine nayn2(iboth,ntot_a,listpt_a,
     &              ntot_i,listpt_i,nneigh,listnay,distmas)
!*********************************************************************
!  You see a maze of little twisted treewalks, all different         *
!                                                                    *
!  Do an n**2/2 search for particle neighbors in listpt_a and        *
!  listpt_i. This is faster if the number of                         *
!  particles in a clump is smaller than some critical number.        *
!  The critical number is related to how many nodes the tree         *
!  traversal must examine in order to get the full list.             *
!  If it has to open every node for every particle, then it          *
!  will take n(2n-1) operations in the tree traversal alone.         *
!  This is much larger than the n**2/2 of the direct search.         *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************

      integer, intent(in):: iboth
      integer, intent(in):: ntot_a,listpt_a(maxclumpsize)
      integer, intent(in):: ntot_i,listpt_i(maxclumpsize)
      integer, intent(inout):: nneigh(maxclumpsize)
      integer, intent(inout):: listnay(maxbite,maxclumpsize)
      double precision, intent(inout):: 
     &               distmas(ndim3,maxbite,maxclumpsize)
      integer j,jsphpt,nb1,jtreept
      integer k,ksphpt,nb2,ktreept
      double precision dx,dy,dz,d2,h2

#include "treefuncs"

      do j=1,ntot_a
         jsphpt=listpt_a(j)
         nb1=0
         if(jsphpt.gt.natom_sph)nb1=1

         if(iboth.eq.1 .or. nb1.eq.0)then
            jtreept=linksphtotree(jsphpt)

            do k=j+1,ntot_a
               ksphpt=listpt_a(k)
               ktreept=linksphtotree(ksphpt)
               nb2=0
               if(ksphpt.gt.natom_sph)nb2=1
               if(mod(nb1+nb2,2).eq.0)then
                  dx=dxi(jtreept,ktreept)
                  dy=dyi(jtreept,ktreept)
                  d2=dx**2 + dy**2 + small
#if NDIM==3
                  dz=dzi(jtreept,ktreept)
                  d2=d2 + dz**2
#endif
                  h2=hsum2(jtreept,ktreept)
                  if(d2.lt.h2)then
                     nneigh (              j)=  nneigh(j)+1
                     listnay(    nneigh(j),j)=  ksphpt
                     distmas(ir2,nneigh(j),j)=  d2
                     distmas(ih2,nneigh(j),j)=  h2*fourth
                     distmas(idx,nneigh(j),j)=  dx
                     distmas(idy,nneigh(j),j)=  dy
                     distmas(ims,nneigh(j),j)=  treept(ims,ktreept)

                     nneigh (              k)=  nneigh(k)+1
                     listnay(    nneigh(k),k)=  jsphpt
                     distmas(ir2,nneigh(k),k)=  d2
                     distmas(ih2,nneigh(k),k)=  h2*fourth
                     distmas(idx,nneigh(k),k)= -dx
                     distmas(idy,nneigh(k),k)= -dy
                     distmas(ims,nneigh(k),k)=  treept(ims,jtreept)
#if NDIM==3
                     distmas(idz,nneigh(j),j)=  dz
                     distmas(idz,nneigh(k),k)= -dz
#endif
                  endif
               endif
            enddo

            do k=1,ntot_i
               ksphpt=listpt_i(k)
               ktreept=linksphtotree(ksphpt)
               nb2=0
               if(ksphpt.gt.natom_sph)nb2=1
               if(mod(nb1+nb2,2).eq.0)then
                  dx=dxi(jtreept,ktreept)
                  dy=dyi(jtreept,ktreept)
                  d2=dx**2 + dy**2 + small
#if NDIM==3
                  dz=dzi(jtreept,ktreept)
                  d2=d2 + dz**2
#endif
                  h2=hsum2(jtreept,ktreept)
                  if(d2.lt.h2)then
                     nneigh(j)               =  nneigh(j)+1
                     listnay(    nneigh(j),j)=  ksphpt
                     distmas(ir2,nneigh(j),j)=  d2
                     distmas(ih2,nneigh(j),j)=  h2*fourth
                     distmas(idx,nneigh(j),j)=  dx
                     distmas(idy,nneigh(j),j)=  dy
                     distmas(ims,nneigh(j),j)=  treept(ims,ktreept)
#if NDIM==3
                     distmas(idz,nneigh(j),j)=  dz
#endif
                  endif
               endif
            enddo
         endif
      enddo

      return
      end subroutine nayn2
!----------------------------------------------------------------------
      subroutine allnayn2(ntot_a,listpt_a,
     &              ntot_i,listpt_i,nneigh,listnay,distmas)
!*********************************************************************
!  You see a maze of little twisted treewalks, all different         *
!                                                                    *
!  Do an n**2/2 search for particle neighbors in listpt_a and        *
!  listpt_i. This is faster if the number of                         *
!  particles in a clump is smaller than some critical number.        *
!  The critical number is related to how many nodes the tree         *
!  traversal must examine in order to get the full list.             *
!  If it has to open every node for every particle, then it          *
!  will take n(2n-1) operations in the tree traversal alone.         *
!  This is much larger than the n**2/2 of the direct search.         *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************

      integer, intent(in):: ntot_a,listpt_a(maxclumpsize)
      integer, intent(in):: ntot_i,listpt_i(maxclumpsize)
      integer, intent(inout):: nneigh(maxclumpsize)
      integer, intent(inout):: listnay(maxbite,maxclumpsize)
      double precision, intent(inout)::
     &               distmas(ndim3,maxbite,maxclumpsize)
      integer j,jsphpt,jtreept
      integer k,ksphpt,ktreept
      double precision dx,dy,dz,d2,h2

#include "treefuncs"

      do j=1,ntot_a

         jsphpt=listpt_a(j)
         jtreept=linksphtotree(jsphpt)

         do k=j+1,ntot_a
            ksphpt=listpt_a(k)
            ktreept=linksphtotree(ksphpt)
            dx=dxi(jtreept,ktreept)
            dy=dyi(jtreept,ktreept)
            d2=dx**2 + dy**2 + small
#if NDIM==3
            dz=dzi(jtreept,ktreept)
            d2=d2 + dz**2
#endif
            h2=hsum2(jtreept,ktreept)

            if(d2.lt.h2)then
               nneigh (              j)=  nneigh(j)+1
               listnay(    nneigh(j),j)=  ksphpt
               distmas(ir2,nneigh(j),j)=  d2
               distmas(ih2,nneigh(j),j)=  h2*fourth
               distmas(idx,nneigh(j),j)=  dx
               distmas(idy,nneigh(j),j)=  dy
               distmas(ims,nneigh(j),j)=  treept(ims,ktreept)

               nneigh (              k)=  nneigh(k)+1
               listnay(    nneigh(k),k)=  jsphpt
               distmas(ir2,nneigh(k),k)=  d2
               distmas(ih2,nneigh(k),k)=  h2*fourth
               distmas(idx,nneigh(k),k)= -dx
               distmas(idy,nneigh(k),k)= -dy
               distmas(ims,nneigh(k),k)=  treept(ims,jtreept)
#if NDIM==3
               distmas(idz,nneigh(j),j)=  dz
               distmas(idz,nneigh(k),k)= -dz
#endif
            endif
         enddo

         do k=1,ntot_i
            ksphpt=listpt_i(k)
            ktreept=linksphtotree(ksphpt)
            dx=dxi(jtreept,ktreept)
            dy=dyi(jtreept,ktreept)
            d2=dx**2 + dy**2 + small
#if NDIM==3
            dz=dzi(jtreept,ktreept)
            d2=d2 + dz**2
#endif
            h2=hsum2(jtreept,ktreept)
            if(d2.lt.h2)then
               nneigh(j)               =  nneigh(j)+1
               listnay(    nneigh(j),j)=  ksphpt
               distmas(ir2,nneigh(j),j)=  d2
               distmas(ih2,nneigh(j),j)=  h2*fourth
               distmas(idx,nneigh(j),j)=  dx
               distmas(idy,nneigh(j),j)=  dy
               distmas(ims,nneigh(j),j)=  treept(ims,ktreept)
#if NDIM==3
               distmas(idz,nneigh(j),j)=  dz
#endif
            endif
         enddo
      enddo

      return
      end subroutine allnayn2
!----------------------------------------------------------------------
      subroutine getbothnaybs(itreept,ncl,neigclumps,nla,latoms)
!*********************************************************************
!  You see a little twisted maze of treewalks, not quite the same    *
!                                                                    *
!  Do a search for particle neighbors in node inode.                 *
!  This is set up only the gravity calculation, which means that it  *
!  calculates a true distance and doesn't account for periodic bnds. *
!  (See discussion in treefuncs)                                     *
!                                                                    *
!  Currently not used.                                               *  
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************

      integer, intent(in):: itreept, ncl, neigclumps(ncl)
      integer, intent(inout):: nla,latoms(idim)
      integer n,me,my_sib,my_dau,inode_sib

#include "treefuncs"

      do n=1,ncl

         me=neigclumps(n)
         inode_sib=link(isib,me)

         if(itreept.lt.me .or. itreept.ge.abs(inode_sib))then
!
! Don't worry about filtering out self (not in my own inode)
!
            do while(me.ne.inode_sib)
               my_sib=link(isib,me)
               my_dau=link(idau,me)
               if(hsum2(itreept,me).gt.rr2(itreept,me))then
                  if(my_dau.ne.0)then
                     me=my_dau
                  else
                     nla        = nla+1
                     latoms(nla)= me
                  endif
                  me=my_sib
               else
                  me=my_sib
               endif
            enddo

         else
!
! Must filter out self when found (in my own inode)
!
            do while(me.ne.inode_sib)
               my_sib=link(isib,me)
               my_dau=link(idau,me)
               if(hsum2(itreept,me).gt.rr2(itreept,me))then
                  if(my_dau.ne.0)then
                     me=my_dau
                  else
                     if(me.ne.itreept)then
                        nla        = nla+1
                        latoms(nla)= me
                     endif
                     me=my_sib
                  endif
               else
                  me=my_sib
               endif
            enddo

         endif

      enddo

      return
      end subroutine getbothnaybs
!----------------------------------------------------------------------
      subroutine getparticles(inode,n,ntot_a,ntot_i,listpt_a,listpt_i)
!*********************************************************************
!                                                                    *
!  list all the particles in inode, in SPH numbering,                *
!      separating out active and inactive                            *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************

      integer, intent(in) :: inode,n
      integer, intent(out):: ntot_a,listpt_a(n)
      integer, intent(out):: ntot_i,listpt_i(n)
      integer me,my_dau,inode_sib

      ntot_i=0
      ntot_a=0
      inode_sib=abs(link(isib,inode))
      me=inode
      do while(me.ne.inode_sib)
         my_dau=abs(link(idau,me))
         if(my_dau.eq.0)then
            if(active(me).eq.1)then
               ntot_a=ntot_a+1
               listpt_a(ntot_a)=linktreetosph(me)
            else
               ntot_i=ntot_i+1
               listpt_i(ntot_i)=linktreetosph(me)
            endif
            me=abs(link(isib,me))
         else
            me=my_dau
         endif
      enddo

      return
      end subroutine getparticles
!----------------------------------------------------------------------
      subroutine getallparticles(inode,n,ntot,nact,listpt)
!*********************************************************************
!                                                                    *
!  list all the particles in inode, separating out                   *
!      active and inactive                                           *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************

      integer, intent(in) :: inode,n
      integer, intent(out):: nact,ntot,listpt(n)
      integer me,my_dau,inode_sib

      ntot=0
      nact=0
      inode_sib=abs(link(isib,inode))
      me=inode
      do while(me.ne.inode_sib)
         my_dau=abs(link(idau,me))
         if(my_dau.eq.0)then
            ntot=ntot+1
            listpt(ntot)=me
            if(active(me).eq.1)nact=nact+1
            me=abs(link(isib,me))
         else
            me=my_dau
         endif
      enddo

      return
      end subroutine getallparticles
!----------------------------------------------------------------------
      end module tree
