      module treemake
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!****************************************************************
!  This file contains the VINE module to build a binary tree    *
!  structure from a list of particles, using the `Press tree'   *
!  formalism. It builds the tree with all relevant links, and   *
!  calculates multipole moments of each node in that tree       *
!  if gravity is turned on, via calls to the forcelaw module.   *
!                                                               *
!  WARNING: The exact tree that this module builds depends      *
!  on the number of processors used to build it and the         *
!  number of other things running on the system at the time.    *
!  This is normal and is due to the fact that tree nodes are    *
!  placed in the memory in segments that correspond to          *
!  nodes that were defined by one processor or another. The     *
!  order of the placement can vary, depending on which          *
!  processor gets done with its work first.                     *
!                                                               *
!  This has implications for tests of the accuracy of the       *
!  gravity calculation, because of the intermittent reordering  *
!  that we do on the particles (see subroutine orderparticles). *
!  You have to make sure that you are determining the error of  *
!  the force on the same particle that you use for your         *
!  reference. Use the ipindx array for this purpose.            *
!                                                               *
!  The four externally visible routines are:                    *
!    mtree: build the tree                                      *
!      Arguments:                                               *
!         igrape   : Use grape or not                           *
!         npart    : Total number of particles                  *
!         npart_sph: Number of SPH particles                    *
!         sepmax   : Maximum allowed separation of two `clumps' *
!         bsepmax  : Maximum allowed separation of two `bunches'*
!         boxlims  : Limits of the simulation box (if ibctype>0)*
!                    Put this quantity into internal-to-tree    *
!                    variables bndbox,boxsize,boxinv            *
!         posmas   : Particle position and mass                 *
!         h        : Particle smoothing length/softening length *
!         numclumps: Number of clumps                           *
!         lstclumps: Their locations in the tree                *
!                                                               *
!    rtree: revise the tree                                     *
!      Arguments:                                               *
!         imac     : Multipole acceptance criterion             *
!                    imac=1 Geometric                           *
!                    imac=2 Warren and Salmon 1994 (abs. err)   *
!                    imac=3 Gadget, octupole mac                *
!         isoft    : Type of gravitational softening            *
!         igrape   : Whether to use GRAPE coprocessors          *
!         ifullextrap: Whether to input all particle data, or   *
!                    only data for active particles             *
!         irevise  : Whether to allow partial tree revisions    *
!         npart    : Total number of particles                  *
!         nactive  : Total number of active particles           *
!         iactive  : List of active particles                   *
!         nactive_s: Total number of active SPH particles       *
!         iactive_s: List of active SPH particles               *
!         eps      : Plummer softening radius                   *
!         treeacc  : Accuracy parameter for geometric MAC       *
!                    Also used for the others, but in           *
!                    different ways, with different meanings.   *
!         boxlims  : Limits of the simulation box (if ibctype>0)*
!                    Put this quantity into internal-to-tree    *
!                    variables bndbox,boxsize,boxinv            *
!         posmas   : Particle position and mass                 *
!         h        : Particle smoothing length                  *
!         numclumps: Number of clumps                           *
!         lstclumps: Their locations in the tree                *
!         hclmp_is_exact: Whether the current values of the     *
!                         clump sizes are exact or overestimates*
!                                                               *
!    rebuildcheck: determine whether to rebuild or revise       *
!                  the tree at the next opportunity             *
!      Arguments:                                               *
!         hclmp_is_exact: Whether the current values of the     *
!                         clump sizes are exact or overestimates*
!         clhfrac  : Fraction of initial clump size that        *
!                    a clump can increase before triggering a   *
!                    rebuild                                    *
!         numclumps: Number of clumps                           *
!         lstclumps: Their locations in the tree                *
!         rebuild  : return logical value, T-rebuild,F-revise   *
!                                                               *
!    resethnode: Reset the values of hnode in the tree,         *
!                Used if non-varkern softening is used          *
!      Arguments:                                               *
!         irevise  : Whether to allow partial tree revisions    *
!         npart    : Total number of particles                  *
!         h        : Smoothing lengths of particles             *
!         numclumps: Number of clumps                           *
!         lstclumps: Their locations in the tree                *
!                                                               *
!  All other routines in this file are private to the           *
!  build itself, and should not be used by other parts          *
!  of VINE.                                                     *
!                                                               *
!  Written by Andy Nelson, Markus Wetzstein and Thorsten Naab   *
!                                                               *
!****************************************************************
      use dimens
      use treevars
      use constants
      use stubs
#if SELFFORCE
      use forcelaw, only:revpole
#endif

      implicit none

      private
      public mtree,rtree,rebuildcheck,resethnode

      contains
!---------------------------------------------------------------------
      subroutine mtree(isoft,igrape,npart,npart_sph,sepmax,bsepmax,eps,
     &                 boxlims,posmas,h,numclumps,lstclumps)
!**********************************************************************
!                                                                     *
!  Given the positions, masses and smoothing lengths of npart points, *
!  construct a nearest neighbor tree for use by gravity.              *
!                                                                     *
!  Fills treept, link, linksphtotree, linktreetosph and nroot,        *
!  the pointer to the root of the tree.                               *
!                                                                     *
!  Re-written and parallelized by A. Nelson based on code inherited   *
!  from W. Press, through W. Benz.                                    *
!                                                                     *
!**********************************************************************
      use perform

!$    integer  omp_get_max_threads
!$    external omp_get_max_threads

      integer, intent(in ):: isoft,igrape,npart,npart_sph
      integer, intent(out):: numclumps,lstclumps(npart)
      double precision, intent(in):: sepmax,bsepmax,eps
      double precision, intent(in):: posmas(ndim1,npart), h(npart)
      double precision, intent(in):: boxlims(2,ndim)

      integer newchild(idim)

      integer ivolume,nvolumes
      integer id,j,new,newold,new_mycpu,nremain,nremainold,nstart
      integer i

      double precision tme1,tme2
      character(len=20), parameter:: where =('mtree')

      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

#if PERIODIC
!
! Periodic boundaries
!
      do id=1,ndim
         bndbox(1,id)=boxlims(1,id)
         bndbox(2,id)=boxlims(2,id)
         boxsize (id)=boxlims(2,id)-boxlims(1,id)
         boxinv  (id)=one/(boxsize(id)*(one+epsilon(1d0)))
      enddo
#endif

      natom=npart         ; ntotnodes   =2*natom-1    
      natom_sph=npart_sph ; ntotsphnodes=2*natom_sph-1
      new=natom+1
      ntreelev=0
      nroot_sib=ntotnodes+1
      nay(nroot_sib)=nroot_sib  !reset later.

!$omp parallel default(none) private (j)
!$omp& shared(natom,natom_sph,nay,link,treept,posmas,h)
!$omp& shared(nroot_sib)
!$omp do
      do j=1,natom
         nay   (       j)=nroot_sib
         treept(id2nay,j)=huge(1d0)       !id2nay=ihnode  :reuse ihnode space 
         treept(ims   ,j)=posmas(ims,j)   !                to store useful 
         treept(1     ,j)=posmas(1  ,j)   !                build data
         treept(2     ,j)=posmas(2  ,j)
#if NDIM==3
         treept(3     ,j)=posmas(3  ,j)
#endif
      enddo
!$omp end parallel

!
! Initialize hash grid and linked lists of particles
! iwork,icount,work1,work2 are used as dummy arrays
!
      call makegrid(natom,iwork,icount,work1,work2,nvolumes)
      call initlists

!$omp parallel do default(none) private (j) shared(natom,icount,link)
      do j=1,natom
         icount(   j)=1   !count of number of particles in this node
         link(isib,j)=0   !link is previously used as iwork via equiv.
         link(idau,j)=0
      enddo
!
! Do not use work1/2 after this point, or iwork1/iwork 
! in this while loop, due to equivalances
!
      nremain=natom
      do while(new.le.ntotnodes)
         newold=new
         nremainold=nremain

!$omp parallel default(none) if(nvolumes.gt.1)
!$omp&  num_threads(min(nvolumes,omp_get_max_threads()))
!$omp&  private(ivolume,new_mycpu,nstart,newchild)
!$omp&  shared(nvolumes,new)
!$omp do schedule(guided,1)
         do ivolume=1,nvolumes
            call getnearest(ivolume)
         enddo

         new_mycpu=0

!$omp do schedule(guided,1)
         do ivolume=1,nvolumes
            call findnodes(ivolume,new_mycpu,newchild)
         enddo
!$omp enddo nowait

!$omp critical
         nstart=new                    !parallel treebuild is not
         new=new+new_mycpu             !deterministic due to races here
!$omp end critical

         call makenodes(new_mycpu,nstart,newchild)

!$omp end parallel

         nremain=nremain - (new - newold)
         ntreelev=ntreelev+1

         call revlists(newold,new,nvolumes)
         if(nremain.le.nhalf)then
            call revgrid(natom,newchild,nvolumes)  !newchild is
         endif                                     !workspace here

         if(idbg.ge.3)then
            print *,'Tree Level      :',ntreelev
            print *,'Nodes Associated:',new-1
            print *,'Nodes Remaining :',nremain
            print *,' '
         endif
         if(ntreelev+1.gt.maxtreelev .or. new.eq.newold)then
            call failbuild(ntreelev,maxtreelev,nremain,nremainold)
         endif
      enddo
!
! Now make a ordered list of the tree nodes, in order to get speed 
! optimized ordering for memory accesses: tree numbering is sorted 
! for improved cache and TLB speed performance
!
! Finish relinking (started in findnodes)
!
      call relinkrootsibs
!
! Make links between the new `tree' numbering and old sph numbering
!
      call sorttree                             !icount used here, discard on return
      call updatelinks(ntotnodes,iwork1,icount) !icount is workspace here 
!
! Get a set of sub-root nodes that define tree branches 
! that can be updated independently in parallel
!
      call getbranchroots
!
! Get an access speed optimized list of clumps, and reset
! the links to know about clumps, then save their initial
! positions and sizes, for tree rebuild testing
!
      call getclumps(igrape,natom,ntotnodes,sepmax,bsepmax,
     &               numclumps,lstclumps)
      call saveclumps(isoft,eps,npart,h,numclumps,lstclumps)
!
! Done with tree building
!
      if(idbg.ge.2)then
         if(igrape.eq.0)then
            print *,'done ',where,ntotnodes,ntreelev,
     &         dble(natom)/dble(numclumps)
         else
            print *,'done ',where,ntotnodes,ntreelev,
     &         dble(natom)/dble(numclumps),
     &         dble(natom)/dble(numbunches)
         endif
      endif

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_mtree)=timing(itme_mtree)+tme2-tme1
         ncalls(itme_mtree)=ncalls(itme_mtree)+1
         timing(itme_gravity)=timing(itme_gravity)+tme2-tme1
!         ncalls(itme_gravity) only increment in subr gravity
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine mtree
!---------------------------------------------------------------------
      subroutine rtree(igrav,imac,isoft,igrape,ifullextrap,irevise,
     &                 npart,nactive,iactive,nactive_s,iactive_s,
     &                 eps,treeacc,boxlims,posmas,h,numclumps,lstclumps,
     &                 hclmp_is_exact)
!*********************************************************************
!                                                                    *
!  This routine updates node positions, sizes (h values), multipole  *
!  moments and convergence radii, assuming that particle positions   *
!  h values have changed.                                            *
!                                                                    *
!  As an optional second step, it does exact calculations of node    *
!  sizes (h values) for nodes contained in clumps and bunches.       *
!  This can be efficient because it reduces the search times in      *
!  the gravity and SPH routines. It is NOT required, since an        *
!  approximate (too large) h calculations is done in revpole/revnodes*
!  We do this additional step conditioned on whether there are       *
!  enough particles active to make the cost worth it--if only a few  *
!  particles need force updates, then exact sizing of all clumps     *
!  can cost more than the update itself, even done less efficiently. *
!                                                                    *
!  There is nothing magic about the fraction of active particles     *
!  defined below. Other values might be better. The 1% and 10%       *
!  were set assuming that a full gravity calc is about 100x more     *
!  expensive than a full rtree calc, including the exact hnodes      *
!  stuff.                                                            *
!                                                                    *
!  Also, if we are using any softening except variable kernel        *
!  (id_varkern) then we will have to reset these h values at         *
!  the end of gravity subroutine, so don't we don't need to set      *
!  exact hnodes here, *except* for the clumps themselves, which      *
!  may still be useful to do.                                        *
!                                                                    *
!  Options for size of particles in the tree                         *
!    isoft.eq.id_varkern (variable kernel softening)                 *
!       htree=h                                                      *
!       Ensures finding all neighbors and correctly spline           *
!       softening them                                               *
!    isoft.eq.id_fixkern (constant kernel softening, overrides sph h)*
!       htree=eps for sph   particles                                *
!       htree=h   for nbody particles                                *
!       Ensures finding spline neighbors and                         *
!       Ensures multipole calc need not worry about softening.       *
!       Must reset to h after gravity, to get correct                *
!       neighbor lists for sph calculations                          *
!    isoft.eq.id_Plummer (constant Plummer softening)                *
!       htree=eps                                                    *
!       Don't care about spline softening gravity.                   *
!       Must reset to h after gravity, to get correct                *
!       neighbor lists                                               *
!                                                                    *
!  Written and parallelized by A. Nelson, based on code              *
!  inherited from W. Press, through W. Benz                          *
!                                                                    *
!*********************************************************************
      use perform

      logical, intent(out):: hclmp_is_exact
      integer, intent(in):: igrav,imac,isoft,igrape,irevise,ifullextrap
      integer, intent(in):: npart
      integer, intent(in):: nactive_s,iactive_s(nactive_s)
      integer, intent(in):: nactive  ,iactive  (nactive)
      integer, intent(in):: numclumps, lstclumps(numclumps)
      double precision, intent(in):: eps,treeacc,boxlims(2,ndim)
      double precision, intent(in):: posmas(ndim1,npart), h(npart)

      integer i
!
! Factors for calling/skipping gethnodes/gethclumps routines,
! which are not required for correctness, but are requried for
! maximum speed. When only a few particles are active, the
! speed gain in other parts of the code is offset by the cost of
! calling these routines, so skip them.
!
!  Itanic processors in parallel do horribly badly with these
!  routines. Factors of more than 2 performance degradation
!  in 2cpu parallel vs serial. Get back some on more processors,
!  but still do pretty badly. Therefore: If ITAN, then
!  skip them more readily than on other processors.
!
      double precision fracactive
#ifdef ITAN
      double precision, parameter:: critfrac1=2.5d-1
      double precision, parameter:: critfrac2=1d-1
#else
      double precision, parameter:: critfrac1=1d-1
      double precision, parameter:: critfrac2=1d-2
#endif

      double precision tme1,tme2
      character(len=20), parameter:: where = ('rtree')

      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

      if(natom.ne.npart)then
         print *,'Tree   thinks there are ',natom,' particles'
         print *,'Driver thinks there are ',npart,' particles'
         stop'Inconsistent input data in rtree'
      endif

      if(idbg.ge.1 .and. irevise.eq.id_partrevise)then
         print *,'You have selected to do partial tree revisions'
         print *,'but they are not currently implemented'
         print *,'Caveat Emptor'
      endif
!
! Copy data into tree arrays
!
      call copyin(isoft,ifullextrap,npart,nactive,iactive,
     &            nactive_s,iactive_s,boxlims,eps,posmas,h)
!
!  Revise the tree nodes data, including multipoles if needed
!
      if(igrav.eq.1)then
         call revpole(igrape,irevise,imac,treeacc)
      else
         call revnodes(irevise)
      endif
!
!  Get exact h values if it is efficient to do so, and
!  report if we did so to the caller.
!
      hclmp_is_exact=.false.
      fracactive=dble(nactive)/dble(npart)
      if(fracactive.gt.critfrac1)then
         hclmp_is_exact=.true.

         if(nactive_s.gt.0 .and. isoft.eq.id_varkern)then
            call gethnodes(irevise,numclumps,lstclumps)
         else
            call gethclumps(irevise,numclumps,lstclumps)
         endif
         if(igrape.eq.1)call gethbunches

      elseif(fracactive.gt.critfrac2)then
         hclmp_is_exact=.true.

         if(nactive_s.gt.0 .and. isoft.eq.id_varkern)then
            call gethclumps(irevise,numclumps,lstclumps)
            if(igrape.eq.1)call gethbunches
         endif

      endif

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_rtree)=timing(itme_rtree)+tme2-tme1
         ncalls(itme_rtree)=ncalls(itme_rtree)+1
         timing(itme_gravity)=timing(itme_gravity)+tme2-tme1
!         ncalls(itme_gravity) only increment in subr gravity
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine rtree
!---------------------------------------------------------------------
      subroutine rebuildcheck(hclmp_is_exact,clhfrac,
     &                        numclumps,lstclumps,rebuild)
!*********************************************************************
! Check whether or not to rebuild the tree, or just revise it.       *
!                                                                    *
! Currently the definition of rebuild or not means to check          *
! the size of the clump. If it increases by some percentage,         *
! rebuild.                                                           *
!                                                                    *
! Put another criterion here if you want.                            *
! One example might be to trigger on how far clumps have moved       *
! since the last treebuild. If they have moved a long way            *
! (user settable), then rebuild.                                     *
!                                                                    *
! Note that this routine is currently normally called from a point   *
! in the evolution when the tree as it is referenced here is         *
! slightly out of date. In other words, the tree values here         *
! reflect conditions at the last deriv calculations, while the       *
! call to this routine happens at the end of a step. Not a big       *
! deal really, since to build or not to build is a user tuneable     *
! thing anyway.                                                      *
!                                                                    *
! The input parameter clhfrac is the fraction of the initial         *
! size of the clump, that a clump can increase or decrease before    *
! it triggers a rebuild                                              *
!                                                                    *
! Because the rtree routine can return either exact or               *
! approximate (overestimated) calculations of the clump size         *
! we allow clumps to be an extra 30% larger if we only have          *
! approximate values. This value may need additional tuning.         *
!                                                                    *
! Subroutine by M. Wetzstein and A. Nelson                           *
!*********************************************************************
      logical, intent(in):: hclmp_is_exact
      integer, intent(in):: numclumps, lstclumps(numclumps)
      double precision, intent(in)::clhfrac
      logical, intent(out):: rebuild

      integer jclump,iclump
      double precision clhfrac_eff,clhfracinv
      character(len=20), parameter:: where = ('rebuildcheck')

      if(idbg.ge.3)print *,where

      rebuild=.false.

      clhfrac_eff=clhfrac
      if(.not.hclmp_is_exact)clhfrac_eff=1.3d0*clhfrac

      clhfracinv=1d0/clhfrac_eff

      do jclump=1,numclumps
         iclump=lstclumps(jclump)
         if(treept(ihnode,iclump).ge.clhfrac_eff*clmpsize0(jclump))then
            rebuild=.true.
         endif
         if(treept(ihnode,iclump).le.clhfracinv*clmpsize0(jclump))then
            rebuild=.true.
         endif
      enddo

      return
      end subroutine rebuildcheck
!---------------------------------------------------------------------
      subroutine resethnode(irevise,npart,h,numclumps,lstclumps)
!*********************************************************************
!                                                                    *
! Reset smoothing lengths to non-zero values in prep for neighbor    *
! finding in case where fixed softening (fixed spline or Plummer)    *
! is used. Ordinarily only called from subr. gravity                 * 
!                                                                    *
! See the comment header of revpole/revnodes for a description of    *
! n1, n2 and how the update works.                                   *
!                                                                    *
! Subroutine by A. Nelson                                            *
!                                                                    *
!*********************************************************************
      use perform

      integer, intent(in):: irevise
      integer, intent(in):: npart,numclumps,lstclumps(numclumps)
      double precision, intent(in):: h(npart)

      logical fixup

      integer icpu,ntopbranch,nbotbranch
      integer jbranch,n1,n2,i,j,me,my_dau,m1,m2,itreept
      double precision emratm1,emratm2,rmag

      double precision tme1,tme2
      character(len=20), parameter:: where = ('resethnode')

#include "treefuncs"

      if(idbg.ge.2)print *,where
      if(itme.ge.1)call cpu_time(tme1)

!$omp parallel default(none) private(j,itreept)
!$omp& shared(natom_sph,natom,linksphtotree,h,treept)
!$omp do
      do j=1,natom_sph
         itreept =linksphtotree(j)
         treept(ihnode,itreept)= h(j)
      enddo
!$omp do 
      do j=natom_sph+1,natom
         itreept =linksphtotree(j)
         treept(ihnode,itreept)= zero
      enddo
!$omp end parallel

      fixup=.false.

      do i=1,2

!$omp parallel if(.not.fixup) default(none) shared(fixup,nroot)
!$omp& shared(numbranches,lstbranches,link,treept,ntotnodes)
!$omp& shared(nbrnchstart,numcpus_sav)
!$omp& private(icpu,ntopbranch,nbotbranch)
!$omp& private(jbranch,n1,n2,rmag,emratm1,emratm2,m1,m2,me,my_dau)
!$omp do 
         do icpu=numcpus_sav,1,-1
            ntopbranch=nbrnchstart(2,icpu)
            nbotbranch=nbrnchstart(1,icpu)
            do jbranch=ntopbranch,nbotbranch,-1

               call setlooplims(fixup,jbranch,n1,n2)

               do me=n1,n2,-1
                  if(link(idau,me).eq.0)cycle
                  m1=abs(link(idau,me))
                  m2=abs(link(isib,m1))
                  emratm1=treept(ims,m1)/treept(ims,me)
                  emratm2=one - emratm1
                  rmag=sqrt(rr2(m1,m2))
                  treept(ihnode,me)=max(emratm2*rmag+treept(ihnode,m1),
     &                                  emratm1*rmag+treept(ihnode,m2))
               enddo
            enddo
         enddo
!$omp end parallel

      fixup=.true.

      enddo
!
!--get hnode final definition
!
      call gethnodes(irevise,numclumps,lstclumps)

      if(itme.ge.1)then
         call cpu_time(tme2)
         timing(itme_rtree)=timing(itme_rtree)+tme2-tme1
         ncalls(itme_rtree)=ncalls(itme_rtree)+1
         timing(itme_gravity)=timing(itme_gravity)+tme2-tme1
!         ncalls(itme_gravity) only increment in subr gravity
         if(itme.ge.2)print *,where,xcpus_inv*(tme2-tme1)
      endif

      return
      end subroutine resethnode
!---------------------------------------------------------------------
      subroutine copyin(isoft,ifullextrap,npart,nactive,iactive,
     &                  nactive_s,iactive_s,boxlims,eps,posmas,h)
!*********************************************************************
!                                                                    *
!  This routine copies all of the driver data into the tree arrays   *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: isoft,ifullextrap,npart
      integer, intent(in):: nactive  , iactive(nactive)
      integer, intent(in):: nactive_s, iactive_s(nactive_s)
      double precision, intent(in):: eps,boxlims(2,ndim)
      double precision, intent(in):: posmas(ndim1,npart), h(npart)

      integer id,j,k,itreept

#if PERIODIC
      do id=1,ndim
         bndbox(1,id)=boxlims(1,id)
         bndbox(2,id)=boxlims(2,id)
         boxsize (id)=boxlims(2,id)-boxlims(1,id)
         boxinv  (id)=1d0/(boxsize(id)*(one+epsilon(1d0)))
      enddo
#endif
!
! Yes, duplicating all this code below multiple
! times does make a big speed difference.
!
      select case(ifullextrap)
      case(1)

!$omp parallel default(none)
!$omp& private(j,itreept) 
!$omp& shared(eps,posmas,h,isoft)
!$omp& shared(natom,natom_sph,treept,linksphtotree)

         select case(isoft)
      
         case(id_varkern)                    !variable kernel softening
!$omp do 
            do j=1,natom
               itreept =linksphtotree(j)
               treept(ihnode ,itreept)= h(j)
               treept(iconvrg,itreept)= treept(ihnode,itreept)
               treept(ims    ,itreept)= max(posmas(ims,j),tiny(1d0))
               treept(1      ,itreept)= posmas(1  ,j)
               treept(2      ,itreept)= posmas(2  ,j)
#if NDIM==3
               treept(3      ,itreept)= posmas(3  ,j)
#endif
            enddo

         case(id_fixkern)                    !fixed kernel softening
!$omp do
            do j=1,natom_sph
               itreept =linksphtotree(j)
               treept(ihnode ,itreept)= eps
               treept(iconvrg,itreept)= treept(ihnode,itreept)
               treept(ims    ,itreept)= max(posmas(ims,j),tiny(1d0))
               treept(1      ,itreept)= posmas(1  ,j)
               treept(2      ,itreept)= posmas(2  ,j)
#if NDIM==3
               treept(3      ,itreept)= posmas(3  ,j)
#endif
            enddo
!$omp end do nowait
!$omp do
            do j=natom_sph+1,natom
               itreept =linksphtotree(j)
               treept(ihnode ,itreept)= h(j)
               treept(iconvrg,itreept)= treept(ihnode,itreept)
               treept(ims    ,itreept)= max(posmas(ims,j),tiny(1d0))
               treept(1      ,itreept)= posmas(1  ,j)
               treept(2      ,itreept)= posmas(2  ,j)
#if NDIM==3
               treept(3      ,itreept)= posmas(3  ,j)
#endif
            enddo

         case(id_Plummer)                    !Plummer softening
!$omp do
            do j=1,natom
               itreept =linksphtotree(j)
               treept(ihnode ,itreept)= eps
               treept(iconvrg,itreept)= treept(ihnode,itreept)
               treept(ims    ,itreept)= max(posmas(ims,j),tiny(1d0))
               treept(1      ,itreept)= posmas(1  ,j)
               treept(2      ,itreept)= posmas(2  ,j)
#if NDIM==3
               treept(3      ,itreept)= posmas(3  ,j)
#endif
            enddo
         case default
            stop'rtree:unknown softening'
         end select
!$omp end parallel


      case(0)

!$omp parallel default(none)
!$omp& private(j,k,itreept) 
!$omp& shared(eps,posmas,h,isoft,nactive,iactive,nactive_s,iactive_s)
!$omp& shared(natom,natom_sph,treept,linksphtotree)

         select case(isoft)
         case(id_varkern)                    !variable kernel softening
!$omp do 
            do k=1,nactive
               j=iactive(k)
               itreept =linksphtotree(j)
               treept(ihnode ,itreept)= h(j)
               treept(iconvrg,itreept)= treept(ihnode,itreept)
               treept(ims    ,itreept)= max(posmas(ims,j),tiny(1d0))
               treept(1      ,itreept)= posmas(1  ,j)
               treept(2      ,itreept)= posmas(2  ,j)
#if NDIM==3
               treept(3      ,itreept)= posmas(3  ,j)
#endif
            enddo

         case(id_fixkern)                    !fixed kernel softening
!$omp do
            do k=1,nactive_s
               j=iactive_s(k)
               itreept =linksphtotree(j)
               treept(iconvrg,itreept)= treept(ihnode,itreept)
               treept(ims    ,itreept)= max(posmas(ims,j),tiny(1d0))
               treept(1      ,itreept)= posmas(1  ,j)
               treept(2      ,itreept)= posmas(2  ,j)
#if NDIM==3
               treept(3      ,itreept)= posmas(3  ,j)
#endif
            enddo
!$omp end do nowait
!$omp do
            do j=1,natom_sph
               itreept =linksphtotree(j)
               treept(ihnode ,itreept)= eps
            enddo
!$omp end do nowait
!$omp do
            do k=nactive_s,nactive
               j=iactive(k)
               itreept =linksphtotree(j)
               treept(ihnode ,itreept)= h(j)
               treept(iconvrg,itreept)= treept(ihnode,itreept)
               treept(ims    ,itreept)= max(posmas(ims,j),tiny(1d0))
               treept(1      ,itreept)= posmas(1  ,j)
               treept(2      ,itreept)= posmas(2  ,j)
#if NDIM==3
               treept(3      ,itreept)= posmas(3  ,j)
#endif
            enddo

         case(id_Plummer)                    !Plummer softening
!$omp do
            do k=1,nactive
               j=iactive(k)
               itreept =linksphtotree(j)
               treept(iconvrg,itreept)= treept(ihnode,itreept)
               treept(ims    ,itreept)= max(posmas(ims,j),tiny(1d0))
               treept(1      ,itreept)= posmas(1  ,j)
               treept(2      ,itreept)= posmas(2  ,j)
#if NDIM==3
               treept(3      ,itreept)= posmas(3  ,j)
#endif
            enddo
!$omp end do nowait
!$omp do
            do j=1,natom
               itreept =linksphtotree(j)
               treept(ihnode ,itreept)= eps
            enddo

         case default
            stop'copyin:unknown softening'
         end select
!$omp end parallel

      case default
         stop'Unkown extrapolation in copyin'
      end select


      return
      end subroutine copyin
!---------------------------------------------------------------------
      subroutine getallsubnodes(inode,is,isubnode,na,list,lim)
!*********************************************************************
!                                                                    *
! Find all of the subnodes in the current node                       *
! ie all of the nodes that are not particles                         *
! and a list of all atoms in the current node.                       *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************

      integer, intent(in ):: inode 
      integer, intent(out):: is
      integer, intent(out):: na,list(maxclumpsize)
      integer, intent(out):: isubnode(maxclumpsize-1)
      integer, intent(out):: lim(2,maxclumpsize-1)
      integer me,inode_sib,my_dau

      is=0
      na=1
      inode_sib=abs(link(isib,inode))

      if(inode_sib-inode.le.3)return

      do me=inode,inode_sib-1

         my_dau=link(idau,me)
         if(my_dau.ne.0)then
            is=is+1
            isubnode(is)=me
            lim(1,is)=na
            lim(2,is)=na + (abs(link(isib,me))-me)/2
         else
            list(na)=me
            na=na+1
         endif
      enddo

      return
      end subroutine getallsubnodes
!---------------------------------------------------------------------
      subroutine findnodes(ivolume,new_mycpu,newchild)
!*********************************************************************
!                                                                    *
! Put all mutual nearest neighbors on this level into nodes,         *
! storing them for further definition in makenodes.                  *
!                                                                    *
! Secondly, do a little post-processing here, since doing it now     *
! in a piecemeal fashion means that it can be done in parallel.      *
! Rather than afterwards, where it must be done serially.            *
! Specifically: update the siblings of all of the extreme right      *
! daughter nodes of the newly created left child node, l. To         *
! get the ordered list we want, they all need to point at ll,        *
! which is l's sibling. For more information, see                    *
!     Makino 1990 JCP 87, 148                                        *
!     Dubinski 1996 New Astronomy 1, 133                             *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in   ):: ivolume
      integer, intent(inout):: newchild(idim/2)
      integer, intent(inout):: new_mycpu

      integer my_dau,my_dausib
      integer l,ll,ncel
      integer ixvol,ixlo,ixhi,ixgrid
      integer jyvol,jylo,jyhi,jygrid
      integer kzvol,kzlo,kzhi,kzgrid
      integer kplane,jkcoord

      ixvol=1+mod( ivolume-1,nxvol)
      jyvol=1+mod((ivolume-1)/nxvol,nyvol)
      kzvol=1+(ivolume-1)/(nxvol*nyvol)
      ixlo=1+(ixvol-1)*nznvol ; ixhi= min(nznx,ixvol*nznvol)
      jylo=1+(jyvol-1)*nznvol ; jyhi= min(nzny,jyvol*nznvol)
      kzlo=1+(kzvol-1)*nznvol ; kzhi= min(nznz,kzvol*nznvol)

      do kzgrid=kzlo,kzhi
         kplane=(kzgrid-1)*nplane
      do jygrid=jylo,jyhi
         jkcoord=(jygrid-1)*nznx + kplane
      do ixgrid=ixlo,ixhi

         ncel= ixgrid + jkcoord

         l=ihead(1,ncel)
         do while(l.ne.0)
            if(link(isib,l).eq.0)then
               ll=nay(l)
               if(nay(ll).eq.l .and. l.lt.ll)then
                  link(isib,ll)=l                 !set siblings
                  link(isib,l)=ll

                  new_mycpu=new_mycpu+1           !save identity 
                  newchild(new_mycpu)=l           !for parent defn

                  my_dau=link(idau,l)             !update children's
                  do while(my_dau.ne.0)           !siblings
                     my_dausib=link(isib,my_dau)
                     link(isib,my_dausib)=ll

                     my_dau=link(idau,my_dausib)
                  enddo
               endif
            endif
            l=next(l)
         enddo
      enddo
      enddo
      enddo

      return
      end subroutine findnodes
!---------------------------------------------------------------------
      subroutine getclumps(igrape,natom,ntotnodes,sepmax,bsepmax,
     &                     numclumps,lstclumps)
!*********************************************************************
! Get a list of clumps of particles and order them for speed.        *
! Treewalks are then done for clumps of particles rather than        *
! for individual particles, to save duplicate work: close particles  *
! in the tree have similar neighbor lists etc, so require similar    *
! treewalks. Also the cache/TLB hit rate improves dramatically.      *
! This works quite well!                                             *
!                                                                    *
! Note about sepmax: sepmax is the maximum center of mass            *
!  distance of two child nodes allowed when their parent is          *
!  to be defined as a clump. If the distance is greater than this,   *
!  the two child nodes (or their children etc) are defined as clumps,*
!  overriding considerations about number of particles in the clump. *
!  This condition is useful to avoid clumps made up of particles     *
!  in low density regions of the simulation, far from the `action'   *
!  from getting huge numbers of neighbor clumps and making the       *
!  whole simulation slow down.                                       *
!                                                                    *
! Note about the number of particles in a node calculation:          *
!  We use the fact that the tree is now ordered and the fact that    *
!  any given node contains exactly 2n-1 nodes and subnodes in        *
!  in it, where n is the number of atoms in that node, to do the     *
!  calculation of the number of atoms in the node very quickly.      *
!  The number of subnodes in a given node will be the difference     *
!  between its index value in the list and its sibling's index       *
!  value: (my_sib-me)=2n-1.                                          *
!                                                                    *
! For the tree traversal stuff (specifically closewalk and           *
! getneigh), we need to know whether we are examining a clump        *
! or not in order to give them special treatment. Therefore, give    *
! negative indices to pointers to clumps (idau or isib), and         *
! check the sign.                                                    *
!                                                                    *
! If grape is turned on then get a similar list, but on a            *
! larger scale. The grape likes larger lists of particles when it    *
! calculates gravity. Don't need to negate the indices like we did   *
! for the clumps though. It is sufficient just to have a list,       *
! for the tree walk.                                                 *
!                                                                    *
! Note that treept has not been updated to the tree ordering yet,    *
! while the indices refering to it have been updated. That means     *
! that id2nay references need to switch to the sph numbering.        *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: igrape,natom,ntotnodes
      integer, intent(out):: numclumps,lstclumps(natom)
      double precision, intent(in):: sepmax,bsepmax

      integer i,me,mesph,iroot_sib,my_sib,my_sibsph,num,iclump
      double precision sepmax2,bsepmax2
      integer my_lev,mybranch(maxtreelev),lastnode,my_dau,my_dausib
      integer notused
      character(len=20), parameter:: where = ('getclumps')

#include "treefuncs"

      if(idbg.ge.3)print *,where

      sepmax2 =sepmax**2

! This should be paralyzable using the same list of branches 
! that is used in rtree. It takes so little cpu time though,
! that its a pretty low priority. Someday... 

      me=nroot
      mesph =linktreetosph(me)
      iroot_sib=link(isib,me)

      treept(id2nay,                  mesph )=two*sepmax2
      treept(id2nay,linktreetosph(iroot_sib))=two*sepmax2

      numclumps=0
      do while(me.ne.iroot_sib)
         mesph =linktreetosph(me)
         if(treept(id2nay,mesph).gt.sepmax2)then
            if(link(idau,me).ne.0)then
               me       =link(idau,me)
            else
               numclumps=numclumps+1
               lstclumps(numclumps)=me
               me=link(isib,me)
            endif
         else
            my_sib=link(isib,me)
            num= (my_sib-me)/2 + 1
            if(num.gt.maxclumpsize)then
               me       =link(idau,me)
            else
               numclumps=numclumps+1
               lstclumps(numclumps)=me
               me=link(isib,me)
            endif
         endif
      enddo
!
! Do same thing to get grape bunches
!
      if(igrape.eq.1)then

         if(maxclumpsize.gt.maxbunchsize)then
            print *,'Clumps cannot be bigger than bunches when grape'
            print *,'is used. The code will crash.'
            print *,'maxclumpsize=',maxclumpsize
            print *,'maxbunchsize=',maxbunchsize
            stop'Recompile with maxclumpsize.le.maxbunchsize in dimens'
         endif

         bsepmax2=bsepmax**2

         me=nroot
         iroot_sib=link(isib,me)
         numbunches=0
         do while(me.ne.iroot_sib)
            mesph =linktreetosph(me)
            if(treept(id2nay,mesph).gt.bsepmax2)then
               if(link(idau,me).ne.0)then
                  me       =link(idau,me)
               else
                  numbunches=numbunches+1
                  lstbunches(numbunches)=me
                  me=link(isib,me)
               endif
            else
               my_sib=link(isib,me)
               num= (my_sib-me)/2 + 1
               if(num.gt.maxbunchsize)then
                  me       =link(idau,me)
               else
                  numbunches=numbunches+1
                  lstbunches(numbunches)=me
                  me=link(isib,me)
               endif
            endif
         enddo

      endif
!
! Set up negative indices for pointers to clumps
!
!$omp parallel default(none)
!$omp& private(my_lev,mybranch,notused)
!$omp& private(i,iclump,me,lastnode,my_dau,my_dausib)
!$omp& shared(nroot,numclumps,lstclumps,link)
       my_lev=1
       mybranch(my_lev)=nroot

!$omp do
      do i=1,numclumps
         iclump=lstclumps(i)
         lastnode=iclump-1

         if(link(idau,lastnode).eq.iclump)then        !left child: only
                                                      !adjust parent's child

            link(idau,lastnode)=-link(idau,lastnode)
         else                                         !right child: adjust
                                                      !all right siblings
                                                      !pointing at me

            call findbranch(iclump,my_lev,notused,mybranch)

            me=abs(link(idau,mybranch(my_lev-1)))
            my_dau=abs(link(idau,me))

            link(isib,me)=-link(isib,me)

            do while(my_dau.ne.0)
               my_dausib=abs(link(isib,my_dau))

               link(isib,my_dausib)=-link(isib,my_dausib)

               me=my_dausib
               my_dau=abs(link(idau,me))
            enddo

         endif
      enddo
!$omp end parallel

      return
      end subroutine getclumps
!----------------------------------------------------------------------
      subroutine findbranch(itreept,my_lev,levstart,mybranch)
!*********************************************************************
!  Get all of the parent nodes of itreept and the level (levstart)   *
!   where the branch is different from the previous branch we        *
!   worked on.                                                       *
!                                                                    *
!  This is duplicated code from the tree traversal file tree.F       *
!  Duplicated to preserve the independence of code in                *
!  the two files treemake.F and tree.F.                              *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************
      integer, intent(in):: itreept
      integer, intent(inout):: my_lev,levstart,mybranch(maxtreelev)

      integer m,my_dau,my_dausib

      if(my_lev.gt.1)then

         do while(abs(          mybranch(my_lev)) .gt.itreept .or.
     &            abs(link(isib,mybranch(my_lev))).le.itreept)
            my_lev=my_lev-1
         enddo
      else
         mybranch(my_lev)=nroot
      endif

      levstart=my_lev

      do while(itreept.gt.mybranch(my_lev))
         my_dau=abs(link(idau,mybranch(my_lev)))
         my_dausib=abs(link(isib,my_dau))

         my_lev=my_lev+1
         if(itreept.ge.my_dausib)then
            mybranch(my_lev)=my_dausib
         else
            mybranch(my_lev)=my_dau
         endif
      enddo

      do m=my_lev,ntreelev+1
         mybranch(m)=mybranch(my_lev)
      enddo

      return
      end subroutine findbranch
!---------------------------------------------------------------------
      subroutine saveclumps(isoft,eps,npart,h,numclumps,lstclumps)
!*********************************************************************
!                                                                    *
! Save initial positions and size of clumps, for use in              *
! determining whether or not to rebuild the tree.                    *
!                                                                    *
! Recall that treept has not yet been translated to                  *
! the ordered tree numbering system, so we have to use               *
! linktreetosph to find where the clumps and their                   *
! atoms are in the array.                                            *
!                                                                    *
! This version defines sph particles to have a fixed size given      *
! by the value of gravitational softening length, eps, if any        *
! softening other than variable kernel softening is used.            *
! This definition is inconsistent the way that sph particles are     *
! defined in the sense that their interactions really should         *
! extent all the way out to the edge of their kernel. It shouldn't   *
! affect the rebuild frequency all that terribly much because        *
! ordinarily we would expect to have eps and h not that much         *
! different anyway.                                                  *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************
      integer, intent(in):: isoft
      integer, intent(in):: npart
      integer, intent(in):: numclumps, lstclumps(numclumps)
      double precision, intent(in):: eps,h(npart)

      integer i,jclump,iclump,indx_clmp,icl_sib,me,my_dau
      double precision rmax,rad, epsuse
      character(len=20), parameter:: where = ('saveclumps')

#include "treefuncs"

      epsuse=eps    !sgi doesn't like intent in firstprivate

!$omp parallel do default(none)
!$omp& firstprivate(epsuse,natom_sph)
!$omp& shared(isoft,numclumps,lstclumps,linktreetosph)
!$omp& shared(treept,clmpsize0,link,h)
!$omp& private(i,jclump,iclump,indx_clmp,icl_sib,me,my_dau,rmax,rad)
      do jclump=1,numclumps

         iclump=lstclumps(jclump)
         indx_clmp=linktreetosph(iclump)
         icl_sib=abs(link(isib,iclump))

         me=iclump
         rmax=ZERO

         if(isoft.ne.id_varkern)then

            do while(me.ne.icl_sib)
               my_dau=abs(link(idau,me))
               if(my_dau.eq.0)then
                  i=linktreetosph(me)
                  if(i.le.natom_sph)then
                     rad=sqrt(rr2(indx_clmp,i)) + epsuse
                  else
                     rad=sqrt(rr2(indx_clmp,i)) + h(i)
                  endif
                  rmax=max(rmax,rad)

                  me=abs(link(isib,me)) 
               else
                  me=my_dau
               endif
            enddo

         else

            do while(me.ne.icl_sib)
               my_dau=abs(link(idau,me))
               if(my_dau.eq.0)then
                  i=linktreetosph(me)
                  rad=sqrt(rr2(indx_clmp,i)) + h(i)
                  rmax=max(rmax,rad)

                  me=abs(link(isib,me)) 
               else
                  me=my_dau
               endif
            enddo
         endif

         clmpsize0(jclump)=rmax
      enddo

      return
      end subroutine saveclumps
!---------------------------------------------------------------------
      subroutine revnodes(irevise)
!*********************************************************************
!                                                                    *
! Just like revpole but does not update node multipole moments.      *
! Useful when we are doing non self gravitating simulations and      *
! multipoles are not required.                                       *
!                                                                    *
! treept(ihnode ,i)= physical size of the node. At the end of rtree  *
!             (call gethnode) it is redefined for clumps and         *
!             their children, if the clumps are small enough         *
!             to make this redefinition efficient [[need to specify]]*
!                                                                    *
! Use a backwards (end to root) traversal to update the tree.        *
! In other words, start at ntotnodes and proceed to nroot,           *
! hitting all children before their parents.                         *
!                                                                    *
! n1 and n2 are the last and first nodes to be examined in an        *
! independent branch. That means that n1 is the last atom in         *
! the branch and n2 is the branch root.  In the fixup stage,         *
! the roles are reversed. We are looking at orphan nodes in          *
! between independent branches. That means that n1 is the node       *
! immediately preceeding the branch root, and n2 is either the       *
! last atom in the previous branch, or the root of the entire        *
! tree.                                                              *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************

      integer, intent(in):: irevise

      logical fixup
      integer icpu,nbotbranch,ntopbranch
      integer n1,n2,i
      integer imacabs,j,jbranch,me,my_dau,m1,m2
      double precision emratm1,emratm2,emred,difx,dify,difz,rmag
      character(len=20), parameter:: where = ('revnodes')

      if(idbg.ge.3)print *,where

      fixup=.false.

      do i=1,2

!$omp parallel if(.not.fixup) default(none)
!$omp& shared(ntotnodes)
!$omp& shared(fixup,nroot,numbranches,lstbranches,link,treept)
!$omp& shared(nbrnchstart,numcpus_sav)
!$omp& private(icpu,ntopbranch,nbotbranch)
!$omp& private(jbranch,n1,n2,me,my_dau,m1,m2)
!$omp& private(emratm1,emratm2,emred,rmag,difz,dify,difx)
!$omp do 
      do icpu=numcpus_sav,1,-1
         ntopbranch=nbrnchstart(2,icpu)
         nbotbranch=nbrnchstart(1,icpu)
         do jbranch=ntopbranch,nbotbranch,-1

            call setlooplims(fixup,jbranch,n1,n2)

            do me=n1,n2,-1
               if(link(idau,me).eq.0)cycle

               m1=abs(link(idau,me))
               m2=abs(link(isib,m1))
               treept(ims,me)=treept(ims,m1) + treept(ims,m2)
               emratm1      =treept(ims,m1)/treept(ims,me)
               emratm2      =one - emratm1
               emred        =emratm1*emratm2*treept(ims,me)
               difx         =treept(1,m2) - treept(1,m1)
               dify         =treept(2,m2) - treept(2,m1)
               treept(1,me) =treept(1,m1) + emratm2*difx
               treept(2,me) =treept(2,m1) + emratm2*dify
#if NDIM==2
               rmag         =sqrt(difx*difx + dify*dify)
#else
               difz         =treept(3,m2) - treept(3,m1)
               treept(3,me) =treept(3,m1) + emratm2*difz
               rmag         =sqrt(difx*difx + dify*dify + difz*difz)
#endif
               treept(ihnode,me)=max(emratm2*rmag+treept(ihnode,m1),
     &                               emratm1*rmag+treept(ihnode,m2))
            enddo   !nodes in branch

         enddo      !branches
      enddo         !icpu
!$omp end parallel

      fixup=.true.

      enddo

      return
      end subroutine revnodes
!---------------------------------------------------------------------
      subroutine setlooplims(fixup,jbranch,n1,n2)

      logical, intent(in ):: fixup
      integer, intent(in ):: jbranch
      integer, intent(out):: n1,n2

      integer me,my_dau


      if(.not.fixup)then
         if(jbranch.eq.numbranches)then
            n1=ntotnodes
         else
            me=lstbranches(jbranch+1)
            my_dau=abs(link(idau,me))
            do while(my_dau.ne.0)
               me=me-1
               my_dau=abs(link(idau,me))
            enddo
            n1=me
         endif
         n2=lstbranches(jbranch)
      else
         n1=lstbranches(jbranch)-1
         if(jbranch.eq.1)then
            n2=nroot
         else
            me=n1
            my_dau=abs(link(idau,me))
            do while(my_dau.ne.0)
               me=me-1
               my_dau=abs(link(idau,me))
            enddo
            n2=me
         endif
      endif

      end subroutine setlooplims
!---------------------------------------------------------------------
      subroutine makegrid(n,list,iwork,work1,work2,nvolumes)
!***************************************************************
!                                                              *
!  Make a hash grid to roughly locate nodes, then map          *
!  these grid cells into one dimension and give each particle  *
!  a number defining its cell in this 1d coordinate system     *
!                                                              *
!   key(np) is the 1d coordinate of node np in the hash grid.  *
!   It gets the value:                                         *
!   2D: key(np) = izn + (jzn-1)*nznx                           *
!   3D: key(np) = izn + (jzn-1)*nznx + (kzn-1)*nplane          *
!                                                              *
!   the hash grid grid(i,1|2|3) contains the x,y or z position *
!   of the next grid boundary larger than the node refered to  *
!   by key                                                     *
!                                                              *
!  The izn/jzn/kzn loops in this subroutine are parallelizable,*
!  but experiments on an o3k and an o2k show that the speedup  *
!  is horrible...the loops slow down by a factor 10 or so.     *
!  That means we just leave them as serial loops. Will revisit *
!  this some day                                               *
!                                                              *
!  Note that the various arrays here are dimensioned ndim*idim *
!  In fact, current code only uses the first idim section.     *
!  The reason for putting in the ndim*idim is to allow for     *
!  the future optimization of doing all ndim particle sorts    *
!  at the same time.                                           *
!                                                              *
!  Subroutine by A. Nelson, based on code inherited            *
!  from W. Press, through W. Benz                              *
!                                                              *
!***************************************************************
      use treesort

      integer, intent(in   ):: n
      integer, intent(inout):: list(ndim*n),iwork(ndim*n)
      integer, intent(out  ):: nvolumes
      double precision, intent(inout):: work1(ndim*n),work2(ndim*n)

      integer(kind=ik8) iznlong,jznlong,kznlong,nlong

      integer nzn,i,j,izn,ilow,ihigh,jzn,jlow,jhigh,joffset
      integer k,kzn,klow,khigh,koffset
      integer nxmin,nxmax,nymin,nymax,nzmin,nzmax
      double precision xlimit,ylimit,zlimit,distmax
      character(len=20), parameter:: where = ('makegrid')

      if(idbg.ge.3)print *,where

      nzn=max(1,int(dble(n/nhash)**diminv))
      nznx=nzn ; nzny=nzn ; nznz=nzn
      if(ndim.eq.2)nznz=1
      nxvol=1+nznx/nznvol ; nyvol=1+nzny/nznvol ; nzvol=1+nznz/nznvol
      if(ndim.eq.2)nzvol=1
      ihalf=0
      nhalf=max(1,n/2)

      nplane=nznx*nzny
      ncells=nplane*nznz
      nvolumes=nxvol*nyvol*nzvol

      nlong=n

      xlimit=zero ; ylimit=zero ; zlimit=zero

      if(nzn.gt.nznmax)stop'nzn gt nznmax'
!
! sort the points in x, fill grid(i,1) and get x part of coordinate
!

      call pquick(n,1,treept,list,iwork,work1,work2)

      do izn=1,nznx
         iznlong=izn
         ilow =((iznlong-1)*nlong)/nznx + 1
         ihigh=(iznlong*nlong)/nznx
         ihigh=min(n,ihigh)
         if(ihigh.ne.n)then
             grid(izn,1)=half*(treept(1,list(ihigh  )) +
     &                         treept(1,list(ihigh+1)) )
         endif
         do i=ilow,ihigh
            key(list(i))=izn
         enddo
      enddo
      nxmin=list(1)
      nxmax=list(n)
      xlimit = treept(1,nxmax) - treept(1,nxmin)
!      
! sort the points in y, fill grid and get y part of coordinate
!
      call pquick(n,2,treept,list,iwork,work1,work2)

      do jzn=1,nzny
         jznlong=jzn
         jlow =((jznlong-1)*nlong)/nzny + 1
         jhigh=(jznlong*nlong)/nzny
         jhigh=min(n,jhigh)
         if(jhigh.ne.n)then
             grid(jzn,2)=half*(treept(2,list(jhigh  )) +
     &                         treept(2,list(jhigh+1)) )
         endif
         joffset=nznx*(jzn-1)
         do j=jlow,jhigh
            key(list(j))=key(list(j))+joffset
         enddo
      enddo
      nymin = list(1)
      nymax = list(n)
      ylimit = treept(2,nymax) - treept(2,nymin)

#if NDIM==3
!
! sort the points in z, fill grid(i,3) and get z part of coordinate
!
      call pquick(n,3,treept,list,iwork,work1,work2)

      do kzn=1,nznz
         kznlong=kzn
         klow =((kznlong-1)*nlong)/nznz + 1
         khigh=(kznlong*nlong)/nznz
         khigh=min(n,khigh)
         if(khigh.ne.n)then
             grid(kzn,3)=half*(treept(3,list(khigh  )) +
     &                         treept(3,list(khigh+1)) )
         endif
         koffset= (kzn-1)*nplane
         do k=klow,khigh
            key(list(k))=key(list(k))+koffset
         enddo
      enddo
      nzmin = list(1)
      nzmax = list(n)
      zlimit = treept(3,nzmax) - treept(3,nzmin)

#endif
!
!  put box edges twice as far away as furthest possible point.
!  (avoid lots of nasty if(not at box edge) in subr. nearest)
!   note if ndim=2 then zlimit=zero as defined at the top of
!   this subr.
!
      distmax=sqrt(xlimit**2 + ylimit**2 + zlimit**2)
      grid(0,1)   = treept(1,nxmin) - two*distmax
      grid(nznx,1)= treept(1,nxmax) + two*distmax
      grid(0,2)   = treept(2,nymin) - two*distmax
      grid(nzny,2)= treept(2,nymax) + two*distmax
#if NDIM==3
      grid(0,3)   = treept(3,nzmin) - two*distmax
      grid(nznz,3)= treept(3,nzmax) + two*distmax
#endif

      if(xlimit.gt.huge(1d0) .or. xlimit.lt.-huge(1d0))then
          print *,'You may have problems with particle position'
          print *,'damage. The limiting distance is very large'
          print *,'xlimit=',xlimit
      endif
      if(ylimit.gt.huge(1d0) .or. ylimit.lt.-huge(1d0))then
          print *,'You may have problems with particle position'
          print *,'damage. The limiting distance is very large'
          print *,'ylimit=',ylimit
      endif
      if(zlimit.gt.huge(1d0) .or. zlimit.lt.-huge(1d0))then
          print *,'You may have problems with particle position'
          print *,'damage. The limiting distance is very large'
          print *,'zlimit=',zlimit
      endif

      return
      end subroutine makegrid
!---------------------------------------------------------------------
      subroutine initlists
!*********************************************************************
!                                                                    *
! Create a set of zero terminated, linked lists of all               *
! particles in a given hash grid cell. The first entry               *
! of each list is ihead(1,m), additional entries are found           *
! in the next( ) array. This contains the first of any newly         *
! created nodes in this zone. ihead(2,m) contains the first *old*    * 
! node in zone m. In other words, it starts the linked               *
! list somewhat later on in the chain. We can therefore search       *
! through all particles in the zone by starting at ihead(1,m) and    *
! stopping when the value of next( ) = 0, or we can search           *
! through only newly created nodes, by starting at ihead(1,m)        *
! and stopping when next( ) = ihead(2, ).                            *
!                                                                    *
! Put zero in ihead for the linked list terminator for all           *
! cells. When a particle is found in cell m, fill next with          *
! the current ihead value and replace ihead with the new             *
! particle.                                                          *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************
      integer m,j

!$omp parallel do default(none) private(m)
!$omp& shared(ncells,ihead)
      do m=1,ncells
         ihead(1,m)=0
         ihead(2,m)=0
      enddo
      do j=1,natom
         m=key(j)
         next(j)=ihead(1,m)
         ihead(1,m)=j
      enddo

      return
      end subroutine initlists
!---------------------------------------------------------------------
      subroutine makenodes(new_mycpu,nstart,newchild)
!*********************************************************************
!                                                                    *
!  Calculate positions of new_mycpu newly created nodes,             *
!  found in the newchild array                                       *
!                                                                    *
!  Then find the grid key value for each new node                    *
!  (Note: need to add one to x, y and z grid coordinates to get      *
!   from 1..nzn? rather than 0..nzn?-1)                              *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************
      integer, intent(in   ):: new_mycpu,nstart
      integer, intent(in   ):: newchild(new_mycpu)

      integer n,nn,l,ll,mm1,mm2
      integer ix1,ix2,ilow,ihi,izn
      integer jy1,jy2,jlow,jhi,jzn
      integer kz1,kz2,klow,khi,kzn
      double precision fll,difx,dify,difz

      n=nstart-1
      do nn=1,new_mycpu
         l =newchild(nn) 
         ll=link(isib,l)
         n=n+1
         link(idau,n)=newchild(nn)
         link(isib,n)=0

         nay(l )=nroot_sib
         nay(ll)=nroot_sib
         nay(n )=nroot_sib
         treept(id2nay,n)=huge(1d0)
         icount(n)=icount(l) + icount(ll)

         treept(ims,n)=treept(ims,l)+treept(ims,ll)
         fll=treept(ims,ll)/treept(ims,n)
         difx=treept(1,ll)-treept(1,l)
         dify=treept(2,ll)-treept(2,l)
         treept(1,n)=treept(1,l)+fll*difx
         treept(2,n)=treept(2,l)+fll*dify
#if NDIM==3
         difz=treept(3,ll)-treept(3,l)
         treept(3,n)=treept(3,l)+fll*difz
#endif

         mm1=key(l )-1  ; mm2=key(ll)-1

         ix1=mod(mm1     ,nznx) + 1 ; ix2=mod(mm2     ,nznx) + 1
         jy1=mod(mm1/nznx,nzny) + 1 ; jy2=mod(mm2/nznx,nzny) + 1

         ilow=min(ix1,ix2) ; ihi=max(ix1,ix2)
         jlow=min(jy1,jy2) ; jhi=max(jy1,jy2)

         do izn=ilow,ihi
            if(treept(1,n).lt.grid(izn,1))exit
         enddo
         izn= min(izn,ihi)
         key(n)=izn

         do jzn=jlow,jhi
            if(treept(2,n).lt.grid(jzn,2))exit
         enddo
         jzn=min(jzn,jhi)
         key(n)= key(n) + (jzn-1)*nznx

#if NDIM==3

         kz1=mm1/nplane + 1 ; kz2=mm2/nplane + 1
         klow=min(kz1,kz2)  ; khi=max(kz1,kz2)

         do kzn=klow,khi
            if(treept(3,n).lt.grid(kzn,3))exit
         enddo
         kzn=min(kzn,khi)
         key(n)= key(n) + (kzn-1)*nplane
#endif

      enddo

      return
      end subroutine makenodes
!---------------------------------------------------------------------
      subroutine getnearest(ivolume)
!*********************************************************************
!                                                                    *
!  Find nearest neighbor particle of each particle or node           *
!  still active.                                                     *
!                                                                    *
!  Subroutine by A. Nelson, based on code inherited through W Benz   *
!  from W. Press.  The structure of this routine is mostly Press.    *
!  Most of the (re)implementation is Nelson.                         *
!                                                                    *
! ihead(1,m)  : treept( ) index of the first node in the list        *
!                    of unassociated nodes contained in zone m       *
! ihead(2,m)  : treept( ) index of the first node in the list        *
!                    of previously created (i.e. on a previous loop  *
!                    through the nearest/findnods/makenodes process  *
! next(inode) : treept( ) index of the next node in grid zone m      *
! nay (inode) : treept( ) index of inode's nearest neighbor. On      *
!                    the first call here, nay(inode)=inode. After    *
!                    that, nay(inode)=inode only if inode has        *
!                    already been associated.                        *
!                                                                    *
!*********************************************************************
      integer, intent(in ):: ivolume

      integer ixvol,jyvol,kzvol
      integer ixlo,ixhi,ixgrid
      integer jylo,jyhi,jygrid
      integer kzlo,kzhi,kzgrid
      integer kplane,jkcoord

      logical prev_is_taken
      integer ncel,inode,nc,icoord,id,idir,k,kend
      integer jx,jy,jz,jjy,jjz,iw,idold,iwall
      integer limbox(ndim,2),newbox(ndim,2)
      double precision d2min,d2,dmin,dwall(ndim,2)

#include "treefuncs"
!
! loop over grid cells to find nearest neighbors
!
      ixvol=1+mod( ivolume-1,nxvol)
      jyvol=1+mod((ivolume-1)/nxvol,nyvol)
      kzvol=1+(ivolume-1)/(nxvol*nyvol)
      ixlo=1+(ixvol-1)*nznvol ; ixhi= min(nznx,ixvol*nznvol)
      jylo=1+(jyvol-1)*nznvol ; jyhi= min(nzny,jyvol*nznvol)
      kzlo=1+(kzvol-1)*nznvol ; kzhi= min(nznz,kzvol*nznvol)

      do kzgrid=kzlo,kzhi
      kplane=(kzgrid-1)*nplane
      do jygrid=jylo,jyhi
      jkcoord=(jygrid-1)*nznx + kplane
      do ixgrid=ixlo,ixhi

         ncel= ixgrid + jkcoord
         inode=ihead(1,ncel)

         do while(inode.ne.0)
!
! initialize the zone limits and distances to the walls
! You'd think that the (much easier to read) looped 1..ndim form
! of the following would run just fast as this manual unrolling.
! But you'd be wrong. -AFN
!
! x
            limbox(1,1)=ixgrid-1            !add 1 below
            limbox(1,2)=limbox(1,1)
            newbox(1,1)=limbox(1,1)
            newbox(1,2)=limbox(1,2)
            dwall(1,1)=treept(1,inode)-grid(limbox(1,1),  1)
            dwall(1,2)=grid(limbox(1,2)+1,1)-treept(1,inode)
! y
            limbox(2,1)=jygrid-1
            limbox(2,2)=limbox(2,1)
            newbox(2,1)=limbox(2,1)
            newbox(2,2)=limbox(2,2)
            dwall(2,1)=treept(2,inode)-grid(limbox(2,1),  2)
            dwall(2,2)=grid(limbox(2,2)+1,2)-treept(2,inode)
! z
#if NDIM==3
            limbox(3,1)=kzgrid-1
            limbox(3,2)=limbox(3,1)
            newbox(3,1)=limbox(3,1)
            newbox(3,2)=limbox(3,2)
            dwall(3,1)=treept(3,inode)-grid(limbox(3,1),  3)
            dwall(3,2)=grid(limbox(3,2)+1,3)-treept(3,inode)
#endif
!
! Find the closest wall
!
            dmin=huge(1d0)
            do iwall=1,2
            do idir=1,ndim
               if(dwall(idir,iwall).lt.dmin)then
                  iw=iwall
                  id=idir
                  dmin=dwall(idir,iwall)
               endif
            enddo
            enddo
            idold=id
!
! check nodes in current cell, either all or only the new ones,
! depending on whether previous nay remains available. 
!
            k=ihead(1,ncel)
            if(nay(nay(inode)).eq.nroot_sib)then
               kend=0                      !previous nay is now associated.
               nc=0                        !must check every node 
               d2min=huge(1d0)
               prev_is_taken=.true.
            else                           !previous nay remains available.
               kend=ihead(2,ncel)          !we remember old nay data, so we
               nc=nay(inode)               !only need to check newly created nodes
               d2min=treept(id2nay,inode)
               prev_is_taken=.false.
            endif

            do while(k.ne.kend)
               if(k.ne.inode)then
                  d2=rr2(k,inode)
                  if(d2.lt.d2min)then
                     d2min=d2
                     nc=k
                  endif
               endif
            k=next(k)
            enddo

            do while(dmin*dmin.lt.d2min)
!
! reset search volume to full box (no-op first time through)
!
               newbox(idold,1)=limbox(idold,1)
               newbox(idold,2)=limbox(idold,2)
!
! and then expand it
!
               if(iw.eq.1)then
                  limbox(id,1)= limbox(id,1)-1
                  dwall (id,1)= treept(id,inode)-grid(limbox(id,1),id)
               else
                  limbox(id,2)= limbox(id,2)+1
                  dwall (id,2)= grid(limbox(id,2)+1,id)-treept(id,inode)
               endif
               newbox(id,1)=limbox(id,iw)
               newbox(id,2)=limbox(id,iw)
!
! find closest particle in the new volume
!
               jjz=0
#if NDIM==3
               do jz=newbox(3,1),newbox(3,2)
                  jjz=nplane*jz
#endif
                  do jy=newbox(2,1),newbox(2,2)
                     jjy=nznx*jy + jjz
                     do jx=newbox(1,1),newbox(1,2)
                        icoord=jx+1 + jjy
                        k   =ihead(1,icoord)
                        kend=ihead(2,icoord) ; if(prev_is_taken)kend=0 
                        do while(k.ne.kend)
                           d2=rr2(k,inode)
                           if(d2.lt.d2min)then
                              d2min=d2
                              nc=k
                           endif
                        k=next(k)
                        enddo
                     enddo
                  enddo
#if NDIM==3
               enddo
#endif
!
! find the new closest wall
!
               idold=id
               dmin=huge(1d0)
               do iwall=1,2
               do idir=1,ndim
                  if(dwall(idir,iwall).lt.dmin)then
                     iw=iwall
                     id=idir
                     dmin=dwall(idir,iwall)
                  endif
               enddo
               enddo
            enddo          !loop over nearest wall/node 
            nay(inode)=nc
            treept(id2nay,inode)=d2min

         inode=next(inode)
         enddo             !loop over nodes in current cell
      enddo                !x loop over hash grid cells
      enddo                !y loop over hash grid cells
      enddo                !z loop over hash grid cells

      return
      end subroutine getnearest
!---------------------------------------------------------------------
      subroutine revlists(newold,new,nvolumes)
!*********************************************************************
!                                                                    *
!  Remove nodes that have newly found siblings from                  *
!  the linked lists of unassociated nodes for each hash              *
!  grid cell, and put the newly created nodes onto those             *
!  linked lists.                                                     *
!                                                                    *
!  Subroutine by A. Nelson                                           *
!                                                                    *
!*********************************************************************
      integer, intent(in   ):: new,newold,nvolumes
      integer ncel,l,lold,m,n

!$    integer omp_get_max_threads
!$    external omp_get_max_threads

!
! Remove newly attached nodes 
!   Two steps:
!     1) determine if the current head of the list for this 
!        cell has been attached. If so, go down the linked list 
!        to find first unattached node or the end. Put in ihead 
!     2) determine if each successive node in the linked
!        chain of next(l) has been attached, if so, bypass it
!        and put the next next( ) there instead.
!
!$omp parallel default(none) if(nvolumes.gt.1)
!$omp& num_threads(min(ncells,omp_get_max_threads()))
!$omp&  private(ncel,l,lold)
!$omp&  shared(ncells,ihead,next,link)
!$omp do
      do ncel=1,ncells
         l=ihead(1,ncel)

         do while(l.ne.0)               !end of linked list?
            if(link(isib,l).eq.0)exit   !attached?
            l=next(l)                   !yes, go on to next on list
         enddo
         ihead(1,ncel)=l

         l=ihead(1,ncel)
         lold=l
         do while(l.ne.0)
            if(link(isib,l).eq.0)then   !not attached, advance current
               lold=l                   !pointer lold to l
            else                        !attached, jump over l and
               next(lold)=next(l)       !point link to next(l).
            endif
            l=next(l)
         enddo

         ihead(2,ncel)=ihead(1,ncel)    !ihead(2 ) is first 'old' node
      enddo
!$omp end parallel
!
! Put new nodes on linked lists
!
      do n=newold,new-1
         m=key(n)
         next(n)=ihead(1,m)
         ihead(1,m)=n
      enddo

      return
      end subroutine revlists
!---------------------------------------------------------------------
      subroutine relinkrootsibs
!*********************************************************************
!  Finish post-processing of tree nodes into ordered list.           *
!  Form more information see the findnodes routine and also          *
!     Makino 1990 JCP 87, 148                                        *
!     Dubinski 199? New Astronomy 1, 133                             *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer ntot1,me,my_dau,my_dausib
      character(len=20), parameter:: where = ('relinkrootsibs')

      if(idbg.ge.3)print *,where
!
! In original treebuild the root node is in the last
! (ntotnodes) array position, so we start from there
! and define its new sibling to be one greater than the
! largest node in the tree.
!
      me=ntotnodes
      ntot1=ntotnodes+1
      link(isib,me)=ntot1
      link(isib,ntot1)=ntotnodes

! now fix rightmost root children

      my_dau=link(idau,me)
      do while(my_dau.ne.0) 
         my_dausib=link(isib,my_dau)
         link(isib,my_dausib)=ntot1

         my_dau=link(idau,my_dausib)
      enddo

      return
      end subroutine relinkrootsibs
!---------------------------------------------------------------------
      subroutine revgrid(natom,iheadtmp,nvolumes)
!*********************************************************************
!                                                                    *
!  Revise the hash grid roughly defining the node positions, and     *
!  renumber those nodes in the 1d numbering system defining its      *
!  position in the grid                                              *
!                                                                    *
!  It would be nice to be able to use the num_threads clause in the  *
!  omp parallel directives below, like this:                         *
!                                                                    *
!  num_threads(min(nzny,omp_get_max_threads()))                      *
!                                                                    *
!  but it appears that this triggers some bug. Compiler or           *
!  otherwise, I haven't got the time to chase it right now.          *
!  Symptoms are a segvio after running through a few iterations      *
!  of the loop, usually the second loop (decreasing the y zones)     *
!  putting print statements in, changes this behavior in several     *
!  ways. The bug apparently doesn't hit if debugging options are     *
!  turned on. The print statement behavior change makes me think     *
!  memory or stack corruption somehow.                               *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in   ):: natom
      integer, intent(inout):: iheadtmp(natom)
      integer, intent(inout):: nvolumes

      integer ncel(3),ncellsold
      integer nznxold,nznyold,nznzold,ncelnew,nplaneold
      integer izn,jzn,kzn,i,inode,nextinode,lim2
      character(len=20), parameter:: where = ('revgrid')

      if(idbg.ge.3)print *,where

      if(ncells.eq.1)return

      nhalf=max(1,nhalf/2)

      if(mod(ihalf,ndim).eq.0)then
         nznxold=nznx
         nznx=max(1,nznx/2)
         nplaneold=nplane
         nplane=nznx*nzny
         do izn=1,nznx
            grid(izn,1)=grid(2*izn,1)
         enddo
         grid(nznx,1)=grid(nznxold,1)

!$omp parallel do if(nvolumes.gt.8) default(none)
!$omp& shared(key,next,ihead,iheadtmp,ncells)
!$omp& firstprivate(nznxold,nplaneold,nplane,nznx,nzny,nznz)
!$omp& private(i,inode,izn,jzn,kzn,ncel,ncelnew,nextinode,lim2)
         do jzn=1,nzny
         do kzn=1,nznz  !possibly nznz=1 if 2d
         do izn=1,nznx

            ncelnew=   izn   + (jzn-1)*nznx    + (kzn-1)*nplane
            ncel(1)= 2*izn-1 + (jzn-1)*nznxold + (kzn-1)*nplaneold
            ncel(2)= 2*izn   + (jzn-1)*nznxold + (kzn-1)*nplaneold
            ncel(3)= 2*izn+1 + (jzn-1)*nznxold + (kzn-1)*nplaneold

            lim2=2
            if(2*izn+1.eq.nznxold)lim2=3

            iheadtmp(ncelnew)=0
            do i=1,lim2
               inode=ihead(2,ncel(i))
               do while(inode.ne.0)
                  nextinode=next(inode)
                  key     (inode  ) = ncelnew
                  next    (inode  ) = iheadtmp(ncelnew)
                  iheadtmp(ncelnew) = inode
                  inode=nextinode
               enddo
            enddo
            iheadtmp(ncells+ncelnew)=iheadtmp(ncelnew)
            do i=1,lim2
               inode=ihead(1,ncel(i))
               do while(inode.ne.ihead(2,ncel(i)))
                  nextinode=next(inode)
                  key     (inode  ) = ncelnew
                  next    (inode  ) = iheadtmp(ncelnew)
                  iheadtmp(ncelnew) = inode
                  inode=nextinode
               enddo
            enddo
         enddo
         enddo
         enddo

      elseif(mod(ihalf,ndim).eq.1)then
         nznyold=nzny
         nzny=max(1,nzny/2)
         nplaneold=nplane
         nplane=nznx*nzny

         do jzn=1,nzny
            grid(jzn,2)=grid(2*jzn,2)
         enddo
         grid(nzny,2)=grid(nznyold,2)

!$omp parallel do if(nvolumes.gt.8) default(none)
!$omp&  shared(key,next,ihead,iheadtmp,ncells)
!$omp&  firstprivate(nplaneold,nplane,nznx,nzny,nznz,nznyold)
!$omp& private(i,inode,izn,jzn,kzn,ncel,ncelnew,nextinode,lim2)
         do izn=1,nznx
         do kzn=1,nznz  !possibly nznz=1 if 2d
         do jzn=1,nzny

            ncelnew= izn +       (jzn-1)*nznx + (kzn-1)*nplane
            ncel(1)= izn + ((2*jzn-1)-1)*nznx + (kzn-1)*nplaneold
            ncel(2)= izn + ((2*jzn  )-1)*nznx + (kzn-1)*nplaneold
            ncel(3)= izn + ((2*jzn+1)-1)*nznx + (kzn-1)*nplaneold

            lim2=2
            if(2*jzn+1.eq.nznyold)lim2=3

            iheadtmp(ncelnew)=0
            do i=1,lim2
               inode=ihead(2,ncel(i))
               do while(inode.ne.0)
                  nextinode=next(inode)
                  key     (inode  ) = ncelnew
                  next    (inode  ) = iheadtmp(ncelnew)
                  iheadtmp(ncelnew) = inode
                  inode=nextinode
               enddo
            enddo
            iheadtmp(ncells+ncelnew)=iheadtmp(ncelnew)
            do i=1,lim2
               inode=ihead(1,ncel(i))
               do while(inode.ne.ihead(2,ncel(i)))
                  nextinode=next(inode)
                  key     (inode  ) = ncelnew
                  next    (inode  ) = iheadtmp(ncelnew)
                  iheadtmp(ncelnew) = inode
                  inode=nextinode
               enddo
            enddo
         enddo
         enddo
         enddo

      else

#if NDIM==3

         nznzold=nznz
         nznz=max(1,nznz/2)
         do kzn=1,nznz
            grid(kzn,3)=grid(2*kzn,3)
         enddo
         grid(nznz,3)=grid(nznzold,3)

!$omp parallel do if(nvolumes.gt.8) default(none)
!$omp& shared(key,next,ihead,iheadtmp,ncells)
!$omp& firstprivate(nplane,nznx,nzny,nznz,nznzold)
!$omp& private(inode,izn,jzn,kzn,ncel,ncelnew,nextinode,lim2,i)
         do jzn=1,nzny
         do izn=1,nznx
         do kzn=1,nznz

            ncelnew= izn + (jzn-1)*nznx +       (kzn-1)*nplane
            ncel(1)= izn + (jzn-1)*nznx + ((2*kzn-1)-1)*nplane
            ncel(2)= izn + (jzn-1)*nznx + ((2*kzn  )-1)*nplane
            ncel(3)= izn + (jzn-1)*nznx + ((2*kzn+1)-1)*nplane

            lim2=2
            if(2*kzn+1.eq.nznzold)lim2=3

            iheadtmp(ncelnew)=0
            do i=1,lim2
               inode=ihead(2,ncel(i))
               do while(inode.ne.0)
                  nextinode=next(inode)
                  key     (inode  ) = ncelnew
                  next    (inode  ) = iheadtmp(ncelnew)
                  iheadtmp(ncelnew) = inode
                  inode=nextinode
               enddo
            enddo
            iheadtmp(ncells+ncelnew)=iheadtmp(ncelnew)
            do i=1,lim2
               inode=ihead(1,ncel(i))
               do while(inode.ne.ihead(2,ncel(i)))
                  nextinode=next(inode)
                  key     (inode  ) = ncelnew
                  next    (inode  ) = iheadtmp(ncelnew)
                  iheadtmp(ncelnew) = inode
                  inode=nextinode
               enddo
            enddo
         enddo
         enddo
         enddo
#endif

      endif

      nxvol=1+nznx/nznvol ; nyvol=1+nzny/nznvol ; nzvol=1+nznz/nznvol
      if(ndim.eq.2)nzvol=1
      nvolumes=nxvol*nyvol*nzvol

      ncellsold=ncells

      ncells=nplane*nznz
      ihalf = ihalf + 1
      do i=1,ncells
         ihead(1,i)=iheadtmp(i)
         ihead(2,i)=iheadtmp(ncellsold+i)
      enddo

      return
      end subroutine revgrid
!---------------------------------------------------------------------
      subroutine getbranchroots
!*********************************************************************
!                                                                    *
! Get a list of branches for the tree update (eg. revpoles/revnodes) *
! to work on in parallel. This works well for parallel runs with     *
! non-flat memory.                                                   *
!                                                                    *
! The idea it to divide the work into some small number of           *
! independent branches, update them, then update a few orphan        *
! nodes that live in between these branches (and above them)         *
! The size of these branches would ideally be exactly                *
! ntotnodes/numcpus but this is not possible in practice.            *
! Some particles always end up as `flyaways' from the main           *
! region, and therefore only get put into the tree very high up.     *
! Therefore use ntotnodes/(10*numcpus) as a maximum branch size.     *
! 10 is just a number. We could use a different number.              *
! Tune as needed.                                                    *
!                                                                    *
! lstbranches   A list of tree branches that can be updated          *
!               independently                                        *
! numbranches   The number of such branches                          *
!                                                                    *
! Subroutine by A. Nelson                                            *
!                                                                    *
!*********************************************************************

      integer numcpus,maxbranchsize
      integer me,iroot_sib,num,my_sib
      integer icpu,nodesperproc,numnodes,nodessofar

      character(len=20), parameter:: where = ('getbranchroots')

      if(idbg.ge.2)print *,where

! Get the independent branches, for tree updates (rtree) 
!
      call numprocessors(numcpus)
      numcpus_sav=numcpus         !rtree requires same numcpus as here
                                  !this should be automatic, but lets
                                  !make it more bulletproof

      numbranches=1 ; lstbranches(numbranches)=nroot
      nbrnchstart(1,1)=1
      nbrnchstart(2,1)=1
      if(numcpus.eq.1)return

      maxbranchsize=max(maxclumpsize,ntotnodes/(10*numcpus))
      me=nroot ; iroot_sib=link(isib,me)

      nodesperproc=ntotnodes/numcpus
      nodessofar=0  ;icpu=0

      numbranches=0
      do while(me.ne.iroot_sib)
         my_sib=link(isib,me)
         numnodes=my_sib-me
         num= numnodes/2 + 1
         if(num.gt.maxbranchsize)then
            me=link(idau,me)
            nodessofar=nodessofar+1
         else
            numbranches=numbranches+1
            lstbranches(numbranches)=me
            me=link(isib,me)
            nodessofar=nodessofar+numnodes
            if(nodessofar.ge.icpu*nodesperproc)then
               icpu=icpu+1
               nbrnchstart(1,icpu)=numbranches
               nbrnchstart(2,icpu-1)=numbranches-1
            endif
         endif
      enddo
      nbrnchstart(2,numcpus)=numbranches

      if(numbranches.gt.maxbranches)then
         stop'getbranchroots:Need bigger branch array in treevars'
      endif

      return
      end subroutine getbranchroots
!---------------------------------------------------------------------
      subroutine sorttree
!*********************************************************************
!                                                                    *
! Sort the tree data to produce consecutive ordering.                *
! `Consecutive' means in the same order as nodes are                 *
! accessed if every node in the tree is opened one after             *
! the other. This speeds up the force/neighbor calculation by        *
! organizing data which is accessed frequently nearby in memory.     *
!                                                                    *
! Produce two arrays linking the original sph particle               *
! ordering to the new sorted ordering and back                       *
!                                                                    *
!   linksphtotree: linksphtotree(i)=n  means that the i'th memory    *
!                    location in the list of sph particles is the    *
!                    same as the n'th position in the list of tree   *
!                    nodes. Use this when converting from sph        *
!                    format to tree format                           *
!                                                                    *
!   linktreetosph: linktreetosph(i)=n  means that the i'th memory    *
!                     location in the list of tree positions is      *
!                     the same as the n'th position in the list      *
!                     of sph particles. Use this when converting     *
!                     from tree format to sph format.                *
!                                                                    *
!   Note that in both linksphtotree and linktreetosph the list of    *
!   sph points is ill-defined for particle `numbers' greater than    *
!   natom. For particles > natom, the positions correspond to        *
!   the node's original position in the unsorted tree.               *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************

      integer my_cpu,numcpus
      integer nbegin,nend,nsofar,n_in_me
      integer me,my_dau,my_sib
      integer n

      character(len=20), parameter:: where = ('sorttree')
      if(idbg.ge.2)print *,where
!
! make the links between the original unsorted sph ordering
!   and the reordered tree, starting at the root node, which
!   is numbered ntotnodes in the originally created (sph) numbering 
!
!$omp parallel default(none)
!$omp& private(my_cpu,numcpus,nbegin,nend,n,nsofar,n_in_me)
!$omp& private(me,my_dau,my_sib)
!$omp& shared(ntotnodes,icount,link,linksphtotree,linktreetosph)
!
! Limits of the node partition allocated to my_cpu
!
      call get_mycpu(my_cpu) ; call numthreads(numcpus)
      nbegin= ((my_cpu-1)*ntotnodes)/numcpus + 1
      nend  = (    my_cpu*ntotnodes)/numcpus
      if(my_cpu.eq.numcpus)nend=ntotnodes
!
! Find first node in my_cpu's partition, called 'me'
!
      me=ntotnodes
      nsofar=1
      do while(nsofar.lt.nbegin)
          my_dau=link(idau,me)
          my_sib=link(isib,me)

          n_in_me=2*icount(me)-1
          if(nsofar+n_in_me.gt.nbegin)then
             nsofar=nsofar+1
             me=my_dau
          else
             nsofar=nsofar+n_in_me
             me=my_sib
          endif
      enddo
!
! Make translation links between sph and tree ordering
!
      do n=nbegin,nend 
         linksphtotree(me)=n
         linktreetosph(n)=me

         my_dau=link(idau,me)
         my_sib=link(isib,me)
         if(my_dau.eq.0)then
            me=my_sib
         else
            me=my_dau
         endif
      enddo
!$omp end parallel

!
! take care of root's `sibling'...the end of list flag
!
      linksphtotree(ntotnodes+1)=ntotnodes+1
      linktreetosph(ntotnodes+1)=ntotnodes+1

      return
      end subroutine sorttree
!---------------------------------------------------------------------
      subroutine updatelinks(ntotnodes,newdau,newsib)
!*********************************************************************
!                                                                    *
! Update the links to the sibling and daughter nodes                 *
! for the new ordering, determined in sorttree                       *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in)   :: ntotnodes
      integer, intent(inout)::newdau(ntotnodes),newsib(ntotnodes)

      integer i,j
      integer my_dauold,my_sibold

      character(len=20), parameter:: where = ('updatelinks')
      if(idbg.ge.2)print *,where
!
! set nroot to be first node in tree. 
!
      nroot=1

!$omp parallel default(none) 
!$omp& private (i,j,my_dauold,my_sibold)
!$omp& shared(ntotnodes,link,newdau,newsib)
!$omp& shared(linktreetosph,linksphtotree)
!$omp do
      do i=1,ntotnodes
         j=linktreetosph(i)        !get the original node id
         my_dauold=link(idau,j)    !then find the daughter and
         my_sibold=link(isib,j)    !sibling on the original list 

         if(my_dauold.eq.0)then
            newdau(i)=0
         else
            newdau(i)=linksphtotree(my_dauold)
         endif
         newsib(i)=linksphtotree(my_sibold)
      enddo
!$omp do 
      do i=1,ntotnodes
         link(idau,i)=newdau(i)
         link(isib,i)=newsib(i)
         newdau(i)=0             !required cleanup needed for 
         newsib(i)=0             !equivalanced array 'active'
      enddo
!$omp end parallel

      return
      end subroutine updatelinks
!---------------------------------------------------------------------
      subroutine gethclumps(irevise,numclumps,lstclumps)
!*********************************************************************
!                                                                    *
!  Get exact sizes all clumps, useful for reducing time for          *
!  treewalks for gravity.                                            *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: irevise
      integer, intent(in):: numclumps, lstclumps(numclumps)

!$    integer  omp_get_max_threads
!$    external omp_get_max_threads

      integer numcpus,my_cpu
      integer ibot,itop,jbot,jtop

      integer jclump,iclump,me,icl_sib
      double precision rmax,rad
      character(len=20), parameter:: where = ('gethclumps')

#include "treefuncs"

      if(idbg.ge.3)print *,where

!$omp parallel default(none) 
!$omp& num_threads(min(numclumps,omp_get_max_threads()))
!$omp&  shared(ntotnodes,numclumps,lstclumps,link,treept)
!$omp&  private (jclump,iclump,me,rad,rmax)
!$omp&  private(my_cpu,numcpus,ibot,itop,jbot,jtop)

      call get_mycpu(my_cpu)
      call numthreads(numcpus)

! divide clumps between processors and get limits for my_cpu's loop

      jbot= 1 + ((my_cpu-1)*numclumps)/numcpus
      jtop=         (my_cpu*numclumps)/numcpus
      jtop=min(numclumps,jtop)

      ibot=lstclumps(jbot)
      if(my_cpu.ne.numcpus)then
         itop=lstclumps(jtop+1)-1
      else
         itop=ntotnodes
      endif

      jclump=jtop
      iclump=lstclumps(jclump)

      rmax=zero
      do me=itop,ibot,-1
         if(link(idau,me).eq.0)then
            rad=sqrt(rr2(iclump,me)) + treept(ihnode,me)
            rmax=max(rmax,rad)
         endif
         if(me.eq.iclump)then
            treept(ihnode,iclump)=rmax
            jclump=max(1,jclump-1)     !special case for last iter to
            iclump=lstclumps(jclump)   !avoid access to 0 element
            rmax=zero                  !(which never gets used anyway)
         endif
      enddo
!$omp end parallel

      return
      end subroutine gethclumps
!---------------------------------------------------------------------
      subroutine gethbunches
!*********************************************************************
!                                                                    *
!  Get final values for the sizes of all of the bunches              *
!  of particles grouped together for treewalking with grape.         *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer my_cpu,numcpus
      integer ibot,itop,jbot,jtop

!$    integer  omp_get_max_threads
!$    external omp_get_max_threads

      integer jclump,iclump
      integer me
      double precision rmax,rad
      character(len=20), parameter:: where = ('gethbunches')

#include "treefuncs"

      if(idbg.ge.3)print *,where

!$omp parallel default(none) 
!$omp& num_threads(min(numbunches,omp_get_max_threads()))
!$omp&  shared(ntotnodes,numbunches,lstbunches,treept,link)
!$omp&  private (jclump,iclump,me,rad,rmax)
!$omp&  private(my_cpu,numcpus,ibot,itop,jbot,jtop)

      call get_mycpu(my_cpu)
      call numthreads(numcpus)

! divide clumps between processors and get limits for my_cpu's loop

      jbot= 1 + ((my_cpu-1)*numbunches)/numcpus
      jtop=         (my_cpu*numbunches)/numcpus
      jtop=min(numbunches,jtop)

      ibot=lstbunches(jbot)
      if(my_cpu.ne.numcpus)then
         itop=lstbunches(jtop+1)-1
      else
         itop=ntotnodes
      endif

      jclump=jtop
      iclump=lstbunches(jclump)

      rmax=zero
      do me=itop,ibot,-1
         if(link(idau,me).eq.0)then
            rad=sqrt(rr2(iclump,me)) + treept(ihnode,me)
            rmax=max(rmax,rad)
         endif
         if(me.eq.iclump)then
            treept(ihnode,iclump)=rmax
            jclump=max(1,jclump-1)        !reason for max: see
            iclump=lstbunches(jclump)     !comment in gethclumps
            rmax=zero
         endif
      enddo
!$omp end parallel

      return
      end subroutine gethbunches
!---------------------------------------------------------------------
      subroutine gethnodes(irevise,numclumps,lstclumps)
!*********************************************************************
!                                                                    *
!  Get final values for the sizes of all of the subnodes             *
!  of the clumps grouped together for treewalking.                   *
!  Could do without this routine but it speeds up getneigh           *
!  to do it this way.                                                *
!                                                                    *
!  If the clumps are very big, just reduce the clump sizes           *
!  themselves, not any of the subnodes, in order to save time.       *
!                                                                    *
!  Subroutine by A. Nelson.                                          *
!                                                                    *
!*********************************************************************
      integer, intent(in):: irevise
      integer, intent(in):: numclumps, lstclumps(numclumps)

      integer jclump,iclump
      integer isn,isnode,me,i
      integer na,list(maxclumpsize)
      integer is,isubnode(maxclumpsize-1)
      integer lim(2,maxclumpsize-1)
      double precision rmax,rad
      character(len=20), parameter:: where = ('gethnodes')

#include "treefuncs"

      if(idbg.ge.3)print *,where

      if(maxclumpsize.gt.max_exact_h)then

         call gethclumps(irevise,numclumps,lstclumps)

      else

!$omp parallel do default(none)
!$omp& shared(numclumps,lstclumps,treept)
!$omp& private (jclump,iclump,is,isubnode,na,list,lim)
!$omp& private (i,me,rmax,rad,isn,isnode)
         do jclump=1,numclumps
            iclump=lstclumps(jclump)

            call getallsubnodes(iclump,is,isubnode,na,list,lim)
            if(na.le.2)cycle
            do isn=1,is
               if(lim(2,isn)-lim(1,isn).eq.1)cycle   !pairs already exact
               isnode=isubnode(isn)
               rmax=zero
               do i=lim(1,isn),lim(2,isn)
                  me=list(i)
                  rad=sqrt(rr2(isnode,me)) + treept(ihnode,me)
                  rmax=max(rmax,rad)
               enddo
               treept(ihnode,isnode)=rmax
            enddo
         enddo

      endif

      return
      end subroutine gethnodes
!---------------------------------------------------------------------
      subroutine failbuild(ntreelev,maxtreelev,nremain,nremainold)

      integer, intent(in)::ntreelev,maxtreelev,nremain,nremainold

      print *,' '
      print *,' '
      print *,'THERE WAS A PROBLEM BUILDING THE TREE'
      print *,' '
      print *,'Number of tree levels so far    :',ntreelev
      print *,'Coded maximum number of levels  :',maxtreelev
      print *,'Number of unassociated nodes    :',nremain
      print *,'Old number of unassociated nodes:',nremainold
      print *,' '
      print *,'You have either '
      print *,'  -A very strange particle distribution'
      print *,'     Are your particles exactly positioned on a'
      print *,'     lattice? If so, nearest neighbors may not be'
      print *,'     unique. Circular chains could develop. Try'
      print *,'     giving their positions a random perturbation of'
      print *,'     magnitude 5e-16 or so. Other pathological'
      print *,'     distributions are very unlikely but may be'
      print *,'     happening if you find that ntreelev=maxtreelev'
      print *,'     solve by increasing the maxtreelev parameter'
      print *,' '
      print *,'  -A very large number of particles'
      print *,'     It may be possible to overflow some integers'
      print *,'     in the tree build if you have more than about'
      print *,'     1e7 particles in 3d or 2e6 particles in 2d'
      print *,'     I''ve found some locations for this and made'
      print *,'     them into long integers. There may be others'
      print *,' '
      print *,'  -A VERY, VERY large number of particles'
      print *,'     You will probably want to change every other'
      print *,'     integer array to long integer if this is the'
      print *,'     case, also you need to increase the' 
      print *,'     maxtreelev parameter'
      print *,' '
      print *,'  -Particles with damage in their positions'
      print *,'     likely if nremain=nremainold and' 
      print *,'               ntreelev<maxtreelev'
      print *,'     solve by fixing the damage'
      print *,' '
      print *,'  -Run out of stack'
      print *,'     VINE, like many fortran programs, requires a large'
      print *,'     stack to run and it can crash with strange'
      print *,'     unrepeatable errors, especially in the treebuild,'
      print *,'     but also elsewhere if you don''t have enough'
      print *,'     stackspace. Try the limit command (in unix) and do'
      print *,'     ''limit stackspace unlimited'''
      print *,'     There are other variants of this fix that may be'
      print *,'     needed for OMP parallel operations for the slave'
      print *,'     stacks. See discussion in the relevant'
      print *,'     INCLUDE/makefile.<whatever> for your compiler'
      print *,' '
      print *,'  -Found a bug in the tree build'
      print *,'     solve by fixing the bug'
      print *,'       Likely candidates: '
      print *,'         -Are the sorts sorting correctly?'
      print *,'            print out and plot these data. Sorted?'
      print *,'            try on different cpu count. Works?'
      print *,'            Try serial sorts (i.e. not just with'
      print *,'            ncpus=1, but with PARALLEL=0 in the makefile'
      print *,'            Works? Contact your compiler vendor. The'
      print *,'            sort is insidiously difficult for compilers'
      print *,'            to do correctly. We''ve found lots of'
      print *,'            compiler bugs from that source'
      print *,'         -Is the hash grid revision working correctly?'
      print *,'            verify that the nodes are still inside'
      print *,'            their assigned cell'
      print *,'         -There are probably others '
      print *,' '
      print *,'  -Found a nasty compiler, openmp synchronization, or'
      print *,'     hardware bug. The treebuild is extraordinarily'
      print *,'     sensitive to this issue, but it can occur '
      print *,'     elsewhere too, with a bus error or segvio'
      print *,'     try rerunning. Does it continue right through'
      print *,'          where it aborted this time?'
      print *,'     try running on a different architecture or compiler'
      print *,'          Does it run there?'
      print *,'     try doing some RAM memory sanity checks.'
      print *,'          Do you have some crappy 3rd party RAM?'
      print *,' '
      print *,'     beyond that, you might not find a solution'
      print *,'     unless you are willing to do a lot of digging'
      print *,'     and then scream at your compiler or hardware vendor'
      print *,' '
      print *,'In any case, you need a bottle of something'
      print *,'strong because the present tree build has failed'
      print *,' '

      stop'IT IS A SAD DAY: The tree build has failed'

      end subroutine failbuild

!---------------------------------------------------------------------
      end module treemake
