      module treevars
!****************************************************************
! This module is a component of VINE, copyright by the authors, *
! as listed below. Please see the COPYING and LICENSE files     *
! in the parent directory for additional information.           *
! This notification is present as an attempt to follow the      *
! recommendations for licensing published on the FSF page.      * 
!************************************************************************
!  This file contains the VINE module defining the variables            *
!  and arrays used for the tree building and traversal                  *
!  routines.                                                            *
!                                                                       *
!  Written by Andy Nelson, Markus Wetzstein and Thorsten Naab           *
!                                                                       *
!************************************************************************
      use dimens, only:ims,idim,ndim,ndim3,maxcpus
      use dimens, only:maxclumpsize,maxbunchsize
      use dimens, only:ik8

      implicit none

      private ims,idim,ndim,ndim3,maxcpus
      private maxclumpsize,maxbunchsize
      private ik8
      public
      save

!************************************************************************
!************************************************************************
!************************************************************************
!***************       TREE INFORMATION       ***************************
!************************************************************************
!************************************************************************
!************************************************************************
!
!
!************************************************************************
!  TREE BUILD                                                           *
!************************************************************************
!
! maximum size for the tree arrays. Technically this could be
! 2idim-1, but we use the end space of the arrays for some useful info, 
! like a fictitious root sibling and memory saving equivalence stuff 
! in the tree build
!
      integer, parameter:: mmax=2*idim
!
! maximum size of 1d hash grid for tree build
!
      integer, parameter:: nznmax=2000
!
! initial average number of nodes per hash zone
!
      integer, parameter:: nhash=3   !DO NOT SET TO 1!!!
                                     !ihead could overflow (theoretically)
!
! number of hash grid zones per direction in each region 
! This number appears to affect high cpu scaling slightly and is
! also slightly problem dependent, but 16 seems to be a reasonable number. 
!
      integer, parameter:: nznvol=16
!
! maximum tree depth. Adjust as needed for larger
! numbers of particles or pathological distributions.
!
      integer, parameter:: maxtreelev=2000
      integer, parameter:: maxbranches=15*maxcpus
!
! Identity flags for full or partial tree revision options
!
      integer, parameter:: id_fullrevise=1,id_partrevise=0
!
!************************************************************************
!  TREE WALK                                                            *
!************************************************************************
!
! array locations for tree searching quantities
!
      integer, parameter:: ihnode =ims+1   !size of node
      integer, parameter:: iconvrg=ims+2   !convergence criterion
      integer, parameter:: isib=1,idau=2

      integer, parameter:: id2nay = ihnode   !reuse of ihnode
                                             !space during tree build.
!
! maximum size of work to bite off at one time
!
      integer, parameter:: maxn2=75         !Maximum clumpsize to calculate
                                            !n squared accelerations
                                            !for gravity and
                                            !for neighbor lists 
      integer, parameter:: max_exact_h=500  !maxparticles to calculate 
                                            !exact clump sizes for
                                            !if it happens you reset
                                            !maxclumpsize for some
                                            !strange reason.
!
!************************************************************************
!  module contents                                                      *
!************************************************************************
!                                                                       *
!  treept    : position, mass, node size and critical radius for        * 
!              node opening of the current node                         *
!  clmpsize0 : Initial size of the clumps. Used for                     *
!              judging when to rebuild the tree.                        *
!  amag      : Magnitude of old acceleration (for some MACs)            *
!              in main code numbering (i.e. needs linksphtotree to      *
!              reference in the tree itself.                            *
!  bndbox    : Boundaries of a box (used only for non-free boundaries)  *
!              first index (1 or 2) is left or right side respectively  *
!              (might be useful someday to generalize to any surface,   *
!               like say, the surface of a sphere)                      *
!  rbox      : Size of the box rbox(id)= bndbox(2,id)-bndbox(1,id)      *
!  rboxinv   : One over size of box 1/rbox(id)                          *
!  rboxhalf  : Half of rbox                                             *
!  natom     : number of atoms in the tree                              *
!  natom_sph : number of atoms in the tree that are SPH particles       *
!  nroot     : index of root particle in the tree                       *
!  nroot_sph : index of root particle for sph particles in the tree     *
!  nroot_nbod: index of root particle for Nbody particles in the tree   *
!  ntotnodes : total number of nodes (incl particles) in tree           *
!  ntotsphnodes: total number of SPH nodes (incl particles) in tree     *
!  link      : the daughter or sibling of the current node              *
!  ntreelev  : number of levels in the tree                             *
!  active    : Logical flag to say whether the particle or clump        *
!              needs its force calculated or neighbors found etc        *
!                                                                       *
!  Sorted tree info:                                                    *
!                                                                       *
!  linksphtotree: linksphtotree(i)=n  means that the i'th memory        *
!                    location in the list of sph particles is the       *
!                    same as the n'th position in the list of tree      *
!                    nodes. Use this when converting from sph           *
!                    format to tree format                              *
!                                                                       *
!  linktreetosph: linktreetosph(i)=n  means that the i'th memory        *
!                     location in the list of tree positions is         *
!                     the same as the n'th position in the list         *
!                     of sph particles. Use this when converting        *
!                     from tree format to sph format.                   *
!                                                                       *
!  Note that in both cases the list of sph points is ill-defined        *
!  for particle `numbers' greater than natom. For particles > natom,    *
!  the positions correspond to the node's original position in the      *
!  unsorted tree.                                                       *
!************************************************************************
      integer nznx,nzny,nznz,nplane,ncells,nhalf,ihalf
      integer nxvol,nyvol,nzvol
      double precision grid(0:nznmax+1,ndim)

      double precision treept(ndim3,mmax),   !equiv temparr in orderparticles
     &                    clmpsize0(idim),   !and work1/work2 in mtree
     &                         amag(idim),
     &   bndbox(2,ndim),boxsize(ndim),boxinv(ndim),boxhalf(ndim)

      integer   active(mmax),    !equiv next  in mtree
     &          link(2,mmax),    !equiv iwork in mtree and itemparr in orderparticles 
     &   linksphtotree(mmax),    !equiv nay   in mtree
     &   linktreetosph(mmax),    !equiv key   in mtree
     &      lstbunches(idim),    !equiv ihead in mtree, isort in orderparticles
     &   lstbranches(maxbranches),numbranches,
     &   nbrnchstart(2,0:maxcpus+1),numcpus_sav, 
     &   nroot     ,nroot_sib,
     &   nroot_sph ,nroot_sphsib,
     &   nroot_nbod,nroot_nbodsib,
     &   ntotnodes,ntotsphnodes,natom,natom_sph,ntreelev

      integer, save:: numbunches=0  !initing to zero eliminates some grape ifdefs

      integer icount(ndim*idim)     !this is a work array for treemake

!************************************************************************
!                                                                       *
! No new arrays are declared below this point in the file.              *
!                                                                       *
! Below are a number of equivalences for conserving space in various    *
! parts of the tree build and the particle reordering process           *
!                                                                       *
!************************************************************************
!
!===================================================================
!  reuse tree space for temp storage in module vineutils 
!
      integer itemparr(4,idim),isort(idim)
      equivalence (link      ,itemparr)      !4*idim
      equivalence (lstbunches,isort   )      !idim

      double precision temparr(ndim3,mmax)
      equivalence (treept    ,temparr)       !full size
!===================================================================
! reuse tree space needed only after the tree is built
! in the treebuild itself. 
! 
!  work1/work2 are used in treebuild pre processing, before nodes
!  are created: put start of work1 immediately following last
!  possible particle. put work2 right after that.

      double precision work1(ndim*idim),work2(ndim*idim)
      equivalence (treept(1,1 + idim                      ) ,work1)  !ndim*idim
      equivalence (treept(1,1 + idim + 1 + ndim*idim/ndim3) ,work2)  !ndim*idim

!used during tree construction:
      integer ihead(2,idim/2),next(mmax),nay(mmax),key(mmax)
      equivalence (lstbunches   ,ihead    )  !idim
      equivalence (linksphtotree,nay      )  !mmax
      equivalence (linktreetosph,key      )  !mmax
      equivalence (active       ,next     )  !mmax

!iwork used in pre and iwork1 used in post processing of tree construction 
      integer iwork(ndim*idim), iwork1(mmax)
      equivalence (link         ,iwork    )  !ndim*idim
      equivalence (active       ,iwork1   )  !mmax

      end module treevars
